/* @(#) All Right Reserved, Copyright (c) 1997 Australian National University */
/* @(#)                     Copyright (c) 1997 Fujitsu Laboratories Ltd. */
/*  */
/*        */
/*                        Permission Notice */
/*  */
/*                Australian National University */
/*                Canberra ACT 0200 Australia */
/*                           and */
/*                Fujitsu Laboratories Ltd. */
/*                4-1-1 Kamikodanaka Nakahara */
/*                Kawasaki 211-88 Japan */
/*  */
/*   Permission to use, copy, modify and distribute this software (including  */
/*   its documentation) without fee is hereby granted under the following terms */
/*   and conditions: */
/*  */
/*   1. This permission notice and all of the copyright notices in this */
/*      software must appear in all copies and derived versions of all */
/*      or a part of this software. */
/*  */
/*   2. Sales of this software, whether incorporated in other products as a */
/*      part thereof or not, is prohibited. */
/*  */
/*   3. On publication or presentation in any form of the results obtained */
/*      by use of this software, an explicit reference to this software and */
/*      its copyright owner shall be made. */
/*  */
/*   4. THIS SOFTWARE IS PROVIDED 'AS IS' AND THE AUSTRALIAN NATIONAL UNIVERSITY */
/*      AND FUJITSU LABORATORIES LTD. MAKE NO WARRANTIES, EITHER EXPRESSED OR */
/*      IMPLIED, AS TO ANY MATTER WHATSOEVER REGARDING THIS SOFTWARE,  */
/*      INCLUDING, WITHOUT LIMITATION, WARRANTIES OF MERCHANTABILITY OR FITNESS */
/*      FOR ANY PARTICULAR PURPOSE. */
/*      IN NO EVENT SHALL THE AUSTRALIAN NATIONAL UNIVERSITY OR FUJITSU */
/*      LABORATORIES LTD. BE LIABLE FOR ANY CLAIM FROM A THIRD PARTY, OR */
/*      FOR SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES WHATSOEVER RESULTING */
/*      FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN ACTION OF CONTRACT, */
/*      NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION */
/*      WITH THE USE, COPYING OR MODIFICATION OF THIS SOFTWARE. */
/*  */

/************************************************************************/
/*                              Cell.h 	                                */
/************************************************************************/
/* SCCS INFO @(#)Cell.h 1.7         last modified P. Strazdins, 97/05/19
   first created: P. Strazdins 95/09
*/
/* set up definitions for cell programs wrt communication with host (if any)
 * and other cells.
 * compilatation:
 *	if FTN_BLACS defined, set up cell for Fortran BLACS communication.
 *	if APLIB_* defined, set up cell for APLIB (APx000) communication.
 *	if none defined, set up as for single workstation.
 *	if HOST defined (valid only if FTN_BLACS or APLIB_*),
 *	  (argv, argc) and date strings are b/c from a host program
 *	  and overall output (in string form) is sent to host for printing
 *	else (assumes Unix I/O calls implemented on cells with main(argv, argc))
 *	  if APLIB_* || FTN_BLACS defined, print main stdout into a file
 *	  else print to default stdout.
 */
#ifndef __Cell_Incl__
#define __Cell_Incl__

#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>

#if defined(APLIB_STD) || defined(APLIB_STRIDE) || defined(APLIB_LINUX) || defined(APLIB_PLUS_STRIDE)
# define APLIB
#endif 


#ifdef FTN_BLACS
# define CELL_INIT			

# ifdef HOST				/* these 2 ap1000-dept*/
#  include <ccell.c7.h>			/* needed for host-cell messaging*/
#  define main cell_main
# else
int getncelx(),  getncely(); 
# endif

static int BLACSCntxt = -1;	
static int CID, CID_I, CID_J, NP, NP_I, NP_J;
# include       "FtnBLACS.h"

					/* modify next dir. for non APx000*/
# define GET_PNUM(Ncely, Ncelx)		/* find out main grid size */ 	\
	{ (Ncelx) = getncelx(); 	(Ncely) = getncely(); }

# define COMMUN_INIT(Ncely, Ncelx)					\
    { int BLACSDefHandle=0;						\
      GET_PNUM(Ncely, Ncelx);						\
      NP_I = (Ncely); NP_J = (Ncelx);					\
      FTN_BLACS_GET(&BLACSDefHandle, &BLACSDefHandle, &BLACSCntxt);	\
      FTN_BLACS_PINFO(&CID, &NP);					\
      FTN_BLACS_GRIDINIT(&BLACSCntxt, "Row", &NP_I, &NP_J);		\
      FTN_BLACS_GRIDINFO(&BLACSCntxt, &NP_I, &NP_J, &CID_I, &CID_J);	\
      printf("BLACS cntxt %d, cell (%d,%d) of %d x %d grid\n", 		\
      		BLACSCntxt, CID_I, CID_J, NP_I, NP_J);			\
      BLACSCntxtSet(BLACSCntxt); CommSetUp(NP_I, NP_J);			\
      /*above 2 needed if FDATACommun procedures are called directly*/	\
    }                            

# include "FDATACommun.h"
# include <string.h>
static FDATA a_comm_[16];	/* typically CELL_SEND/RECV() have size <= 8 */
# define CELL_COMM_MTYPE 1010
# define CELL_SEND(i, j, a, size)					\
	/*send to cell (i,j) size bytes from address a*/		\
	{ int n_ = ((size)+sizeof(FDATA)-1)/sizeof(FDATA);		\
	  memcpy((void *) a_comm_, (void *) (a), (size_t) (size));	\
	  SendMat((NP_I+(i)-CID_I)%NP_I, (NP_J+(j)-CID_J)%NP_J, 1, n_, 	\
	  	   a_comm_, n_, CELL_COMM_MTYPE); 			}
# define CELL_RECV(i, j, a, size)					\
	/*receive from cell (i,j) size bytes at address a */		\
	{ int n_ = ((size)+sizeof(FDATA)-1)/sizeof(FDATA);		\
	  RecvMat((NP_I+(i)-CID_I)%NP_I, (NP_J+(j)-CID_J)%NP_J, 1, n_, 	\
	  	   a_comm_, n_, CELL_COMM_MTYPE);			\
	  memcpy((void *) (a), (void *) a_comm_, (size_t) (size));	}


#elif defined(APLIB)
# ifdef APLIB_LINUX	
#   define CELL_INIT		{ ap_init(); ap_use_signals(0); sync(0,0); }
# else		
#   define CELL_INIT
# endif

# include <ccell.c7.h>
# define COMMUN_INIT(Ncely, Ncelx)	GET_PNUM(Ncely, Ncelx)
# ifndef APLIB_LINUX
# define main cell_main
# endif
# define CID     		getcid()
# define CID_I   		getcidy()
# define CID_J   		getcidx()
# define NP			getncel()
# define NP_I    		getncely()
# define NP_J    		getncelx()
# define GET_PNUM(Ncely, Ncelx)	{ (Ncely) = NP_I; (Ncelx) = NP_J; }
# define COMMUN_INIT(Ncely, Ncelx) GET_PNUM(Ncely, Ncelx)	
# define CELL_SEND(i, j, a, size) r_asend(j, i, gettid(), 0, (MSGP) a, size)
# define CELL_RECV(i, j, a, size)					\
	{ r_arecv(j, i, gettid(), ANY_TYPE);	readmsg((MSGP) a, size);}

#else
# define CELL_INIT
# define CID     		0
# define CID_I   		0
# define CID_J   		0
# define NP_I    		1
# define NP_J    		1
# define NP			1
# define GET_PNUM(Ncely, Ncelx)	{ (Ncely) = 1; (Ncelx) = 1; }	
# define CELL_RECV(i, j, a, size)
# define CELL_SEND(i, j, a, size)					
# define COMMUN_INIT(Ny, Nx)	GET_PNUM(Ny, Nx)
#endif


#define EXTERN extern
#include "LogIO.h"			/* defines GetDate, STDOUT_MSGTYPE */


#if defined(HOST)				
					/* modify next 3 dirs. for non APx000*/
# define HostSend(type, a, size) 	/* send size bytes at addr. a to host*/\
	{ h_send(type, a, size); }
# define COMMUN(v)        		/* receive value of var. v from host*/\
	{ h_recv(ANY_TYPE); readmsg(&v, sizeof(v)); }
# define COMMUN_ARR(a, r) 		/*receive r bytes from host to addr. a*/\
	{ h_recv(ANY_TYPE); if (r < 0) r = getmsize(); readmsg(a, r); }

static char HostString[512];
# define InitArgs(argc, argv)						\
	{ int r=-1;  char *argvcpy;					\
	  argv = (char **) malloc(128*sizeof(char *));			\
	  COMMUN_ARR(HostString, r);					\
	  argvcpy = malloc(r);	memcpy(argvcpy, HostString, r);		\
	  Str2Chv(argvcpy, r, argc, argv);				\
	}
# undef GetDate
# define GetDate(Date)							\
	{ int r=-1; COMMUN_ARR(Date, r); Date[r] = '\0'; }

# define GatherSend(v, size) 		CELL_SEND(0, 0, &(v), size)
# define GatherMsg(Ny, Nx, arr, size) 	GatherMessage(Ny, Nx, arr, size)

# define STDOUT 			HostString
# define STDOUT_SETUP(Stdout)
# define HPRINT(FPRINTF_ARGS)						\
	{ if (CID==0) {							\
	    sprintf FPRINTF_ARGS;					\
	    HostSend( STDOUT_MSGTYPE, STDOUT, strlen(STDOUT)+1);	\
	} }
# define LOGFILE_APPEND(LogFile, Msg)					\
	{ if (CID==0) 							\
	   HostSend( LOGFILE_MSGTYPE, Msg, strlen(Msg)+1 );		\
	}
# else

# define HostSend(type, v, size) 
# define COMMUN(v)        
# define COMMUN_ARR(v, r)        

# define InitArgs(argc,argv)	
# if !defined(APLIB)  && !defined (FTN_BLACS)
#   define GatherMsg(Ny, Nx, arr, size)
#   define GatherSend(v, size) 		HostSend(STDOUT_MSGTYPE, &(v), size)

#   define STDOUT 			stdout
#   define STDOUT_SETUP(Stdout)		
# else
#   define GatherSend(v, size) 		CELL_SEND(0, 0, &(v), size)
#   define GatherMsg(Ny, Nx, arr, size) GatherMessage(Ny, Nx, arr, size)
#   ifdef APLIB_LINUX
#     define STDOUT 			stdout
#     define STDOUT_SETUP(Stdout)		
#   else  	
static FILE * stdoutLog;
#     define STDOUT 			stdoutLog 
#     define STDOUT_SETUP(Stdout)		stdoutLog = fopen("stdout.cell", "a")
#   endif
# endif
# define HPRINT(FPRINTF_ARGS)		{if (CID==0) fprintf FPRINTF_ARGS;}
# define LOGFILE_APPEND(LogFile, Msg)					\
	   LogFileAppend(LogFile, Msg)					
#endif /*HOST*/

#endif /*__Cell_Incl__*/
