/* @(#) All Right Reserved, Copyright (c) 1997 Australian National University */
/* @(#)                     Copyright (c) 1997 Fujitsu Laboratories Ltd. */
/*  */
/*        */
/*                        Permission Notice */
/*  */
/*                Australian National University */
/*                Canberra ACT 0200 Australia */
/*                           and */
/*                Fujitsu Laboratories Ltd. */
/*                4-1-1 Kamikodanaka Nakahara */
/*                Kawasaki 211-88 Japan */
/*  */
/*   Permission to use, copy, modify and distribute this software (including  */
/*   its documentation) without fee is hereby granted under the following terms */
/*   and conditions: */
/*  */
/*   1. This permission notice and all of the copyright notices in this */
/*      software must appear in all copies and derived versions of all */
/*      or a part of this software. */
/*  */
/*   2. Sales of this software, whether incorporated in other products as a */
/*      part thereof or not, is prohibited. */
/*  */
/*   3. On publication or presentation in any form of the results obtained */
/*      by use of this software, an explicit reference to this software and */
/*      its copyright owner shall be made. */
/*  */
/*   4. THIS SOFTWARE IS PROVIDED 'AS IS' AND THE AUSTRALIAN NATIONAL UNIVERSITY */
/*      AND FUJITSU LABORATORIES LTD. MAKE NO WARRANTIES, EITHER EXPRESSED OR */
/*      IMPLIED, AS TO ANY MATTER WHATSOEVER REGARDING THIS SOFTWARE,  */
/*      INCLUDING, WITHOUT LIMITATION, WARRANTIES OF MERCHANTABILITY OR FITNESS */
/*      FOR ANY PARTICULAR PURPOSE. */
/*      IN NO EVENT SHALL THE AUSTRALIAN NATIONAL UNIVERSITY OR FUJITSU */
/*      LABORATORIES LTD. BE LIABLE FOR ANY CLAIM FROM A THIRD PARTY, OR */
/*      FOR SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES WHATSOEVER RESULTING */
/*      FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN ACTION OF CONTRACT, */
/*      NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION */
/*      WITH THE USE, COPYING OR MODIFICATION OF THIS SOFTWARE. */
/*  */

/************************************************************************/
/*	FDATAAuxBlkMatrix: FDATA Auxillary Distributed matrix procedures*/
/************************************************************************/
/* SCCS INFO @(#)FDATAAuxBlkMatrix.c 3.3         last modified P. Strazdins, 97/04/08
   first created: P. Strazdins 94/09
*/
/* Auxilliary routines for initializing block-scattered matrice
 * compilation: macro FDATATYPE must be defined
*/
#include "FDATASetUp.h"
#include "FDATAOps.h"
#include "BLASOptions.h"
#include "LocMat.h"
#include "MatDim.h"
#include "DistMat.h"

static void MakeScattered(RandElts, I0, J0, m, n, UnitAdd, Trans, A)
  int RandElts, m, n; FDATA *UnitAdd; char *Trans; DistMat A;
/* post: DM_Loc(A)[i, j] = f(i', j') where i'=I0+L2Gd(i,A.DI), j'=J0+L2Gd(j,A.DJ)
 * 	 	for 0<=i<m, 0<=j<n
 * 	 where  f(i, j) = (RandElts? random(a*i+b*j): a*i+b*j) +
 *			  (i==j? UnitAdd: 0)
 *  		(a,b) = Trans=='S'? (x,x): (Trans==NoTrandpose? (x,y): (y,x))
 *		for some x,y>0.0
 * notes: Trans=='S' is used to generate symmetric matrices.
 */
{ int i, j, iG, jG;
  for (i=0; i < m; i++) {
    iG = I0 + L2Gd(i, DM_D0(A));
    for (j=0; j < n; j++) {
      FDATA v;
      jG = J0 + L2Gd(j, DM_D1(A));
      if (RandElts) {
      	if (*Trans=='S')
      	  srand(1024*MAX(iG,jG)+MIN(iG,jG)); 
	else if (*Trans==NoTranspose)	       /* nb. dangerous to factor out */
    	  srand(3*iG+jG);
  	else
    	  srand(3*jG+iG);
        v = (FD_BASE) (rand() % 10000) * 0.0001 + 0.001;
      } else {
      	if (*Trans=='S')
      	  v = (FD_BASE) (iG+1) * 0.1 + (FD_BASE) (jG+1) * 0.1;
	else if (*Trans==NoTranspose)
          v = (FD_BASE) (iG+1) * 0.1 + (FD_BASE) (jG+1) * 0.001;
        else
          v = (FD_BASE) (iG+1) * 0.001 + (FD_BASE) (jG+1) * 0.1;
      }
      if (iG == jG)
	v += *UnitAdd;
      *LM_Index(DM_Loc(A), i, j) = v;
    } /* for (j...) */
  } /* for (i...) */
} /*MakeScattered*/


/*static*/ int CheckScattered(RandElts, I0, J0, m, n, UnitAdd, Trans, A,
			  MxDiff, AvDiff, MaxI, MaxJ)
  int RandElts, m, n; FDATA *UnitAdd; char *Trans; DistMat A;
  FD_BASE *MxDiff, *AvDiff; int *MaxI, *MaxJ; 
/* check m x n local matrix of matrix A is same as that generated by: 
 *	 MakeScattered(RandElts, I0, J0, m, n, UnitAdd, Trans, B)
 * 	 where DM_Dim(i, B) = DM_Dim(i, A)
 * post: r.v. & setting of MxDiff, AvDiff, MaxI, MaxJ as per
 *	 MatDiff(m, n, DM_LOc(A), DM_Loc(B), MxDiff, AvDiff, MaxI, MaxJ)
 */
{ int i, iG; 			int NErrs = 0; FD_BASE Av, Mx; int MxI, MxJ;
  FDATA *Bx;			DistMat Bi;
  *MxDiff = 0.0; *MaxI = 0; *MaxJ = 0;
  *AvDiff = 0.0;
  if (n <= 0)
    return(NErrs);

  if (DM_Trans(A))	/* ensures iG refers to DM_D0(A) in SubMat(A, iG, 0) */
    A = Trans(A);  
  MatMalloc("CheckScattered", 1, &Bx, n);
  for (i = 0; i < m; i++) {
    DistMat_RVAlloc(1);	
    int nej;	
    iG = I0 + L2Gd(i, DM_D0(A));	
    Bi = SubMat(A, iG, 0);
    LM_Init(DM_Loc(Bi), Bx, n, 1);		/* 1 x n */

    MakeScattered(RandElts, iG, J0, 1, n, UnitAdd, Trans, Bi);
    nej = MatDiff(1, n, DM_Loc(Bi), LocSubMat(DM_Loc(A), i, 0),
		     &Mx, &Av, &MxI, &MxJ); 
    if (nej)
      printf("-%d (%d); ", i, nej);

    NErrs += nej;
    if (Mx > *MxDiff) {
      *MxDiff = Mx; *MaxI = i; *MaxJ = MxJ;
    }
    *AvDiff += Av;
  } /* for (i...) */
  if (m > 0) *AvDiff /= ((FD_BASE) m);
  MatFree(Bx);
  if (NErrs)
    printf("\n");
  return (NErrs);
} /* CheckScattered() */


static void TrPreMultDiag(UpLo, Diag, m, n, alpha, beta, C)
  char *UpLo, *Diag; int m, n; FDATA *alpha, *beta; DistMat C;
/* pre:  UpLoOK(*UpLo)
 *       C <- *alpha * C 
 *	 for proper UpLo triangular part of local m x n matrix of C
 *       if Diag!=Unit, diag(C) = *beta
 */
{ int i; 				pMatDim Di, Dj;
  LocMat lC;
  Di = DM_D0(C);			Dj = DM_D1(C);
  lC = DM_Loc(C);
  if (*UpLo==Upper) {
    SWAP(int, m, n); 			SWAP(pMatDim, Di, Dj);
    SWAP(int, LM_RInc(lC), LM_CInc(lC)); 
  }
  for (i=0; i < m; i++) {
    int iG = L2Gd(i, Di), 	ji = G2Ld(iG, Dj), j;
    if (*Diag==Unit && ji<n && SubMatP0d(iG, Di)==0 && SubMatP0d(iG, Dj)==0) 
      *LM_Index(lC, i, ji) = *beta;
    ji = MIN(ji, n);
    for (j=0; j < ji; j++)
      *LM_Index(lC, i, j)  *= *alpha;
  } /* for (i...) */
} /* TrPreMultDiag() */
 

