/* @(#) All Right Reserved, Copyright (c) 1997 Australian National University */
/* @(#)                     Copyright (c) 1997 Fujitsu Laboratories Ltd. */
/*  */
/*        */
/*                        Permission Notice */
/*  */
/*                Australian National University */
/*                Canberra ACT 0200 Australia */
/*                           and */
/*                Fujitsu Laboratories Ltd. */
/*                4-1-1 Kamikodanaka Nakahara */
/*                Kawasaki 211-88 Japan */
/*  */
/*   Permission to use, copy, modify and distribute this software (including  */
/*   its documentation) without fee is hereby granted under the following terms */
/*   and conditions: */
/*  */
/*   1. This permission notice and all of the copyright notices in this */
/*      software must appear in all copies and derived versions of all */
/*      or a part of this software. */
/*  */
/*   2. Sales of this software, whether incorporated in other products as a */
/*      part thereof or not, is prohibited. */
/*  */
/*   3. On publication or presentation in any form of the results obtained */
/*      by use of this software, an explicit reference to this software and */
/*      its copyright owner shall be made. */
/*  */
/*   4. THIS SOFTWARE IS PROVIDED 'AS IS' AND THE AUSTRALIAN NATIONAL UNIVERSITY */
/*      AND FUJITSU LABORATORIES LTD. MAKE NO WARRANTIES, EITHER EXPRESSED OR */
/*      IMPLIED, AS TO ANY MATTER WHATSOEVER REGARDING THIS SOFTWARE,  */
/*      INCLUDING, WITHOUT LIMITATION, WARRANTIES OF MERCHANTABILITY OR FITNESS */
/*      FOR ANY PARTICULAR PURPOSE. */
/*      IN NO EVENT SHALL THE AUSTRALIAN NATIONAL UNIVERSITY OR FUJITSU */
/*      LABORATORIES LTD. BE LIABLE FOR ANY CLAIM FROM A THIRD PARTY, OR */
/*      FOR SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES WHATSOEVER RESULTING */
/*      FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN ACTION OF CONTRACT, */
/*      NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION */
/*      WITH THE USE, COPYING OR MODIFICATION OF THIS SOFTWARE. */
/*  */

/************************************************************************/
/*                              GMM.Parameters.c                        */
/************************************************************************/
/* SCCS INFO @(#)GMM.Parameters.c 3.2         last modified P. Strazdins, 97/05/21
   first created: P. Strazdins 93/01/21
*/
/* input command line parameters into variables for general distributed
 * matrix multiplication  program performing C := alpha*op(A)*op(B)+beta*C
 *        
 *  with the usage:
 *	PROGNAME
 *  followed optionally by following USAGE string:                      */
#define USAGE \
"[-a [-ULT] [-0 AI0 AJ0] [-C]] [-b [-ULT] [-0 BI0 BJ0] [-C]]   \n\
	[-c [-UL] [-0 CI0 CJ0] [-C]] [-B BM BN BK] [-l lvl] [-2] [-E] [-X rptmult] \n\
	[-p PrintRes] [-M n v] [-t trlvl trtype] [-v AlgVers] [-e ry rx] [-m my mx]\n\
	[-r NReps] [-s StepM StepN StepK] M N K [alpha [beta]]"
/*  with default values:						*/
#define DEFAULTS \
"AI0=AJ0=BI0=BJ0=CI0=CJ0=0; BM=BN=BK=1; alpha=beta=1.0	lvl=2\n\
	NReps=1, StepM=M, StepN=N, StepK=K; rptmult=1\n\
	PrintRes=0 AlgVers=0 trlvl=-1 trtype=0\n\
	ry=2.0, rx=1.0; my = mx = (int) SQRT_FD_MAXMANT/sqrt(K)/3"
/* and constraints:							*/
#define CONTRAINTS						\
"0 <= M,N,K; 0<=AI0,BI0,CI0, 0<=AJ0,BJ0,CJ0; my, mx >= 1	\n\
	-a -UL (A symm.) ==> M=K BM=BK AI0=AJ0 my=mx ry=rx; 	\n\
	-b -UL (B symm.) ==> K=N BN=BK BI0=BJ0 my=mx ry=rx;	\n\
	-c -UL (C symm.) ==> M=N BM=BN CI0=CJ0 my=mx ry=rx"
/* and matrix form options corresponding to the variables:
 *	-a -U	: FormA = Upper		(upper triangular matrix)
 *	-a -L	: FormA = Lower         (lower triangular matrix)
 *	-a -T	: TransA = Transpose
 *		  (defaults: FormA = Rectangular, TransA = NoTranspose) 
 *	and similarly for -b, -c
 *	nb. -U or -L options signify that only that half of matrix be accessed
 * and combined block-cell offset parameters
 *	-a -0 AI0 AJ0
 *	and similarly for -b, -c
 * and local matrix storage option -C (column-major storage; 
 * row major is default).
 * The special option -2 signifies a SYR2K() computation.
 * The -M option may be repeated with differing n values; all these
 * can set the DBLAS machine parameters before the compuation begins.
 * The -t option is used for DBLAS trace level setting.
 * The -v AlgVers option is used to force a version of a DBLAS algorithm,
 * is chosse MM alg that `spreads' across the AlgVers'th matrix dimension. 
 * The -l LvL option can be used to force a level 1 (Lvl=1) or 2 (Lvl=2) 
 * computation, if the other parameters allow this.
 * The -E option can be used to suprress DBLAS parameter checking.
 * Matrices of a global index (i,j) are initialized to ry*(i%my) + rx*(j%mx).
 * PrintRes>=0 is a debug option to print out matrices for debugging.
 */
#include "FDATAOps.h"

#if (FDATATYPE == SINGLE || FDATATYPE == COMPLEX) 
  /* = 2^(# bits in FD_BASE's mantissa / 2); 
     can be used to limit mx, my so that result matrix is representable 
     exactly by FDATA for the given value of K */
# define SQRT_FD_MAXMANT 0x00001000
#else
# define SQRT_FD_MAXMANT 0x00100000
#endif
  double sqrt();

/* command line parameters; see header comment */
  static int M, N, K, Ncely, Ncelx, AI0, AJ0, BI0, BJ0, CI0, CJ0, 
             BM, BN, BK, ColMajA, ColMajB, ColMajC, Level, DoubleMult;
  static int NReps, StepM, StepN, StepK, rptmult;
  static char TransA, FormA, TransB, FormB, FormC;
  static FDATA alpha, beta; 
  static FD_BASE rx, ry; static int mx, my;

#include "FDATAAssert.h"   /* extern int PrintRes, TraceLevel, TraceType, ForceAlgVers*/
#include "MachineParams.h"      /* for -M oprtion */


static void OptionError(), PrintUsage();

static void GetParameters(argc, argv)
  int argc; char *argv[];
/* post: set GMM command line variables according to argv[0..argc-1] */
{ extern char *optarg;
  extern int optind;
  
  int c; int StepOpt=0, ModOpt=0; char MatSpecOpt=' '; 
  /* set up default values */
  AI0 = AJ0 = BI0 = BJ0 = CI0 = CJ0 = 0;
  BM = BN = BK = 1; 
  TransA = TransB = NoTranspose; 
  FormA = FormB = FormC = Rectangular;	ColMajA = ColMajB = ColMajC = 0;
  NReps = 1; rptmult=1; DoubleMult=0;  Level = 3;
  ry=2.0; rx=1.0; 			my = mx = 1;
  FD_SET_REAL(alpha, 1.0);  FD_SET_REAL(beta, 1.0);

  if (argc==0)
    return;

  while ((c = getopt(argc, argv, "abcULT0CB2r:sl:eEmp:X:tMv:")) != -1) 
    /* parse options */
    switch (c) {
      case 'a': case 'b': case 'c':
        MatSpecOpt = c;							break;
      case 'U': case 'L':
        switch (MatSpecOpt) {
          case 'a':
	    FormA = c=='R'? Rectangular: (c=='U'? Upper: Lower);	break;
          case 'b':
	    FormB = c=='R'? Rectangular: (c=='U'? Upper: Lower);	break;
          case 'c':
	    FormC = c=='R'? Rectangular: (c=='U'? Upper: Lower);	break;
	  case ' ':
	    OptionError(c, "not preceeded by matrix specifier option (-abc)");
	}								break;
      case 'T':
        switch (MatSpecOpt) {
          case 'a':
	    TransA = Transpose;						break;
          case 'b':
	    TransB = Transpose;						break;
          case 'c':
	    OptionError(c, "invalid after matrix specifier option -c"); break;
	  case ' ':
	    OptionError(c, "not preceeded by matrix specifier option (-abc)");
	}								break;
      case '0':
        if (optind + 1 < argc) {
          switch (MatSpecOpt) {
            case 'a':
	      AI0 = atoi(argv[optind]);	AJ0 = atoi(argv[optind+1]);	break; 
            case 'b':
	      BI0 = atoi(argv[optind]);	BJ0 = atoi(argv[optind+1]);	break; 
            case 'c':
	      CI0 = atoi(argv[optind]);	CJ0 = atoi(argv[optind+1]);	break; 
	    case ' ':
	      OptionError(c, "not preceeded by matrix specifier option (-abc)");
	    }
	  optind += 2;	
	} else
	  OptionError(c, "I0 or J0 parameters are missing");		break;
      case 'C':
          switch (MatSpecOpt) {
            case 'a':
	      ColMajA = 1; 			break; 
            case 'b':
	      ColMajB = 1;			break; 
            case 'c':
	      ColMajC = 1;			break; 
	    case ' ':
	      OptionError(c, "not preceeded by matrix specifier option (-abc)");
	    }					break;
      case 'B':
        if (optind + 2 < argc) {
	  BM = atoi(argv[optind]);	BN = atoi(argv[optind+1]);
	  BK = atoi(argv[optind+2]);	optind+=3;
        } else
	  OptionError(c, "BM, BN or BK parameters are missing");
							 		break;
      case '2':
	DoubleMult = 1;							break;
      case 'l':
	Level = atoi(optarg);						break;
      case 'r':
	NReps = atoi(optarg);						break;
      case 's':
        if (optind + 2 < argc) {
	  StepM = atoi(argv[optind]); 	StepN = atoi(argv[optind+1]); 
	  StepK = atoi(argv[optind+2]); optind += 3; 
          StepOpt=1;
        } else 
	  OptionError(c, "StepM, StepN or StepK parameters are missing");
								 	break;
      case 'X':
	rptmult = atoi(optarg);						break;
      case 'e':
        if (optind + 1 < argc) {
	  ry = atof(argv[optind]); 	rx = atof(argv[optind+1]); 
	  optind += 2;
        } else 
	  OptionError(c, "ry or rx parameters are missing");
								 	break;
      case 'm':
        if (optind + 1 < argc) {
	  my = atoi(argv[optind]); 	mx = atoi(argv[optind+1]); 
	  optind += 2;			ModOpt = 1;
        } else 
	  OptionError(c, "my or mx parameters are missing");
								 	break;
      case 'E':
	NoOuterErrCheck=1;						break;
      case 't':
        if (optind + 1 < argc) {
	  TraceLevel = atoi(argv[optind]);
	  TraceType = atoi(argv[optind+1]);
	  optind += 2;		
        } else 
	  OptionError(c, "trave level & type parameters are missing");
								 	break;
      case 'v':
	ForceAlgVers = atoi(optarg);					break;
      case 'M':
        if (optind + 1 < argc) {
          MachParamSet(atoi(argv[optind]), atoi(argv[optind+1]));
	  optind += 2;		
        } else 
	  OptionError(c, "n or v parameters are missing");
								 	break;
      case 'p':
	PrintRes = atoi(optarg);					break;
      case '?':
        PrintUsage();
        exit (1);
    } /* switch (c) */

  if (optind+2 >= argc || argc <= 3) { 
    if (argc > 1) 
      HPRINT((STDOUT, "%s: insufficient parameters M N K given.\n", 
	      PROGNAME));
    PrintUsage();
    exit (argc > 1? 2: 0);
  }

  /* get non-optional parameters */
  M = atoi(argv[optind]);		N = atoi(argv[optind+1]);
  K = atoi(argv[optind+2]);
  if (optind+3 < argc) {
    FD_SET_REAL(alpha, atof(argv[optind+3]));
  }
  if (optind+4 < argc) {
    FD_SET_REAL(beta, atof(argv[optind+4]));
  }
  Level = MIN(MAX(Level, 1), 3);
  Level = MAX(Level, (M!=1)+(N!=1)+(K!=1)); 
  if (! StepOpt) {
    StepM = M; StepN = N; StepK = K;
  } 
  if (! ModOpt) {
    if (K==0)
      my = mx = 1;
    else
      my = mx = ((int) SQRT_FD_MAXMANT / (int) (sqrt((double) K)) / 3);
    if (DoubleMult) { my = my / 2; mx = mx / 2;} 
  }
} /* GetParameters() */


static void OptionError(option, msg)
  char *msg, option;
/*post: print out option error message  with msg & option; print usage & exit */
{ HPRINT((STDOUT, "%s: error in parsing option -%c: %s\n", PROGNAME, option, msg));
  PrintUsage();
  exit(2);
}

static void PrintParameters();

static void PrintUsage()
/*post: prints program usage to stdout*/
{ HPRINT((STDOUT, "usage : %s %s\n", PROGNAME, USAGE));
  HPRINT((STDOUT, "where: %s\n", CONTRAINTS));
  HPRINT((STDOUT, "with default values: %s\n", DEFAULTS));
  PrintParameters();
} /* PrintUsage() */


static void CheckParameters()
/* post: check values of parameter variables, aborting if invalid */
{ int A0M=AI0, A0N=AJ0, B0M=BI0, B0N=BJ0;
  if (TransA != NoTranspose) {
     A0M=AJ0; A0N=AI0;
   }
  if (TransB != NoTranspose) {
     B0M=BJ0; B0N=BI0;
  }
  if (M < 0 || N < 0 || K < 0) {
    HPRINT((STDOUT, "%s: invalid (-ve) matrix dimensions (M=%d, N=%d, K=%d)\n",
 	   PROGNAME, M, N, K));
    PrintUsage();
    exit (2);
  }
  if (AI0<0 || AJ0<0) {
    HPRINT((STDOUT, "%s: invalid -a -0 params (AI0=%d,AJ0=%d)\n", PROGNAME, AI0, AJ0));
    PrintUsage();
    exit (2);
  }
  if (BI0<0 || BJ0<0) {
    HPRINT((STDOUT, "%s: invalid -b -0 params (BI0=%d,BJ0=%d)\n", PROGNAME, BI0, BJ0));
    PrintUsage();
    exit (2);
  }
  if (CI0<0 || CJ0<0) {
    HPRINT((STDOUT, "%s: invalid -c -0 params (CI0=%d,CJ0=%d)\n", PROGNAME, CI0, CJ0));
    PrintUsage();
    exit (2);
  }
  if ((BM < 1) || (BN < 1) || (BK < 1)) {
    HPRINT((STDOUT, "%s: block sizes BM=%d, BN=%d, BK=%d must be >= 1\n",
	   PROGNAME, BM, BN, BK));
    PrintUsage();
    exit (2);
  }
  if (my < 1 || mx < 1) {
    HPRINT((STDOUT, "%s: my=%d, mx=%d must both exceed 1\n", PROGNAME, mx, my));
    PrintUsage();
    exit (2);
  }
  if (FormA != Rectangular  &&  (M!=K || BM!=BK || AI0%BM!=AJ0%BK)) {
    HPRINT((STDOUT, "%s: -a -%c means A symmetric but M=%d != K=%d or BM=%d != BK=%d or AI0%%BM=%d != AJ0%%BK=%d\n", 
	   PROGNAME, FormA, M, K, BM, BK, AI0%BM, AJ0%BK));
    PrintUsage();
    exit (2);
  }
  if (FormB != Rectangular  &&  (K!=N || BK!=BN || BI0%BK!=BJ0%BN)) {
    HPRINT((STDOUT, "%s: -b -%c means B symmetric but K=%d != N=%d or BK=%d != BN=%d or BI0%%BK=%d != BJ0%%BN=%d\n", 
	   PROGNAME, FormB, K, N, BK, BN, BI0%BK, BJ0%BN));
    PrintUsage();
    exit (2);
  }
  if ((FormC != Rectangular)  && (M!=N || BM!=BN || CI0%BM!=CJ0%BN)) {
    HPRINT((STDOUT, "%s: -c -%c means C symmetric but M=%d != N=%d or BM=%d != BN=%d or CI0%%BM=%d != CJ0%%BN=%d\n",
	    PROGNAME, FormC, M, N, BM, BN, CI0%BM, CJ0%BN));
    PrintUsage();
    exit (2);
  }
  if (FormA != Rectangular  &&  (rx != ry || mx != my)) {
    HPRINT((STDOUT, "%s: -a -%c means A symmetric but rx!=ry or mx!=my violates this\n",
	   PROGNAME, FormA));
    PrintUsage();
    exit (2);
  }
  if (FormB != Rectangular  && (rx != ry || mx != my)) {
    HPRINT((STDOUT, "%s: -b -%c means B symmetric but rx!=ry or mx!=my violates this\n", 
	   PROGNAME, FormB));
    PrintUsage();
    exit (2);
  }
  if (FormA != Rectangular && (mx < K || my < M) && (AI0 != AJ0)) {
    HPRINT((STDOUT, "%s: Warning: -b -%c means A symmetric but (mx < K || my < M) && (AI0 != AJ0) invalidates checking alg.\n", 
	   PROGNAME, FormA));
  }
  if (FormB != Rectangular && (mx < N || my < K) && (BI0 != BJ0)) {
    HPRINT((STDOUT, "%s: Warning: -b -%c means B symmetric but (mx < N || my < K) && (BI0 != BJ0) may invalidate checking alg.\n",
	   PROGNAME, FormB));
  }
  if ((A0M%BM != CI0%BM) || (A0N%BK != B0M%BK) || (B0N%BN != CJ0%BN)) {
    HPRINT((STDOUT, "%s: warning: block offsets of A, B, C are not consistent across common dimesions\n",
	   PROGNAME));
  }
} /* CheckParameters() */


static void PrintParameters()
/*post: print out current values of variables set by command line to stdout*/
{ HPRINT((STDOUT, "%s: %dx%d distributed level %d global C = %f A%s*B%s%s + %fC where:\n",
	 PROGNAME, Ncely, Ncelx, Level, alpha, (TransA==Transpose)? "^T": "", 
	(TransB==Transpose)? "^T": "", 
	!DoubleMult? "": (TransA==Transpose? " + B^TA ": " + BA^T"), beta)); 
  HPRINT((STDOUT, "	A of form %c, block-cell offset (%d,%d) %s\n", 
	 FormA, AI0, AJ0, ColMajA? "(col-major)": ""));
  HPRINT((STDOUT, "	B of form %c, block-cell offset (%d,%d) %s\n", 
	 FormB, BI0, BJ0, ColMajB? "(col-major)": ""));
  HPRINT((STDOUT, "	C of form %c, block-cell offset (%d,%d) %s\n", 
	 FormC, CI0, CJ0, ColMajC? "(col-major)": ""));
  HPRINT((STDOUT, "	block sizes are BM=%d, BN=%d, BK=%d\n", BM, BN, BK));
  HPRINT((STDOUT, "	initially M=%d, N=%d, K=%d; timings will be averaged over %d repetitions\n",
	 M, N, K, rptmult));
  HPRINT((STDOUT, "Matrices have block-scattered storage with element (i,j) = %f*(i%%%d) + %f(j%%%d)\n",
          ry, my, rx, mx));
} /* PrintParameters() */

