/* @(#) All Right Reserved, Copyright (c) 1997 Australian National University */
/* @(#)                     Copyright (c) 1997 Fujitsu Laboratories Ltd. */
/*  */
/*        */
/*                        Permission Notice */
/*  */
/*                Australian National University */
/*                Canberra ACT 0200 Australia */
/*                           and */
/*                Fujitsu Laboratories Ltd. */
/*                4-1-1 Kamikodanaka Nakahara */
/*                Kawasaki 211-88 Japan */
/*  */
/*   Permission to use, copy, modify and distribute this software (including  */
/*   its documentation) without fee is hereby granted under the following terms */
/*   and conditions: */
/*  */
/*   1. This permission notice and all of the copyright notices in this */
/*      software must appear in all copies and derived versions of all */
/*      or a part of this software. */
/*  */
/*   2. Sales of this software, whether incorporated in other products as a */
/*      part thereof or not, is prohibited. */
/*  */
/*   3. On publication or presentation in any form of the results obtained */
/*      by use of this software, an explicit reference to this software and */
/*      its copyright owner shall be made. */
/*  */
/*   4. THIS SOFTWARE IS PROVIDED 'AS IS' AND THE AUSTRALIAN NATIONAL UNIVERSITY */
/*      AND FUJITSU LABORATORIES LTD. MAKE NO WARRANTIES, EITHER EXPRESSED OR */
/*      IMPLIED, AS TO ANY MATTER WHATSOEVER REGARDING THIS SOFTWARE,  */
/*      INCLUDING, WITHOUT LIMITATION, WARRANTIES OF MERCHANTABILITY OR FITNESS */
/*      FOR ANY PARTICULAR PURPOSE. */
/*      IN NO EVENT SHALL THE AUSTRALIAN NATIONAL UNIVERSITY OR FUJITSU */
/*      LABORATORIES LTD. BE LIABLE FOR ANY CLAIM FROM A THIRD PARTY, OR */
/*      FOR SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES WHATSOEVER RESULTING */
/*      FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN ACTION OF CONTRACT, */
/*      NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION */
/*      WITH THE USE, COPYING OR MODIFICATION OF THIS SOFTWARE. */
/*  */

/************************************************************************/
/*                              HostMsg.c                               */
/************************************************************************/
/* SCCS INFO @(#)HostMsg.c 3.2         last modified P. Strazdins, 96/05/21
   first created: P. Strazdins 95/09
*/
/* host library used for processing error and timing values gathered from cell 
 * programs, and for their display. Here all matrices must be packed in 
 * row-major order, ie A[i,j] = *(A+i*N+j) if A is M x N.
 */


static void PrintCoords(M, N, MxDiffI, MxDiffJ)
/* post: prints MxDiffI,MxDiffJ[0..M-1][0..N-1] in row-major tuple format
 */
int N, M, MxDiffI[], MxDiffJ[];
{ int i, j;
  for (i=0; i < M; i++) {
    for (j=0; j < N; j++)
      HPRINT((STDOUT, "(%d,%d) ", MxDiffI[i*N+j], MxDiffJ[i*N+j]));
    HPRINT((STDOUT, "\n"));
  }
} /* PrintCoords() */


static void PrintIntMat(M, N, A)
int N, M, A[];
/* post: prints A[0..M-1][0..N-1] in row-major format
 */
{ int i, j;
  for (i=0; i < M; i++) {
    for (j=0; j < N; j++)
      HPRINT((STDOUT, " %4d", A[i*N+j]));
    HPRINT((STDOUT, "\n"));
  }
} /* PrintIntMat() */

static int SumIntMat(M, N, A)
  int N, M, A[];
/* post: returns Sum: 0 <= i < M 0 <= j , N A[i,j]
 */
{ int i, j, s=0; 
  for (i=0; i < M; i++) 
    for (j=0; j < N; j++)
      s += A[i*N+j];
  return (s);    
} /* SumIntMat() */
 

static void GatherMessage(Ncely, Ncelx, Array, SizeElts)
int Ncelx, Ncely, SizeElts; char Array[];
/* pre:  SizeElts > 0
 *  post: Array[i*Ncelx+j] = message of size SizeElts received from task tid 
 *        on cell (i,j)
 */
{ int i, j;
  for (i=0; i < Ncely; i++)
    for (j=0; j < Ncelx; j++) {
      CELL_RECV(i, j, &Array[(Ncelx*i+j)*SizeElts], SizeElts);
    }
} /* GatherMessage() */


static void PrintMatrix(Name, Array, M, N, Ld)
  char Name; int M, N, Ld; FD_BASE Array[];
{ int i, j;
  for (i=0; i < M; i++) {
    HPRINT((STDOUT, "%c ", Name));	
    for (j=0; j < N; j++)
      HPRINT((STDOUT, "%2.3f ", Array[i*Ld+j]));
    HPRINT((STDOUT, "\n")); 
    Name = ' ';
  }
} /* PrintMatrix() */


static FD_BASE MaxMatrix(Array, Ncely, Ncelx)  
  int Ncelx, Ncely; FD_BASE Array[];
{ int i; FD_BASE Mx=Array[0];
  for (i=1; i < Ncelx*Ncely; i++) 
    if (Array[i] > Mx)
         Mx = Array[i];
  return(Mx);  
} /* MaxMatrix() */


static FD_BASE SumMatrix(Array, Ncely, Ncelx)  
  int Ncelx, Ncely; FD_BASE Array[];
{ int i; FD_BASE Sum=0.0;
  for (i=0; i < Ncelx*Ncely; i++) 
    Sum += Array[i];
  return(Sum);  
} /* SumMatrix() */
