/* @(#) All Right Reserved, Copyright (c) 1997 Australian National University */
/* @(#)                     Copyright (c) 1997 Fujitsu Laboratories Ltd. */
/*  */
/*        */
/*                        Permission Notice */
/*  */
/*                Australian National University */
/*                Canberra ACT 0200 Australia */
/*                           and */
/*                Fujitsu Laboratories Ltd. */
/*                4-1-1 Kamikodanaka Nakahara */
/*                Kawasaki 211-88 Japan */
/*  */
/*   Permission to use, copy, modify and distribute this software (including  */
/*   its documentation) without fee is hereby granted under the following terms */
/*   and conditions: */
/*  */
/*   1. This permission notice and all of the copyright notices in this */
/*      software must appear in all copies and derived versions of all */
/*      or a part of this software. */
/*  */
/*   2. Sales of this software, whether incorporated in other products as a */
/*      part thereof or not, is prohibited. */
/*  */
/*   3. On publication or presentation in any form of the results obtained */
/*      by use of this software, an explicit reference to this software and */
/*      its copyright owner shall be made. */
/*  */
/*   4. THIS SOFTWARE IS PROVIDED 'AS IS' AND THE AUSTRALIAN NATIONAL UNIVERSITY */
/*      AND FUJITSU LABORATORIES LTD. MAKE NO WARRANTIES, EITHER EXPRESSED OR */
/*      IMPLIED, AS TO ANY MATTER WHATSOEVER REGARDING THIS SOFTWARE,  */
/*      INCLUDING, WITHOUT LIMITATION, WARRANTIES OF MERCHANTABILITY OR FITNESS */
/*      FOR ANY PARTICULAR PURPOSE. */
/*      IN NO EVENT SHALL THE AUSTRALIAN NATIONAL UNIVERSITY OR FUJITSU */
/*      LABORATORIES LTD. BE LIABLE FOR ANY CLAIM FROM A THIRD PARTY, OR */
/*      FOR SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES WHATSOEVER RESULTING */
/*      FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN ACTION OF CONTRACT, */
/*      NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION */
/*      WITH THE USE, COPYING OR MODIFICATION OF THIS SOFTWARE. */
/*  */

/************************************************************************/
/*                              LogIO                                   */
/************************************************************************/
/* SCCS INFO @(#)LogIO.h 1.2         last modified P. Strazdins, 97/05/19
 * first created: P. Strazdins 93/12/17
 */
/* general macros for string and file manipulation;
   used for IO communication for cell and host programs
*/

#ifndef __LogIO_Incl__
#define __LogIO_Incl__

#include <stdio.h>
#include <string.h>
#define APPEND_FILE "a"

#include <time.h>

#define GetDate(Date)                                                   \
        { time_t t = time(NULL);         char * date = ctime (&t);      \
          strcpy(Date, date);                                           }

					/* types for cell -> host messages */
#define STDOUT_MSGTYPE          0	/* to be printed on stdout */
#define LOGFILE_MSGTYPE         7	/* to be printed on log file */

#define GetLogFileName(CellProgName, Date, LogFileName)			\
	/* form a name of a log file out of the cell program name and date */\
	{ char month[4], day[4];   					\
	  sscanf(Date, "%*s %s %s", month, day);			\
	  sprintf(LogFileName, "Log%s.%s%s", CellProgName, month, day);	\
	}

#define LogFileAppend(LogFileName, str)					\
	/* append str on to the file of name LogFileName */		\
	{ FILE *logfile;						\
  	  logfile = fopen (LogFileName, APPEND_FILE);			\
  	  if (logfile != NULL) {					\
    	    fputs(str, logfile);					\
    	    fclose(logfile);						\
  	  }								\
	}								

#define Chv2Str(chv, i1, i2, sep, str, len)				\
	/* str[0..len-1] = concat. of chv[i1..i2-1], using separator sep*/\
	{ int i_; (len) = 0; 						\
	  for (i_ = (i1); i_ < (i2); i_++) {				\
	    int li = strlen((chv)[i_]);					\
	    strcpy(&(str)[len], (chv)[i_]);				\
	    (str)[(len)+li] = (sep);					\
	    (len) += li+1;						\
	  } 								\
	  (str)[len] = '\0';						\
	}
	     	
#define Str2Chv(str, len, chc, chv)					\
	/* chv[0..chc-1] points to the chc strings in str[0..len-1] */	\
	{ int r_ = (len);		char *s_ = (str);		\
	  (chc) = 0;							\
	  while (r_ > 0) { 						\
	    (chv)[chc] = s_;		(chc)++;			\
	    r_ -= strlen(s_) + 1;	s_ += strlen(s_) + 1;		\
	  }								\
	}

#endif /*__LogIO_Incl__*/
