/* @(#) All Right Reserved, Copyright (c) 1997 Australian National University */
/* @(#)                     Copyright (c) 1997 Fujitsu Laboratories Ltd. */
/*  */
/*        */
/*                        Permission Notice */
/*  */
/*                Australian National University */
/*                Canberra ACT 0200 Australia */
/*                           and */
/*                Fujitsu Laboratories Ltd. */
/*                4-1-1 Kamikodanaka Nakahara */
/*                Kawasaki 211-88 Japan */
/*  */
/*   Permission to use, copy, modify and distribute this software (including  */
/*   its documentation) without fee is hereby granted under the following terms */
/*   and conditions: */
/*  */
/*   1. This permission notice and all of the copyright notices in this */
/*      software must appear in all copies and derived versions of all */
/*      or a part of this software. */
/*  */
/*   2. Sales of this software, whether incorporated in other products as a */
/*      part thereof or not, is prohibited. */
/*  */
/*   3. On publication or presentation in any form of the results obtained */
/*      by use of this software, an explicit reference to this software and */
/*      its copyright owner shall be made. */
/*  */
/*   4. THIS SOFTWARE IS PROVIDED 'AS IS' AND THE AUSTRALIAN NATIONAL UNIVERSITY */
/*      AND FUJITSU LABORATORIES LTD. MAKE NO WARRANTIES, EITHER EXPRESSED OR */
/*      IMPLIED, AS TO ANY MATTER WHATSOEVER REGARDING THIS SOFTWARE,  */
/*      INCLUDING, WITHOUT LIMITATION, WARRANTIES OF MERCHANTABILITY OR FITNESS */
/*      FOR ANY PARTICULAR PURPOSE. */
/*      IN NO EVENT SHALL THE AUSTRALIAN NATIONAL UNIVERSITY OR FUJITSU */
/*      LABORATORIES LTD. BE LIABLE FOR ANY CLAIM FROM A THIRD PARTY, OR */
/*      FOR SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES WHATSOEVER RESULTING */
/*      FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN ACTION OF CONTRACT, */
/*      NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION */
/*      WITH THE USE, COPYING OR MODIFICATION OF THIS SOFTWARE. */
/*  */

/************************************************************************/
/*                              TRSM.Parameters.c                       */
/************************************************************************/
/* SCCS INFO @(#)TRSM.Parameters.c 3.2         last modified P. Strazdins, 97/04/08
   first created: P. Strazdins 94/09
*/
/* command line parameters for testing Ncely x Ncelx Distributed 
 * triangular Matrix Update
 * usage:
 *       PROGNAME         # print out options, possible values & defaults
 *       PROGNAME USAGE
 */
#define USAGE \
"[-f FN] [-S Side] [-U UpLo] [-T TransA] [-D Diag] [-B BM BN BA]\n\
	[-0 AI0 AJ0 BI0 BJ0] [-l LdB] [-o MemOffs] [-r NReps] [-s StepM StepN]\n\
	[-M n v] [-t trlvl trtype] [-v AlgVers] [-c AlgVersCheck] [-L LvL ] [-p PrintRes] [-e EMx] M [N [alpha]] \n\
with default values: \n\
	M=N, AI0=AJ0=BI0=BJ0=0; alpha = 1.0, BM=BN=BA=1, \n\
	FN='S'('M'), Side='L'('R'), UpLo='U'('L'), Diag='U'('N'), TransA='N'('T')\n\
	AlgVers=AlgVersCheck=0 trlvl=-1 trtype=0 NReps=1 StepM=M StepN=N, PrintRes=MemOffs=0"

/* The -M option may be repeated with differing n values; all these  
 * can set the DBLAS machine parameters before the compuation begins.
 * The -t option is used for DBLAS trace level setting.   
 * The -v AlgVers option is used to force a version of a DBLAS algorithm,
 * is chose an algorithm that `spreads' across the AlgVers'th matrix dimension.
 * The -c option, if specified, uses an independent BLAS function
 * to check the result; for TRSM, there are accuracy problems for M > 1000.
 * Otherwise, for TRMM (TRSM) I is the initial (final) value of B;
 * this form of checking is thus not as strong as that with the -c option.
 * Setting EMx > 0  will make DBLAS flag an error
 * if received message buffers contain data of greater absolute value;
 * this is only useful if, given A[i,j] in [0,1), all elements of
 * partially computed B can be within a given absolute value.
 */

#include "BLASOptions.h"
#include "FDATAOps.h"

/* command line parameters; see header comment*/
  int M, N, Ncely, Ncelx;
  int AI0=0, AJ0=0, BI0=0, BJ0=0, BM=1, BN=1, BA=1, NReps=1, StepM=0, StepN=0;
  int MemOffs=0, LdBp=0;
  char FN='S', Side=Left, UpLo=Upper, Diag=Unit, TransA=NoTranspose; 
  FDATA alpha; 
  int Level=3, CheckViaBLAS=0, AlgVersCheck=0;;
  
#include "FDATAAssert.h"   /* extern int PrintRes, TraceLevel, TraceType */
#include "MachineParams.h"      /* for -M oprtion */


void PrintUsage()
/* post: prints program usage to stdout	*/
{ HPRINT((STDOUT, "usage: %s %s\n", PROGNAME, USAGE));
}

void OptionError(option, msg)
  char *msg, option;
/* post: print out option error message  with msg & option; print usage & exit*/
{ HPRINT((STDOUT, "%s: error in parsing option -%c: %s\n", PROGNAME, option, msg));
  PrintUsage();
  exit(1);
}

void GetParameters(argc, argv)
  int argc; char *argv[];
/* post: set TRSM command line variables according to argv[0..argc-1] */
{ extern char *optarg;
  extern int optind;
  int c; 
  LdBp = 0;
  while ((c = getopt(argc, argv, "f:S:U:T:D:l:B0w:r:sL:o:p:tMv:e:c")) != -1) 
    /* parse options */
    switch (c) {
      case 'f':
	FN=optarg[0];					break;
      case 'S':
	Side=optarg[0];					break;
      case 'U':
	UpLo=optarg[0];					break;
      case 'T':
	TransA=optarg[0];				break;
      case 'D':
	Diag=optarg[0];					break;
      case 'l':
	LdBp=atoi(optarg);				break;
      case 'r':
	NReps=atoi(optarg);				break;
      case 'o':
	MemOffs=atoi(optarg);				break;       
      case 'L':
	Level=atoi(optarg);				break;       
      case 'B':
        if (optind + 2 < argc) {
          BM = atoi(argv[optind]);   BN = atoi(argv[optind+1]);
	  BA = atoi(argv[optind+2]);
          optind += 3; 
        } else 
          OptionError(c, "BM, BN, BA parameters are missing");
                                                                        break;
      case '0':
        if (optind + 3 < argc) {
          AI0 = atoi(argv[optind]);   AJ0 = atoi(argv[optind+1]);
	  BI0 = atoi(argv[optind+2]); BJ0 = atoi(argv[optind+3]);
          optind += 4; 
        } else 
          OptionError(c, "AI0, AJO, BI0, BJ0 parameters are missing");
                                                                        break;
      case 's':
        if (optind + 1 < argc) {
          StepM = atoi(argv[optind]);   StepN = atoi(argv[optind+1]);
          optind += 2;  
        } else 
          OptionError(c, "StepM, StepN parameters are missing");
                                                                        break;
      case 't':
        if (optind + 1 < argc) {
          TraceLevel = atoi(argv[optind]);
          TraceType = atoi(argv[optind+1]);
          optind += 2;
        } else
          OptionError(c, "trave level & type parameters are missing");
                                                                        break;
      case 'v':
        ForceAlgVers = atoi(optarg);                                    break;
      case 'M':
        if (optind + 1 < argc) {
          MachParamSet(atoi(argv[optind]), atoi(argv[optind+1]));
          optind += 2;
        } else
          OptionError(c, "n or v parameters are missing");
                                                                        break;
      case 'p':  
        PrintRes = atoi(optarg);                                        break;
      case 'e': 
        MsgMaxAbsVal = atof(optarg);
      case 'c': 
	CheckViaBLAS=1;
                                                                        break;
      case '?':
        OptionError(c, "invalid option given");                         break;
    }
        
  if (optind >= argc || argc <= 3) {
    if (argc > 1)
      HPRINT((STDOUT, "%s: insufficient parameters M ... given.\n", PROGNAME));
    PrintUsage();
    exit(1);  
  }
          
  M = atoi(argv[optind]);
  if (argc > optind+1) N=atoi(argv[optind+1]); else N=M;
  if (argc > optind+2) {FD_SET_REAL(alpha, (FD_BASE) atof(argv[optind+2]))}
                   else {FD_SET_REAL(alpha, 1.0)} 

  LdBp = MAX(LdBp, N);
  Level = MIN(MAX(Level, 2), 3);
  Level = MAX(Level, (M!=1)+(N!=1)+((Side==Left? M: N)!=1));
} /* GetParameters() */


void PrintParameters()
{ int  BMA = (Side==Left)? BM: BA,  BNA = (Side==Left)? BA: BN;
  if (TransA==Transpose) {
    int t=BMA; BMA=BNA; BNA=t;
  }
  HPRINT((STDOUT, "%dx%d config level %d B = %1.1f %sA%s%s%s, B is %dx%d with %dx%d blocks & offset (%d,%d) ld=%d;\n",
	 Ncely, Ncelx, Level, alpha, (Side==Left)? "": "B",
         (TransA==Transpose)? "^T": "",(FN=='S')? "^-1": "", 
         (Side==Left)? "B": "", M, N, BM, BN, BI0, BJ0, LdBp));
  HPRINT((STDOUT, "A is %c%c with %dx%d blocks & offset (%d,%d). mem offs of %d\n",
	  UpLo, Diag, BMA, BNA, AI0, AJ0, MemOffs));	
  HPRINT((STDOUT, "%d repetitions with step of %d,%d; alg vers=%d (check via %s)\n",
		   NReps, StepM, StepN, ForceAlgVers, 
		   CheckViaBLAS? (FN=='S'? "TRMM": "GEMM"): "B=I"));
} /* PrintParameters() */


void CheckParameters()
/* post: check values of parameter variables, aborting if invalid */
{  
  if (Ncely<=0 || Ncely<=0) {
    HPRINT((STDOUT, "%s: invalid cell configuration values (Ncely=%d, Ncelx=%d)\n",
           PROGNAME, Ncely, Ncelx));
    PrintUsage();
    exit(2);
  }
  if (M < 0 || N < 0) {
    HPRINT((STDOUT, "%s: invalid (-ve) matrix dimensions (M=%d, N=%d)\n",
           PROGNAME, M, N));
    PrintUsage();
    exit(2);
  }
  if (AI0<0 || AJ0<0) {
    HPRINT((STDOUT, "%s: invalid offsets to A (AI0=%d,AJ0=%d)\n", PROGNAME, AI0, AJ0));
    PrintUsage();
    exit(2);
  }
  if (BI0<0 || BJ0<0) {
    HPRINT((STDOUT, "%s: invalid offsets to B (BI0=%d,BJ0=%d)\n", PROGNAME, BI0, BJ0));
    PrintUsage();
    exit(2);
  }
  if (BM < 1 || BN < 1 || BA < 1) {
    HPRINT((STDOUT, "%s: invalid block sizes for B (BM=%d, BN=%d, BA=%d)\n", PROGNAME, BM, BN, BA));
    PrintUsage();
    exit(2);
  }
  
  if (!SideOK(Side)) {
    HPRINT((STDOUT, "%s: invalid Side specifier (Side=%c)\n", PROGNAME, Side));
    PrintUsage();
    exit(2);
  }
  if (!UpLoOK(UpLo)) {
    HPRINT((STDOUT, "%s: invalid UpLo specifier (UpLo=%c)\n", PROGNAME, UpLo));
    PrintUsage();
    exit(2);
  }
  if (!DiagOK(Diag)) {
    HPRINT((STDOUT, "%s: invalid Unit specifier (Unit=%c)\n", PROGNAME, Diag));
    PrintUsage();
    exit(2);
  }
  if (!TransOK(TransA)) {
    HPRINT((STDOUT, "%s: invalid TransA specifier (TransA=%c)\n", PROGNAME, TransA));
    PrintUsage();
    exit(2);
  }
} /* CheckParameters() */

