/* @(#) All Right Reserved, Copyright (c) 1997 Australian National University */
/* @(#)                     Copyright (c) 1997 Fujitsu Laboratories Ltd. */
/*  */
/*        */
/*                        Permission Notice */
/*  */
/*                Australian National University */
/*                Canberra ACT 0200 Australia */
/*                           and */
/*                Fujitsu Laboratories Ltd. */
/*                4-1-1 Kamikodanaka Nakahara */
/*                Kawasaki 211-88 Japan */
/*  */
/*   Permission to use, copy, modify and distribute this software (including  */
/*   its documentation) without fee is hereby granted under the following terms */
/*   and conditions: */
/*  */
/*   1. This permission notice and all of the copyright notices in this */
/*      software must appear in all copies and derived versions of all */
/*      or a part of this software. */
/*  */
/*   2. Sales of this software, whether incorporated in other products as a */
/*      part thereof or not, is prohibited. */
/*  */
/*   3. On publication or presentation in any form of the results obtained */
/*      by use of this software, an explicit reference to this software and */
/*      its copyright owner shall be made. */
/*  */
/*   4. THIS SOFTWARE IS PROVIDED 'AS IS' AND THE AUSTRALIAN NATIONAL UNIVERSITY */
/*      AND FUJITSU LABORATORIES LTD. MAKE NO WARRANTIES, EITHER EXPRESSED OR */
/*      IMPLIED, AS TO ANY MATTER WHATSOEVER REGARDING THIS SOFTWARE,  */
/*      INCLUDING, WITHOUT LIMITATION, WARRANTIES OF MERCHANTABILITY OR FITNESS */
/*      FOR ANY PARTICULAR PURPOSE. */
/*      IN NO EVENT SHALL THE AUSTRALIAN NATIONAL UNIVERSITY OR FUJITSU */
/*      LABORATORIES LTD. BE LIABLE FOR ANY CLAIM FROM A THIRD PARTY, OR */
/*      FOR SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES WHATSOEVER RESULTING */
/*      FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN ACTION OF CONTRACT, */
/*      NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION */
/*      WITH THE USE, COPYING OR MODIFICATION OF THIS SOFTWARE. */
/*  */

/************************************************************************/
/*                              TRSM.cell.c                             */
/************************************************************************/
/* SCCS INFO @(#)TRSM.cell.c 3.7         last modified P. Strazdins, 97/05/21
   first created: P. Strazdins 94/09
*/
/* cell program for testing Ncely x Ncelx Distributed Matrix Triangular
 * Inverse Matrix Update, per:
 *      D_TR?M(Side, UpLo, Diag, TransA, m, n, alpha, A, LdA, B, LdB,
 *             AI0, AJ0, BI0, BJ0, BM, BN, BA)
 * for i=0..NReps-1; where
 *      n = N+i*StepN; m = M+i*StepM
 *      LdA = (*Side==Left? m: n); LdB = n;
 *      initially B[i,j]=A[i,j] for 0<=i,j<=LdA for FN='S' (TRSM)
 *      initially B[i,j]=(i=j) for 0<=i,j<=LdA for FN='M'  (TRMM)
 * see  TRSM.Parameters.c for how these variables can be set from the command line
 * (externally-defined) cpp FLAGs:
 *      FDATATYPE       SINGLE (default), DOUBLE
 */

#include "Cell.h"
#include "Timer.c"
 
#include "FDATASetUp.h"
#include "FDATAOps.h"
#include "BLASOptions.h"
#include "LocMat.h"
#include "MatDim.h"
#include "DistMat.h"
#include "FDATAMalloc.h"
#include "FDATACellMatrix.h"
#include "FDATACommun.h"
#include "D_BLAS.h"

#include "HostMsg.c"
#include "FDATAAuxMatrix.c"
#include "FDATAAuxBlkMatrix.c"
#define PROGNAME "TRSM.cell"
#include "TRSM.Parameters.c"

/**************************************** main program ******************/
#define RandElts(M,N)			(M+N>=100)

void main (argc, argv)
  int argc; char *argv[];
{ DistMat_RVAlloc(3);
  FDATA *WS; DistMat A, B, B1;		/* B1 is expected result*/
  int MaxN, MaxM;               	/* maximum values of N, M */
  int Na;				/* = M or N; A is Na x Na */
  int BMA, BNA;				/* A has BMAxBNA blocks; 1 of BM,BN,BK*/
  int WSSize;				/* maximum workspace size*/
  int TrA;

  double flops, speed, tcomp;
  char Date[64], LogFileName[64], ResultMsg[256];
 
  FD_BASE DAv=0.0, DMax=0.0;            /* av. & max. differences */
  int NErr=0, DMaxI=0, DMaxJ=0;         /* # errors; position of max. differences*/
  int i;

  CELL_INIT;
  STDOUT_SETUP(STDOUT);
  InitArgs(argc, argv);
  GetDate(Date);
  
  COMMUN_INIT(Ncely, Ncelx);
  HPRINT((STDOUT, "%s", Date));
  GetLogFileName(argv[0], Date, LogFileName);
  sprintf(ResultMsg, "\n%.20s # %dx%d ", Date, Ncely, Ncelx);
  LOGFILE_APPEND(LogFileName, ResultMsg);
  Chv2Str(argv, 0, argc, ' ', ResultMsg, i/*dummy*/);
  printf("%s cell (%d,%d) of %dx%d: %s\n",
          Date, CID_I, CID_J, NP_I, NP_J, ResultMsg);
  LOGFILE_APPEND(LogFileName, ResultMsg);
  
  GetParameters(argc, argv);
  CheckParameters();
  PrintParameters();
                                            
  BMA = (Side==Left)? BM: BA; BNA = (Side==Left)? BA: BN;
  if (TransA==Transpose) {
    int t=BNA; BNA=BMA; BMA=t;
  }

  A = InitDimMat(0, BMA, BNA, CID_I, Ncely, CID_J, Ncelx, AI0, AJ0);
  B = InitDimMat(0, BM, BN, CID_I, Ncely, CID_J, Ncelx, BI0, BJ0);

  /* scale global matrix sizes to local ones */
  MaxM = M + (NReps - 1) * ((StepM>=0)? StepM: -StepM);
  MaxN = N + (NReps - 1) * ((StepN>=0)? StepN: -StepN);
  Na = (Side==Left)? MaxM: MaxN;
  MaxM = G2Ld(MaxM, DM_D0(B)); 		MaxN = G2Ld(MaxN, DM_D1(B));
  LdBp = G2Ld(LdBp, DM_D1(B));		LdBp = MAX(LdBp, 1);
  if (LdBp > MaxN) MaxN = LdBp;

  WSSize = G2Ld(Na, DM_D0(A)) * G2Ld(Na, DM_D1(A)); /*A*/
  if (MaxM*MaxN > WSSize) 
    WSSize = 2 * MaxM*MaxN;		/* A of same size B; need for checking*/
  else
    WSSize += MaxM*MaxN; 		/* B */

  MatMalloc("TRSM.cell", 1, &WS, WSSize+MemOffs);  

  for (i=0; i < NReps; i++) {
    char UpLoTr = (TransA==NoTranspose? UpLo: (UpLo==Upper? Lower: Upper));
    char RevUpLoTr = (UpLoTr==Upper? Lower: Upper);
    int Na = Side==Left? M: N;
    int ma = G2Ld(Na, DM_D0(A)),	na = G2Ld(Na, DM_D1(A)),
        mb = G2Ld(M, DM_D0(B)), 	nb = G2Ld(N, DM_D1(B));
    int LdA = MAX(na, 1), 		LdB = MAX(nb, LdBp);
    int TrB = Side!=Left;
    int TrA = (TransA==Transpose)^TrB;
    char UpLoTrA = !TrA? UpLo: (UpLo==Upper? Lower: Upper);

    LM_Init(DM_Loc(A), WS+MemOffs+mb*LdB, LdA, 1);
    LM_Init(DM_Loc(B), WS+MemOffs, LdB, LdB==1? mb: 1);	

    MakeScattered(RandElts(M,N), 0, 0, ma, na, &FD_1, &TransA, A);
    if (CheckViaBLAS) {
      MakeScattered(RandElts(M,N), 0, 0, mb, nb, &FD_1, "NoTrans", B);
      if (FN=='S')
        D_TRMM(&UpLoTrA, &Diag, !TrB? M: N, !TrB? N: M, alpha,
               !TrA? A: Trans(A), !TrB? B: Trans(B));	
    } else {
      if (FN != 'S')                    /*B=I will update to op(TransA, A)*/
        MatZero(mb, nb, DM_pLoc(B));
      else { 				/* B = A */
        int m0 = (Side==Right && M > N)? G2Ld(N, DM_D0(B)): mb, 
	    n0 = (Side==Left  && N > M)? G2Ld(M, DM_D1(B)): nb;
        LocMat lBm;
        MakeScattered(RandElts(M,N), 0, 0, mb, nb, &FD_1, "NoTrans", B);
        MatZero(mb-m0, nb, pLocSubMat_(&lBm, DM_pLoc(B), m0, 0));
        MatZero(mb, nb-n0, pLocSubMat_(&lBm, DM_pLoc(B), 0, n0));
      }
      TrPreMultDiag(&RevUpLoTr, FN!='S'?"Unit":&Diag, mb, nb, &FD_0, &FD_1, B);
    }
    MatPrint("A", MIN(ma, PrintRes),  MIN(na, PrintRes), DM_Loc(A));
    MatPrint("B", MIN(mb, PrintRes),  MIN(nb, PrintRes), DM_Loc(B));
      
    start_timer(); 
    if (Level==2 && alpha==1.0)
      (FN=='S'? D_TRSV: D_TRMV)(&UpLoTrA, &Diag, !TrB? M: N, 
                                !TrA? A: Trans(A), !TrB? B: Trans(B));
    else
      (FN=='S'? D_TRSM: D_TRMM)(&UpLoTrA, &Diag, !TrB? M: N, !TrB? N: M, alpha,
			        !TrA? A: Trans(A), !TrB? B: Trans(B));
    tcomp = end_timer(1);

    flops = 1.0*M*N*(Na-(Diag==Unit)+1-FD_EQ(alpha,FD_1));
    speed = flops / (1.0e6*tcomp);
    HPRINT((STDOUT, "Computed %dx%d tri update (%.0f FLOPs) %Gs at %.2f MFLOPs (%.3f/cell)\n",
            M, N, flops, tcomp, speed, speed/Ncelx/Ncely));
    sprintf(ResultMsg, " #s: %.3f %.3f", tcomp, speed/Ncelx/Ncely);
    LOGFILE_APPEND(LogFileName, ResultMsg);

    B1 = SubMat(B, 0, 0);		/* use same memory as A */
    LM_ADR(DM_Loc(B1)) = LM_ADR(DM_Loc(A));
    if (CheckViaBLAS) {
      if (FN!='S') {			/* make A into equiv rectangular matrix 
					   and check via DGEMM() */
        char RevUpLo = UpLo==Upper? Lower: Upper;
	DistMat_RVAlloc(2); 		
	DistMat B0 = SubMat(B, 0, 0),	C = SubMat(B, 0, 0);	
	MatMalloc("TRSM.cell", 1, &LM_ADR(DM_Loc(C)), 2*mb*LdB);
	LM_ADR(DM_Loc(B0)) = LM_ADR(DM_Loc(C)) + mb*LdB;
        MakeScattered(RandElts(M,N), 0, 0, mb, nb, &FD_1, "NoTrans", B0);
        MatPrint("B0", MIN(mb, PrintRes),  MIN(nb, PrintRes), DM_Loc(B0));
        TrPreMultDiag(&RevUpLo, &Diag, ma, na, &FD_0, &FD_1, A);
        MatPrint("A", MIN(ma, PrintRes),  MIN(na, PrintRes), DM_Loc(A));
	if (!TrB)
	  D_GEMM(M, N, M, alpha, !TrA? A: Trans(A), B0, 0.0, C);
	else
	  D_GEMM(M, N, N, alpha, B0, TrA? A: Trans(A), 0.0, C);
	MatCopy(mb, nb, DM_pLoc(C), NULL, DM_pLoc(B1));
	MatFree(LM_ADR(DM_Loc(C)));
      } else {   
#if 0
        MatPrint("B", MIN(mb, PrintRes),  MIN(nb, PrintRes), DM_Loc(B));
        (FN!='S'? D_TRSM: D_TRMM)(&UpLoTrA, &Diag, !TrB? M: N, !TrB? N: M, alpha,
			        !TrA? A: Trans(A), !TrB? B: Trans(B));      
#endif
        MakeScattered(RandElts(M,N), 0, 0, mb, nb, &FD_1, "NoTrans", B1);
      }
    } else {
      if (FN=='S') {                    /*expecting alpha I*/
        MatZero(mb, nb, DM_pLoc(B1));
        TrPreMultDiag(&UpLoTr, "Unit", mb, nb, &FD_0, &alpha, B1);
      } else {				/* will update to alpha op(Trans, A)*/
        int m0 = (Side==Right && M > N)? G2Ld(N, DM_D0(B)): mb, 
	    n0 = (Side==Left  && N > M)? G2Ld(M, DM_D1(B)): nb;
        LocMat lBm;
        MakeScattered(RandElts(M,N), 0, 0, mb, nb, &FD_1, "NoTrans", B1);
        MatZero(mb-m0, nb, pLocSubMat_(&lBm, DM_pLoc(B1), m0, 0));
        MatZero(mb, nb-n0, pLocSubMat_(&lBm, DM_pLoc(B1), 0, n0));
        TrPreMultDiag(&RevUpLoTr, &Diag, mb, nb, &FD_0, &FD_1, B1);
        MatScale(mb, nb, &alpha, DM_pLoc(B1));
      }    
    }
    MatPrint("B", MIN(mb, PrintRes),  MIN(nb, PrintRes), DM_Loc(B));
    MatPrint("B1", MIN(mb, PrintRes),  MIN(nb, PrintRes), DM_Loc(B1));

    NErr = MatDiff(mb, nb, DM_Loc(B), DM_Loc(B1), &DMax, &DAv, &DMaxI, &DMaxJ);
#define E_MAX *LM_Index(DM_Loc(B), DMaxI, DMaxJ)
    if (mb>0 && nb>0)
      printf("%d errs; Av: %f; Max Error: B[%d,%d] = %f; |error| = %f\n",
             NErr, DAv, DMaxI, DMaxJ, M*N>0? E_MAX: 0.0, DMax);
#undef E_MAX
      
    GatherSend(NErr, sizeof(int));
    GatherSend(DMax, sizeof(FD_BASE));
    GatherSend(DAv,  sizeof(FD_BASE));
    GatherSend(DMaxI, sizeof(int));
    GatherSend(DMaxJ, sizeof(int));
      
    if (CID==0) {                       /* accumulate error results */
      int *NErrs, *MxDiffIs, *MxDiffJs;         FD_BASE *MxDiffs, *AvDiffs;  
      NErrs = (int *) malloc(3*Ncely*Ncelx*sizeof(int));
      MxDiffIs = NErrs + Ncely*Ncelx;   MxDiffJs = MxDiffIs + Ncely*Ncelx;
      MxDiffs = (FD_BASE *) memalign(sizeof(FD_BASE), 2*Ncely*Ncelx*sizeof(FD_BASE));
      AvDiffs = MxDiffs + Ncely*Ncelx;
      
      NErrs[0] = NErr;                  /* needed for non-parallel programs */
      MxDiffs[0] = DMax;                AvDiffs[0] = DAv;
      MxDiffIs[0] = DMaxI;              MxDiffJs[0] = DMaxJ;
      GatherMsg(Ncely, Ncelx,  NErrs, sizeof(int));
      NErr = SumIntMat(Ncely, Ncelx, NErrs);
      GatherMsg(Ncely, Ncelx,  MxDiffs, sizeof(FD_BASE));
      DMax = MaxMatrix(MxDiffs, Ncely, Ncelx);
      GatherMsg(Ncely, Ncelx,  AvDiffs, sizeof(FD_BASE));
      DAv = SumMatrix(AvDiffs, Ncely, Ncelx)/(Ncely*Ncelx);
      HPRINT((STDOUT, "%d global errors, max=%G, av=%G\n", NErr, DMax, DAv));
      sprintf(ResultMsg, " #e: %d %.3f %.3f", NErr, DMax, DAv);
      LOGFILE_APPEND(LogFileName, ResultMsg);
 
      GatherMsg(Ncely, Ncelx,  MxDiffIs, sizeof(int));
      GatherMsg(Ncely, Ncelx,  MxDiffJs, sizeof(int));
      if (NErr) {
        PrintIntMat(Ncely, Ncelx, NErrs);
        PrintMatrix('a', MxDiffs, Ncely, Ncelx, Ncelx);
        PrintMatrix('v', AvDiffs, Ncely, Ncelx, Ncelx);
        PrintCoords(Ncely, Ncelx, MxDiffIs, MxDiffJs);
      }
      free(NErrs);                      free(MxDiffs);
      HPRINT((STDOUT, "\n\n"));
    }
    if (NErr)
      exit(1);
    N += StepN; M += StepM;
  } /* for (i=(...) */
} /* main() */


