/* @(#) All Right Reserved, Copyright (c) 1997 Australian National University */
/* @(#)                     Copyright (c) 1997 Fujitsu Laboratories Ltd. */
/*  */
/*        */
/*                        Permission Notice */
/*  */
/*                Australian National University */
/*                Canberra ACT 0200 Australia */
/*                           and */
/*                Fujitsu Laboratories Ltd. */
/*                4-1-1 Kamikodanaka Nakahara */
/*                Kawasaki 211-88 Japan */
/*  */
/*   Permission to use, copy, modify and distribute this software (including  */
/*   its documentation) without fee is hereby granted under the following terms */
/*   and conditions: */
/*  */
/*   1. This permission notice and all of the copyright notices in this */
/*      software must appear in all copies and derived versions of all */
/*      or a part of this software. */
/*  */
/*   2. Sales of this software, whether incorporated in other products as a */
/*      part thereof or not, is prohibited. */
/*  */
/*   3. On publication or presentation in any form of the results obtained */
/*      by use of this software, an explicit reference to this software and */
/*      its copyright owner shall be made. */
/*  */
/*   4. THIS SOFTWARE IS PROVIDED 'AS IS' AND THE AUSTRALIAN NATIONAL UNIVERSITY */
/*      AND FUJITSU LABORATORIES LTD. MAKE NO WARRANTIES, EITHER EXPRESSED OR */
/*      IMPLIED, AS TO ANY MATTER WHATSOEVER REGARDING THIS SOFTWARE,  */
/*      INCLUDING, WITHOUT LIMITATION, WARRANTIES OF MERCHANTABILITY OR FITNESS */
/*      FOR ANY PARTICULAR PURPOSE. */
/*      IN NO EVENT SHALL THE AUSTRALIAN NATIONAL UNIVERSITY OR FUJITSU */
/*      LABORATORIES LTD. BE LIABLE FOR ANY CLAIM FROM A THIRD PARTY, OR */
/*      FOR SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES WHATSOEVER RESULTING */
/*      FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN ACTION OF CONTRACT, */
/*      NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION */
/*      WITH THE USE, COPYING OR MODIFICATION OF THIS SOFTWARE. */
/*  */

/************************************************************************/
/*                              Timer.c                                 */
/************************************************************************/
/* SCCS INFO @(#)Timer.c 3.5         last modified P. Strazdins, 96/05/31
 *  first created: P. Strazdins 95/09
 */
/* module to provide simplified timer interface for Solaris 1 & 2
 * workstations or AP cells.   
 * note: currently FTN_BLACS versions must run on AP cells
 */
  
#if defined (FTN_BLACS)
# include 	"FtnBLACS.h"
# define 	SYNCH 	FTN_BLACS_BARRIER(&BLACSCntxt, "A")
# define	DTIMER	FTN_DWALLTIME00
#elif defined(APLIB)
# define 	SYNCH 	sync(0,0)
# define	DTIMER	dgettime
#else
# define 	SYNCH
# ifdef SOL2
#   undef FD_SET		/* avoid clash in /usr/include/sys/types.h */
#   include 	<sys/time.h>
#   include 	<sys/times.h>
#   include 	<limits.h>
#   undef FD_SET		/* avoid clash in /usr/include/sys/types.h */
# else
#   include 	<sys/types.h>
#   include 	<sys/time.h>
#   include 	<sys/timeb.h>
#   include 	<sys/times.h>
#   define HZ 60		/* unit for time in times.h */
# endif SOL2
#endif APLIB

double UserEndTimerVal;		/* records user time for last EndTimer() call */

#if defined(APLIB) || defined (FTN_BLACS) 
static double tim[3]; 
#else
# ifdef SOL2
static hrtime_t t0, t1, t2;
static struct tms ut0, ut1;
# else
static struct timeb t0, t1, t2;
static struct tms ut0, ut1;
# endif 
#endif /*APLIB*/


static void start_timer()
{ SYNCH;	
#if defined(APLIB) || defined (FTN_BLACS)
  tim[0] = DTIMER();
#else 
# ifdef SOL2 
  times(&ut0);
  t1 = gethrtime();
# else
  times(&ut0);
  ftime(&t1);
# endif 
#endif /*APLIB*/
} /* start_timer() */


static double end_timer(s) 
  int s;     
{	
#if defined(APLIB) || defined (FTN_BLACS)
/*double t = dgettime(); extern char ResultMsg[];*/
  if (s)
    SYNCH; 
  tim[1] = DTIMER();
  if (s)
    SYNCH;
  tim[2] = DTIMER();
  UserEndTimerVal = tim[1] - tim[0];
  return(/*t-tim[0]*/tim[1]-tim[0]-(tim[2]-tim[1])); 
#else
# ifdef  SOL2
  t2 = gethrtime();
  t0 = gethrtime() - t2;  
  times(&ut1);
  UserEndTimerVal = (double) (ut1.tms_utime - ut0.tms_utime) / (double) CLK_TCK;
  return( (double)(t2 - t1 - t0) / 1.0E9 );
# else
  ftime(&t2);
  times(&ut1);
  UserEndTimerVal = (double) (ut1.tms_utime - ut0.tms_utime) / (double) HZ;  
  return (double) (1000.0*t2.time + t2.millitm - 1000.0*t1.time - t1.millitm)/
                       1000.0;
# endif /*SOL2*/
#endif /*APLIB*/
} /* end_timer() */

