/* @(#) All Right Reserved, Copyright (c) 1997 Australian National University */
/* @(#)                     Copyright (c) 1997 Fujitsu Laboratories Ltd. */
/*  */
/*        */
/*                        Permission Notice */
/*  */
/*                Australian National University */
/*                Canberra ACT 0200 Australia */
/*                           and */
/*                Fujitsu Laboratories Ltd. */
/*                4-1-1 Kamikodanaka Nakahara */
/*                Kawasaki 211-88 Japan */
/*  */
/*   Permission to use, copy, modify and distribute this software (including  */
/*   its documentation) without fee is hereby granted under the following terms */
/*   and conditions: */
/*  */
/*   1. This permission notice and all of the copyright notices in this */
/*      software must appear in all copies and derived versions of all */
/*      or a part of this software. */
/*  */
/*   2. Sales of this software, whether incorporated in other products as a */
/*      part thereof or not, is prohibited. */
/*  */
/*   3. On publication or presentation in any form of the results obtained */
/*      by use of this software, an explicit reference to this software and */
/*      its copyright owner shall be made. */
/*  */
/*   4. THIS SOFTWARE IS PROVIDED 'AS IS' AND THE AUSTRALIAN NATIONAL UNIVERSITY */
/*      AND FUJITSU LABORATORIES LTD. MAKE NO WARRANTIES, EITHER EXPRESSED OR */
/*      IMPLIED, AS TO ANY MATTER WHATSOEVER REGARDING THIS SOFTWARE,  */
/*      INCLUDING, WITHOUT LIMITATION, WARRANTIES OF MERCHANTABILITY OR FITNESS */
/*      FOR ANY PARTICULAR PURPOSE. */
/*      IN NO EVENT SHALL THE AUSTRALIAN NATIONAL UNIVERSITY OR FUJITSU */
/*      LABORATORIES LTD. BE LIABLE FOR ANY CLAIM FROM A THIRD PARTY, OR */
/*      FOR SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES WHATSOEVER RESULTING */
/*      FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN ACTION OF CONTRACT, */
/*      NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION */
/*      WITH THE USE, COPYING OR MODIFICATION OF THIS SOFTWARE. */
/*  */

/************************************************************************/
/* 	BLASInnerLoops: BLAS-2 & BLAS-3 inner loop procedures	*/
/************************************************************************/
/* SCCS INFO @(#)BLASInnerLoops.h 2.1         last modified P. Strazdins, 97/01/10
   first created: P. Strazdins 92/01
*/
/* set of procedures for SPARC processors implementing fast BLAS inner 
   loops; this header file can be used for all FDATATYPE values;
   the corresponding object procedures are in the archive 
   $FDATATYPE/BLASInnerLoops.a
   
   requires FDATATYPE to be defined
   
   written by P. Strazdins, from 91/02/12
*/

#ifndef __BLASInnerLoops_INCL__

#define __BLASInnerLoops_INCL__

#include "FDATASetUp.h"

#if (FDATATYPE==SINGLE)
# define MatVecMul	SMatVecMul
# define Rank1Update 	SRank1Update
# define Rank1UpdateA 	SRank1UpdateA
# define RankKUpdate 	SRankKUpdate
# define UpdateRect 	SUpdateRect
# define UpdateRectTr1 	SUpdateRectTr1
# define UpdateRectTr2 	SUpdateRectTr2
# define VecMatMul 	SVecMatMul
#elif (FDATATYPE==DOUBLE)
# define MatVecMul	DMatVecMul
# define Rank1Update 	DRank1Update
# define RankKUpdate 	DRankKUpdate
# define Rank1UpdateA 	DRank1UpdateA
# define UpdateRect 	DUpdateRect
# define UpdateRectTr1 	DUpdateRectTr1
# define UpdateRectTr2 	DUpdateRectTr2
# define VecMatMul 	DVecMatMul
# define Axpy1		DAxpy1
#elif (FDATATYPE==COMPLEX)
# define MatVecMul	CMatVecMul
# define Rank1Update 	CRank1Update
# define RankKUpdate 	CRankKUpdate
# define Rank1UpdateA 	CRank1UpdateA
# define UpdateRect 	CUpdateRect
# define UpdateRectTr1 	CUpdateRectTr1
# define UpdateRectTr2 	CUpdateRectTr2
# define VecMatMul 	CVecMatMul
#else
# define MatVecMul	ZMatVecMul
# define Rank1Update 	ZRank1Update
# define RankKUpdate 	ZRankKUpdate
# define Rank1UpdateA 	ZRank1UpdateA
# define UpdateRect 	ZUpdateRect
# define UpdateRectTr1 	ZUpdateRectTr1
# define UpdateRectTr2 	ZUpdateRectTr2
# define VecMatMul 	ZVecMatMul
#endif 
 
/* main matrix multiply routines UpdateRect & UpdateRectTr1 update a
   RECT_Y x RECT_x rectangle of the result matrix */
#if (FDATATYPE==SINGLE)
# define RECT_Y 4
# define RECT_X 4
#elif (FDATATYPE==DOUBLE || FDATATYPE==COMPLEX)   
# define RECT_Y 4
# define RECT_X 2
#else 
# define RECT_Y 2
# define RECT_X 2
#endif

/* crossover point for K where more efficient to base matrix multiply
 * on UpdateRect procedures rather than rank-1 updates 
 */
#define MIN_UPDATE_RECT_K 4

#define RKSIZE 3		/* size */

extern void MatVecMul STDC_PARAMS(
	(int k, int m, int bounds[], FDATA A[], FDATA B[], FDATA C[])
);
/* pre:  k <= LdA, 1 <= LdB, 1 <= LdC, 
 *	 2 | LdA if FDTATYPE==SINGLE
 *       {LdA, LdB, LdC} = bounds[0..2]
 * post: C += A B where A is m by k, B is k by 1, C is m by 1,
 * (perform matrix-vector multiplication, as for BLAS-2
 *  _GEMV('N', k, m, 1.0, A, bounds[0], B, bounds[1], 1.0, C, bounds[2]))
 */

extern void Rank1Update STDC_PARAMS(
	(int m, int n, int bounds[], FDATA A[], FDATA B[], FDATA C[])
);	
/* pre:  1 <= LdA, 1 <= LdB, n <= LdC, 
 * 	 2 | LdC if FDTATYPE==SINGLE,
 *       {LdA, LdB, LdC} = bounds[0..2]
 * post: C += A B where A is m by 1, B is 1 by n (stride LdB), C is m by n,
 * (perform column-based rank 1 matrix update, as for BLAS-2
 *  _GER(m, n, 1.0, A, LdA, B, LdB, C, LdC))
*/

extern void Rank1UpdateA STDC_PARAMS(
	(int m, int n, int bounds[], FDATA A[], FDATA B[], FDATA C[])
);
/* pre:  1 <= LdA, 1 = LdB, n <= LdC, 
 *	 2 | LdC if FDTATYPE==SINGLE
 *       {LdA, LdB, LdC} = bounds[0..2]
 * post: C += A B where A is m by 1, B is 1 by n (stride LdB), C is m by n,
 * (perform row-based rank 1 matrix update, as for BLAS-2
 *  _GER(m, n, 1.0, A, LdA, B, 1, C, LdC))
 */

extern void UpdateRect STDC_PARAMS(
	(int k, int TrC, int bounds[], FDATA A[], FDATA B[], FDATA C[])
);
/* pre:  k <= LdA, RECT_X <= LdB, 1 <= LdC, 
 *	 2 | LdA, 2 | LdB if FDATATYPE=SINGLE,
 *       {LdA, LdB, LdC} = bounds[0..2]
 * post: C += A B where C is RECT_Y by RECT_X, if TrC == 0
 *       C^T += A B where C is RECT_X by RECT_Y, o.w.
 *       and A is RECT_Y by k, B is k by RECT_X, 
 * (perform matrix-matrix multiplication, as for BLAS-3
 *  _GEMM('N', 'N', RECT_Y, RECT_X, k, 1.0, A, bounds[0], B, bounds[1], 1.0, C, bounds[2])
 *  if TrC == 0, and
 *  _GEMM('T', 'T', RECT_X, RECT_Y, k, 1.0, B, bounds[1], A, bounds[2], 1.0, C, bounds[2])
 *  otherwise, using fact that C^T += BA equivalent to C = A^TB^T)
 */

extern void UpdateRectTr1 STDC_PARAMS( 
	(int k, int bounds[], FDATA A[], FDATA B[], FDATA C[])
);
/* pre: 4 <= LdA, 4 <= LdB, 1 <= LdC, 
 *	2 | LdA && 2 | LdB if FDTATYPE==SINGLE,
 *      {LdA, LdB, LdC} = bounds[0..2]
 * post: C += A^T B where A is k by RECT_Y, B is k by RECT_X, C is RECT_Y by RECT_X
 * (perform matrix-matrix multiplication, as for BLAS-3
 *  _GEMM('T', 'N', RECT_Y, RECT_X, k, 1.0, A, bounds[0], B, bounds[1], 1.0, C, bounds[2]))
 */


extern void UpdateRectTr2 STDC_PARAMS(
	(int k, int bounds[], FDATA A[], FDATA B[], FDATA C[])
);	
/* pre:  k <= LdA, k <= LdB, 1 <= LdC, 
 *	 2 | LdA && 2 | LdB if FDTATYPE==SINGLE,
 *       {LdA, LdB, LdC} = bounds[0..2]
 * post: C += A B^T where A is RECT_Y by k, B is RECT_X by k, C is RECT_Y by RECT_X
 * (perform matrix-matrix multiplication, as for BLAS-3
 *  _GEMM('N', 'T', RECT_Y, RECT_X, k, 1.0, A, bounds[0], B, bounds[1], 1.0, C, bounds[2]))
 */

extern void VecMatMul STDC_PARAMS(
	(int k, int n, int bounds[], FDATA A[], FDATA B[], FDATA C[])
);
/* pre:  1 <= LdA, n <= LdB, 1 <= LdC, 
 *	 2 | LdB if FDTATYPE==SINGLE,
 *       {LdA, LdB, LdC} = bounds[0..2]
 * post: C += A B where A is 1 by k (stride LdA), B is k by n, C is 1 by n,
 * (perform vector-matrix multiplication, as for BLAS-2
 *  _GEMV('T', n, k, 1.0, B, bounds[1], A, bounds[0], 1.0, C, bounds[2]))
 */

extern void MatVecMul STDC_PARAMS(
	(int m, int k, int bounds[], FDATA A[], FDATA B[], FDATA C[])
);
/* pre:  1 <= LdA, n <= LdB, 1 <= LdC, 
 *	 2 | LdB if FDTATYPE==SINGLE,
 *       {LdA, LdB, LdC} = bounds[0..2], 
 * post: C += A B where A is m by k, B is k by 1, C is m by 1,
 * (perform vector-matrix multiplication, as for BLAS-2
 *  _GEMV('N', m, k, 1.0, A, bounds[0], B, bounds[1], 1.0, C, bounds[2]))
 */


extern void RankKUpdate STDC_PARAMS(
	(int m, int n, int bounds[], FDATA A[], FDATA B[], FDATA *alpha, FDATA C[])
);
/* pre:  {RIncA, RIncB, RIncC, CIncA, CIncB, CIncC} = bounds[0..5],
 * post: as for MatMul(m, n, RKSIZE, alpha, {A, RIncA, CIncA}, {B, RIncB, CIncC},
 * 		       &FD_1, {C, RIncC, CIncC});
 */


extern void Axpy1 STDC_PARAMS(
	(int n, FDATA *alpha, FDATA X[], int IncX, FDATA Y[],  int IncY)
);
/* pre:  X (Y) is a vector of stride IncX (IncY); IncX, IncY > 0
 * post: Y[0..n-1] += *alpha X[0..n-1]
 */	      
#endif /*__BLASInnerLoops_INCL__*/
