/* @(#) All Right Reserved, Copyright (c) 1997 Australian National University */
/* @(#)                     Copyright (c) 1997 Fujitsu Laboratories Ltd. */
/*  */
/*        */
/*                        Permission Notice */
/*  */
/*                Australian National University */
/*                Canberra ACT 0200 Australia */
/*                           and */
/*                Fujitsu Laboratories Ltd. */
/*                4-1-1 Kamikodanaka Nakahara */
/*                Kawasaki 211-88 Japan */
/*  */
/*   Permission to use, copy, modify and distribute this software (including  */
/*   its documentation) without fee is hereby granted under the following terms */
/*   and conditions: */
/*  */
/*   1. This permission notice and all of the copyright notices in this */
/*      software must appear in all copies and derived versions of all */
/*      or a part of this software. */
/*  */
/*   2. Sales of this software, whether incorporated in other products as a */
/*      part thereof or not, is prohibited. */
/*  */
/*   3. On publication or presentation in any form of the results obtained */
/*      by use of this software, an explicit reference to this software and */
/*      its copyright owner shall be made. */
/*  */
/*   4. THIS SOFTWARE IS PROVIDED 'AS IS' AND THE AUSTRALIAN NATIONAL UNIVERSITY */
/*      AND FUJITSU LABORATORIES LTD. MAKE NO WARRANTIES, EITHER EXPRESSED OR */
/*      IMPLIED, AS TO ANY MATTER WHATSOEVER REGARDING THIS SOFTWARE,  */
/*      INCLUDING, WITHOUT LIMITATION, WARRANTIES OF MERCHANTABILITY OR FITNESS */
/*      FOR ANY PARTICULAR PURPOSE. */
/*      IN NO EVENT SHALL THE AUSTRALIAN NATIONAL UNIVERSITY OR FUJITSU */
/*      LABORATORIES LTD. BE LIABLE FOR ANY CLAIM FROM A THIRD PARTY, OR */
/*      FOR SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES WHATSOEVER RESULTING */
/*      FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN ACTION OF CONTRACT, */
/*      NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION */
/*      WITH THE USE, COPYING OR MODIFICATION OF THIS SOFTWARE. */
/*  */

/************************************************************************/
/*              D_BLAS.h: Distributed BLAS header file		*/
/************************************************************************/
/* SCCS INFO @(#)D_BLAS.h 4.4         last modified P. Strazdins, 97/05/14
   first created: P. Strazdins 93/03/23
*/
/* 
Header file for Distributed BLAS (DBLAS) procedures; 
for compilation, the macro FDATATYPE (having the values SINGLE or DOUBLE)
must be defined:
	if FDATATYPE=SINGLE, FDATA=float
	if FDATATYPE=DOUBLE, FDATA=double
Below, the underscores in the procedure name should be replaced by 'S',
if FDATATYPE=SINGLE (the DSBLAS procedures), and 'D', if FDATATYPE=DOUBLE
(the DDBLAS procedures). eg. for FDATATYPE=SINGLE, D_GEMM becomes DSGEMM.


This implementation of Distributed BLAS (DBLAS) is made for block-cyclic
matrices, with an interface for DistMat matrices (see files DistMat.h and
LocMat.h for a description) replacing array addresses and stride pairs
for non-distributed BLAS.  Note that the LocMat matrix descriptors can
encompass both row-major and column-major storage conventions.  It also
has similar error handling characteristics to the normal BLAS.
     
The DBLAS procedures all assume global matrices are distributed over a
P[0] x P[1] processor grid in rectangular block-cyclic storage,  with
array indices beginning from 0.  Thus, for the global M x N matrix A:
	A = {Tr, {Ax, Inc[0..1]}, D[0..1]}, where D[i] ={B0[i],B[i],P0[i],P[i]}
then A has B[0] x B[1] blocks with the global top-left element at
address Ax in cell (I0[0], I0[1]) with block offset (B0[0], B0[1]),
where:
	I0[i] = (P[i] + P0[i] - I[i])%P[i]	, for i = 0,1
and: 
	(I[0],I[1]) is the absolute cell id of the current cell.
The procedures also require that:
	Tr = {0, 1}
	0<=B0[i]<B[i] and 0<=P0[i]<P[i]		,  for i=0,1.
	MIN(Inc[0], Inc[1]) = 1 
and that the MAX(Inc[0], Inc[1]) at least as large as its the complementary
local matrix size. This is coded as:
	DistMatOK(M, N, A)
	 
The global element A[i,j] is thus the top left element of SubMat(A, i, j), 
where SubMat(A, i, j) is defined in DistMat.h.  In the case of
triangular matrices, A[0,0] is top-left element of the enclosing
rectangular sub-matrix for that cell, rather than the top-left element
of the triangular sub-matrix for that cell.  On a 1x1 configuration,
these two are the same; however on general configurations, the latter is
different and difficult to calculate.  Thus, to keep the interface
simple, the former was chosen.  Similarly the P0 sub-fields refer to
cell offsets relative to the cell containing the top-left element of the
enclosing global rectangular matrix. 

If the transposition specifier Tr=1, the matrix A is regarded as
`transposed'.  That is, if, in the context of a computation, A is
normally regarded as an M x N matrix, the `M-dimension' (`N-dimension')
of A would correspond to D[!Tr] (D[Tr]).  This obviates the Trans and
Side parameters of the BLAS procedures.  The convention chosen will be
that computations corresponding to the 'Left' side will be regarded as
`normal'; to express a BLAS computation corresponding to the `Right'
side, the transpose fields of all matrix operands should be reversed. 
The notation A^T (= Trans(A)) signifies merely the toggling of the Tr
field of A. 

As well as checking the internal validity of each DistMat parameter, the
DBLAS procedures also checks between DistMat parameters.  For example,
consider the computation C = AB where A is M x K and B is K x N. 
Firstly, the cell dimension specifiers P[0..1] within each of A, B and C
must be the same, with the exception that P[i]=1 if the corresponding
matrix dimension is `pre-spread' (see below).  This is coded as:
	P_Compat(A, C), P_Compat(B, C) 
Secondly, with the 3 matrix dimensions M, N, K, we must have alignment
in at least 2 out of 3 of the corresponding dimension specifiers of A,
B, and C.  For example, in the case of non-transposed A, B and C, D[0]
of A and C correspond to the M-dimension, D[1] of A and D[0] of B
correspond in the K-dimension.  `Alignment' means that the block sizes
and offsets are the same.  It also requires that the corresponding cell
dimension specifier P[i] from A is the same as the P[i] from C.  This is
coded as:
	MDimAligned(A,C), NDimAligned(B,C), KDimAligned(A,B,C)

The input matrices (only), ie.  A or B, may be `pre-spread' across one
of their matrix dimensions.  This corresponds to a vector replicated
across cell rows or columns, except that it can be a proper matrix. 
This means that a cell holds the complete portion of the global index
for that matrix, in which case the dimension specifier D for that
dimension is given by:
	 D = {0,1,0,1}. 
Note that the original block-scattered structure of dimension is lost. 
Allowing `pre-spread' matrices is useful in that communication can be
reduced. 
 
The Distributed BLAS procedures perform BLAS operations on the
respective global matrices provided (a given task on) all cells are
simultaneously executing (in SPMD mode) the corresponding DBLAS
procedure. 

In all DBLAS procedures, matrix size, shape, and scaling, specifiers
(eg.  M, *UpLo, alpha) refer to the corresponding global computation and
must have identical values across all cells. 

The Tr, B0, and B and P dimension sub-fields of A must similarly be
the same across all cells.  The P0  sub-fields must be consistent in
that all cells will compute from them the same I0[i] values from them. 
The local matrix specifier of A, eg. {Ax, Inc[0..1]}, can be different
on the cells.  Note that local element A[i_loc,j_loc] is at address 
	(Ax + i_loc*Inc[0] + j_loc*Inc[1]). 

The DBLAS procedures check the validity of each of its parameters on
each cell (checking of corresponding parameters across cells is not
implemented); if one is found invalid, a message is written to the cells
standard output, indicating the procedure's name, the number of the 1st
parameter from which an error was found, and the values of all
parameters.  Where the error is with respect to several parameters. 
(such as an alignment restriction is not met), the numbers of the
associated parameters are also given.  Note that symmetric matrices must
have the same block-alignment along both matrix dimensions.  The
procedure on this cell then aborts. 

The Level 3 DBLAS procedures all partition the computation across
(generally) the smallest matrix dimension in units large enough to
achieve good level 3 cell computations (see MachineParams.h). 
Workspaces are dynamically allocated from heap memory; their size can be
restricted to not exceed MAX_DBLAS_WS_SIZE.  Partitioning and if needed
synchronisation is also performed on messaging so that in systems where
incoming messages have to be stored in system buffers, the size of
unread messages should be not exceed 2*OPT_MSG_SIZE.  If the DBLAS
procedures are compiled with the -DTUNE option, all MachineParams
constants may be set dynamically. 

*/   

#ifndef __D_BLAS_INCL__
#define __D_BLAS_INCL__

#include "FDATASetUp.h"
#include "LocMat.h"
#include "DistMat.h"

#if (FDATATYPE==SINGLE)		/*define procedure identifiers and names*/
# define D_GEMM  	DSGEMM
# define D_SYMM  	DSSYMM
# define D_SYRK  	DSSYRK
# define D_SYR2K 	DSYR2K
# define D_TRMM  	DSTRMM
# define D_TRSM  	DSTRSM

# define D_GEMV	 	DSGEMV	 
# define D_SYMV	 	DSSYMV	 
# define D_TRMV	 	DSTRMV	 
# define D_TRSV	 	DSTRSV	 
# define D_GER	 	DSGER	 
# define D_SYR	 	DSSYR	 
# define D_SYR2	 	DSSYR2	 

# define D_SWAP	 	DSSWAP	 
# define D_SCAL	 	DSSCAL	 
# define D_COPY	 	DSCOPY	 
# define D_AXPY	 	DSAXPY	 
# define D_DOT	 	DSDOT	 
# define D_NRM2	 	DSNRM2	 
# define D_ASUM	 	DSASUM	 
# define D_AMAX		DSAMAX	 

# define D_GetElt	DSGetElt	 
# define D_SetElt	DSSetElt	 
# define D_CellHolds	DSCellHolds	 
# define D_ScaleAdd	DSScaleAdd

#elif (FDATATYPE==DOUBLE)
# define D_GEMM  	DDGEMM
# define D_SYMM  	DDSYMM
# define D_SYRK  	DDSYRK
# define D_SYR2K 	DDSYR2K
# define D_TRMM  	DDTRMM
# define D_TRSM  	DDTRSM

# define D_GEMV	 	DDGEMV	 
# define D_SYMV	 	DDSYMV	 
# define D_TRMV	 	DDTRMV	 
# define D_TRSV	 	DDTRSV	 
# define D_GER	 	DDGER	 
# define D_SYR	 	DDSYR	 
# define D_SYR2	 	DDSYR2	 

# define D_SWAP	 	DDSWAP	 
# define D_SCAL	 	DDSCAL	 
# define D_COPY	 	DDCOPY	 
# define D_AXPY	 	DDAXPY	 
# define D_DOT	 	DDDOT	 
# define D_NRM2	 	DDNRM2	 
# define D_ASUM	 	DDASUM	 
# define D_AMAX		DDAMAX	 

# define D_GetElt	DDGetElt	 
# define D_SetElt	DDSetElt	 
# define D_CellHolds	DDCellHolds	 
# define D_ScaleAdd	DDScaleAdd
#endif /*FDATATYPE*/

/********************* DistMat predicates *******************************/

#define DistMatOK(M, N, A)						\
	(DM_OK(A) && DistMatIncOK(M, N, A))				
#define DistMatIncOK(M, N, A)						\
	IncOK(G2Ld(M, DM_DM(A)), G2Ld(N, DM_DN(A)), 			\
	      LM_Inc(DM_Trans(A), DM_Loc(A)), LM_Inc(1-DM_Trans(A), DM_Loc(A)))

#define P_Compat(A, C)							\
	(Dim_PCompat(DM_D0(A), DM_D0(C)) && Dim_PCompat(DM_D1(A), DM_D1(C)))

#define P_Same(A, C)							\
	(Dim_PEq(DM_D0(A), DM_D0(C)) && Dim_PEq(DM_D1(A), DM_D1(C)))

#define OuterDimAligned(d, A, C)  					\
	(Dim_BCompat(DM_D((d)!=DM_Trans(A),A), DM_D((d)!=DM_Trans(C),C)) &&\
	 Dim_PEq(    DM_D((d)!=DM_Trans(A),A), DM_D((d)!=DM_Trans(A),C)))
#define MDimAligned(A, C)						\
	OuterDimAligned(0, A, C)  
#define NDimAligned(B, C)						\
	OuterDimAligned(1, B, C)
#define KDimAligned(A, B, C)						\
	(Dim_BCompat(DM_D(!DM_Trans(A), A), DM_D( DM_Trans(B), B)) &&	\
	 Dim_PEq(    DM_D(!DM_Trans(A), A), DM_D(!DM_Trans(A), C)) &&	\
	 Dim_PEq(    DM_D( DM_Trans(B), B), DM_D( DM_Trans(B), C)))

#define MDimEq(A, C)							\
	Dim_Eq(DM_DM(A), DM_DM(C))
	 
/*********************** level 3 procedures ******************************/

void D_GEMM STDC_PARAMS(
	(int M, int N, int K, FDATA alpha, DistMat A, DistMat B, 
	 FDATA beta, DistMat C)
);
/* pre:  M, N, K >= 0; 
 *	 DistMatOK(M, K, A),  DistMatOK(K, N, B),  DistMatOK(M, N, C);
 *	 P_Compat(A, C), P_Compat(B, C);
 *	 at least 2 of MDimAligned(A,C), NDimAligned(B,C), KDimAligned(A,B,C)
 *	 where A' = op(*TransA, A), B' =  op(*TransB, B)
 * post: performs the distributed   matrix operation:
 *	 	C = alpha*A*B + beta*C
 *	 where A is M x K, B is K x N, C is M x N 
 */   

void D_SYMM STDC_PARAMS(
	(char *UpLo, int M, int N, FDATA alpha, DistMat A, DistMat B, 
	 FDATA beta, DistMat C)
);	
/* pre:  *UpLo = 'U', 'u', 'L', 'l'; M, N >= 0;
 *	 DistMatOK(M, M, A),  DistMatOK(M, N, B),  DistMatOK(M, N, C).
 *	 P_Compat(A, C), P_Compat(B, C), 
 *	 at least 2 of MDimAligned(A,C), NDimAligned(B,C), KDimAligned(A,B,C)
 *	 A is a symmetric matrix stored as an upper (if *UpLo = 'u', 'U')
 *	 or lower (otherwise) triangular matrix, and: 
 *	 	Dim_BCompat(DM_D0(A), DM_D1(A)) 
 * post: performs the distributed   matrix operation:
 *	 	C = alpha*A*B + beta*C
 *	 where A is M x M, B, C are M x N
 * note: if DM_Trans(A), *UpLo should be reversed from its absolute sense.
 * note: to perform the equivalent of the BLAS:
 *		_SYMM(UpLo, "Right", M, N, alpha, A, B, beta, C)
 *	 use:
 *		D_SYMM(UpLo, N, M, alpha, A, Trans(B), beta, Trans(C))
 */   

void D_SYRK STDC_PARAMS(
	(char *UpLo, int N, int K, 
	 FDATA alpha, DistMat A, FDATA beta, DistMat C)
);	
/* pre:  *Uplo = 'U', 'u', 'L', 'l'; N, K >= 0; 
 *	 DistMatOK(N, K, A), DistMatOK(N, N, C); 
 *	 P_Compat(A', C), MDimAligned(A', C)
 *	 C is a symmetric matrix stored as an upper (if *UpLo = 'u', 'U')
 *	 or lower (otherwise) triangular matrix, and: 
 *	 	Dim_BCompat(DM_D0(C), DM_D1(C)) 
 * post: performs the distributed   matrix operation:
 *	 	C = alpha*A*A^T + beta*C
 *	 where A is N x K, C is N x N 
 * note: to perform the equivalent of the BLAS:
 *		_SYRK(UpLo, "Trans", N, K, alpha, A, beta, C)
 *	 use:
 *		D_SYRK(UpLo, N, M, alpha, Trans(A), beta, C)
 */

void D_SYR2K STDC_PARAMS(
	(char *UpLo, int N, int K, 
	 FDATA alpha, DistMat A, DistMat B, FDATA beta, DistMat C)
);	
/* pre:  *UpLo = 'U', 'u', 'L', 'l'; N, K >= 0; 
 *	 DM_Trans(A)==DM_Trans(B)
 *	 DistMatOK(N, K, A), DistMatOK(N, K, B), DistMatOK(N, N, C); 
 *	 P_Compat(A, C),  P_Compat(B, C)
 *	 at least 2 of MDimAligned(A,C), NDimAligned(B,C), KDimAligned(A,B,C)
 *	 C is a symmetric matrix stored as an upper (if *UpLo = 'u', 'U')
 *	 or lower (otherwise) triangular matrix, and: 
 *	 	Dim_BCompat(DM_D0(C), DM_D1(C)) 
 * post: performs the distributed   matrix operation:
 *	 	C = alpha*(A*B^T + B*A^T) + beta*C
 *	 where A, B is N x K, C is N x N 
 * note: to perform the equivalent of the BLAS:
 *		_SYR2K(UpLo, "Trans", N, K, alpha, A, beta, C)
 *	 use:
 *		D_SYR2K(UpLo, N, K, alpha, Trans(A), Trans(B), beta, C)
 */   

void D_TRMM STDC_PARAMS(
	(char *UpLo, char *Diag, int M, int N, FDATA alpha, DistMat A, 
	 DistMat B)
); 
/* pre:  *UpLo = 'U', 'u', 'L', 'l'; *Diag = 'U', 'u', 'N', 'n';   M, N >= 0; 
 *	 DistMatOK(M, M, A), DistMatOK(M, N, B), P_Compat(A, B)
 *	 MDimAligned(A,B) || KDimAligned(A,B,B)
 *	 A is an upper (if *UpLo = 'u', 'U') or lower (otherwise) triangular 
 *	 matrix with a diagonal taken to be unity iff Diag = 'U', 'u'.
 * post: performs the distributed   matrix operation:
 *	 	B = alpha* A B
 *	 where A is M x M, B is M x N
 * note: if DM_Trans(A), *UpLo should be reversed from its absolute sense.
 * note: to perform the equivalent of the BLAS:
 *		_TRMM("Right", UpLo, Trans, Diag, M, N, alpha, A, B)
 *	 use:
 *		D_TRMM(UpLo', Diag, N, M, alpha, A', Trans(B))
 *	 where A' = *Trans="T",'t'? A: Trans(A), 
 *	       *UpLo' = *Trans="T",'t'? *UpLo: TransForm(UpLo)
 */   


void D_TRSM STDC_PARAMS(
	(char *Uplo, char *Diag, int M, int N,  FDATA alpha, DistMat A, 
	 DistMat B)
); 
/* pre:  *UpLo = 'U', 'u', 'L', 'l'; *Diag = 'U', 'u', 'N', 'n';   M, N >= 0; 
 *	 DistMatOK(M, M, A), DistMatOK(M, N, B), P_Compat(A, B)
 *	 MDimAligned(A,B) || KDimAligned(A,B,B)
 *	 A is an upper (if *UpLo = 'u', 'U') or lower (otherwise) triangular 
 *	 matrix with a diagonal taken to be unity iff Diag = 'U', 'u'.
 * post: performs the distributed   matrix operation:
 *	 	B = alpha* A B
 *	 where A is M x M, B is M x N
 * note: if DM_Trans(A), *UpLo should be reversed from its absolute sense.
 * note: to perform the equivalent of the BLAS:
 *		_TRSM("Right", UpLo, Trans, Diag, M, N, alpha, A, B)
 *	 use:
 *		D_TRSM(UpLo', Diag, N, M, alpha, A', Trans(B))
 *	 where A' = *Trans="T",'t'? A: Trans(A)
 *	       *UpLo' = *Trans="T",'t'? *UpLo: TransForm(UpLo)
 */   

/*********************** level 2 procedures ******************************/ 
/* 
In the Level 2 and 1 procedures, vector operands are taken to be column
vectors, if their transposition bit is not set, and row vectors, otherwise.
*/

void D_GEMV STDC_PARAMS(
	(int M, int N, FDATA alpha, DistMat A, DistMat X, FDATA beta, DistMat Y)
);
/* pre:  M, N >= 0;
 *	 DistMatOK(M, N, A), DistMatOK(N, 1, X),  DistMatOK(M, 1, Y), 
 *	 P_Compat(X, A), P_Compat(Y, A);
 *	 NDimAligned(X, A), MDimAligned(Y, A)
 * post: performs the distributed   matrix operation:
 *	 	A = alpha*X*Y + A
 *	 where A is M x N, X is N x 1, Y is 1 x M 
 * note: to multiply by a vector on the left of an M x N A, use:
 *	 	D_GEMV(N, M, alpha, Trans(A), Trans(X), beta, Trans(Y)
 */

void D_SYMV STDC_PARAMS(
	(char  *UpLo, int N, FDATA alpha, DistMat A, 
	 DistMat X, FDATA beta, DistMat Y)
);
/* pre:  *UpLo = 'U', 'u', 'L', 'l'; N >= 0
 *	 DistMatOK(M, N, A), DistMatOK(N, 1, X),  DistMatOK(M, 1, Y), 
 *	 P_Compat(X, A), P_Compat(Y, A);
 *	 MDimAligned(X, A), NDimAligned(Y, A)
 *       A is a symmetric matrix stored as an upper (if *UpLo = 'u', 'U')
 *       or lower (otherwise) symmetric matrix, and:
 *              Dim_BCompat(DM_D0(A), DM_D1(A))
 * post: performs the distributed matrix operation:
 *	 	A = alpha*X*Y + A
 *	 where A is N x N, X is N x 1, Y is N x 1
 * note: if DM_Trans(A), *UpLo should be reversed from its absolute sense.
 */

void D_TRMV STDC_PARAMS(
        (char *UpLo, char *Diag, int N, DistMat A, DistMat X)
);
/* pre:  *UpLo = 'U', 'u', 'L', 'l'; N >= 0
 *	 DistMatOK(N, N, A), DistMatOK(N, 1, X);
 *	 P_Compat(X, A);     MDimAligned(X, A)
 *       A is an upper (if *UpLo = 'u', 'U') or lower (otherwise) triangular
 *       matrix with a diagonal taken to be unity iff Diag = 'U', 'u'.
 * post: performs the distributed matrix operation:
 *	 	X  = A*X
 *	 where A is N x N, X is N x 1
 * note: if DM_Trans(A), *UpLo should be reversed from its absolute sense.
 */

void D_TRSV STDC_PARAMS(
        (char *UpLo, char *Diag, int N, DistMat A, DistMat X)
);
/* pre:  *UpLo = 'U', 'u', 'L', 'l'; N >= 0
 *	 DistMatOK(N, N, A), DistMatOK(N, 1, X);
 *	 P_Compat(X, A);     MDimAligned(X, A)
 *       A is an upper (if *UpLo = 'u', 'U') or lower (otherwise) triangular
 *       matrix with a diagonal taken to be unity iff Diag = 'U', 'u'.
 * post: performs the distributed matrix operation:
 *	 	X  = (A^-1)*X
 *	 where A is N x N, X is N x 1
 * note: if DM_Trans(A), *UpLo should be reversed from its absolute sense.
 */

void D_GER STDC_PARAMS(
	(int M, int N,  FDATA alpha, DistMat X, DistMat Y, DistMat A)
);
/* pre:  DistMatOK(M, 1, X), DistMatOK(N, 1, Y), DistMatOK(M, N, A);
 *	 P_Compat(X, A), P_Compat(Y, A);
 *	 MDimAligned(X, A), NDimAligned(Trans(Y), A)
 * post: performs the distributed matrix operation:
 *	 	A = alpha*X*Y^T + A
 *	 where X is M x 1, Y is N x 1, A is M x N 
 */   

void D_SYR STDC_PARAMS(
	(char *UpLo, int N, FDATA alpha, DistMat X, DistMat A)
);	
/* pre:  *Uplo = 'U', 'u', 'L', 'l'; N >= 0; 
 *	 DistMatOK(N, 1, X), DistMatOK(N, N, A); P_Compat(X, A), 
 *	 MDimAligned(X, A),
 *	 A is a symmetric matrix stored as an upper (if *UpLo = 'u', 'U')
 *	 or lower (otherwise) triangular matrix, and: 
 *	 	Dim_BCompat(DM_D0(A), DM_D1(A)) 
 * post: performs the distributed matrix operation:
 *	 	A = alpha*X*(X^T) + A
 *	 where X is N x 1, A is N x N 
 */   

void D_SYR2 STDC_PARAMS(
	(char *UpLo, int N, FDATA alpha, DistMat X, DistMat Y, DistMat A)
);	
/* pre:  *Uplo = 'U', 'u', 'L', 'l'; N >= 0; 
 *	 DistMatOK(N, 1, X), DistMatOK(N, 1, Y), DistMatOK(N, N, A); 
 *	 P_Compat(X, A), P_Compat(Y, A);
 *	 MDimAligned(X, A), MDimAligned(Y, A)
 *	 A is a symmetric matrix stored as an upper (if *UpLo = 'u', 'U')
 *	 or lower (otherwise) symmetric matrix, and: 
 *	 	Dim_BCompat(DM_D0(A), DM_D1(A)) 
 * post: performs the distributed matrix operation:
 *	 	A = alpha*(X*Y^T + Y*X^T) + A
 *	 where X, Y are N x 1, A is N x N 
 */

/*********************** level 1 procedures ******************************/ 
/* 
Only the cells having part of the vectors, 
or requiring the results, need be participating in the calls. 
Furthermore, the DistMat parameters can have their `scope' reduced to a
row of cells, a column or even a single cell, using:
	RowVec(X), ColVec(X) or UnitVec(X)
This is useful for the level 1 functions where, for efficiency reasons,
the scope of the cells requiring the return value can be restricted
to that of the cells holding X. In this case, the scope of the cells
participating in the call must be similarly reduced.
*/

void D_SWAP STDC_PARAMS( (int N, DistMat X, DistMat Y) );
/* pre:  N>= 0; DM_Trans(X) = DM_Trans(Y); 
 *	 DistMatOK(N, 1, X), DistMatOK(N, 1, Y), MDimEq(X,Y), P_Same(X,Y) 
 * post: performs the distributed matrix operation:
 *		X <-> Y
 *	 where X, Y are N x 1
 */

void D_SCAL STDC_PARAMS( (int N, FDATA alpha, DistMat X) );
/* pre:  N>= 0; 
 *	 DistMatOK(N, 1, X)
 * post: performs the distributed matrix operation:
 *		X = alpha*X
 *	 where X is N x 1
 */

void D_COPY STDC_PARAMS(
	(int N, DistMat X, DistMat Y)
);
/* pre:  N>= 0; DM_Trans(X) = DM_Trans(Y); 
 *	 DistMatOK(N, 1, X), DistMatOK(N, 1, Y), MDimAligned(X,Y), P_Same(X,Y)
 * post: performs the distributed matrix operation:
 *		Y = X
 *	 where X, Y are N x 1
 */

void D_AXPY STDC_PARAMS(  (int N, FDATA alpha, DistMat X, DistMat Y) );
/* pre:  N>= 0; DM_Trans(X) = DM_Trans(Y);
 *	 DistMatOK(N, 1, X), DistMatOK(N, 1, Y), MDimAligned(X,Y), P_Same(X,Y)
 * post: performs the distributed matrix operation:
 *		Y = Y + alpha*X
 *	 where X, Y are N x 1
 */

FDATA D_DOT STDC_PARAMS( (int N, DistMat X, DistMat Y) );
/* pre:  N>= 0; DM_Trans(X) = DM_Trans(Y);
 *	 DistMatOK(N, 1, X), DistMatOK(N, 1, Y), MDimAligned(X,Y) P_Same(X,Y)
 * post: returns the value of X^T*Y to all cells in scope of X
 *	 where X, Y are N x 1
 */

FDATA D_NRM2 STDC_PARAMS( (int N,  DistMat X) );	
/* pre:  N>= 0; 
 *	 DistMatOK(X)
 * post: returns the value of ||X||_2 to all cells in the scope of X
 *	 where X is N x 1
 */

FDATA D_ASUM STDC_PARAMS( (int N,  DistMat X) );
/* pre:  N>= 0; 
 *	 DistMatOK(N, 1, X)
 * post: returns the value of ||X||_1 to all cells in the scope of X
 *	 where X is N x 1
 */

int D_AMAX STDC_PARAMS( (int N, DistMat X, FDATA *Max) );
/* pre:  N>= 0; 
 *	 DistMatOK(N, 1, X)
 * post: returns the value of the first global index i of X s.t. 
 *	  	|X[i]| >= |X[j]|	, for 0<=j<N
 *	 and, if Max != NULL, sets *MAX = X[i], to all cells in the scope of X
 *	 where X is N x 1
 */

/********************** new parallel level 1 procedures  ********************/

void D_ScaleAdd STDC_PARAMS(
	(int M, int N, FDATA alpha, DistMat A, FDATA beta, DistMat C)
);
/* pre:  M, N > 0;  DistMatOK(M, N, A), DistMatOK(M, N, C), P_Same(A, C)
 * post: C = beta * C + alpha * A
 * notes: for DM_Trans(C) != DM_Trans(A), this procedure performs a
 *	  transpose operation. for alpha = 0.0, it performs a scaling operation.
 *	  for beta - 0.0, it performs a copy operation.
 */

FDATA D_GetElt STDC_PARAMS( (DistMat A) );
/* pre:  DistMatOK(1, 1, X)
 * post: returns the value of global A[0,0] to all cells in the scope of A
 */

void D_SetElt STDC_PARAMS( (DistMat A, FDATA v) );
/* pre:  DistMatOK(1, 1, X)
 * post: sets the value of global A[0,0] to v
 */

int D_CellHolds STDC_PARAMS( (int M, int N, DistMat A) );
/* pre:  DistMatOK(M, N, A)
 * post: returns whether this cell holds any portion of the 
 *	 M x N distributed matrix A
 */
          
#endif /*__D_BLAS_INCL__*/

