/* @(#) All Right Reserved, Copyright (c) 1997 Australian National University */
/* @(#)                     Copyright (c) 1997 Fujitsu Laboratories Ltd. */
/*  */
/*        */
/*                        Permission Notice */
/*  */
/*                Australian National University */
/*                Canberra ACT 0200 Australia */
/*                           and */
/*                Fujitsu Laboratories Ltd. */
/*                4-1-1 Kamikodanaka Nakahara */
/*                Kawasaki 211-88 Japan */
/*  */
/*   Permission to use, copy, modify and distribute this software (including  */
/*   its documentation) without fee is hereby granted under the following terms */
/*   and conditions: */
/*  */
/*   1. This permission notice and all of the copyright notices in this */
/*      software must appear in all copies and derived versions of all */
/*      or a part of this software. */
/*  */
/*   2. Sales of this software, whether incorporated in other products as a */
/*      part thereof or not, is prohibited. */
/*  */
/*   3. On publication or presentation in any form of the results obtained */
/*      by use of this software, an explicit reference to this software and */
/*      its copyright owner shall be made. */
/*  */
/*   4. THIS SOFTWARE IS PROVIDED 'AS IS' AND THE AUSTRALIAN NATIONAL UNIVERSITY */
/*      AND FUJITSU LABORATORIES LTD. MAKE NO WARRANTIES, EITHER EXPRESSED OR */
/*      IMPLIED, AS TO ANY MATTER WHATSOEVER REGARDING THIS SOFTWARE,  */
/*      INCLUDING, WITHOUT LIMITATION, WARRANTIES OF MERCHANTABILITY OR FITNESS */
/*      FOR ANY PARTICULAR PURPOSE. */
/*      IN NO EVENT SHALL THE AUSTRALIAN NATIONAL UNIVERSITY OR FUJITSU */
/*      LABORATORIES LTD. BE LIABLE FOR ANY CLAIM FROM A THIRD PARTY, OR */
/*      FOR SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES WHATSOEVER RESULTING */
/*      FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN ACTION OF CONTRACT, */
/*      NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION */
/*      WITH THE USE, COPYING OR MODIFICATION OF THIS SOFTWARE. */
/*  */

/************************************************************************/
/*         D_BLAS2.c: Level 2 Distributed BLAS implementation file	*/
/************************************************************************/
/* SCCS INFO @(#)D_BLAS2.c 4.6         last modified P. Strazdins, 97/05/21
   first created: P. Strazdins 93/03/23
 */
/* implementation file for Distributed Level-2 BLAS procedures; 
   See DBLAS.h for a general description of this module.
   (externally-defined) cpp FLAGs:
        FDATATYPE       SINGLE (default), DOUBLE, COMPLEX or DOUBLE_COMPLEX
        ASSERT, TRACE   turn assertion checking and procedure tracing on
                        in major internal procedure calls
*/   
#include "D_BLAS.h"
#include "D_BLASCheck.h"

#include "BLASOptions.h"
#include "FDATASetUp.h"
#include "FDATAAssert.h"

#include "LocMat.h"
#include "MatDim.h"
#include "DistMat.h"
#include "FDATACommun.h"  		/* CommSetUp() */


#ifdef COMP_D_GEMV
#include "GenMult.h"

void D_GEMV (M, N, alpha, A, X, beta, Y)
  int M; int N; FDATA alpha; DistMat A;
  DistMat X; FDATA beta; DistMat Y;
{ int ParamErr=0;			int  AlignM, AlignN;
  AlphaBetaDecln();  

  if (!NoOuterErrCheck) {
    CheckAlign(ParamErr, AlignN, N, DM_Inc(Tr(A), A), DM_DN(A),
               DM_Inc(1-Tr(X), X), DM_DM(X), 0x245);
    CheckParam(ParamErr, AlignN && d_P(DM_DM(X))==DM_P(Tr(X), A) &&
                         P_OK(d_P(DM_DN(X)), DM_P(1-Tr(X), A)), 0x45);
    CheckAlign(ParamErr, AlignM, M, DM_Inc(1-Tr(A), A),  DM_DM(A),
               DM_Inc(1-Tr(Y), Y),   DM_DM(Y), 0x147);
    CheckParam(ParamErr, AlignM && d_P(DM_DM(Y))==DM_P(Tr(Y), A) &&
                         P_OK(d_P(DM_DN(Y)), DM_P(1-Tr(Y), A)), 0x47);
  }

  SetUpComm(A);
  ERROR_CHECK_TR(
    (TraceStr, "(%d,%d,%1.1f,%s,%s,%1.1f,%s)", M, N, 
     alpha, DM_TRACE(0, A), DM_TRACE(1, X), beta, DM_TRACE(2, Y)),
     ParamErr, "GEMV", 1
  );  

  AlphaSet(alpha, pAlpha);              BetaSet(beta, pBeta);
  DistVecMatMult(1, M, N, pAlpha, Trans(X), Trans(A), pBeta, Trans(Y));
  TRACE_EXIT;
} /* D_GEMV() */

#endif /* COMP_D_GEMV */


#ifdef COMP_D_SYMV
#include "SYMMult.h"

void D_SYMV (UpLo, N, alpha, A, X, beta, Y)
  char  *UpLo; int N; FDATA alpha; DistMat A;
  DistMat X; FDATA beta; DistMat Y;
{ char UpLoA = UpperCase(*UpLo);
  int ParamErr=0;
  AlphaBetaDecln();
  int AlignM=1, AlignN=1;

  if (!NoOuterErrCheck) {
    CheckParam(ParamErr, UpLoOK(UpLoA), 0x1);
    CheckParam(ParamErr, Dim_BCompat(DM_D0(A), DM_D1(A)), 0x44);
    CheckAlign(ParamErr, AlignM, N, DM_Inc(1-Tr(A), A),   DM_DM(A),
               DM_Inc(1-Tr(Y), Y),   DM_DM(Y), 0x247);
    CheckParam(ParamErr, AlignM && d_P(DM_DM(A))==DM_P(Tr(A), Y) &&
                                   d_P(DM_DN(A))==DM_P(1-Tr(A), Y), 0x47);
    CheckAlign(ParamErr, AlignM, N, DM_Inc(Tr(A), A),     DM_DN(A),
               DM_Inc(1-Tr(X), X),  DM_DM(X), 0x45);
    CheckParam(ParamErr, AlignM, 0x45);
    CheckParam(ParamErr,
               PAlignCheck(d_P(DM_DN(X)), DM_P(1-Tr(X), Y), &AlignN), 0x57);
  }

  SetUpComm(Y);
  ERROR_CHECK_TR(
    (TraceStr, "(%c,%d,,%1.1f,%s,%s,%1.1f,%s)",
     *UpLo, N, alpha, DM_TRACE(0, A), DM_TRACE(1, X), beta, DM_TRACE(2, Y)),
    ParamErr, "SYMV", 0            
  );
    
  AlphaSet(alpha, pAlpha);              BetaSet(beta, pBeta);
  SYMMult(N, 1, pAlpha, UpLoA, A, X, pBeta, Y);
  TRACE_EXIT;

} /* D_SYMV() */

#endif /* COMP_D_SYMV */


#ifdef COMP_D_TRMV
#include "TriMatUpdate.h"

void D_TRMV (UpLo, Diag, N, A, X)
  char *UpLo; char *Diag; int N;
  DistMat A; DistMat X;
{ char UpLoA = UpperCase(*UpLo), 	DiagA = UpperCase(*Diag);
  int ParamErr=0;
  int AlignN=1;

  if (!NoOuterErrCheck) {
    CheckParam(ParamErr, UpLoOK(UpLoA), 0x1);
    CheckParam(ParamErr, DiagOK(DiagA), 0x2);
    CheckAlign(ParamErr, AlignN, N, DM_Inc(Tr(A), A),   DM_DN(A),
               DM_Inc(1-Tr(X), X),  DM_DM(X), 0x345);
    CheckDim(ParamErr, N, DM_Inc(1-Tr(A), A), DM_DM(A), 0x34);
    CheckParam(ParamErr, AlignN && d_P(DM_DM(A))==DM_P(Tr(A), X) &&
               		 DM_P(1-Tr(A), X)==d_P(DM_DN(A)), 0x45);
    CheckDim(ParamErr, N, DM_Inc(1-Tr(A), A), DM_DM(A), 0x35);
  }

  SetUpComm(A);
  ERROR_CHECK_TR(  
    (TraceStr, "(%c,%c,%d,%s,%s)", *UpLo, *Diag, 
     N, DM_TRACE(0, A), DM_TRACE(1,X)),
     ParamErr, "TRMV", 1
  );    

  if (DM_Trans(A)^DM_Trans(X))
    UpLoA = TransForm(UpLoA);
  TriMatUpdate(UpLoA, DiagA, N, 1, &FD_1, A, X);
  TRACE_EXIT;
} /* D_TRMV() */
#endif /* COMP_D_TRMV */


#ifdef COMP_D_TRSV
#include "TriMatSolve.h"

void D_TRSV (UpLo, Diag, N, A, X)
  char *UpLo; char *Diag; int N;
  DistMat A; DistMat X;
{ char UpLoA = UpperCase(*UpLo), 	DiagA = UpperCase(*Diag);
  int ParamErr=0;
  int AlignN=1;

  if (!NoOuterErrCheck) {
    CheckParam(ParamErr, UpLoOK(UpLoA), 0x1);
    CheckParam(ParamErr, DiagOK(DiagA), 0x2);
    CheckAlign(ParamErr, AlignN, N, DM_Inc(Tr(A), A),   DM_DN(A),
               DM_Inc(1-Tr(X), X),  DM_DM(X), 0x345);
    CheckDim(ParamErr, N, DM_Inc(1-Tr(A), A), DM_DM(A), 0x34);
    CheckParam(ParamErr, AlignN && d_P(DM_DM(A))==DM_P(Tr(A), X) &&
               		 DM_P(1-Tr(A), X)==d_P(DM_DN(A)), 0x45);
    CheckDim(ParamErr, N, DM_Inc(1-Tr(A), A), DM_DM(A), 0x35);
  }

  SetUpComm(A);
  ERROR_CHECK_TR(  
    (TraceStr, "(%c,%c,%d,%s,%s)", *UpLo, *Diag, 
     N, DM_TRACE(0, A), DM_TRACE(1,X)),
     ParamErr, "TRSV", 1
  );    

  if (DM_Trans(A)^DM_Trans(X))
    UpLoA = TransForm(UpLoA);
  TriMatSolve1Blk(UpLoA, DiagA, N, 1, &FD_1, A, X);
  TRACE_EXIT;
} /* D_TRSV() */
#endif /* COMP_D_TRSV */


#ifdef COMP_D_GER
#include "GenMult.h"

void D_GER (M, N,  alpha, X, Y, A)
  int M; int N;  FDATA alpha; DistMat X; DistMat Y; DistMat A;
{ int ParamErr=0;
  int AlignM, AlignN;
  AlphaDecln();

  if (!NoOuterErrCheck) {
    CheckAlign(ParamErr, AlignM, M, DM_Inc(1-Tr(X), X),   DM_DM(X),
               DM_Inc(1-Tr(A), A),   DM_DM(A), 0x146);
    CheckParam(ParamErr, AlignM && d_P(DM_DM(X))==DM_P(Tr(X), A) &&
                         P_OK(d_P(DM_DN(X)), DM_P(1-Tr(X), A)), 0x46);
    CheckAlign(ParamErr, AlignN, N, DM_Inc(1-Tr(Y), Y), DM_DM(Y),
               DM_Inc(Tr(A), A), DM_DN(A), 0x256);
    CheckParam(ParamErr, AlignN && d_P(DM_DM(Y))==DM_P(Tr(Y), A) &&
                         P_OK(d_P(DM_DN(Y)), DM_P(1-Tr(Y), A)), 0x56);
  }

  SetUpComm(A);
  ERROR_CHECK_TR(   
    (TraceStr, "(%d,%d,%1.1f,%s,%s,%s)", M, N, alpha, 
    DM_TRACE(0, X), DM_TRACE(1, Y), DM_TRACE(2, A)),
    ParamErr,  "GER", 1
  );

  AlphaSet(alpha, pAlpha); 
  DistRankKMult(M, N, 1, pAlpha, X, Trans(Y), NULL, A);
  TRACE_EXIT;
} /* D_GER() */

#endif /*COMP_D_GER*/


#ifdef COMP_D_SYR
#include "SYRKMult.h"

void D_SYR (UpLo, N, alpha, X, A)
  char *UpLo; int N; FDATA alpha; DistMat X; DistMat A;
{ int ParamErr=0;			int AlignN;
  char UpLoA = UpperCase(*UpLo); 
  AlphaDecln();

  if (!NoOuterErrCheck) {
    CheckParam(ParamErr, UpLoOK(UpLoA), 0x1);  
    CheckAlign(ParamErr, AlignN, N, DM_Inc(1-Tr(X), X),   DM_DM(X),
               DM_Inc(1-Tr(A), A),   DM_DM(A), 0x145);
    CheckParam(ParamErr, AlignN && d_P(DM_DM(X))==DM_P(Tr(X), A) &&
                         P_OK(d_P(DM_DN(X)), DM_P(1-Tr(X), A)), 0x45);
    CheckDim(ParamErr, N, DM_Inc(Tr(A), A), DM_DN(A), 0x25);
    CheckParam(ParamErr, Dim_BCompat(DM_D0(A), DM_D1(A)), 0x55);
  }

  SetUpComm(A);
  ERROR_CHECK_TR(
    (TraceStr, "(%c,%d,%1.1f,%s,%s)", 
     *UpLo, N, alpha, DM_TRACE(0, X), DM_TRACE(1,A)),
    ParamErr, "SYR", 1
  );

  AlphaSet(alpha, pAlpha);
  DistSYRKMult(N, 1, pAlpha, X, NULL, UpLoA, A);
  TRACE_EXIT;
} /* D_SYR() */

#endif /* COMP_D_SYR */


#ifdef COMP_D_SYR2
#include "SYR2KMult.h"

void D_SYR2 (UpLo, N, alpha, X, Y, A)
  char *UpLo; int N; FDATA alpha; DistMat X; DistMat Y; DistMat A;
{ char UpLoA = UpperCase(*UpLo); 
  int ParamErr=0;
  AlphaBetaDecln();
  int AlignN=1;

  if (!NoOuterErrCheck) {
    CheckParam(ParamErr, UpLoOK(UpLoA), 0x1);
    CheckAlign(ParamErr, AlignN, N, DM_Inc(1-Tr(X), X),   DM_DM(X),
               DM_Inc(1-Tr(A), A),   DM_DM(A), 0x146);
    CheckParam(ParamErr, AlignN && d_P(DM_DM(X))==DM_P(Tr(X), A) &&
                         P_OK(d_P(DM_DN(X)), DM_P(1-Tr(X), A)), 0x46);
    CheckDim(ParamErr, N, DM_Inc(1-Tr(Y), Y), DM_DM(Y), 0x25);
    CheckParam(ParamErr, Dim_BCompat(DM_DM(Y), DM_DM(A)) &&
                         d_P(DM_DM(Y))==DM_P(Tr(Y), A) &&
                         P_OK(d_P(DM_DN(Y)), DM_P(1-Tr(Y), A)), 0x56);
    CheckDim(ParamErr, N, DM_Inc(Tr(A), A), DM_DN(A), 0x26);
    CheckParam(ParamErr, Dim_BCompat(DM_D0(A), DM_D1(A)), 0x66);
  }  

  SetUpComm(A);
  ERROR_CHECK_TR(   
    (TraceStr, "(%c,%d,%1.1f,%s,%s,%s)", *UpLo, N, alpha, 
     DM_TRACE(0, X), DM_TRACE(1,Y), DM_TRACE(2, A)),
     ParamErr, "SYR2", 1
  ); 

  AlphaSet(alpha, pAlpha);
  DistSYR2KMult(N, 1, pAlpha, X, Y, NULL, UpLoA, A);
  TRACE_EXIT;
} /* D_SYR2() */
#endif /* COMP_D_SY2R */
