/* @(#) All Right Reserved, Copyright (c) 1997 Australian National University */
/* @(#)                     Copyright (c) 1997 Fujitsu Laboratories Ltd. */
/*  */
/*        */
/*                        Permission Notice */
/*  */
/*                Australian National University */
/*                Canberra ACT 0200 Australia */
/*                           and */
/*                Fujitsu Laboratories Ltd. */
/*                4-1-1 Kamikodanaka Nakahara */
/*                Kawasaki 211-88 Japan */
/*  */
/*   Permission to use, copy, modify and distribute this software (including  */
/*   its documentation) without fee is hereby granted under the following terms */
/*   and conditions: */
/*  */
/*   1. This permission notice and all of the copyright notices in this */
/*      software must appear in all copies and derived versions of all */
/*      or a part of this software. */
/*  */
/*   2. Sales of this software, whether incorporated in other products as a */
/*      part thereof or not, is prohibited. */
/*  */
/*   3. On publication or presentation in any form of the results obtained */
/*      by use of this software, an explicit reference to this software and */
/*      its copyright owner shall be made. */
/*  */
/*   4. THIS SOFTWARE IS PROVIDED 'AS IS' AND THE AUSTRALIAN NATIONAL UNIVERSITY */
/*      AND FUJITSU LABORATORIES LTD. MAKE NO WARRANTIES, EITHER EXPRESSED OR */
/*      IMPLIED, AS TO ANY MATTER WHATSOEVER REGARDING THIS SOFTWARE,  */
/*      INCLUDING, WITHOUT LIMITATION, WARRANTIES OF MERCHANTABILITY OR FITNESS */
/*      FOR ANY PARTICULAR PURPOSE. */
/*      IN NO EVENT SHALL THE AUSTRALIAN NATIONAL UNIVERSITY OR FUJITSU */
/*      LABORATORIES LTD. BE LIABLE FOR ANY CLAIM FROM A THIRD PARTY, OR */
/*      FOR SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES WHATSOEVER RESULTING */
/*      FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN ACTION OF CONTRACT, */
/*      NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION */
/*      WITH THE USE, COPYING OR MODIFICATION OF THIS SOFTWARE. */
/*  */

/************************************************************************/
/*         D_BLASCheck: Distributed BLAS error checking macros		*/
/************************************************************************/
/* SCCS INFO @(#)D_BLASCheck.h 1.1         last modified P. Strazdins, 97/05/14
   first created: P. Strazdins 97/02
 */
/* common macro definitions used to simplify error checking in outer level
 * DBLAS procedures
*/   

#include "LocMat.h"
#include "MatDim.h"
#include "DistMat.h"

#ifndef _D_BLASCheck_INCL_
#define _D_BLASCheck_INCL_

#define UpperCase(c) (('a' <= (c) && (c) <= 'z')? (c)+'A'-'a': (c))

#define DimOK(M) ((M)>=0)

#define Tr(A)		DM_Trans(A)
#define DM_Inc(t, A)	LM_Inc(t, DM_Loc(A))
#define DM_P(t, C)	d_P(DM_D(t, C))

#define CheckParam(ParamErr, ParamCond, ParamNum) 		\
	(ParamErr) = ((ParamErr)==0 && !(ParamCond))? (ParamNum): (ParamErr);

#define CheckAlign(ParamErr, AlignN, N, MIncA, DNA, MIncB, DNB, ParamNums) \
	if (!ParamErr) {						\
	  (ParamErr) = (ParamNums);					\
	  AlignN = CommonDimAlignCheck(N, MIncA, DNA, MIncB, DNB, &(ParamErr));\
	}

#define CheckDim(ParamErr, N, MIncA, DNA, ParamNums) 			\
	if (!ParamErr) {						\
	  (ParamErr) = (ParamNums);					\
	  CommonDimAlignCheck(N, MIncA, DNA, 0, NULL, &(ParamErr));\
	}

#define P_OK(P_ADN, P_Aref)	((P_ADN)==(P_Aref) || (P_ADN)==1)

static inline int PAlignCheck(P_ADN, P_Aref, NAlign)
  int P_ADN, P_Aref, *NAlign;
{ int NAlignA = P_ADN==P_Aref;
  *NAlign &= NAlignA;
  return (NAlignA || P_ADN==1);
}
	
#define DM_TRACE(n, A)	PrintDistMat(A, TmpTraceStrs[n])

#define SetUpComm(C)	CommSetUp(d_P(DM_D0(C)), d_P(DM_D1(C)))

/*copy of parameter alpha etc needed for workaround apparent Sun Sys V cc bug*/
#define AlphaDecln()			FDATA Alpha, *pAlpha
#define BetaDecln()			FDATA Beta, *pBeta
#define AlphaBetaDecln()		AlphaDecln(); BetaDecln()
#define AlphaSet(alpha, pAlpha)		Alpha=alpha, pAlpha=&Alpha
#define BetaSet(beta, pBeta)		pBeta= beta==1.0? NULL: (Beta=beta, &Beta)

#endif _D_BLASCheck_INCL_

