/* @(#) All Right Reserved, Copyright (c) 1997 Australian National University */
/* @(#)                     Copyright (c) 1997 Fujitsu Laboratories Ltd. */
/*  */
/*        */
/*                        Permission Notice */
/*  */
/*                Australian National University */
/*                Canberra ACT 0200 Australia */
/*                           and */
/*                Fujitsu Laboratories Ltd. */
/*                4-1-1 Kamikodanaka Nakahara */
/*                Kawasaki 211-88 Japan */
/*  */
/*   Permission to use, copy, modify and distribute this software (including  */
/*   its documentation) without fee is hereby granted under the following terms */
/*   and conditions: */
/*  */
/*   1. This permission notice and all of the copyright notices in this */
/*      software must appear in all copies and derived versions of all */
/*      or a part of this software. */
/*  */
/*   2. Sales of this software, whether incorporated in other products as a */
/*      part thereof or not, is prohibited. */
/*  */
/*   3. On publication or presentation in any form of the results obtained */
/*      by use of this software, an explicit reference to this software and */
/*      its copyright owner shall be made. */
/*  */
/*   4. THIS SOFTWARE IS PROVIDED 'AS IS' AND THE AUSTRALIAN NATIONAL UNIVERSITY */
/*      AND FUJITSU LABORATORIES LTD. MAKE NO WARRANTIES, EITHER EXPRESSED OR */
/*      IMPLIED, AS TO ANY MATTER WHATSOEVER REGARDING THIS SOFTWARE,  */
/*      INCLUDING, WITHOUT LIMITATION, WARRANTIES OF MERCHANTABILITY OR FITNESS */
/*      FOR ANY PARTICULAR PURPOSE. */
/*      IN NO EVENT SHALL THE AUSTRALIAN NATIONAL UNIVERSITY OR FUJITSU */
/*      LABORATORIES LTD. BE LIABLE FOR ANY CLAIM FROM A THIRD PARTY, OR */
/*      FOR SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES WHATSOEVER RESULTING */
/*      FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN ACTION OF CONTRACT, */
/*      NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION */
/*      WITH THE USE, COPYING OR MODIFICATION OF THIS SOFTWARE. */
/*  */

/************************************************************************/
/*	DistMat: distributed block-cyclic matrices definition		*/
/************************************************************************/
/* SCCS INFO @(#)DistMat.c 4.6         last modified P. Strazdins, 97/04/18
   first created: P. Strazdins 94/05/17
*/

#include "DistMat.h"
#include "MatDim.h"
#include "FDATAAssert.h"
#include "FDATAMalloc.h"

DistMat InitDimMat_(A, Trans, B_0, B_1, P0_0, P_0, P0_1, P_1, di, dj)
  DistMat A; int Trans, B_0, B_1, P0_0, P_0, P0_1, P_1, di, dj;
{ FDATA *Ax;
  pInitDim(DM_D0(A), 0, B_0, P0_0, P_0);
  pSubDim_(DM_D0(A), DM_D0(A), di);
  pInitDim(DM_D1(A), 0, B_1, P0_1, P_1);
  pSubDim_(DM_D1(A), DM_D1(A), dj);
  DM_SetTrans(Trans, A);
  return (A);
} /* InitDimMat_() */


DistMat SubMat_(Aij, A, i, j)
  DistMat Aij, A; int i, j;
{ if (DM_Trans(A)) 
    SWAP(int, i, j);    
  if (Aij != A) {
    *DM_Ptr(Aij) = *DM_Ptr(A);
    DM_SetTrans(DM_Trans(A), Aij);
  }
#define Ax LM_ADR(DM_Loc(Aij))
  if (i!=0) {
      if (i==-1)
        i = G2LSubDimM1(DM_D0(Aij));	
      else 
        i = G2LSubDim(i, DM_D0(Aij));
      if (Ax!=NULL) {
        Ax += FAST_MUL2(i, LM_Inc(0, DM_Loc(Aij)));    
      }                      
    }  /* if (i!=0) */ 
  if (j!=0) {
      if (j==-1)
        j = G2LSubDimM1(DM_D1(Aij));	
      else  
        j = G2LSubDim(j, DM_D1(Aij));
      if (Ax!=NULL) {
        Ax += FAST_MUL2(j, LM_Inc(1, DM_Loc(Aij)));    
      }                      
    }  /* if (i!=0) */ 
  return (Aij);
} /* SubMat_() */


DistMat ReduceDimMat_( VA, RowCol, A )
  int RowCol; DistMat VA, A;
{ 
  DM_SetTrans(DM_Trans(A), VA);
  DM_Loc(VA) = DM_Loc(A);
  if (RowCol==-1) {
    MatDimUnit(DM_D0(VA));
    MatDimUnit(DM_D1(VA)); 
    return(VA);
  }   	
  RowCol = RowCol^DM_Trans(A);
  MatDimUnit(DM_D(RowCol, VA));
  MatDimCp(DM_D(1-RowCol, VA), DM_D(1-RowCol, A));
  return (VA); 
} /* ReduceDimMat_() */


DistMat Spread2DistMat_(B, RowCol, Trans, DN, sB)
  DistMat B; int  RowCol, Trans; pMatDim DN; pLocMat sB;
{ DM_SetTrans(Trans, B);
  DM_Loc(B) = *sB;  
  MatDimUnit(DM_D(1-RowCol, B));
  MatDimCp(DM_D(RowCol, B), DN);
  return(B);
} /*Spread2DistMat_()*/
 

DistMat NewMat_(NewC, M, N, C, WS)
  int M, N; DistMat NewC, C; FDATA *WS;
{ int m = G2Ld(M, DM_DM(C)),    n = G2Ld(N, DM_DN(C));
  LocMat lC; 
  if (n==1) n=2;			/* avoid degenerate row-major strides */
  if (WS==NULL)
    MatMalloc_F("NewMat", 1, &WS, m*n);
  if (n==0) 
    n = 2;
  if (DM_Trans(C))  
    LM_Init(DM_Loc(NewC), WS, 1, n);
  else
    LM_Init(DM_Loc(NewC), WS, n, 1);
  MatDimCp(DM_D0(NewC), DM_D0(C));
  MatDimCp(DM_D1(NewC), DM_D1(C));
  DM_SetTrans(DM_Trans(C), NewC);
  return(NewC);
} /*NewMat_()*/


char *PrintDistMat( A, pA )
  DistMat A; char *pA; 
{ 
  sprintf(pA, DM_SPRINTF_FORMAT, DM_SPRINTF_PARAMS(A)); 
  return (pA);
} /* PrintDistMat() */ 
