/* @(#) All Right Reserved, Copyright (c) 1997 Australian National University */
/* @(#)                     Copyright (c) 1997 Fujitsu Laboratories Ltd. */
/*  */
/*        */
/*                        Permission Notice */
/*  */
/*                Australian National University */
/*                Canberra ACT 0200 Australia */
/*                           and */
/*                Fujitsu Laboratories Ltd. */
/*                4-1-1 Kamikodanaka Nakahara */
/*                Kawasaki 211-88 Japan */
/*  */
/*   Permission to use, copy, modify and distribute this software (including  */
/*   its documentation) without fee is hereby granted under the following terms */
/*   and conditions: */
/*  */
/*   1. This permission notice and all of the copyright notices in this */
/*      software must appear in all copies and derived versions of all */
/*      or a part of this software. */
/*  */
/*   2. Sales of this software, whether incorporated in other products as a */
/*      part thereof or not, is prohibited. */
/*  */
/*   3. On publication or presentation in any form of the results obtained */
/*      by use of this software, an explicit reference to this software and */
/*      its copyright owner shall be made. */
/*  */
/*   4. THIS SOFTWARE IS PROVIDED 'AS IS' AND THE AUSTRALIAN NATIONAL UNIVERSITY */
/*      AND FUJITSU LABORATORIES LTD. MAKE NO WARRANTIES, EITHER EXPRESSED OR */
/*      IMPLIED, AS TO ANY MATTER WHATSOEVER REGARDING THIS SOFTWARE,  */
/*      INCLUDING, WITHOUT LIMITATION, WARRANTIES OF MERCHANTABILITY OR FITNESS */
/*      FOR ANY PARTICULAR PURPOSE. */
/*      IN NO EVENT SHALL THE AUSTRALIAN NATIONAL UNIVERSITY OR FUJITSU */
/*      LABORATORIES LTD. BE LIABLE FOR ANY CLAIM FROM A THIRD PARTY, OR */
/*      FOR SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES WHATSOEVER RESULTING */
/*      FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN ACTION OF CONTRACT, */
/*      NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION */
/*      WITH THE USE, COPYING OR MODIFICATION OF THIS SOFTWARE. */
/*  */

/************************************************************************/
/*                     FDATAAssPre.h		                        */
/************************************************************************/
/* SCCS INFO @(#)FDATAAssPre.h 2.1         last modified P. Strazdins, 97/01/10
   first created: P. Strazdins DCS ANU 92/02
*/
/* header file for BLAS-3 Assembly routine calls with precondition
	checking (if ASSERT_ASS defined) & tracing
	(needed as these cannot be done within assembler leaf routines)

	assumes the following (macros) have been declared in-line:
		char TR_STR[128];
		int ASSERT_PROC(int cond, char *str);

  	written by P. Strazdins, from 15/12/91				*/
/************************************************************************/

#ifndef __FDATAAssPre_INCL__
#define __FDATAAssPre_INCL__

   
#ifdef ASSERT_ASS

#  define TR_Rank1Update(m, n, bounds, A, B, C) \
	sprintf(TR_STR, "(%d,%d,[%d,%d,%d],a%d,a%d,a%d)", \
		m, n, bounds[0], bounds[1], bounds[2], \
		(int)(A)/sizeof(FDATA), (int)(B)/sizeof(FDATA), (int)(C)/sizeof(FDATA) ); \
	ASSERT_PROC(1<=bounds[0] && 1<=bounds[1] && 1<=bounds[2] && \
		(FDATATYPE!=SINGLE || bounds[2]%2==0), \
		"Rank1Update"); \
	Rank1Update(m, n, bounds, A, B, C);
	
#  define TR_Rank1UpdateA(m, n, bounds, A, B, C) \
	sprintf(TR_STR, "(%d,%d,[%d,%d,%d],a%d,a%d,a%d)", \
		m, n, bounds[0], bounds[1], bounds[2], \
		(int)(A)/sizeof(FDATA), (int)(B)/sizeof(FDATA), (int)(C)/sizeof(FDATA)); \
	ASSERT_PROC(1<=bounds[0] && 1<=bounds[1] && (n)<=bounds[2] && \
		(FDATATYPE!=SINGLE || bounds[2]%2==0), "Rank1UpdateA"); \
	Rank1UpdateA(m, n, bounds, A, B, C);

#  define TR_MatVecMul(m, k, bounds, A, B, C) \
	sprintf(TR_STR, "(%d,%d,[%d,%d,%d],%d,%d,%d)", \
		m, k, bounds[0], bounds[1], bounds[2], \
		(int)(A)/sizeof(FDATA), (int)(B)/sizeof(FDATA), (int)(C)/sizeof(FDATA)); \
	ASSERT_PROC(k<=bounds[0] && 1<=bounds[1] && 1<=bounds[2] && \
		(FDATATYPE!=SINGLE || bounds[0]%2==0), "MatVecMul"); \
	MatVecMul(m, k, bounds, A, B, C);

#  define TR_VecMatMul(k, n, bounds, A, B, C) \
	sprintf(TR_STR, "(%d,%d,[%d,%d,%d],a%d,a%d,a%d)", \
		k, n, bounds[0], bounds[1], bounds[2], \
		(int)(A)/sizeof(FDATA), (int)(B)/sizeof(FDATA), (int)(C)/sizeof(FDATA)); \
	ASSERT_PROC(1<=bounds[0] && n<=bounds[1] && 1<=bounds[2] \
		&& (FDATATYPE!=SINGLE || bounds[1]%2==0), \
		"VecMatMul"); \
	VecMatMul(k, n, bounds, A, B, C);

#  define TR_UpdateRect(k, TrC, bounds, A, B, C) \
	sprintf(TR_STR, "(%d,%d,[%d,%d,%d],a%d,a%d,a%d)", \
		k, TrC, bounds[0], bounds[1], bounds[2],\
		(int)(A)/sizeof(FDATA), (int)(B)/sizeof(FDATA), (int)(C)/sizeof(FDATA)); \
	ASSERT_PROC((k)<=bounds[0] && RECT_X<=bounds[1] && RECT_X<=bounds[2] && \
		(FDATATYPE!=SINGLE || bounds[0]%2+bounds[1]%2==0), "UpdateRect"); \
	UpdateRect(k, TrC, bounds, A, B, C);

#  define TR_UpdateRectTr1(k, bounds, A, B, C) \
	sprintf(TR_STR, "(%d,[%d,%d,%d],a%d,a%d,a%d)", \
		k, bounds[0], bounds[1], bounds[2], \
		(int)(A)/sizeof(FDATA), (int)(B)/sizeof(FDATA), (int)(C)/sizeof(FDATA)); \
	ASSERT_PROC(RECT_Y<=bounds[0] && RECT_X<=bounds[1] && \
		RECT_X<=bounds[2] && \
		(FDATATYPE!=SINGLE || bounds[0]%2+bounds[1]%2==0), "UpdateRectTr1"); \
	UpdateRectTr1(k, bounds, A, B, C);

#  define TR_UpdateRectTr2(k, bounds, A, B, C) \
	sprintf(TR_STR, "(%d,[%d,%d,%d],a%d,a%d,a%d)", \
		k, bounds[0], bounds[1], bounds[2], \
		(int)(A)/sizeof(FDATA), (int)(B)/sizeof(FDATA), (int)(C)/sizeof(FDATA)); \
	ASSERT_PROC(RECT_X<=bounds[2] && \
		(FDATATYPE!=SINGLE || bounds[0]%2+bounds[1]%2==0), "UpdateRectTr2"); \
	UpdateRectTr2(k, bounds, A, B, C);

#else

#  define TR_Rank1Update(mLdA, n, bounds, A, B, C) \
	Rank1Update(mLdA, n, bounds, A, B, C);
	
#  define TR_Rank1UpdateA(m, n, bounds, A, B, C) \
	Rank1UpdateA(m, n, bounds, A, B, C);

#  define TR_MatVecMul(m, k, bounds, A, B, C) \
	MatVecMul(m, k, bounds, A, B, C);

#  define TR_VecMatMul(kLdA, n, bounds, A, B, C) \
	VecMatMul(kLdA, n, bounds, A, B, C);

#  define TR_UpdateRect(k, TrC, bounds, A, B, C) \
	UpdateRect(k, TrC, bounds, A, B, C);

#  define TR_UpdateRectTr1(kLdA, bounds, A, B, C) \
	UpdateRectTr1(kLdA, bounds, A, B, C);

#  define TR_UpdateRectTr2(k, bounds, A, B, C) \
	UpdateRectTr2(k, bounds, A, B, C);

#endif /*ASSERT_ASS*/
 
#endif /*__FDATAAssPre_INCL__*/
