/* @(#) All Right Reserved, Copyright (c) 1997 Australian National University */
/* @(#)                     Copyright (c) 1997 Fujitsu Laboratories Ltd. */
/*  */
/*        */
/*                        Permission Notice */
/*  */
/*                Australian National University */
/*                Canberra ACT 0200 Australia */
/*                           and */
/*                Fujitsu Laboratories Ltd. */
/*                4-1-1 Kamikodanaka Nakahara */
/*                Kawasaki 211-88 Japan */
/*  */
/*   Permission to use, copy, modify and distribute this software (including  */
/*   its documentation) without fee is hereby granted under the following terms */
/*   and conditions: */
/*  */
/*   1. This permission notice and all of the copyright notices in this */
/*      software must appear in all copies and derived versions of all */
/*      or a part of this software. */
/*  */
/*   2. Sales of this software, whether incorporated in other products as a */
/*      part thereof or not, is prohibited. */
/*  */
/*   3. On publication or presentation in any form of the results obtained */
/*      by use of this software, an explicit reference to this software and */
/*      its copyright owner shall be made. */
/*  */
/*   4. THIS SOFTWARE IS PROVIDED 'AS IS' AND THE AUSTRALIAN NATIONAL UNIVERSITY */
/*      AND FUJITSU LABORATORIES LTD. MAKE NO WARRANTIES, EITHER EXPRESSED OR */
/*      IMPLIED, AS TO ANY MATTER WHATSOEVER REGARDING THIS SOFTWARE,  */
/*      INCLUDING, WITHOUT LIMITATION, WARRANTIES OF MERCHANTABILITY OR FITNESS */
/*      FOR ANY PARTICULAR PURPOSE. */
/*      IN NO EVENT SHALL THE AUSTRALIAN NATIONAL UNIVERSITY OR FUJITSU */
/*      LABORATORIES LTD. BE LIABLE FOR ANY CLAIM FROM A THIRD PARTY, OR */
/*      FOR SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES WHATSOEVER RESULTING */
/*      FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN ACTION OF CONTRACT, */
/*      NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION */
/*      WITH THE USE, COPYING OR MODIFICATION OF THIS SOFTWARE. */
/*  */

/************************************************************************/
/*       FDATACellBlkMatrix.c: FDATA Block Matrix procedures for cells  */
/************************************************************************/
/* SCCS INFO @(#)FDATACellBlkMatrix.c 4.2	last modified P. Strazdins, 97/02/19
   first created: P. Strazdins 94/08/19 
*/
/* set of non-communicating matrix routines used for DBLAS.
 * These routines can be run on cells separately, but the operations
 * assume block-cyclic storage, eg. stride block copy routines.
 * compilation: macro FDATATYPE must be defined;
 *              if macro ASSERT defined, assertion checking is turned on
 */
#include "FDATACellBlkMatrix.h"
#include "FDATAAssert.h"  
#include "MatDim.h"
#include "FDATACellMatCopy.h"
#define NO_ALIGN
#undef __FDATAMatrixCopy_INCL__
#include "FDATACellMatCopy.h"
#undef NO_ALIGN

/********************* rectangular block copy routines  *******************/ 

void BlkMatCopy(Add, M, N, A, alpha, B,
		RB0, RB, RBIncA, RBIncB, CB0, CB, CBIncA, CBIncB)
  int Add, M, N; FDATA *alpha; pLocMat A, B;
  int RB0, RB, RBIncA, RBIncB, CB0, CB, CBIncA, CBIncB;
{ int i, j;  int BlkDim;
  void (*CopyFn) STDC_PARAMS((int, int, pLocMat, FDATA *, pLocMat));
  FDATA *Ax = pLM_ADR(A), *Bx = pLM_ADR(B);
  
  COMMON_ASSERT_TR_E(
    (TraceStr, "(%d,%dx%d,a%d %d %d,%1.1f,a%d %d %d,%d %d %d|%d,%d %d %d|%d)",
     Add, M, N, FD_ADR(pLM_ADR(A)), pLM_RInc(A), pLM_CInc(A), pFD_VAL(alpha), 
     FD_ADR(pLM_ADR(B)), pLM_RInc(B), pLM_CInc(B), 
     RB0, RB, RBIncA, RBIncB, CB0, CB, CBIncA, CBIncB),
    pLM_BlkIncOK(M, N, B, RB0, RB, RBIncB, CB0, CB, CBIncB) &&
    pLM_BlkIncOK(M, N, A, RB0, RB, RBIncA, CB0, CB, CBIncA),
    0, "BlkMatCopy", 3, TRACE_COMPUTE
  ); 
  TRACE_LEVEL_SHIFT(1);			/* suppress traces in MatCopy()*/
  
  if (((int) Ax)%sizeof(FDATA) || ((int) Bx)%sizeof(FDATA))
     CopyFn = Add? MatAddW: MatCopyW;
   else
     CopyFn = Add? MatAdd: MatCopy;
       
  BlkDim = 3;  				/* optimize for unit block stride;
  					   typically, at least 1 will be unit */
  if (M+RB0<=RB || (RBIncA==1 && RBIncB==1)) { 
    RB0 = 0; RB = 1;		RBIncA = RBIncB = 1;
    BlkDim-=2;
  }
  if (N+CB0<=CB || (CBIncA==1 && CBIncB==1)) { 
    CB0 = 0; CB = 1;		CBIncA = CBIncB = 1;	
    BlkDim--;
  }
  if (BlkDim == 0)			/* degenerate case of block copy */
    (*CopyFn)( M, N, A, alpha, B );        
  else {
    int RIncA=pLM_RInc(A), CIncA=pLM_CInc(A), 
        RIncB=pLM_RInc(B), CIncB=pLM_CInc(B);        
    LocMat A_, B_;			/* matrices of expanded stride */
    int RInc, CInc;
#define GET_STRIDE_BLK(Inc, B, BInc, RInc) \
    (Inc) = FAST_MUL(B, BInc); (Inc) = FAST_MUL(Inc, RInc);
         
    GET_STRIDE_BLK(RInc, RB, RBIncA, RIncA);
    GET_STRIDE_BLK(CInc, CB, CBIncA, CIncA);
    A = pMkLocMat(&A_, Ax, RInc, CInc);	

    GET_STRIDE_BLK(RInc, RB, RBIncB, RIncB);
    GET_STRIDE_BLK(CInc, CB, CBIncB, CIncB);
    B = pMkLocMat(&B_, Bx, RInc, CInc);

    for (i=0; i < RB; i++) {
      int m = FAST_DVD(M-i + RB-1, RB);		if (m==0) break;
      if (i+RB0 == RB) {
        Ax += RB*(RBIncA-1)*RIncA;
        Bx += RB*(RBIncB-1)*RIncB;
      }
      pLM_ADR(A) = Ax; 			pLM_ADR(B) = Bx;	
      for (j=0; j < CB; j++) {
        int n = FAST_DVD(N-j + CB-1, CB);	if (n==0) break;
        if (j+CB0 == CB) {
          pLM_ADR(A) += CB*(CBIncA-1)*CIncA;
          pLM_ADR(B) += CB*(CBIncB-1)*CIncB;
        }  
        (*CopyFn)( m, n, A, alpha, B );
        pLM_ADR(A) += CIncA; 		pLM_ADR(B) += CIncB;		    
      }
      Ax += RIncA; 			Bx += RIncB;
    } /* for (i...) */ 
  }
  TRACE_LEVEL_SHIFT(-1);
  TRACE_EXIT;
} /*BlkMatCopy()*/
 

