/* @(#) All Right Reserved, Copyright (c) 1997 Australian National University */
/* @(#)                     Copyright (c) 1997 Fujitsu Laboratories Ltd. */
/*  */
/*        */
/*                        Permission Notice */
/*  */
/*                Australian National University */
/*                Canberra ACT 0200 Australia */
/*                           and */
/*                Fujitsu Laboratories Ltd. */
/*                4-1-1 Kamikodanaka Nakahara */
/*                Kawasaki 211-88 Japan */
/*  */
/*   Permission to use, copy, modify and distribute this software (including  */
/*   its documentation) without fee is hereby granted under the following terms */
/*   and conditions: */
/*  */
/*   1. This permission notice and all of the copyright notices in this */
/*      software must appear in all copies and derived versions of all */
/*      or a part of this software. */
/*  */
/*   2. Sales of this software, whether incorporated in other products as a */
/*      part thereof or not, is prohibited. */
/*  */
/*   3. On publication or presentation in any form of the results obtained */
/*      by use of this software, an explicit reference to this software and */
/*      its copyright owner shall be made. */
/*  */
/*   4. THIS SOFTWARE IS PROVIDED 'AS IS' AND THE AUSTRALIAN NATIONAL UNIVERSITY */
/*      AND FUJITSU LABORATORIES LTD. MAKE NO WARRANTIES, EITHER EXPRESSED OR */
/*      IMPLIED, AS TO ANY MATTER WHATSOEVER REGARDING THIS SOFTWARE,  */
/*      INCLUDING, WITHOUT LIMITATION, WARRANTIES OF MERCHANTABILITY OR FITNESS */
/*      FOR ANY PARTICULAR PURPOSE. */
/*      IN NO EVENT SHALL THE AUSTRALIAN NATIONAL UNIVERSITY OR FUJITSU */
/*      LABORATORIES LTD. BE LIABLE FOR ANY CLAIM FROM A THIRD PARTY, OR */
/*      FOR SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES WHATSOEVER RESULTING */
/*      FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN ACTION OF CONTRACT, */
/*      NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION */
/*      WITH THE USE, COPYING OR MODIFICATION OF THIS SOFTWARE. */
/*  */

/************************************************************************/
/*       	FDATACellBlkMatrix: FDATA Matrix procedures for cells   */
/************************************************************************/
/* SCCS INFO @(#)FDATACellBlkMatrix.h 4.1         last modified P. Strazdins, 97/02/11
   first created: P. Strazdins 93/03/23 
*/
/* set of non-communicating matrix routines used for DBLAS.
 * These routines can be run on cells separately, but the operations
 * assume block-cyclic storage, eg. stride block copy routines.
 * See FDATACellMatrix.h for description of matrix strides.
 * A block-stride is defined by:
 *       B0 (offset within 1st block),
 *       B  (width of block)
 *       BInc (block-stride)
 * See BlkIncOK() for contraints in the stride abd block parameters.
 */
#ifndef __FDATACellBlkMat_INCL__
#define __FDATACellBlkMat_INCL__  

#include "FDATASetUp.h"
#include "LocMat.h"
#include "MatDim.h"

#define LM_BlkIncOK(M, N, A, RB0, RB, RBInc, CB0, CB, CBInc) 		\
	BlkIncOK(M, N, LM_RInc(A), LM_CInc(A), RB0, RB, RBInc, CB0, CB, CBInc) 

#define pLM_BlkIncOK(M, N, A, RB0, RB, RBInc, CB0, CB, CBInc) 		\
	BlkIncOK(M, N, pLM_RInc(A), pLM_CInc(A), RB0, RB, RBInc, CB0, CB, CBInc) 


#define BlkMatStridePack(Add, M, N, A, alpha, B, RB0, RB, RBInc, CB0, CB, CBInc)\
	BlkMatCopy(Add, M, N, A, alpha, B, RB0, RB, RBInc, 1, CB0, CB, CBInc, 1)

#define BlkMatStrideUnpack(Add, M, N, A, alpha, B, RB0, RB, RBInc, CB0, CB, CBInc)\
	BlkMatCopy(Add, M, N, A, alpha, B, RB0, RB, 1, RBInc, CB0, CB, 1, CBInc)


#ifdef CELLBLAS		/* bypass this module; assume RBInc* = CBInc* = 1) */

#define BlkIncOK(M, N, RInc, CInc, RB0, RB, RBInc, CB0, CB, CBInc) 	1

#define BlkMatCopy(Add, M, N, A, alpha, B, RB0, RB, RBIncA, RBIncB,
                  CB0, CB, CBIncA, CBIncB)
        (((Add)? MatAdd: MatCopy)(M, N, A, alpha, B))          

#else

# define BlkIncOK(M, N, RInc, CInc, RB0, RB, RBInc, CB0, CB, CBInc) 	\
        /* ensures no 2 elts of block-strided matrix map to self */     \
        (0<=RB0 && RB0<RB  &&  0<=CB0 && CB0<CB && 0<RBInc && 0<CBInc &&\
         (RInc > L2G(N-1, CB0, CB, 0, CBInc)*CInc ||                  	\
          CInc > L2G(M-1, RB0, RB, 0, RBInc)*RInc ))

# if FDATATYPE==SINGLE
#   define BlkMatCopy		DSBlkMatCopy
# else
#   define BlkMatCopy		DDBlkMatCopy
# endif

void BlkMatCopy STDC_PARAMS( 
	(int Add, int M, int N, pLocMat A, FDATA *alpha, pLocMat B,
	 int RB0, int RB, int RBIncA, int RBIncB, 
	 int CB0, int CB, int CBIncA, int CBIncB)
);
/* pre:  LM_BlkIncOK(M, N, A, RB0, RB, RBIncA, CB0, CB, CBIncA) &&  
 *       LM_BlkIncOK(M, N, B, RB0, RB, RBIncB, CB0, CB, CBIncB)
 * post: B[iB,jB] = (Add)*B[iB,jB] + (*alpha)*A[iA,jA], 0<=i<M, 0<=j<N
 *       where iA=L2G(i, RB0, RB, 0, RBIncA), jA = L2G(j, RB0, RB, 0, RBIncA)
 *       where iB=L2G(i, RB0, RB, 0, RBIncB), jB = L2G(j, RB0, RB, 0, RBIncB)
 */

#endif /* CELLBLAS */

#endif /*__FDATACellBlkMat_INCL__*/
