/* @(#) All Right Reserved, Copyright (c) 1997 Australian National University */
/* @(#)                     Copyright (c) 1997 Fujitsu Laboratories Ltd. */
/*  */
/*        */
/*                        Permission Notice */
/*  */
/*                Australian National University */
/*                Canberra ACT 0200 Australia */
/*                           and */
/*                Fujitsu Laboratories Ltd. */
/*                4-1-1 Kamikodanaka Nakahara */
/*                Kawasaki 211-88 Japan */
/*  */
/*   Permission to use, copy, modify and distribute this software (including  */
/*   its documentation) without fee is hereby granted under the following terms */
/*   and conditions: */
/*  */
/*   1. This permission notice and all of the copyright notices in this */
/*      software must appear in all copies and derived versions of all */
/*      or a part of this software. */
/*  */
/*   2. Sales of this software, whether incorporated in other products as a */
/*      part thereof or not, is prohibited. */
/*  */
/*   3. On publication or presentation in any form of the results obtained */
/*      by use of this software, an explicit reference to this software and */
/*      its copyright owner shall be made. */
/*  */
/*   4. THIS SOFTWARE IS PROVIDED 'AS IS' AND THE AUSTRALIAN NATIONAL UNIVERSITY */
/*      AND FUJITSU LABORATORIES LTD. MAKE NO WARRANTIES, EITHER EXPRESSED OR */
/*      IMPLIED, AS TO ANY MATTER WHATSOEVER REGARDING THIS SOFTWARE,  */
/*      INCLUDING, WITHOUT LIMITATION, WARRANTIES OF MERCHANTABILITY OR FITNESS */
/*      FOR ANY PARTICULAR PURPOSE. */
/*      IN NO EVENT SHALL THE AUSTRALIAN NATIONAL UNIVERSITY OR FUJITSU */
/*      LABORATORIES LTD. BE LIABLE FOR ANY CLAIM FROM A THIRD PARTY, OR */
/*      FOR SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES WHATSOEVER RESULTING */
/*      FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN ACTION OF CONTRACT, */
/*      NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION */
/*      WITH THE USE, COPYING OR MODIFICATION OF THIS SOFTWARE. */
/*  */

/************************************************************************/
/*			 	FDATACellMatCopy			*/
/************************************************************************/
/* SCCS INFO @(#)FDATACellMatCopy.c 4.4         last modified P. Strazdins, 97/03/10
 *	first created P. Strazdins DCS ANU 95/09
 */
/* general scaled matrix copy routine with arbitrary row/column strides.
 * compilation: macro FDATATYPE must be defined;
 *              if macro ASSERT defined, assertion checking is turned on
 *
 *              if macro VANILLA defined, `vanilla' versions of the routines
 *              are used (slower, but useful for debugging or vectorization.
 *		if macro NO_ALIGN defined, create version where 
 *		double word alignment in A or B cannot be assumed.
 *			(should be compiled with appropriate non-align. flags)
 *              if macro EXT_FTN_BLAS defined, Fortran BLAS is used whenever
 *              possible; 
 */ 
/************************************************************************/
#include "FDATACellMatCopy.h"
#include "FDATAOps.h"
#include "FDATAAssert.h"
#include "FDATACellMatrix.h"
#include "MachineParams.h"

#ifdef NO_ALIGN
# define VecCopy VecCopyW
# define VecAdd  VecAddW
# define MatCopy MatCopyW
# define MatAdd  MatAddW
# undef EXT_FTN_BLAS
#endif
 
#ifdef EXT_FTN_BLAS
# include "FtnBLAS.h"
#elif !defined(NO_ALIGN)
# include "BLASInnerLoops.h"
#endif


void VecCopy(N, A, IncA, B, IncB)
  int N, IncA, IncB; FDATA *A, *B;
{ int i, n; 

  COMMON_ASSERT_TR(
    (TraceStr, "(%d,a%d %d,,a%d %d)", N, FD_ADR(A), IncA, FD_ADR(B), IncB),
     1, 0, "VecCopy", 4, TRACE_COMPUTE
  );

#ifdef EXT_FTN_BLAS
  FTN_COPY(&N, A, &IncA, B, &IncB);
#else
  i=0; n=0;
#ifndef VANILLA     
   n=8*(N/8);	
   if (i < n) {
     register int IncA2=IncA+IncA, IncA3=IncA2+IncA, IncA4=IncA2+IncA2;
     register int IncB2=IncB+IncB, IncB3=IncB2+IncB, IncB4=IncB2+IncB2;
     register FDATA t0, t1, t2, t3, b0, b1, b2, b3;  
     do {
          t0 = *A;
          t1 = *(A+IncA);
          i+=8;
          t2 = *(A+IncA2);
          t3 = *(A+IncA3);
          A += IncA4; 
          b0 = *A;
          b1 = *(A+IncA);
          b2 = *(A+IncA2);
          b3 = *(A+IncA3);
          *B = t0; *(B+IncB) = t1;
          A += IncA4; 
          *(B+IncB2) = t2; *(B+IncB3) = t3;
          B += IncB4;
          *B = b0; *(B+IncB) = b1;
          *(B+IncB2) = b2; *(B+IncB3) = b3;
          B += IncB4;          
    } while (i < n);
    
    if (ForceAlgVers==-1 && (t0+t1+t2+t3 > b0+b1+b2+b3)) {
      /*existance of 2nd condition forces gcc to put t0-b3 in SPARC FPU registers !!!*/
        CommonAssert(1, 0, "VecSwap", 10, TRACE_ALL);
    }
  } /* if (i< n)*/ 
#endif VANILLA    
  for (i=n; i < N; i++) {
    *B = *A;
    A += IncA; B += IncB;
  }
#endif /*not defined EXT_FTN_BLAS*/
} /* VecCopy() */


void MatCopy(M, N, A, alpha, B)
  int M, N; FDATA *alpha; pLocMat A, B;
{ int i; 
  int RIncA=pLM_RInc(A), CIncA=pLM_CInc(A), RIncB=pLM_RInc(B), CIncB=pLM_CInc(B);
  FDATA *Ax = pLM_ADR(A), *Bx = pLM_ADR(B);

  COMMON_ASSERT_TR(
    (TraceStr, "(%dx%d,a%d %d %d,%1.1f,a%d %d %d)", M, N,
     FD_ADR(Ax), RIncA, CIncA, pFD_VAL(alpha), FD_ADR(Bx), RIncB, CIncB),
    pLM_IncOK(M, N, A) && pLM_IncOK(M, N, B), 0, "MatCopy", 4, TRACE_COMPUTE
  );

  if (M_INNER(M, N, MAX(ABS(RIncA), ABS(RIncB)), MAX(ABS(CIncA), ABS(CIncB)))){
    SWAP(int, M, N);	SWAP(int, RIncA, CIncA);	SWAP(int, RIncB, CIncB);
  }
      
#ifdef VANILLA
  if (alpha==NULL) alpha = &FD_1;
  for (i=0; i < M; i++) {
    int j; register FDATA *ax = Ax, *bx = Bx;
    FD_ALPHA_REGS(*alpha);
    for (j=0; j < N; j++) {
      FD_ALPHA_MULSET(*bx, =, *ax);
      ax += CIncA; bx += CIncB;
    }
    Ax += RIncA; Bx += RIncB;
  } /* for (i...) */
#elif defined(EXT_FTN_BLAS)
  for (i=0; i < M; i++) {
    FTN_COPY(&N, Ax, &CIncA, Bx, &CIncB);
    if (! pFD_UNITY(alpha))
      FTN_SCAL(&N, alpha, Bx, &CIncB);          	
    Ax += RIncA; Bx += RIncB;
  } /* for (i...) */  
#else
  { register int CIncA2, CIncB2, CIncA3, CIncB3, CIncA4, CIncB4;
    register FDATA *ax, *bx, a, t0, t1, t2, t3, b0, b1, b2, b3;  
    /* this forces gcc to map t0-t3 to FPU registers, gen. faster for alpha=1*/ 
    register int j, n;

     if (N >= 4) {			
       CIncA2=CIncA+CIncA, CIncA3=CIncA2+CIncA, CIncA4=CIncA2+CIncA2;
       CIncB2=CIncB+CIncB, CIncB3=CIncB2+CIncB, CIncB4=CIncB2+CIncB2;
     }
    
    if (pFD_UNITY(alpha)) {
      n=8*(N/8);	
      for (i=0; i<M; i++) {
        ax = Ax; bx = Bx;
        j=0;
        if (j < n) do {
          t0 = *ax;
          t1 = *(ax+CIncA);
          j+=8;
          t2 = *(ax+CIncA2);
          t3 = *(ax+CIncA3);
          ax += CIncA4; 
          b0 = *ax;
          b1 = *(ax+CIncA);
          b2 = *(ax+CIncA2);
          b3 = *(ax+CIncA3);
          *bx = t0; *(bx+CIncB) = t1;
          ax += CIncA4; 
          *(bx+CIncB2) = t2; *(bx+CIncB3) = t3;
          bx += CIncB4;
          *bx = b0; *(bx+CIncB) = b1;
          *(bx+CIncB2) = b2; *(bx+CIncB3) = b3;
          bx += CIncB4;          
        } while (j < n);
        for (j=n; j < N; j++) {
          *bx = *ax;
          ax += CIncA; bx += CIncB;
        }
        Ax += RIncA; Bx += RIncB;
      } /* for (i...)*/
    } else {
     a = *alpha;
     n=4*(N/4);	
     if (n >= 4) n = n-4;               /* avoid accessing beyond bounds */
     for (i=0; i<M; i++) {
        ax = Ax; bx = Bx;
        if (n >= 4) {
                t0 = *ax;
                t1 = *(ax+CIncA);
                                        FD_SETOP(b0, a, *, t0);
                t2 = *(ax+CIncA2);
                                        FD_SETOP(b1, a, *, t1);
          j=0;
          do {                          /*4 way loop unrolling*/
                t3 = *(ax+CIncA3);
                                        FD_SETOP(b2, a, *, t2);
	     ax += CIncA4;
                                                        *bx = b0;
                t0 = *(ax);
                                        FD_SETOP(b3, a, *, t3);
                                                        *(bx+CIncB) = b1;
             j += 4;
                t1 = *(ax+CIncA);
                                        FD_SETOP(b0, a, *, t0);
                                                        *(bx+CIncB2) = b2;
 
                t2 = *(ax+CIncA2);
                                        FD_SETOP(b1, a, *, t1);
                                                        *(bx+CIncB3) = b3;
             bx += CIncB4;
          } while (j < n);
        } /* if (n....) */
        for (j=n; j < N; j++) {   
          t0 = *(ax);
                                        FD_SETOP(b0, a, *, t0);
                                                        *bx = b0;
          ax += CIncA; bx += CIncB;
        }
        Ax += RIncA; Bx += RIncB;
      } /* for (i...) */
    } /* if (*alpha...) */
  }    
#endif /*VANILLA*/
} /* MatCopy() */


void VecAdd(N, A, IncA, B, IncB)
  int N, IncA, IncB; FDATA *A, *B;
{ int i, n; 

  COMMON_ASSERT_TR(
    (TraceStr, "(%d,a%d %d,,a%d %d)", N, FD_ADR(A), IncA, FD_ADR(B), IncB),
     1, 0, "VecAdd", 4, TRACE_COMPUTE
  );

#ifdef EXT_FTN_BLAS
  FTN_AXPY(&N, &FD_1, A, &IncA, B, &IncB);
#else
  i=0; n=0;
# ifndef VANILLA     
   n = 4*(N/4);
   if (n > 0) {
     register int IncA2=IncA+IncA, IncA3=IncA2+IncA, IncA4=IncA2+IncA2;
     register int IncB2=IncB+IncB, IncB3=IncB2+IncB, IncB4=IncB2+IncB2;
     register FDATA t0, t1, t2, t3, t0b, t1b, t2b, t3b, b0, b1, b2, b3;  
    
		t0 = *(A);		t0b = *(B);
		t1 = *(A+IncA);	t1b = *(B+IncB);
				FD_SETOP(b0, t0b, +, t0);
		t2 = *(A+IncA2);	t2b = *(B+IncB2);
				FD_SETOP(b1, t1b, +, t1);
     for (i=4; i<n; i+=4) {
		t3 = *(A+IncA3);	t3b = *(B+IncB3);
	     A += IncA4; 
				FD_SETOP(b2, t2b, +, t2);
							*(B) = b0;
		t0 = *(A);		t0b = *(B+IncB4);
				FD_SETOP(b3, t3b, +, t3);
							*(B+IncB) = b1;
	     B += IncB2;
		t1 = *(A+IncA);	t1b = *(B+IncB3);
				FD_SETOP(b0, t0b, +, t0);
							*(B) = b2;
		t2 = *(A+IncA2);	t2b = *(B+IncB4);
				FD_SETOP(b1, t1b, +, t1);
							*(B+IncB) = b3;
	     B += IncB2;
	  } /* for (i...) */
		t3 = *(A+IncA3);	t3b = *(B+IncB3);
		A += IncA4; 
				FD_SETOP(b2, t2b, +, t2);
							*(B) = b0;
				FD_SETOP(b3, t3b, +, t3);
							*(B+IncB) = b1;
		B += IncB2;
							*(B) = b2;
							*(B+IncB) = b3;
		B += IncB2;
  } /* if (n>0) */
#endif VANILLA    
  for (i=n; i < N; i++) {
      *B += *A;
      A += IncA; B += IncB;
  } 
#endif /*not defined EXT_FTN_BLAS*/
} /* VecAdd() */


void  MatAdd(M, N, A, alpha, B)
  int M, N; FDATA *alpha; pLocMat A, B;
{ int i; 
  int RIncA=pLM_RInc(A), CIncA=pLM_CInc(A), RIncB=pLM_RInc(B), CIncB=pLM_CInc(B);
  FDATA *Ax = pLM_ADR(A), *Bx = pLM_ADR(B);
  COMMON_ASSERT_TR(
    (TraceStr, "(%dx%d,a%d %d %d,%1.1f,a%d %d %d)", M, N,
     FD_ADR(Ax), RIncA, CIncA, pFD_VAL(alpha), FD_ADR(Bx), RIncB, CIncB),
    pLM_IncOK(M, N, A) && pLM_IncOK(M, N, B), 0, "MatAdd", 4, TRACE_COMPUTE
  );
  if (M_INNER(M, N, MAX(ABS(RIncA), ABS(RIncB)), MAX(ABS(CIncA), ABS(CIncB)))){
    SWAP(int, M, N);	SWAP(int, RIncA, CIncA);	SWAP(int, RIncB, CIncB);
  }    
  if (alpha==NULL) alpha = &FD_1;
  if (FD_EQ(*alpha, FD_0))
    return;
#ifdef VANILLA
  else { 
    FD_ALPHA_REGS(*alpha); 
    for (i=0; i < M; i++) {
      int j; register FDATA *ax = Ax, *bx = Bx; 
      for (j=0; j < N; j++) {
        FD_ALPHA_SCALEADD(*bx, *bx, *ax);
        ax += CIncA; bx += CIncB;
      }
      Ax += RIncA; Bx += RIncB;
    } /* for (i...) */
  }
#elif defined(EXT_FTN_BLAS)
  else
    for (i=0; i < M; i++) {
      FTN_AXPY(&N, alpha, Ax, &CIncA, Bx, &CIncB); 	
      Ax += RIncA; Bx += RIncB;
    } /* for (i...) */  
#else
  else if (! FD_EQ(FD_ABS(*alpha), FD_1)) {  /* series of proper saxpys */
    LocMat alpham; 		LM_Init(alpham, alpha, 1, 1);
    TRACE_LEVEL_SHIFT(2);
    for (i=0; i<M; i++) {      	
# ifndef NO_ALIGN
#   if (FDATATYPE==SINGLE)
      LM_ADR(A) = Ax;                   LM_ADR(B) = Bx;
      MatMult(1, N, 1, &FD_1, &alpham, A, &FD_1, B);
#   else
      Axpy1(N, alpha, Ax, CIncA, Bx, CIncB);
#   endif
# else 
      int j; FDATA *ax = Ax, *bx = Bx;	FD_ALPHA_REGS(*alpha); 
      for (j=0; j < N; j++){
	FD_ALPHA_MULSET(*bx, +=, *ax);
	ax += CIncA; bx += CIncB;
      }	   
# endif	
      Ax += RIncA; Bx += RIncB;
    } /* for (i...) */     
    TRACE_LEVEL_SHIFT(-2);
  } else 
  { register int CIncA2=CIncA+CIncA, CIncA3=CIncA2+CIncA, CIncA4=CIncA2+CIncA2;
    register int CIncB2=CIncB+CIncB, CIncB3=CIncB2+CIncB, CIncB4=CIncB2+CIncB2;
    register int n = 4*(N/4), j;
    register FDATA t0, t1, t2, t3, t0b, t1b, t2b, t3b, b0, b1, b2, b3;
    if (FD_EQ(*alpha,FD_1)) 
      for (i=0; i<M; i++) {
        register FDATA *ax = Ax, *bx = Bx;
        if (n > 0) {
		t0 = *(ax);		t0b = *(bx);
		t1 = *(ax+CIncA);	t1b = *(bx+CIncB);
				FD_SETOP(b0, t0b, +, t0);
		t2 = *(ax+CIncA2);	t2b = *(bx+CIncB2);
				FD_SETOP(b1, t1b, +, t1);
	  for (j=4; j<n; j+=4) {
		t3 = *(ax+CIncA3);	t3b = *(bx+CIncB3);
	     ax += CIncA4; 
				FD_SETOP(b2, t2b, +, t2);
							*(bx) = b0;
		t0 = *(ax);		t0b = *(bx+CIncB4);
				FD_SETOP(b3, t3b, +, t3);
							*(bx+CIncB) = b1;
	     bx += CIncB2;
		t1 = *(ax+CIncA);	t1b = *(bx+CIncB3);
				FD_SETOP(b0, t0b, +, t0);
							*(bx) = b2;
		t2 = *(ax+CIncA2);	t2b = *(bx+CIncB4);
				FD_SETOP(b1, t1b, +, t1);
							*(bx+CIncB) = b3;
	     bx += CIncB2;
	  } /* for (j...) */
		t3 = *(ax+CIncA3);	t3b = *(bx+CIncB3);
		ax += CIncA4; 
				FD_SETOP(b2, t2b, +, t2);
							*(bx) = b0;
				FD_SETOP(b3, t3b, +, t3);
							*(bx+CIncB) = b1;
		bx += CIncB2;
							*(bx) = b2;
							*(bx+CIncB) = b3;
		bx += CIncB2;
	}
        for (j=n; j < N; j++) {
          t0 = *(ax); 		t0b = *(bx);  	FD_SETOP(b0, t0b, +, t0); 
          ax += CIncA; 	*(bx) = b0;	bx += CIncB;
        }
        Ax += RIncA; Bx += RIncB;
      } /* for (i...) */
    else if (FD_EQ(*alpha,FD_M1)) 
      for (i=0; i<M; i++) {
        register FDATA *ax = Ax, *bx = Bx;
        if (n > 0) {
		t0 = *(ax);		t0b = *(bx);
		t1 = *(ax+CIncA);	t1b = *(bx+CIncB);
				FD_SETOP(b0, t0b, -, t0);
		t2 = *(ax+CIncA2);	t2b = *(bx+CIncB2);
				FD_SETOP(b1, t1b, -, t1);
          for (j=4; j < n; j+= 4) {	
		t3 = *(ax+CIncA3);	t3b = *(bx+CIncB3);
	    ax += CIncA4; 
				FD_SETOP(b2, t2b, -, t2);
							*(bx) = b0;
		t0 = *(ax);		t0b = *(bx+CIncB4);
				FD_SETOP(b3, t3b, -, t3);
							*(bx+CIncB) = b1;
	    bx += CIncB2;
		t1 = *(ax+CIncA);	t1b = *(bx+CIncB3);
				FD_SETOP(b0, t0b, -, t0);
							*(bx) = b2;
		t2 = *(ax+CIncA2);	t2b = *(bx+CIncB4);
				FD_SETOP(b1, t1b, -, t1);
							*(bx+CIncB) = b3;
	    bx += CIncB2;
	  } /* for (j...) */ 
		t3 = *(ax+CIncA3);	t3b = *(bx+CIncB3);
		ax += CIncA4; 
				FD_SETOP(b2, t2b, -, t2);
							*(bx) = b0;
				FD_SETOP(b3, t3b, -, t3);
							*(bx+CIncB) = b1;
		bx += CIncB2;
							*(bx) = b2;
							*(bx+CIncB) = b3;
		bx += CIncB2;
        }
        for (j=n; j < N; j++) {
          t0 = *(ax); 		t0b = *(bx);  	FD_SETOP(b0, t0b, -, t0); 
          ax += CIncA; 		*(bx) = b0;	bx += CIncB;
        }
        Ax += RIncA; Bx += RIncB;
      }
  } /* alpha =+/-1 */       
#endif /*VANILLA*/
} /*  MatAdd() */


