/* @(#) All Right Reserved, Copyright (c) 1997 Australian National University */
/* @(#)                     Copyright (c) 1997 Fujitsu Laboratories Ltd. */
/*  */
/*        */
/*                        Permission Notice */
/*  */
/*                Australian National University */
/*                Canberra ACT 0200 Australia */
/*                           and */
/*                Fujitsu Laboratories Ltd. */
/*                4-1-1 Kamikodanaka Nakahara */
/*                Kawasaki 211-88 Japan */
/*  */
/*   Permission to use, copy, modify and distribute this software (including  */
/*   its documentation) without fee is hereby granted under the following terms */
/*   and conditions: */
/*  */
/*   1. This permission notice and all of the copyright notices in this */
/*      software must appear in all copies and derived versions of all */
/*      or a part of this software. */
/*  */
/*   2. Sales of this software, whether incorporated in other products as a */
/*      part thereof or not, is prohibited. */
/*  */
/*   3. On publication or presentation in any form of the results obtained */
/*      by use of this software, an explicit reference to this software and */
/*      its copyright owner shall be made. */
/*  */
/*   4. THIS SOFTWARE IS PROVIDED 'AS IS' AND THE AUSTRALIAN NATIONAL UNIVERSITY */
/*      AND FUJITSU LABORATORIES LTD. MAKE NO WARRANTIES, EITHER EXPRESSED OR */
/*      IMPLIED, AS TO ANY MATTER WHATSOEVER REGARDING THIS SOFTWARE,  */
/*      INCLUDING, WITHOUT LIMITATION, WARRANTIES OF MERCHANTABILITY OR FITNESS */
/*      FOR ANY PARTICULAR PURPOSE. */
/*      IN NO EVENT SHALL THE AUSTRALIAN NATIONAL UNIVERSITY OR FUJITSU */
/*      LABORATORIES LTD. BE LIABLE FOR ANY CLAIM FROM A THIRD PARTY, OR */
/*      FOR SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES WHATSOEVER RESULTING */
/*      FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN ACTION OF CONTRACT, */
/*      NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION */
/*      WITH THE USE, COPYING OR MODIFICATION OF THIS SOFTWARE. */
/*  */

/************************************************************************/
/*       	FDATACellMatCopy: FDATA Matrix add/copy procedures     	*/
/************************************************************************/
/* SCCS INFO @(#)FDATACellMatCopy.h 4.3         last modified P. Strazdins, 97/03/10
   first created: P. Strazdins 97/02
*/
/* Set of scaled matrix copy & add routines used for DBLAS; 
 * to give 2 versions of MatAdd() and MatCopy() where double word
 * alignment can & cannot be guaranteed. 
 * The latter requires NO_ALIGN to be defined; 
 * in which case, compiler flags should be set up so that alignment is not 
 * assumed.
 */
#ifndef __FDATAMatrixCopy_INCL__
#define __FDATAMatrixCopy_INCL__  

#include "FDATASetUp.h"
#include "LocMat.h"

/********************* exported symbols  ***********************************/
#ifndef NO_ALIGN

# ifndef CELLBLAS

#if FDATATYPE==SINGLE
#define VecCopy		DSVecCopy   
#define VecAdd		DSVecAdd   
#define MatCopy		DSMatCopy   
#define MatAdd		DSMatAdd   
#else
#define VecCopy		DDVecCopy   
#define VecAdd		DDVecAdd   
#define MatCopy		DDMatCopy   
#define MatAdd		DDMatAdd   
#endif

# else	/* rename to avoid linker clash */

#if FDATATYPE==SINGLE
#define VecCopy		SVecCopy   
#define VecAdd		SVecAdd   
#define MatCopy		SMatCopy   
#define MatAdd		SMatAdd   
#else
#define VecCopy		DVecCopy   
#define VecAdd		DVecAdd   
#define MatCopy		DMatCopy   
#define MatAdd		DMatAdd   
#endif

# endif /*CELLBLAS*/

#else 

#if FDATATYPE==SINGLE
#define VecCopyW	DSVecCopyW   
#define VecAddW		DSVecAddW   
#define MatCopyW	DSMatCopyW   
#define MatAddW		DSMatAddW   
#else
#define VecCopyW	DDVecCopyW   
#define VecAddW		DDVecAddW   
#define MatCopyW	DDMatCopyW   
#define MatAddW		DDMatAddW   
#endif

#endif /*NO_ALIGN*/


#ifndef NO_ALIGN   
void VecCopy STDC_PARAMS( (int N, FDATA *A, int IncA, FDATA *B, int IncB) );
/* post: B[j] = A[j], 0<=j<N
 * note: low-latency routine
 */

void VecAdd STDC_PARAMS( (int N, FDATA *A, int IncA, FDATA *B, int IncB) );
/* post: B[j] += A[j], 0<=j<N
 * note: low-latency routine
 */

void MatCopy STDC_PARAMS(
        (int M, int N, pLocMat A, FDATA *alpha, pLocMat B)
);
/* pre:  pLM_IncOK(M, N, A) && pLM_IncOK(M, N, B)
 * post: B[i,j] = (alpha)*A[i,j], 0<=i<M, 0<=j<N
 * optimized for M > N, N > M, *alpha = +1.0
 */

void MatAdd STDC_PARAMS(
        (int M, int N, pLocMat A, FDATA *alpha, pLocMat B)
);
/* pre:  pLM_IncOK(M, N, A) && pLM_IncOK(M, N, B)
 * post: B[i,j] += (alpha)*A[i,j], 0<=i<M, 0<=j<N
 * optimized for M > N, N > M, *alpha = +/- 1.0
 */

#else
void VecCopyW STDC_PARAMS( (int N, FDATA *A, int IncA, FDATA *B, int IncB) );

void VecAddW STDC_PARAMS( (int N, FDATA *A, int IncA, FDATA *B, int IncB) );

void MatCopyW STDC_PARAMS(
        (int M, int N, pLocMat A, FDATA *alpha, pLocMat B)
);

void MatAddW STDC_PARAMS(
        (int M, int N, pLocMat A, FDATA *alpha, pLocMat B)
);
#endif /* NO_ALIGN*/

#endif /*__FDATAMatrixCopy_INCL__*/
