/* @(#) All Right Reserved, Copyright (c) 1997 Australian National University */
/* @(#)                     Copyright (c) 1997 Fujitsu Laboratories Ltd. */
/*  */
/*        */
/*                        Permission Notice */
/*  */
/*                Australian National University */
/*                Canberra ACT 0200 Australia */
/*                           and */
/*                Fujitsu Laboratories Ltd. */
/*                4-1-1 Kamikodanaka Nakahara */
/*                Kawasaki 211-88 Japan */
/*  */
/*   Permission to use, copy, modify and distribute this software (including  */
/*   its documentation) without fee is hereby granted under the following terms */
/*   and conditions: */
/*  */
/*   1. This permission notice and all of the copyright notices in this */
/*      software must appear in all copies and derived versions of all */
/*      or a part of this software. */
/*  */
/*   2. Sales of this software, whether incorporated in other products as a */
/*      part thereof or not, is prohibited. */
/*  */
/*   3. On publication or presentation in any form of the results obtained */
/*      by use of this software, an explicit reference to this software and */
/*      its copyright owner shall be made. */
/*  */
/*   4. THIS SOFTWARE IS PROVIDED 'AS IS' AND THE AUSTRALIAN NATIONAL UNIVERSITY */
/*      AND FUJITSU LABORATORIES LTD. MAKE NO WARRANTIES, EITHER EXPRESSED OR */
/*      IMPLIED, AS TO ANY MATTER WHATSOEVER REGARDING THIS SOFTWARE,  */
/*      INCLUDING, WITHOUT LIMITATION, WARRANTIES OF MERCHANTABILITY OR FITNESS */
/*      FOR ANY PARTICULAR PURPOSE. */
/*      IN NO EVENT SHALL THE AUSTRALIAN NATIONAL UNIVERSITY OR FUJITSU */
/*      LABORATORIES LTD. BE LIABLE FOR ANY CLAIM FROM A THIRD PARTY, OR */
/*      FOR SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES WHATSOEVER RESULTING */
/*      FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN ACTION OF CONTRACT, */
/*      NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION */
/*      WITH THE USE, COPYING OR MODIFICATION OF THIS SOFTWARE. */
/*  */

/************************************************************************/
/*       FDATACellMatMult: FDATA cell matrix multiply procedure		*/
/************************************************************************/
/* SCCS INFO @(#)FDATACellMatMult.c 4.2         last modified P. Strazdins, 97/03/18
   first created: P. Strazdins 96/01/16 
*/
/* provides a cell MM routine interfacing to the BLASInnerLoops kernels.
 * This procedure is suitable for scalr-based architectures like the SPARC
 *
 * compilation: macro FDATATYPE must be defined; 
 *   		if macro ASSERT defined, assertion checking turned on.
 *		if macro ASSERT_ASS defined, assertion checking to 
 *		BLASInnerLoops procedures turned on.
 *		if macro MM_VANILLA defined, vanilla version of MatMult() is 
 *		invoked, and BLASInnerLoops module is not referenced.
 */
 
#include "FDATACellMatrix.h"
#include "BLASInnerLoops.h"

#include "FDATAAssert.h"
#ifdef ASSERT_ASS
#  define TR_STR TraceStr
#  define ASSERT_PROC(c1, str) \
	  CommonAssert(c1, 0, str, 6, TRACE_COMPUTE) 
#endif 
#include "FDATAAssPre.h"

void MatMult(M, N, K, alpha, A, B, beta, C)
  int M, N, K; FDATA *alpha, *beta;  pLocMat A, B, C;
/* pre:  pLM_IncOKUnit(M, K, A) && pLM_IncOKUnit(K, N, B) &&
 *       pLM_IncOKUnit(M, N, C) && 0 <= M, N, K,
 *       !MatOverlap(M, K, A, M, N, C) && !MatOverlap(K, N, B, M, N, C)
 *	 *alpha == 1.0
 * post: C := C + (*alpha)A*B;  A is MxK, B is KxN, C is MxN;
 */
{ register int j, n, LdC;  FDATA betaf;
  register int RIncA=pLM_RInc(A), CIncA=pLM_CInc(A), 
	       RIncB=pLM_RInc(B), CIncB=pLM_CInc(B);
  FDATA *Ax = pLM_ADR(A), *Bx= pLM_ADR(B), *Cx = pLM_ADR(C);
  int i, m;			int M0 = M, N0 = N;
  int bounds[3];		/* stride bounds for matrices */

  COMMON_ASSERT_TR_E(
    (TraceStr, "(%dx%dx%d,%1.1f,a%d %d %d,a%d %d %d,%1.1f,a%d %d %d)",
     M, N, K, pFD_VAL(alpha), FD_ADR(pLM_ADR(A)), pLM_RInc(A), pLM_CInc(A),
     FD_ADR(pLM_ADR(B)), pLM_RInc(B), pLM_CInc(B), pFD_VAL(beta),
     FD_ADR(pLM_ADR(C)), pLM_RInc(C), pLM_CInc(C)),
    (N==0 || pLM_IncOKUnit(M, K, A)) && (M==0 || pLM_IncOKUnit(K, N, B)) &&
    pLM_IncOKUnit(M, N, C) && 0<=M && 0<=N && 0<=K, 
    MatOverlap(M, K, A, M, N, C) + 2*MatOverlap(K, N, B, M, N, C),
    "MatMult", 3, TRACE_COMPUTE
  );

  if (alpha!=NULL && FD_EQ(*alpha, FD_0)) {
    MatScale(M, N, beta, C);
    TRACE_RETURN;
  }  
#ifdef MM_VANILLA
# include "FDATAOps.h"
  { int k;
  if (alpha==NULL) alpha = &FD_1; 
  betaf = beta==NULL? FD_1: *beta;
  for (i = 0; i < M; i++) {
    FDATA *b = pLM_ADR(B), *cx = Cx;
    for (j=0; j < N; j++) {   
      FDATA *ax=Ax, *bx=b; FDATA c = FD_0;
      for (k=0; k < K; k++) {
        FD_MULSET(c, +=, *ax, *bx);
        ax += pLM_CInc(A); bx += pLM_RInc(B);
      }
      FD_MULSET(*cx, =, *cx, betaf);
      FD_MULSET(*cx, +=, c, *alpha);
      b += pLM_CInc(B);  cx += pLM_CInc(C);
    } /* for (j...) */
    Cx += pLM_RInc(C); Ax+= pLM_RInc(A);
  } /* for (i...) */
  }     
#else

  if (!pFD_UNITY(alpha) && K>0) {
    betaf = beta==NULL? FD_1: *beta;	
    FD_RECIP(betaf, *alpha);
    beta = &betaf;
  }  
  if (!pFD_UNITY(beta))
    MatScale(M, N, beta, C);
  
  LdC = MAX(pLM_RInc(C), pLM_CInc(C));
  
  if (pLM_RInc(C) == 1) {	/* transpose computation to make CIncC=1 */
    SWAP(FDATAPtr, Ax, Bx);
    SWAP(int, M, N);	
    SWAP(int, RIncA, CIncB);	SWAP(int, CIncA, RIncB);	
  }
    	
  if (K >= MIN_UPDATE_RECT_K) { 
#define RECT_i(a)	(tAB==3? (a RECT_X): (a RECT_Y))
#define RECT_j(a)	(tAB==3? (a RECT_Y): (a RECT_X))
    int tAB = (RIncA==1) + 2*(RIncB==1);
    int di = RECT_i( ),
    	dj = RECT_j( ),
    	Ainc= RECT_i(RIncA *), Binc=RECT_j(CIncB *), Cinc=RECT_i(LdC *); 
    FDATA *b;
         	                      
    /* base multiply on updating each di x dj sub-block;
       remaining cols/rows of C updated using matrix-vector / vector-matrix mult*/ 

    m = M - RECT_i(M %);  n = N - RECT_j(N %); bounds[2] = LdC; 
    if (n < N) {
      b = Bx + (n==0? 0: n*CIncB);	
      if (RIncA==1) {      	
        bounds[0] = RIncB; bounds[1] = CIncA; 
      } else {
	bounds[0] = RIncA; bounds[1] = RIncB; 
      }
      for (j=n; j < N; j++) {
        if (RIncA==1) {
          TR_VecMatMul(K, M, bounds, b, Ax, Cx+j);
        } else {
	  TR_MatVecMul(M, K, bounds, Ax, b, Cx+j);
        }
        b += CIncB;
      }
      N = n;
    }
    bounds[0] = RIncA==1? CIncA: RIncA; bounds[1] = RIncB==1? CIncB: RIncB; 
    if (tAB == 3) {
      SWAP(int, bounds[0], bounds[1]);
    } 
    for (i=0; i < m; i+=di) {
      b = Bx;
      for (j=0; j < n; j+=dj) {
      	if (tAB==1) {
      	  TR_UpdateRectTr1(K, bounds, Ax, b, Cx+j);
      	} else if (tAB==0) {
      	  TR_UpdateRect(K, 0, bounds, Ax, b, Cx+j);  
      	} else if (tAB==2) {
      	  TR_UpdateRectTr2(K, bounds, Ax, b, Cx+j);
      	} else {
      	  TR_UpdateRect(K, 1, bounds, b, Ax, Cx+j);
      	}  
        b += Binc;                	  
      }
      Cx += Cinc; Ax+= Ainc;
    }
    if (m < M) {			/* A, C now at [m,0] */
      bounds[2] = 1;      
      if (RIncB == 1) {
        bounds[0] = CIncB; bounds[1] = CIncA;
      } else { 
        bounds[0] = CIncA; bounds[1] = RIncB; 
      }
      for (i=m; i < M; i++) {
        if (RIncB == 1) {
          TR_MatVecMul(N, K, bounds, Bx, Ax, Cx);
        } else {
          TR_VecMatMul(K, N, bounds, Ax, Bx, Cx); 
        }
        Cx += LdC; Ax += RIncA;
      }
      M = m;
    }   
  } else { 	/* base multiply on rank 1 updates (outer products) */
    bounds[0] = RIncA; bounds[1] = CIncB; bounds[2] = LdC;
    for (j=0; j < K; j++) {	
      TR_Rank1Update(M, N, bounds, Ax, Bx, Cx);
      Ax += CIncA; Bx += RIncB;
    }
  }
  
  if (!pFD_UNITY(alpha) && K>0) 
    MatScale(M0, N0, alpha, C);
#endif /*MM_VANILLA*/
  TRACE_EXIT;
} /* MatMult() */

#ifdef ASSERT_ASS
# undef TR_STR
# undef ASSERT_PROC
#endif
