/* @(#) All Right Reserved, Copyright (c) 1997 Australian National University */
/* @(#)                     Copyright (c) 1997 Fujitsu Laboratories Ltd. */
/*  */
/*        */
/*                        Permission Notice */
/*  */
/*                Australian National University */
/*                Canberra ACT 0200 Australia */
/*                           and */
/*                Fujitsu Laboratories Ltd. */
/*                4-1-1 Kamikodanaka Nakahara */
/*                Kawasaki 211-88 Japan */
/*  */
/*   Permission to use, copy, modify and distribute this software (including  */
/*   its documentation) without fee is hereby granted under the following terms */
/*   and conditions: */
/*  */
/*   1. This permission notice and all of the copyright notices in this */
/*      software must appear in all copies and derived versions of all */
/*      or a part of this software. */
/*  */
/*   2. Sales of this software, whether incorporated in other products as a */
/*      part thereof or not, is prohibited. */
/*  */
/*   3. On publication or presentation in any form of the results obtained */
/*      by use of this software, an explicit reference to this software and */
/*      its copyright owner shall be made. */
/*  */
/*   4. THIS SOFTWARE IS PROVIDED 'AS IS' AND THE AUSTRALIAN NATIONAL UNIVERSITY */
/*      AND FUJITSU LABORATORIES LTD. MAKE NO WARRANTIES, EITHER EXPRESSED OR */
/*      IMPLIED, AS TO ANY MATTER WHATSOEVER REGARDING THIS SOFTWARE,  */
/*      INCLUDING, WITHOUT LIMITATION, WARRANTIES OF MERCHANTABILITY OR FITNESS */
/*      FOR ANY PARTICULAR PURPOSE. */
/*      IN NO EVENT SHALL THE AUSTRALIAN NATIONAL UNIVERSITY OR FUJITSU */
/*      LABORATORIES LTD. BE LIABLE FOR ANY CLAIM FROM A THIRD PARTY, OR */
/*      FOR SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES WHATSOEVER RESULTING */
/*      FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN ACTION OF CONTRACT, */
/*      NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION */
/*      WITH THE USE, COPYING OR MODIFICATION OF THIS SOFTWARE. */
/*  */

/************************************************************************/
/*       FDATACellTri1Matrix.c: FDATA Tri Matrix procedures 		*/
/************************************************************************/
/* SCCS INFO @(#)FDATACellTri1Matrix.h 1.1	last modified P. Strazdins, 97/05/14
   first created: P. Strazdins 97/03 
*/
/* set of non-communication triangular matrix routines used for DBLAS.
 * These routines can be run on cells separately, but the operations
 * assume block matrix storage on a P[0] x P[1] mesh.
 * compilation: macro FDATATYPE must be defined;
 *              if macro ASSERT defined, assertion checking is turned on
 *
 *              if macro NO_PARTITION defined, partitioning methods
 *              in Tri1MatPartMult() are minimized.
 */

#ifndef __FDATACellTri1Mat_INCL__
#define __FDATACellTri1Mat_INCL__

#include "MatDim.h"
#include "DistMat.h"

/********************* exported symbols  ***********************************/
 
#ifndef CELLBLAS
 
#if FDATATYPE==SINGLE
#define Tri1MatMult		DSTri1MatMult
#define Tri1MatPartMult		DSTri1MatPartMult
#else
#define Tri1MatMult             DDTri1MatMult
#define Tri1MatPartMult		DDTri1MatPartMult
#endif

#else   /* rename to avoid name clash */

#if FDATATYPE==SINGLE
#define Tri1MatMult		STri1MatMult
#define Tri1MatPartMult		STri1MatPartMult
#else
#define Tri1MatMult             DTri1MatMult
#define Tri1MatPartMult		DTri1MatPartMult
#endif

#endif /* CELLBLAS */


void Tri1MatMult  STDC_PARAMS(
	(int M, int N, int K, FDATA *alpha, char UpLo, char Diag, 
	 DistMat A, pLocMat B, FDATA *beta, pLocMat C)
);
/* pre:  DistMatOK(M, K, A) && pLM_IncOKUnit(K, N, B) &&
 *       pLM_IncOKUnit(M, N, C) && UpLoOK(UpLo) && 0 <= M, N, K &&
 *       DM_Loc(A) is local MxK block triangular matrix of form UpLo
 * post: C = (*alpha)DM_Loc(A)*B +(*beta)*C;  C is MxN
 * notes:generally A would be the cell view of a global M' x M' tri matrix.
 *       in this case,  M:N ~ d_P(DM_DM(A)):d_P(DM_DN(A)), or more precisely:
 *              M' = L2Gd(M, DM_DM(A)) = L2Gd(N, DM_DN(A))
 *	 Note that elt. (0,0) of A must be on a diagonal; M & K can 
 * 	 be set values corresponding to a trapezoidal matrix, 
 *	 provided this holds.
 */

void Tri1MatMult  STDC_PARAMS(
	(int M, int N, int K, FDATA *alpha, char UpLo, char Diag, 
	 DistMat A, pLocMat B, FDATA *beta, pLocMat C)
);
/* pre/post: as for Tri1MatMult()
 * notes:    performs partitioning over K in units of INNER_VECSIZE
 */
#endif /*__FDATACellTri1Mat_INCL__*/
