/* @(#) All Right Reserved, Copyright (c) 1997 Australian National University */
/* @(#)                     Copyright (c) 1997 Fujitsu Laboratories Ltd. */
/*  */
/*        */
/*                        Permission Notice */
/*  */
/*                Australian National University */
/*                Canberra ACT 0200 Australia */
/*                           and */
/*                Fujitsu Laboratories Ltd. */
/*                4-1-1 Kamikodanaka Nakahara */
/*                Kawasaki 211-88 Japan */
/*  */
/*   Permission to use, copy, modify and distribute this software (including  */
/*   its documentation) without fee is hereby granted under the following terms */
/*   and conditions: */
/*  */
/*   1. This permission notice and all of the copyright notices in this */
/*      software must appear in all copies and derived versions of all */
/*      or a part of this software. */
/*  */
/*   2. Sales of this software, whether incorporated in other products as a */
/*      part thereof or not, is prohibited. */
/*  */
/*   3. On publication or presentation in any form of the results obtained */
/*      by use of this software, an explicit reference to this software and */
/*      its copyright owner shall be made. */
/*  */
/*   4. THIS SOFTWARE IS PROVIDED 'AS IS' AND THE AUSTRALIAN NATIONAL UNIVERSITY */
/*      AND FUJITSU LABORATORIES LTD. MAKE NO WARRANTIES, EITHER EXPRESSED OR */
/*      IMPLIED, AS TO ANY MATTER WHATSOEVER REGARDING THIS SOFTWARE,  */
/*      INCLUDING, WITHOUT LIMITATION, WARRANTIES OF MERCHANTABILITY OR FITNESS */
/*      FOR ANY PARTICULAR PURPOSE. */
/*      IN NO EVENT SHALL THE AUSTRALIAN NATIONAL UNIVERSITY OR FUJITSU */
/*      LABORATORIES LTD. BE LIABLE FOR ANY CLAIM FROM A THIRD PARTY, OR */
/*      FOR SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES WHATSOEVER RESULTING */
/*      FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN ACTION OF CONTRACT, */
/*      NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION */
/*      WITH THE USE, COPYING OR MODIFICATION OF THIS SOFTWARE. */
/*  */

/************************************************************************/
/*       	FDATACellTriMatrix: FDATA Matrix procedures for cells   */
/************************************************************************/
/* SCCS INFO @(#)FDATACellTriMatrix.h 4.1         last modified P. Strazdins, 97/01/10
   first created: P. Strazdins 93/03/23 
*/
/* set of non-communication matrix routines used for DBLAS.
 * These routines can be run on cells separately, but the operations
 * assume block matrix storage on a P[0] x P[1] mesh.
 */
#ifndef __FDATACellTriMat_INCL__
#define __FDATACellTriMat_INCL__  

#include "FDATASetUp.h"
#include "LocMat.h"
#include "DistMat.h"
#include "BLASOptions.h"

/********************* exported symbols  ***********************************/

#ifndef CELLBLAS

#if FDATATYPE==SINGLE
#define TriMatScale  		DSTriMatScale
#define TriMatMult		DSTriMatMult
#define TriMatPartMult 		DSTriMatPartMult
#else
#define TriMatScale  		DDTriMatScale
#define TriMatMult		DDTriMatMult
#define TriMatPartMult 		DDTriMatPartMult
#endif

#else	/* rename to avoid name clash */

#if FDATATYPE==SINGLE
#define TriMatScale  		STriMatScale
#define TriMatMult		STriMatMult
#define TriMatPartMult 		STriMatPartMult
#else
#define TriMatScale  		DTriMatScale
#define TriMatMult		DTriMatMult
#define TriMatPartMult 		DTriMatPartMult
#endif

#endif /* CELLBLAS */

void TriMatScale STDC_PARAMS(
	(int M, int N, FDATA *alpha, char FormC, DistMat C)
);
/* pre:  UpLoOK(FormC) && LM_IncOK(M, N, DM_Loc(C))
 *       DM_Loc(C) is an MxN local triangular block matrix of form FormC
 * post: DM_Loc(C) *= *alpha 
 */

void TriMatMult STDC_PARAMS(
	(int M, int N, int K, FDATA *alpha, pLocMat A, pLocMat B, 
	 FDATA *beta, char FormC, DistMat C)
);
/* pre:  pLM_IncOKUnit(M, K, A) && pLM_IncOKUnit(K, N, B) && 
 *       LM_IncOKUnit(M, N, DM_Loc(C)) && UpLoOK(FormC) && 0 <= M, N, K &&
 *	 !Overlap(M, K, A, M, N, DM_Loc(C)) && !Overlap(K, N, B, M, N, DM_Loc(C))
 *       DM_Loc(C) is local MxN block triangular matrix of form FormC
 * post: DM_Loc(C) = (*alpha)A*B +(*beta)DM_Loc(C);  A is MxK, B is KxN,
 * notes:generally C would be the cell view of a global N' x N' symm, matrix.
 *       in this case,  M:N ~ C.D[0].P:C.D[1].P, or more precisely:
 *              N' = L2Gd(M, C.D[0]) = L2Gd(N, C.D[1])
 *       Only in this case is performance particularly important.
 *       
 */

void TriMatPartMult STDC_PARAMS(
	(int M, int N, int K, FDATA *alpha, pLocMat A, pLocMat B, 
	 FDATA *beta, char FormC, DistMat C)
);
/* pre/post:  as for TriMatMult()
 * notes:see notes for TriMatMult(); this routine also applies
 *	 partitioning methods suitable for large M, N, K
 */

#endif /*__FDATACellTriMat_INCL__*/
