/* @(#) All Right Reserved, Copyright (c) 1997 Australian National University */
/* @(#)                     Copyright (c) 1997 Fujitsu Laboratories Ltd. */
/*  */
/*        */
/*                        Permission Notice */
/*  */
/*                Australian National University */
/*                Canberra ACT 0200 Australia */
/*                           and */
/*                Fujitsu Laboratories Ltd. */
/*                4-1-1 Kamikodanaka Nakahara */
/*                Kawasaki 211-88 Japan */
/*  */
/*   Permission to use, copy, modify and distribute this software (including  */
/*   its documentation) without fee is hereby granted under the following terms */
/*   and conditions: */
/*  */
/*   1. This permission notice and all of the copyright notices in this */
/*      software must appear in all copies and derived versions of all */
/*      or a part of this software. */
/*  */
/*   2. Sales of this software, whether incorporated in other products as a */
/*      part thereof or not, is prohibited. */
/*  */
/*   3. On publication or presentation in any form of the results obtained */
/*      by use of this software, an explicit reference to this software and */
/*      its copyright owner shall be made. */
/*  */
/*   4. THIS SOFTWARE IS PROVIDED 'AS IS' AND THE AUSTRALIAN NATIONAL UNIVERSITY */
/*      AND FUJITSU LABORATORIES LTD. MAKE NO WARRANTIES, EITHER EXPRESSED OR */
/*      IMPLIED, AS TO ANY MATTER WHATSOEVER REGARDING THIS SOFTWARE,  */
/*      INCLUDING, WITHOUT LIMITATION, WARRANTIES OF MERCHANTABILITY OR FITNESS */
/*      FOR ANY PARTICULAR PURPOSE. */
/*      IN NO EVENT SHALL THE AUSTRALIAN NATIONAL UNIVERSITY OR FUJITSU */
/*      LABORATORIES LTD. BE LIABLE FOR ANY CLAIM FROM A THIRD PARTY, OR */
/*      FOR SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES WHATSOEVER RESULTING */
/*      FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN ACTION OF CONTRACT, */
/*      NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION */
/*      WITH THE USE, COPYING OR MODIFICATION OF THIS SOFTWARE. */
/*  */

/************************************************************************/
/*       	FDATACellVector: FDATA Vector procedures for cells    	*/
/************************************************************************/
/* SCCS INFO @(#)FDATACellVector.c 1.1         last modified P. Strazdins, 97/05/14
   first created: P. Strazdins 93/03/23 
*/
/* Set of support vector routines used for DBLAS, independent of distributed
 * memory issues.
 * Efficiency: the use of the following techniques for scalar arhcitectures 
 * are used (if EXT_FTN_BLAS and VANILLA not defined):
 *    1. 2-4 way loop unrolling and software pipelining where appropriate   
 *       generally, the SPARC C compiler does well with these optimizations.
 * compilation: macro FDATATYPE must be defined;
 *
 *              if macro VANILLA defined, `vanilla' versions of the routines
 *              are used (slower, but useful for debugging or vectorization.
 *              if macro EXT_FTN_BLAS defined, Fortran BLAS is used whenever
 *              possible; if EXT_FTN_BLAS_3 defined, do so for only level 3 call
 */

#include <math.h>			/* fabs() */
#define FD_SETABS(x)	x = (FDATA) fabs((double) x)	
			/* for sparcs, gcc should compile to an fabs instrn */
/*#define FD_SETABS(x) 	if ((x) < 0.0) (x) = -(x);*/
/*#define FD_SETABS(x)	(x) = ((x) < 0.0)? -(x): (x);*/

#include "FDATAOps.h"
#include "FDATAAssert.h"
#include "FDATACellVector.h"

void VecSwap(n, A, IncA, B, IncB)
  int n, IncA, IncB; FDATA A[], B[];
{ register FDATA a, b;
  COMMON_ASSERT_TR(
    (TraceStr, "(%d,a%d %d,a%d %d)",n, FD_ADR(A), IncA, FD_ADR(B), IncB),
    1, 0, "VecSwap", 4, TRACE_COMPUTE
  ); 
#ifdef VANILLA
  for (; n > 0; n--) {
    a = *A; 	b=*B;
    *B = a;	*A = b;
    A += IncA; 	B += IncB;      	
  } /* for (n...) */
#elif defined EXT_FTN_BLAS
  FTN_SWAP(&n, A, &IncA, B, &IncB); 
#else   
  {
  register int IncA2=IncA+IncA, IncB2=IncB+IncB;
  register FDATA a1, b1; 

  if (n%2==1) {
    a = *A;  b = *B;	
    *B = a;  *A = b;
    B += IncB; A += IncA;
    n--;            
  }
  for (n=-n; n<0; n+=2) {
      a = *A;		a1 = *(A+IncA);	
      b = *B;		b1 = *(B+IncB); 
      *B = a; 		*(B+IncB) = a1;	
      B += IncB2;
      *A = b;		*(A+IncA) = b1;
      A += IncA2;
  }
  } 
#endif /*VANILLA*/
} /* VecSwap() */

  
int VecAbsMaxPos(m, A, IncA)
  FDATA A[];  int m, IncA;
{ register int i, iM; 			register FDATA M0, Max;
  int m4=4*(m/4);
  COMMON_ASSERT_TR(
    (TraceStr, "(%d,a%d %d)", m, FD_ADR(A), IncA),
    1, 0, "VecAbsMaxPos", 4, TRACE_COMPUTE
  );
  i=0; iM=m>0? 0: -1; 			M0=0.0; Max=0.0;
#if !defined(VANILLA) && !defined(EXT_FTN_BLAS)  
  if (m4 > 0) {
    register int IncA2 = IncA+IncA, IncA3=IncA2+IncA, IncA4=IncA2+IncA2;  
    register FDATA M1, M2, M3;		
    register int iM1;			register FDATA Max1;
    M3 = M2 = Max;
    iM1 = iM;				Max1 = Max;
    do {
    	M0 = *A;	
			FD_SETABS(M3);
					if (M2 > Max) { Max = M2; iM=i-2; };
    	M1 = *(A+IncA);	
			FD_SETABS(M0);
					if (M3 > Max1) { Max1 = M3; iM1=i-1; };
    	M2 = *(A+IncA2);
	i += 4;	
			FD_SETABS(M1);
					if (M0 > Max ) { Max = M0;  iM=i-4; };
    	M3 = *(A+IncA3);	
			FD_SETABS(M2);
	A += IncA4; 
					if (M1 > Max1) { Max1 = M1; iM1=i-3; };
    } while (i < m4);	
			FD_SETABS(M3);
					if (M2 > Max) { Max = M2; iM=i-2; };
   					if (M3 > Max1) { Max1 = M3; iM1=i-1; };
    if (Max1 > Max) { Max = Max1; iM = iM1; };	
  } /* if (m4...)*/ 					
#endif 
#if !defined(EXT_FTN_BLAS)
  for (; i < m; i++) {
    M0 = *A;
			FD_SETABS(M0);
    A += IncA;			
					if (M0 > Max) { Max = M0; iM=i; };
  }
  return (iM);
#else
  return( FTN_IAMAX(&m, A, &IncA)-1 );
#endif /*VANILLA*/
} /* VecAbsMaxPos() */ 


FDATA VecAbsSum(m, A, IncA)
  FDATA A[];  int m, IncA;
{ register int i; register FDATA M0, Sum;
  int m4=4*(m/4);
  COMMON_ASSERT_TR(
    (TraceStr, "VecAbsSum(%d,a%d %d)", m, FD_ADR(A), IncA),
    1, 0, "VecAbsSum", 4, TRACE_COMPUTE
  );
  Sum=0.0;
#if !defined(VANILLA) && !defined(EXT_FTN_BLAS)  
  if (m4 > 0) {
    register int IncA2 = IncA+IncA, IncA3=IncA2+IncA, IncA4=IncA2+IncA2;  
    register FDATA M1, M2, M3, Sum1 = 0.0;
    M3 = M2 = 0.0; i=0;   
    do {
    	M0 = *A;	
			FD_SETABS(M3); 
					Sum += M2;
    	M1 = *(A+IncA);	
			FD_SETABS(M0); 
					Sum1 += M3;
    	M2 = *(A+IncA2);
	i += 4;	
			FD_SETABS(M1);
					Sum += M0;
    	M3 = *(A+IncA3);		
			FD_SETABS(M2);
					Sum1 += M1; 
	A += IncA4; 
    } while (i < m4);	
			FD_SETABS(M3);
					Sum += M2;
					Sum1 += M3;
    m -= m4;		Sum += Sum1;
  } /* if (m4...)*/
#endif
#if !defined(EXT_FTN_BLAS)
  for (i=0; i < m; i++) {
    M0 = *A;
			FD_SETABS(M0); 
					Sum += M0;
					
    A += IncA;			
  }
  return (Sum);
#else
  return( FTN_ASUM(&m, A, &IncA) );
#endif /*VANILLA*/
} /* VecAbsSum() */ 

