/* @(#) All Right Reserved, Copyright (c) 1997 Australian National University */
/* @(#)                     Copyright (c) 1997 Fujitsu Laboratories Ltd. */
/*  */
/*        */
/*                        Permission Notice */
/*  */
/*                Australian National University */
/*                Canberra ACT 0200 Australia */
/*                           and */
/*                Fujitsu Laboratories Ltd. */
/*                4-1-1 Kamikodanaka Nakahara */
/*                Kawasaki 211-88 Japan */
/*  */
/*   Permission to use, copy, modify and distribute this software (including  */
/*   its documentation) without fee is hereby granted under the following terms */
/*   and conditions: */
/*  */
/*   1. This permission notice and all of the copyright notices in this */
/*      software must appear in all copies and derived versions of all */
/*      or a part of this software. */
/*  */
/*   2. Sales of this software, whether incorporated in other products as a */
/*      part thereof or not, is prohibited. */
/*  */
/*   3. On publication or presentation in any form of the results obtained */
/*      by use of this software, an explicit reference to this software and */
/*      its copyright owner shall be made. */
/*  */
/*   4. THIS SOFTWARE IS PROVIDED 'AS IS' AND THE AUSTRALIAN NATIONAL UNIVERSITY */
/*      AND FUJITSU LABORATORIES LTD. MAKE NO WARRANTIES, EITHER EXPRESSED OR */
/*      IMPLIED, AS TO ANY MATTER WHATSOEVER REGARDING THIS SOFTWARE,  */
/*      INCLUDING, WITHOUT LIMITATION, WARRANTIES OF MERCHANTABILITY OR FITNESS */
/*      FOR ANY PARTICULAR PURPOSE. */
/*      IN NO EVENT SHALL THE AUSTRALIAN NATIONAL UNIVERSITY OR FUJITSU */
/*      LABORATORIES LTD. BE LIABLE FOR ANY CLAIM FROM A THIRD PARTY, OR */
/*      FOR SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES WHATSOEVER RESULTING */
/*      FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN ACTION OF CONTRACT, */
/*      NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION */
/*      WITH THE USE, COPYING OR MODIFICATION OF THIS SOFTWARE. */
/*  */

/************************************************************************/
/*       	FDATACellVector: FDATA vector procedures for cells    	*/
/************************************************************************/
/* SCCS INFO @(#)FDATACellVector.h 1.1         last modified P. Strazdins, 97/05/14
   first created: P. Strazdins 96/12/18 
*/
/* Set of support matrix routines used for DBLAS, independent of distributed
 * memory issues. A vector is represented as an address-stride pair.
 */
#ifndef __FDATACellVector_INCL__
#define __FDATACellVector_INCL__  

#include "FDATASetUp.h"

/********************* exported symbols  ***********************************/

#ifndef CELLBLAS

#if FDATATYPE==SINGLE
#define VecSwap		DSVecSwap
#define VecAbsMaxPos	DSVecAbsMaxPos
#define VecAbsSum	DSVecAbsSum	
#else
#define VecSwap		DDVecSwap
#define VecAbsMaxPos	DDVecAbsMaxPos
#define VecAbsSum	DDVecAbsSum	
#endif

#else 	/* rename to avoid name clash */

#if FDATATYPE==SINGLE
#define VecSwap		SVecSwap
#define VecAbsMaxPos	SVecAbsMaxPos
#define VecAbsSum	SVecAbsSum	
#else
#define VecSwap		DVecSwap
#define VecAbsMaxPos	DVecAbsMaxPos
#define VecAbsSum	DVecAbsSum	
#endif

#endif

#ifdef EXT_FTN_BLAS             /* define Fortran BLAS procedure names called */
#include "FtnBLAS.h"
#endif /*EXT_FTN_BLAS*/

/************************ optimization-related macros ***********************/
 
void VecSwap STDC_PARAMS( (int n, FDATA A[], int IncA, FDATA B[], int IncB) );
/* post: A[0..n-1] <-> B[0..n-1]
 */
 
int VecAbsMaxPos STDC_PARAMS( (int m, FDATA A[], int IncA) );
/* post: returns 1st. 0 <= i < m: A[i] and |A[i]| >= |A[j]| for 0 <= j < m
 */

FDATA VecAbsSum STDC_PARAMS( (int m, FDATA A[], int IncA ) );
/* post: returns sum: 0 <= i < m: |A[i]|
 */   

#endif /*__FDATACellVector_INCL__*/
