/* @(#) All Right Reserved, Copyright (c) 1997 Australian National University */
/* @(#)                     Copyright (c) 1997 Fujitsu Laboratories Ltd. */
/*  */
/*        */
/*                        Permission Notice */
/*  */
/*                Australian National University */
/*                Canberra ACT 0200 Australia */
/*                           and */
/*                Fujitsu Laboratories Ltd. */
/*                4-1-1 Kamikodanaka Nakahara */
/*                Kawasaki 211-88 Japan */
/*  */
/*   Permission to use, copy, modify and distribute this software (including  */
/*   its documentation) without fee is hereby granted under the following terms */
/*   and conditions: */
/*  */
/*   1. This permission notice and all of the copyright notices in this */
/*      software must appear in all copies and derived versions of all */
/*      or a part of this software. */
/*  */
/*   2. Sales of this software, whether incorporated in other products as a */
/*      part thereof or not, is prohibited. */
/*  */
/*   3. On publication or presentation in any form of the results obtained */
/*      by use of this software, an explicit reference to this software and */
/*      its copyright owner shall be made. */
/*  */
/*   4. THIS SOFTWARE IS PROVIDED 'AS IS' AND THE AUSTRALIAN NATIONAL UNIVERSITY */
/*      AND FUJITSU LABORATORIES LTD. MAKE NO WARRANTIES, EITHER EXPRESSED OR */
/*      IMPLIED, AS TO ANY MATTER WHATSOEVER REGARDING THIS SOFTWARE,  */
/*      INCLUDING, WITHOUT LIMITATION, WARRANTIES OF MERCHANTABILITY OR FITNESS */
/*      FOR ANY PARTICULAR PURPOSE. */
/*      IN NO EVENT SHALL THE AUSTRALIAN NATIONAL UNIVERSITY OR FUJITSU */
/*      LABORATORIES LTD. BE LIABLE FOR ANY CLAIM FROM A THIRD PARTY, OR */
/*      FOR SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES WHATSOEVER RESULTING */
/*      FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN ACTION OF CONTRACT, */
/*      NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION */
/*      WITH THE USE, COPYING OR MODIFICATION OF THIS SOFTWARE. */
/*  */

/************************************************************************/
/*	FDATACommun: low-level matrix communication procedures for DBLAS*/
/************************************************************************/
/* SCCS INFO @(#)FDATACommun.c 4.7         last modified P. Strazdins, 97/05/19
   first created: P. Strazdins 96/01/15
 */
/* compilation: macro FDATATYPE must be defined;
 *              if macro ASSERT defined, assertion checking turned on
 *
 *		if macro FTN_BLACS defined, Fortran BLACS calls are used 
 *		if macro APLIB is defined, AP/AP+ specific calls are to be used
 *		if macro APLIB_STD, stardard AP calls are used instead of
 *		modified routines (extra overhead implied).
 *		if none of these defined, communication calls
 *		are omitted; this is suitable for running on a uniprocessor.
 */

#include <stdio.h>
#include <stdlib.h>

#include "FDATACommun.h"
#include "FDATAAssert.h"
#include "FDATAOps.h"
#include "LocMat.h"

#include "FDATAMalloc.h"
#include "FDATACellMatCopy.h"
#include "FDATACellBlkMatrix.h"

#if defined(APLIB_STRIDE) || defined(APLIB_LINUX) || defined(APLIB_STD) || defined (APLIB_PLUS_STRIDE)
# define APLIB
#endif

#ifdef FTN_BLACS
# include "FtnBLACS.h"
# define COMMUN_ERROR_FLAG 	0
#elif defined(APLIB)
# include <ccell.c7.h>
# if defined(APLIB_LINUX)
  int r_rsend_stride(), x_brd_stride(), y_brd_stride();
# endif  
# define COMMUN_ERROR_FLAG 	cap_errno
#define NO_ALIGN
# undef __FDATAMatrixCopy_INCL__
# include "FDATACellMatCopy.h"
# undef NO_ALIGN
# if defined(APLIB_STRIDE)
#   include "FDATACellSplitUnpack.c"
# endif
#else
#define COMMUN_ERROR_FLAG 	0 
#endif

#ifdef APLIB
#if (FDATATYPE==SINGLE)
# define Y_SUM y_fsum
# define X_SUM x_fsum
# define Y_AMAX y_famax
# define X_AMAX x_famax
#else
# define Y_SUM y_dsum
# define X_SUM x_dsum
# define Y_AMAX y_damax
# define X_AMAX x_damax
#endif 
int Y_SUM(), X_SUM(), Y_AMAX(), X_AMAX(), y_brd(), x_brd();
#endif /*APLIB*/

/**************** setup of parameters associated with communication ********/ 

static char RC_CHAR[] 		= "CR";
static char BrdTopChar[]	= "  ";
static char CombineTopChar[]	= "  ";
static int  AbsCellId[]		= {0, 0};
static int  NumCells[] 		= {1, 1};
static int  VNumCells[]		= {1, 1};
static int  BLACSCntxt		= 0;

FDATA MsgMaxAbsVal = 0.0;       /*disable checking of FDATA values in messages*/

void BLACSCntxtSet( BLACSContext )
  int BLACSContext;
{ BLACSCntxt = BLACSContext;
} /* BLACSCntxtSet() */

	
void CommSetUp(NCelI, NCelJ)
  int  NCelI, NCelJ;
{ 
#ifdef FTN_BLACS
  if (BLACSCntxt==-1) {		/* illegal context; create new one */
    int BLACSDefHandle=0;
    FTN_BLACS_GET(&BLACSDefHandle, &BLACSDefHandle, &BLACSCntxt);
    FTN_BLACS_GRIDINIT(&BLACSCntxt, "Row", &NCelI, &NCelJ);
  }
  FTN_BLACS_GRIDINFO(&BLACSCntxt, &NumCells[0], &NumCells[1],
  		     &AbsCellId[0], &AbsCellId[1]);
#elif defined(APLIB)
  NumCells[0] = getncely(); 		NumCells[1] = getncelx();
  AbsCellId[0] = getcidy();		AbsCellId[1] = getcidx();      
#else
  NumCells[0] = 1;			NumCells[1] = 1;
  AbsCellId[0] = 0; 			AbsCellId[1] = 0;
#endif
  
  /* setup printing of traces / error messages */
  AssertCId[0] = AbsCellId[0];	 	AssertCId[1] = AbsCellId[1];
  
  COMMON_ASSERT_TR(
    (TraceStr, "(%dx%d): (%d,%d) on %dx%d grid, cntxt %d", NCelI, NCelJ, 
     AbsCellId[0], AbsCellId[1], NumCells[0], NumCells[1], BLACSCntxt),  
    (1==NCelI || NCelI==NumCells[0]) && (1==NCelJ || NCelJ==NumCells[1]),
    0, "CommSetUp", 6, TRACE_COMMUN
  );  
  VNumCells[0] = NCelI;			VNumCells[1] = NCelJ;
  
} /* CommSetUp() */ 

  
void BLACSTopSet( RowCol, Op, Top )
  int RowCol; char *Op, *Top;
{ COMMON_ASSERT_TR(
    (TraceStr, "(%d,%c,%c)", RowCol, *Op, *Top),
    RCOK(RowCol) && (*Op=='B' || *Op=='C'), 0, "BLACSTopSet", 6, 
    TRACE_COMMUN
  );
  if (*Op == 'B')
    BrdTopChar[RowCol] = *Top;
  else  
    CombineTopChar[RowCol] = *Top;
} /*BLACSTopSet() */


/*************** miscellaneous routines **********************************/
 
void Abort( ErrorNum )
  int ErrorNum;
{ PrintLinePrefix();
  printf("DBLAS fatal error encountered, code %d -- aborting\n", ErrorNum); 
#ifdef FTN_BLACS
  FTN_BLACS_ABORT( &BLACSCntxt, &ErrorNum); 
#endif    	
  exit( ErrorNum );
} /* Abort() */   


int GetAbsCId( RowCol )
  int RowCol;
{ return(AbsCellId[RowCol]); 
} /*GetAbsCId() */


int GetNumCells( RowCol )
  int RowCol;
{ return(VNumCells[RowCol]); 
} /*GetNumCells() */


void BarrierSync( RowCol )
  int RowCol;
{ COMMON_ASSERT_TR(
    (TraceStr, "(%d)", RowCol),
    RCOK(RowCol) || RowCol==-1, 0, "BarrierSync", 4, TRACE_COMMUN
  );
#ifdef FTN_BLACS
  { char c = RowCol==-1? 'A': RC_CHAR[RowCol]; 
    FTN_BLACS_BARRIER( &BLACSCntxt, &c ); 
  }
#elif defined APLIB
  if (RowCol==-1)
    sync(0, 0);  			/* note: global scope only */
  else {
    FDATA x=0.0;
    (RowCol==0? Y_SUM: X_SUM)( x, &x );      	
  }  
#endif  
} /*BarrierSync() */
   

/******************* primitive communication macros **********************/    

/* REL2ABSCID(): 0 <= RCID < NumCells[RowCol] */
#define REL2ABSCID(RowCol, RCID)					\
	CYCLIC_ADD(AbsCellId[RowCol], RCID, NumCells[RowCol])
/* SRCABSCID(): 0 <= P0 < NumCells[RowCol] */  
#define SRCABSCID(RowCol, P0)						\
	CYCLIC_SUB(AbsCellId[RowCol], P0, NumCells[RowCol])

#if defined(FTN_BLACS)
# define SEND(RDEST, CDEST, M, N, A, IncA, M_TYPE) 			\
	{ int aRDEST_ = REL2ABSCID(0, RDEST), aCDEST_ = REL2ABSCID(1, CDEST),\
	      m_ = (M), n_ = (N), inc_ = (IncA);			\
	  FTN_GESD2D(&BLACSCntxt, &n_, &m_, A, &inc_, &aRDEST_, &aCDEST_);\
	}  
	
# define RECV(RSRC, CSRC, M, N, A, IncA, M_TYPE, r) 			\
	{ int aRSRC_ = REL2ABSCID(0, RSRC), aCSRC_ = REL2ABSCID(1, CSRC),\
	      m_ = (M), n_ = (N), inc_ = (IncA);			\
	  FTN_GERV2D(&BLACSCntxt, &n_, &m_, A, &inc_, &aRSRC_, &aCSRC_);\
	  r += FAST_MUL1(m_, n_);					\
	}  

# define RECV_ADD(RSRC, CSRC, M, N, alpha, A, RIncA, CIncA, M_TYPE, r) 	\
	{ int aRSRC_ = REL2ABSCID(0, RSRC), aCSRC_ = REL2ABSCID(1, CSRC),\
	      m_ = (M), n_ = (N), mn = FAST_MUL1(m_, n_); 		\
	      FDATA *A_; LocMat lA_, lB_;				\
          MatMalloc_FT("RECV_ADD", 1, &A_, mn);  			\
    	  FTN_GERV2D(&BLACSCntxt, &n_, &m_, A_, &n_, &aRSRC_, &aCSRC_);	\
	  r += mn;					\
	  MatAdd(m_, n_, pMkLocMat(&lA_, A_, n_, 1), alpha,  pMkLocMat(&lB_, A, RIncA, CIncA));\
	  MatFree(A_);							\
	}  

# define RECV_ADD_(Add, RSRC, CSRC, M, N, alpha, A, RIncA, CIncA, M_TYPE, r)  \
	if (!Add)		/* assert alpha==CIncA==1 */		\
	  RECV(RSRC, CSRC, M, N, A, RIncA, M_TYPE, r) 			\
	else								\
	  RECV_ADD(RSRC, CSRC, M, N, alpha, A, RIncA, CIncA, M_TYPE, r);
	
# define BRD(RowCol, SRCP0, M, N, A, IncA, r)				\
	{ int m_ = (M), n_ = (N), inc_ = (IncA);			\
	  int aRSRC_ = (RowCol)==0? SRCABSCID(0, SRCP0): AbsCellId[0],	\
	      aCSRC_ = (RowCol)==1? SRCABSCID(1, SRCP0): AbsCellId[1];	\
	  if (SRCP0==0) 						\
	    FTN_GEBS2D(&BLACSCntxt, &RC_CHAR[RowCol], &BrdTopChar[RowCol],\
		         &n_, &m_, A, &inc_);				\
	   else								\
	    FTN_GEBR2D(&BLACSCntxt, &RC_CHAR[RowCol], &BrdTopChar[RowCol],\
		         &n_, &m_, A, &inc_, &aRSRC_, &aCSRC_);		\
	  r += FAST_MUL1(m_, n_);					\
	}	                        

# define SUM(RowCol, SRCP0, M, N, A, IncA, r)				\
	{ int m_ = (M), n_ = (N), inc_ = (IncA), 			\
	      rdst_ = (RowCol)==0? SRCABSCID(0, SRCP0): AbsCellId[0],\
	      cdst_ = (RowCol)==1? SRCABSCID(1, SRCP0): AbsCellId[1];\
	  FTN_GSUM2D(&BLACSCntxt, &RC_CHAR[RowCol], &CombineTopChar[RowCol],\
	  	 &n_, &m_, A, &inc_, &rdst_, &cdst_);			\
	}	                        
	
# define AMAX(RowCol, A, indexA)					\
	{ int n_ = 1, src_[2], dst_[2];					\
	  dst_[RowCol] = -1, dst_[1-RowCol] = AbsCellId[1-RowCol];	\
	  FTN_GAMX2D(&BLACSCntxt, &RC_CHAR[RowCol], &CombineTopChar[RowCol],\
	  	 &n_, &n_, A, &n_, &src_[0], &src_[1], &n_, &dst_[0], &dst_[1]);\
	  if (src_[RowCol]==AbsCellId[RowCol]) 				\
	    FTN_IGEBS2D(&BLACSCntxt, &RC_CHAR[RowCol], &BrdTopChar[RowCol],\
		         &n_, &n_, indexA, &n_);			\
	   else								\
	    FTN_IGEBR2D(&BLACSCntxt, &RC_CHAR[RowCol], &BrdTopChar[RowCol],\
		         &n_, &n_, indexA, &n_, &src_[0], &src_[1]);	 \
	}  

#elif defined(APLIB_LINUX)

# define MX_STRIDE_SIZE 128 

# define SEND(RDEST, CDEST, M, N, A, IncA, M_TYPE) 			\
	{ int i_, inc_ = (IncA)*sizeof(FDATA);				\
	  if ((M)==1) 							\
	    r_rsend(CDEST, RDEST, gettid(), M_TYPE, (MSGP) (A), 	\
	    	    (N)*sizeof(FDATA));					\
	  else 								\
	    for (i_=0; i_ < (N); i_+=MX_STRIDE_SIZE) {			\
	      int nb = MIN((N)-i_, MX_STRIDE_SIZE)*sizeof(FDATA);	\
	      r_rsend_stride(CDEST, RDEST, gettid(), M_TYPE, (u_long *) ((A)+i_), nb, inc_, M);\
	  }								\
	}								

# define RECV(RSRC, CSRC, M, N, A, IncA, M_TYPE, r)                   	\
	RECV_ADD_(0, RSRC, CSRC, M, N, NULL, A, IncA, 1, M_TYPE, r)

# define RECV_ADD(RSRC, CSRC, M, N, alpha, A, RIncA, CIncA, M_TYPE, r)  \
	RECV_ADD_(1, RSRC, CSRC, M, N, alpha, A, RIncA, CIncA, M_TYPE, r)

# define RECV_ADD_(Add, RSRC, CSRC, M, N, alpha, A, RIncA, CIncA, M_TYPE, r) \
	{ int i_ = 0;		LocMat lA_, lB_;			\
	  do {								\
	    FDATA *A_1 = (FDATA *) r_rrecv(CSRC, RSRC, gettid(), M_TYPE);\
            int nb = getmsize()/sizeof(FDATA); (r) += nb; 		\
            if ((M)==1) 						\
              (Add? VecAdd: VecCopy)(nb, A_1, 1, (A)+i_*(CIncA), CIncA);\
            else {							\
              if ((M)> 1) nb /= (M);					\
	      (Add? MatAdd: MatCopy)					\
	      (M, nb,  pMkLocMat(&lA_, A_1, nb, 1), alpha, pMkLocMat(&lB_, (A)+i_*(CIncA), RIncA, CIncA ));\
	    }								\
	    i_ += nb;							\
	  } while (i_ < (N));						\
	}
	  
# define BRD(RowCol, SRCP0, M, N, A, IncA, r) 				\
	{ int i_, inc_ = (IncA)*sizeof(FDATA), I0_ = SRCABSCID(RowCol, SRCP0);\
	  if ((M)==1)							\
	    (r) += (RowCol==0? y_brd: x_brd)(I0_, (MSGP) (A), (N)*sizeof(FDATA))/sizeof(FDATA);\
	  else								\
	    for (i_=0; i_ < (N); i_+=MX_STRIDE_SIZE) {			\
	      int nb = MIN((N)-i_, MX_STRIDE_SIZE)*sizeof(FDATA);	\
	      (r) += (RowCol==0? y_brd_stride: x_brd_stride)( I0_, (u_long *) ((A)+i_), nb, inc_, M)/sizeof(FDATA);\
	    }								\
	}

#elif defined (APLIB_PLUS_STRIDE)

# define SEND(RDEST, CDEST, M, N, A, IncA, M_TYPE)                      \
	xy_stride_send(CDEST, RDEST, M_TYPE, M, (N)*sizeof(FDATA), 	\
		       A, (IncA)*sizeof(FDATA))
	
# define RECV(RSRC, CSRC, M, N, A, IncA, M_TYPE, r)                   	\
	RECV_ADD_(0, RSRC, CSRC, M, N, NULL, A, IncA, 1, M_TYPE, r)

# define RECV_ADD(RSRC, CSRC, M, N, alpha, A, RIncA, CIncA, M_TYPE, r)  \
	RECV_ADD_(1, RSRC, CSRC, M, N, alpha, A, RIncA, CIncA, M_TYPE, r)

# define RECV_ADD_(Add, RSRC, CSRC, M, N, alpha, A, RIncA, CIncA, M_TYPE, r) \
	{ int i_ = 0;							\
	  int r_, r1, s_; FDATA *b1, *b2; int nb; LocMat lB_, lA_; 	\
	  do {								\
	    r_ = xy_stride_recv(CSRC, RSRC, M_TYPE, M, (N)*sizeof(FDATA),\
	  		        &s_, &r1, &b1, &b2);			\
 	    r_ /= sizeof(FDATA); nb = r_;				\
	    if ((M)==1)							\
	      (Add? VecAdd: VecCopy)(nb, b1, 1, (A)+i_*(CIncA), CIncA);	\
	    else {							\
	      if ((M) > 1) nb /= (M);					\
	      (Add? MatAdd: MatCopy)(M, nb, pMkLocMat(&lB_, b1, nb, 1), \
	      	    NULL, pMkLocMat(&lA_, (A)+i_*(CIncA), RIncA, CIncA));\
	    }  	    							\
	    i_ += nb; (r) += r_; 					\
	  } while (i_<(N) && FAST_MOD(r_, M)==0);			\
	}

# define BRD(RowCol, SRCP0, M, N, A, IncA, r) 				\
	{ int r_, r1, s_, I0_ = SRCABSCID(RowCol, SRCP0); FDATA *b1, *b2;\
	  if ((SRCP0)==0)  						\
	    r_ = brd_stride_send(RowCol, M, (N)*sizeof(FDATA),		\
	    			 A, (IncA)*sizeof(FDATA)); 		\
	  else { int i_ = 0, nb; LocMat lB_, lA_; 			\
	    do {							\
	      r_ = brd_stride_recv(RowCol, I0_, M, (N)*sizeof(FDATA), 	\
	    			   &s_, &r1, &b1, &b2);  		\
	      r_ /= sizeof(FDATA); nb = r_;				\
	      if ((M)==1)						\
	        VecCopy(nb, b1, 1, (A)+i_, 1);				\
	      else {							\
	        if (M>1) nb /= (M);					\
	        MatCopy(M, nb, pMkLocMat(&lB_, b1, nb, 1), NULL, 	\
	      	        pMkLocMat(&lA_, (A)+i_, IncA, 1));		\
	      }	        						\
	      i_ += nb; (r) += r_; 					\
	    } while (i_<(N) && FAST_MOD(r_, M)==0);			\
	  }								\
	}  

#elif defined(APLIB_STRIDE)

/* this optimization is to reduce software latencies for vectors */	
#define SplitMatUnpack_f(Add, N, r, s, b1, r1, b2, alpha, A, RIncA, CIncA)\
	if ((r)==(r1) && (r)==(N) && (alpha==NULL)) 			\
	  ((Add)? VecAdd: VecCopy)(N, b1, 1, A, CIncA);			\
	else 								\
	  SplitMatUnpack(Add, N, r, s, b1, r1, b2, alpha, A, RIncA, CIncA);

#define SplitMatUnpack_	SplitMatUnpack_f		

# define SEND(RDEST, CDEST, M, N, A, IncA, M_TYPE)                      \
	xy_stride_send(CDEST, RDEST, M_TYPE, M, (N)*sizeof(FDATA), 	\
		       A, (IncA)*sizeof(FDATA))
	
# define RECV(RSRC, CSRC, M, N, A, IncA, M_TYPE, r)                   	\
	RECV_ADD_(0, RSRC, CSRC, M, N, NULL, A, IncA, 1, M_TYPE, r)

# define RECV_ADD(RSRC, CSRC, M, N, alpha, A, RIncA, CIncA, M_TYPE, r)  \
	RECV_ADD_(1, RSRC, CSRC, M, N, alpha, A, RIncA, CIncA, M_TYPE, r)

# define RECV_ADD_(Add, RSRC, CSRC, M, N, alpha, A, RIncA, CIncA, M_TYPE, r) \
	{ int r_, r1, s_; FDATA *b1, *b2;  				\
	  r_ = xy_stride_recv(CSRC, RSRC, M_TYPE, M, (N)*sizeof(FDATA), \
	  		      &s_, &r1, &b1, &b2);			\
	  r_ /= sizeof(FDATA); r1 /= sizeof(FDATA); s_ /= sizeof(FDATA);\
	  SplitMatUnpack_(Add, N, r_, s_, b1, r1, b2, alpha, A, RIncA, CIncA);\
	  (r) += r_; 							\
	}

# define BRD(RowCol, SRCP0, M, N, A, IncA, r) 				\
	{ int r_, r1, s_, I0_ = SRCABSCID(RowCol, SRCP0); FDATA *b1, *b2;\
	  if ((SRCP0)==0)  						\
	    r_ = brd_stride_send(RowCol, M, (N)*sizeof(FDATA),		\
	    			 A, (IncA)*sizeof(FDATA)); 		\
	  else { 							\
	     r_ = brd_stride_recv(RowCol, I0_, M, (N)*sizeof(FDATA), 	\
	    			   &s_, &r1, &b1, &b2);  		\
	     r_ /= sizeof(FDATA); r1 /= sizeof(FDATA); s_ /= sizeof(FDATA);\
	     SplitMatUnpack_(0, N, r_, s_, b1, r1, b2, NULL, A, IncA, 1);\
	     (r) += r_; 						\
	  }								\
	}


#elif defined(APLIB_STD)

# define SEND(RDEST, CDEST, M, N, A, IncA, M_TYPE) 			\
	{ int i; FDATA *Ax_ = (A); int n_ = (N), m_ = (M);		\
	  if ((IncA) == n_ && m_ > 1) { n_ *= m_; m_ = 1;}		\
	  for (i=0; i < m_; i++) {					\
	    r_rsend(CDEST, RDEST, gettid(), M_TYPE, Ax_, n_*sizeof(FDATA)); \
	    Ax_ += (IncA); 						\
	  } /* for (i...) */						\
	}

# define RECV(RSRC, CSRC, M, N, A, IncA, M_TYPE, r)                   	\
	RECV_ADD_(0, RSRC, CSRC, M, N, NULL, A, IncA, 1, M_TYPE, r)

# define RECV_ADD(RSRC, CSRC, M, N, alpha, A, RIncA, CIncA, M_TYPE, r)  \
	RECV_ADD_(1, RSRC, CSRC, M, N, alpha, A, RIncA, CIncA, M_TYPE, r)


# define RECV_ADD_(Add, RSRC, CSRC, M, N, alpha, A, RIncA, CIncA, M_TYPE, r) 	\
	{ int M_ = (M), m_, n_; FDATA *Ax_ = (A); LocMat lA_, lB_;	\
	  while (M_ > 0) {						\
	    FDATA *A_1 = (FDATA *) r_rrecv(CSRC, RSRC, gettid(), M_TYPE);\
	    n_ = getmsize()/sizeof(FDATA); m_ = FAST_DVDEQ(n_, N); (r) += n_;\
	    (Add? (((int) A_1) % sizeof(FDATA)? MatAddW: MatAdd):	\
		  (((int) A_1) % sizeof(FDATA)?  MatCopyW: MatCopy))	\
		(m_, N, pMkLocMat(&lA_, A_1, N, 1), alpha, pMkLocMat(&lB_, Ax_, RIncA, CIncA));\
	    Ax_ += FAST_MUL(m_, RIncA); 	M_ -= m_;		\
	  } /* for (i...) */						\
	}

# define BRD(RowCol, SRCP0, M, N, A, IncA, r) 				\
	{ int i, nb = (N)*sizeof(FDATA), I0 = SRCABSCID(RowCol, SRCP0); \
	  FDATA *Ax_ = (A);  int m_ = (M); 				\
	  /*if ((IncA)==(N) && m_> 1) { nb *= m_; m_ = 1; }*/		\
	  for (i=0; i < m_; i++) {					\
	    (r) += (RowCol==0? y_brd: x_brd)( I0, Ax_, nb )/sizeof(FDATA);\
	    Ax_ += (IncA); 						\
	  } /* for (i...) */						\
	}

#else

# define SEND(RDEST, CDEST, M, N, A, IncA, M_TYPE)
# define RECV(RDEST, CDEST, M, N, A, IncA, M_TYPE, r)
# define RECV_ADD(RDEST, CDEST, M, N, alpha, A, RIncA, CIncA, M_TYPE, r)
# define RECV_ADD_(ADD, RDEST, CDEST, M, N, alpha, A, RIncA, CIncA, M_TYPE, r)
# define BRD(RowCol, SRCP0, M, N, A, IncA, r)
# define SUM(RowCol, SRCP0, M, N, A, IncA, r)	
#endif



/******************* matrix routines ***************************/

#define RIJOK(RI, RJ) (ABS(RI)<NumCells[0] && ABS(RJ)<NumCells[1]) 
#define NORMALIZE(RI, RJ)					\
  { if (VNumCells[0]==1) RI = 0;				\
    if (VNumCells[1]==1) RJ = 0;				\
    if (RI < 0) RI += NumCells[0];				\
    if (RJ < 0) RJ += NumCells[1];				\
  }								
  
void AbsMaxElt(RowCol, A, indexA)
  int RowCol, *indexA; FDATA *A; 
{ COMMON_ASSERT_TR(
    (TraceStr, "(%d,*%1.1f,*%d)", RowCol, *A, *indexA),
    RCOK(RowCol), 0, "AbsMaxElt", 4, TRACE_COMMUN
  );

  if (VNumCells[RowCol]==1)
    return;
#ifdef APLIB 
  (RowCol==0? Y_AMAX: X_AMAX)(*A, *indexA, A, indexA);
#elif defined(FTN_BLACS)
  AMAX(RowCol, A, indexA); 
#endif         

  COMMON_ASSERT(1, COMMUN_ERROR_FLAG, "AbsMaxElt (post)", 6, TRACE_COMMUN);
} /* AbsMaxElt() */


void SendMat(RI, RJ, m, n, A, IncA, MsgType)
  int RI, RJ, m, n, IncA; FDATA A[]; int MsgType;
{ COMMON_ASSERT_TR_E(
    (TraceStr, "(%d %d,%dx%d,a%d %d,%d)", RI, RJ, m, n, FD_ADR(A), IncA, MsgType),
    RIncOK(n, IncA) && RIJOK(RI, RJ), 0, "SendMat", 4, TRACE_COMMUN
  );
  if (m==0 || n==0) {
    TRACE_RETURN;
  }
  NORMALIZE(RI, RJ);
  SEND(RI, RJ, m, n, A, IncA, MsgType);
  TRACE_EXIT;
 } /* SendMat() */


void RecvMat(RI, RJ, m, n, A, IncA, MsgType)
  int RI, RJ, m, n, IncA, MsgType; FDATA A[]; 
{ int r=0;  
  COMMON_ASSERT_TR_E(
    (TraceStr, "(%d %d,%dx%d,a%d %d,%d)", RI, RJ, m, n, FD_ADR(A), IncA, MsgType),
    RIncOK(n, IncA) && RIJOK(RI, RJ), 0, "RecvMat", 4, TRACE_COMMUN
  );

  if (m==0 || n==0) {
    TRACE_RETURN;
  }
  NORMALIZE(RI, RJ);
  RECV(RI, RJ, m, n, A, IncA, MsgType, r);
 
  COMMON_ASSERT_TR(
    (TraceStr, " recv: exp=%d, act=%d", m*n, r), 
    m*n==r, COMMUN_ERROR_FLAG, "RecvMat", 5, TRACE_COMMUN
  );
  TRACE_EXIT;
} /* RecvMat() */


void SumMat(RowCol, P0, P, m, n, A)
  int RowCol, P0, m, n; pLocMat A; 
{ int r=0, IncA = pLM_Ld(A); FDATA *Ax = pLM_ADR(A);
  LocMat lA;

  COMMON_ASSERT_TR_E(
    (TraceStr, "(%d,%d %d,%dx%d,a%d %d %d)", 
     RowCol, P0, P, m, n, FD_ADR(Ax), pLM_RInc(A), pLM_CInc(A)),
     RCOK(RowCol) && pLM_IncOKUnit(m, n, A) && P<=NumCells[RowCol] &&
     (pLM_CInc(A)==1 || P0==0), 0, "SumMat", 4, TRACE_COMMUN
  );

  if (P==1 || m==0 || n==0) {
    TRACE_RETURN;
  }  
#ifdef FTN_BLACS
  if (pLM_CInc(A)!=1 && n>1) {	/* assert P0==0 */
    if (m>1) {
      /* BLACS will not be able to handle reversed format of A in receiver */
      MatMalloc_FT("SumMat", 1, &Ax, m*n);
      IncA = n;  
      MatCopy(m, n, A, NULL, pMkLocMat(&lA, Ax, IncA, 1));
    } else /* interchange m, n to compensate for that within SUM() */
      m = n, n = 1;
  } 
  SUM(RowCol, P0, m, n, Ax, IncA, r);  
  if (pLM_CInc(A)!=1 && n>1 && m>1) {
    MatCopy(m, n, &lA, NULL, A);
    MatFree(Ax);
  }
#elif defined(APLIB)
  if ((m == 1) && (n == 1)) {
    (RowCol==0? Y_SUM: X_SUM)( Ax[0], Ax );	
    TRACE_RETURN;
  } else {
    int DelP0 = 1;	 	  
    while (DelP0 < P) {
      int DelIP0 = RowCol==0? DelP0: 0,  DelJP0 = RowCol==1? DelP0: 0; 	
      if (((P0/DelP0)%2 == 1) && P0<P) {
        SEND(-DelIP0, -DelJP0, m, n, Ax, IncA, SUMMSGTYPE);
      } else if (P0 + DelP0 < P) {
        RECV_ADD(DelIP0, DelJP0, m, n, NULL, Ax, pLM_RInc(A), pLM_CInc(A), SUMMSGTYPE, r);
      } /* if (P0/DelP0...) */
      DelP0 *= 2;	
    } /* while (DelP0 ...) */ 
  } /* if (m... n...) */
#endif
 
  COMMON_ASSERT(1, COMMUN_ERROR_FLAG, "SumMat (post)", 5, TRACE_COMMUN);
  TRACE_EXIT;  
} /* SumMat() */


void BrdMat(RowCol, P0, m, n, A, IncA)
  int RowCol, P0, IncA, m, n; FDATA A[]; 
{ int r=0;  				
  COMMON_ASSERT_TR_E(
    (TraceStr, "(%d,%d,%dx%d,a%d %d)", RowCol, P0, m, n, FD_ADR(A), IncA),
    RCOK(RowCol) && RIncOK(n, IncA), 0, "BrdMat", 4, TRACE_COMMUN
  );

  if (VNumCells[RowCol]==1 || m==0 || n==0) {
    TRACE_RETURN;
  } 
  BRD(RowCol, P0, m, n, A, IncA, r);  

  if (P0 != 0) {
    COMMON_ASSERT_TR(
      (TraceStr, " recv: exp=%d, act=%d", m*n, r), 
       m*n==r, COMMUN_ERROR_FLAG, "BrdMat", 5, TRACE_COMMUN
    );
  }  
  TRACE_EXIT;
} /* BrdMat() */

/******************* [1] block matrix routines ***************************/
	       
void BrdBlkMat(RowCol, P0, m, B0, B, BInc, n, A, IncA)
  int RowCol, P0, IncA, m, B0, B, BInc, n; FDATA A[]; 
{ int r=0;  
#ifdef ASSERT
  int expsize= m*n;
#endif  
  COMMON_ASSERT_TR_E(
    (TraceStr, "(%d,%d,%d[%d %d %d]x%d,a%d %d)",
     RowCol, P0, m, B0, B, BInc, n, FD_ADR(A), IncA),
    RCOK(RowCol) && (m+B0<=B || BInc==1 || IncA==n) && 
    RIncOK(n, IncA) && BlkOK(B0, B, BInc), 0, "BrdBlkMat", 4, TRACE_COMMUN
  );

  if (VNumCells[RowCol]==1 || m==0 || n==0) {
    TRACE_RETURN;
  }  
  if (m+B0<=B || BInc==1) {
    BRD(RowCol, P0, m, n, A, IncA, r);  
#ifdef APLIB_STRDE			/*from here on, m=1 optim. unimportant*/
# undef SplitMatUnpack_
# define SplitMatUnpack_ SplitMatUnpack
#endif
  } else {
    IncA = FAST_MUL1(B, n*BInc);	
    if (B0 > 0) {
      int m1 = MIN(B-B0, m);		
      BRD(RowCol, P0, 1, m1*n, A, IncA, r);
      A += (B*BInc-B0)*n;		m -= m1;
    }
    if (m >= B) {
      int m1 = FAST_DVD(m, B);            	
      BRD(RowCol, P0, m1, FAST_MUL2(n, B), A, IncA, r);
      A += m1*IncA;      		m = FAST_MOD(m, B);	
    }
    if (m > 0)
      BRD(RowCol, P0, 1, (m%B)*n, A, IncA, r);
  }   

  if (P0 != 0) {
    COMMON_ASSERT_TR(
      (TraceStr, " recv: exp=%d, act=%d", expsize, r), 
       expsize==r, COMMUN_ERROR_FLAG, "BrdBlkMat", 5, TRACE_COMMUN
    );
  }  
  TRACE_EXIT;
} /* BrdBlkMat() */


void SendBlkMat(RI, RJ, m, B0, B, BInc, n, A, MsgType)
  int RI, RJ, m, B0, B, BInc, n; FDATA A[]; int MsgType;
{ COMMON_ASSERT_TR_E(
    (TraceStr, "(%d %d,%d[%d %d %d]x%d,a%d,%d)",
     RI, RJ, m, B0, B, BInc, n, FD_ADR(A), MsgType),
    RIJOK(RI, RJ) && BlkOK(B0, B, BInc), 0, "SendBlkMat", 4, TRACE_COMMUN
  );

  if (m==0 || n==0) {
    TRACE_RETURN;
  }
  NORMALIZE(RI, RJ);
  if (m+B0<=B || BInc==1) {
    SEND(RI, RJ, m, n, A, n, MsgType);  
  } else {
    int IncA = FAST_MUL1(B, n*BInc);	
    if (B0 > 0) {
      int m1 = MIN(B-B0, m);
      SEND(RI, RJ, 1, m1*n, A, IncA, MsgType);
      A += (B*BInc-B0)*n;		m -= m1;
    }
    if (m >= B) {
      int m1 = FAST_DVD(m, B);	
      SEND(RI, RJ, m1, FAST_MUL2(n, B), A, IncA, MsgType);
      A += m1*IncA;      		m = FAST_MOD(m, B);	
    }
    if (m > 0)
      SEND(RI, RJ, 1, m*n, A, IncA, MsgType);   
  }    
  TRACE_EXIT;
} /* SendBlkMat() */


void RecvBlkMat(RI, RJ, Add, m, B0, B, BInc, n, alpha, A, MsgType)
  int RI, RJ, Add, m, B0, B, BInc, n; FDATA  *alpha, A[]; int MsgType; 
{ int r=0;				
#ifdef ASSERT
  int expsize= m*n;
#endif  
  COMMON_ASSERT_TR_E(
    (TraceStr, "(%d %d,%d,%d[%d %d %d]x%d,%1.1f,a%d,%d)",
     RI, RJ, Add, m, B0, B, BInc, n, pFD_VAL(alpha), FD_ADR(A), MsgType),
    RIJOK(RI, RJ) && BlkOK(B0, B, BInc), 0, "RecvBlkMat", 4, TRACE_COMMUN
  );

  if (m==0 || n==0) {
    TRACE_RETURN;
  }  
  NORMALIZE(RI, RJ);
  if (m+B0<=B || BInc==1) {
#ifdef APLIB_STRDE			/*here, m=1 optim. important*/
# undef SplitMatUnpack_
# define SplitMatUnpack_ SplitMatUnpack_f
#endif
    RECV_ADD_(Add, RI, RJ, m, n, alpha, A, n, 1, MsgType, r);
#ifdef APLIB_STRDE			/*from here on, m=1 optim. unimportant*/
# undef SplitMatUnpack_
# define SplitMatUnpack_ SplitMatUnpack
#endif
  } else {
    int IncA = FAST_MUL1(B, n*BInc);        	  	
    if (B0 > 0) {
      int m1 = MIN(B-B0, m);
      RECV_ADD_(Add, RI, RJ, 1, m1*n, alpha, A, IncA, 1, MsgType, r); 
      A += (B*BInc-B0)*n;		m -= m1;
    }
    if (m >= B) {
      int m1 = FAST_DVD(m, B);            	
      RECV_ADD_(Add, RI, RJ, m1, FAST_MUL2(n, B), alpha, A, IncA, 1, MsgType, r);  
      A += m1*IncA;      		m = FAST_MOD(m, B);	
    }
    if (m > 0)
      RECV_ADD_(Add, RI, RJ, 1, m*n, alpha, A, IncA, 1, MsgType, r);   
  }  
  COMMON_ASSERT_TR_X(
    (TraceStr, " recv: exp=%d, act=%d", expsize, r),
     expsize==r, COMMUN_ERROR_FLAG, "RecvBlkMat", 5, TRACE_COMMUN
  );
} /* RecvBlkMat() */
	       
/*********************** 2-dim block matrix routines ************************/
	       
void Send2BlkMat(RI, RJ, M, N, alpha, A, RB0, RB, RBInc, CB0, CB, CBInc)
  int RJ, RI, RB0, RB, RBInc, CB0, CB, CBInc; 
  FDATA *alpha; pLocMat A;
{ int r=FAST_MUL1(M, N); 
  COMMON_ASSERT_TR_E(
    (TraceStr, "(%d %d,%dx%d,%1.1f,a%d %d %d,%d %d %d,%d %d %d)", RI, RJ, M, N, 
     pFD_VAL(alpha), FD_ADR(pLM_ADR(A)), pLM_RInc(A), pLM_CInc(A), RB0, RB, RBInc, CB0, CB, CBInc),
    RIJOK(RI, RJ) && pLM_BlkIncOK(M, N, A, RB0, RB, RBInc, CB0, CB, CBInc),
    0, "Send2BlkMat", 4, TRACE_COMMUN
  );	

  if (M==0 || N==0) {
    TRACE_RETURN;
  }  
  NORMALIZE(RI, RJ);
  if (pFD_UNITY(alpha) && ((pLM_CInc(A)==1) && (N+CB0<=CB || CBInc==1) && 
      (M==1 || (N==pLM_RInc(A) && (M+RB0<=RB || RBInc==1))))) {
    SEND(RI, RJ, 1, r, pLM_ADR(A), r, TRANSMSGTYPE);
  } else {
    FDATA *Bx; 				LocMat B;
    MatMalloc_FT("Send2BlkMat", 1, &Bx, r);		
    BlkMatStridePack(0, M, N, A, alpha, pMkLocMat(&B, Bx, N, 1), 
                     RB0, RB, RBInc, CB0, CB, CBInc);              
    SEND(RI, RJ, 1, r, Bx, r, TRANSMSGTYPE);
    MatFree(Bx);
  }  
  TRACE_EXIT;
} /* Send2BlkMat */


void Recv2BlkMat(RI, RJ, Add, M, N, B, RB0, RB, RBInc, CB0, CB, CBInc)
  int RJ, RI, Add, M, N, RB0, RB, RBInc, CB0, CB, CBInc; 
  pLocMat B; 
{ int r=0;   		    		
  FDATA *Ax;				LocMat A;
  COMMON_ASSERT_TR_E(
    (TraceStr, "(%d %d,%d,%dx%d,a%d %d %d,%d %d %d,%d %d %d)",
     RI, RJ, Add, M, N, FD_ADR(pLM_ADR(B)), pLM_RInc(B), pLM_CInc(B), 
     RB0, RB, RBInc, CB0, CB, CBInc),
    RIJOK(RI, RJ) && pLM_BlkIncOK(M, N, B, RB0, RB, RBInc, CB0, CB, CBInc),
    0, "Recv2BlkMat", 4, TRACE_COMMUN
  );	

  if (M==0 || N==0) {
    TRACE_RETURN;
  } 
  NORMALIZE(RI, RJ);
  if (N+CB0<=CB) CBInc==1;
  if (M+RB0<=RB) RBInc==1;
#if defined(FTN_BLACS) 
  if (!Add && (pLM_CInc(B)==1) && (CBInc==1) && (RBInc==1)) {
    RECV(RI, RJ, M, N, pLM_ADR(B), pLM_RInc(B), TRANSMSGTYPE, r);
  } else {			
    int mn = FAST_MUL1(M, N);	
    MatMalloc_FT("Recv2BlkMat", 1, &Ax, mn);		
    RECV(RI, RJ, 1, mn, Ax, mn, TRANSMSGTYPE, r);
    BlkMatStrideUnpack(Add, FAST_DVDEQ(r ,N), N, pMkLocMat(&A, Ax, N, 1), NULL, B,
                       RB0, RB, RBInc, CB0, CB, CBInc);              
    MatFree(Ax);
  }
#elif defined(APLIB_STD) || defined(APLIB_LINUX) || defined(APLIB_PLUS_STRIDE) 
# if defined(APLIB_PLUS_STRIDE)
  { int r1, s; FDATA *b2;
    r = FAST_MUL1(M, N*sizeof(FDATA));
    r = xy_stride_recv(RJ, RI, TRANSMSGTYPE, 1, r, &s, &r1, &Ax, &b2);
    r /= sizeof(FDATA);
  }
# else    
  Ax = (FDATA *) r_rrecv(RJ, RI, gettid(), TRANSMSGTYPE);
  r = getmsize()/sizeof(FDATA);
# endif
  pMkLocMat(&A, Ax, N, 1);
  if (RBInc==1 && CBInc==1 && ((int)Ax% sizeof(FDATA)==0)) 
    (Add? MatAdd: MatCopy)(FAST_DVDEQ(r, N), N, &A, NULL, B);
  else
    BlkMatStrideUnpack(Add, FAST_DVDEQ(r, N), N, &A,
                       NULL, B, RB0, RB, RBInc, CB0, CB, CBInc);              
#elif defined(APLIB_STRIDE) 
  { FDATA *b2; int r1; int s;
    r = FAST_MUL1(M, N*sizeof(FDATA));
    r = xy_stride_recv(RJ, RI, TRANSMSGTYPE, 1, r, &s, &r1, &Ax, &b2)/sizeof(FDATA);
    r1 /= sizeof(FDATA);    		/*assert s=0*/
    if (r1==r && RBInc==1 && CBInc==1) 
      (Add? MatAdd: MatCopy)(FAST_DVDEQ(r, N), N, pMkLocMat(&A, Ax, N, 1), NULL, B);
    else   				/*message was split (AP ring buffer)*/
      SplitBlkMatStrideUnpack(Add, N, r, Ax, r1, b2, N, NULL,
  	                      B, RB0, RB, RBInc, CB0, CB, CBInc); 
  }
#else 
  /*this function should not have been called & following assertion will fail*/;
#endif  
 
  COMMON_ASSERT_TR_X(
    (TraceStr, " recv: exp=%d, act=%d", M*N, r), 
    r==M*N, COMMUN_ERROR_FLAG, "Recv2BlkMat", 5, TRACE_COMMUN
  );
} /* Recv2BlkMat */


void RecvBrd2BlkMat(RowCol, P0, M, N, B, RB0, RB, RBInc, CB0, CB, CBInc)
  int RowCol, P0, M, N, RB0, RB, RBInc, CB0, CB, CBInc; 
  pLocMat B;
{ int r=0;
  COMMON_ASSERT_TR_E(
    (TraceStr, "(%d,%d,%dx%d,a%d %d %d,%d %d %d,%d %d %d)", RowCol, P0, M, N, 
     FD_ADR(pLM_ADR(B)), pLM_RInc(B), pLM_CInc(B), RB0, RB, RBInc, CB0, CB, CBInc),
    RCOK(RowCol) && pLM_BlkIncOK(M, N, B, RB0, RB, RBInc, CB0, CB, CBInc) &&
    VNumCells[RowCol]>1 && P0!=0, 0, "RecvBrd2BlkMat", 4, TRACE_COMMUN
  );	
  
  if (M==0 || N==0) {
    TRACE_RETURN;
  }  
  if ((pLM_CInc(B)==1) && (N+CB0<=CB || CBInc==1) && (M+RB0<=RB || RBInc==1)) {
    BRD(RowCol, P0, M, N, pLM_ADR(B), pLM_Ld(B), r);
  } else {
#if defined(FTN_BLACS) || defined(APLIB_STD) || defined(APLIB_LINUX)
    FDATA *Ax;				LocMat lA_;
    MatMalloc_FT("RecvBrd2BlkMat", 1, &Ax, FAST_MUL1(M, N));		
    BRD(RowCol, P0, M, N, Ax, N, r);
    BlkMatStrideUnpack(0, FAST_DVDEQ(r, N), N,  pMkLocMat(&lA_, Ax, N, 1), 
                       NULL, B, RB0, RB, RBInc, CB0, CB, CBInc);              
    MatFree(Ax);
#elif defined(APLIB_STRIDE)
    FDATA *b1, *b2; int r1; int s;
    r = brd_stride_recv(RowCol, SRCABSCID(RowCol, P0), M, N*sizeof(FDATA), 
    			&s, &r1, &b1, &b2) / sizeof(FDATA);	
    /* assert s=0; also the omission of partitioning over N on AP+ assumes
       that send sender sent from a contiguous message */			
    SplitBlkMatStrideUnpack(0, N, r, b1, r1/sizeof(FDATA), b2, N, NULL,
  	                    B, RB0, RB, RBInc, CB0, CB, CBInc);
#elif defined(APLIB_PLUS_STRIDE)
    FDATA *b1, *b2; int r_, r1; int s;
    int i = 0, nb, I0_ = SRCABSCID(RowCol, P0); LocMat lB_, lA_;
    LM_Copy(0, lB_, *B);
    do {
      r_ = brd_stride_recv(RowCol, I0_, M, (N)*sizeof(FDATA),&s, &r1, &b1, &b2);
      r_ /= sizeof(FDATA); nb = FAST_DVD(r_, M);		
      BlkMatStrideUnpack(0, M, nb, pMkLocMat(&lA_, b1, nb, 1),			
                         NULL, &lB_, RB0, RB, RBInc, CB0, CB, CBInc);  
      if (i < (N)) {                  
        pLocSubMat(&lB_, 0, nb);
        if (CBInc > 1) 
          CB0 = (CB0 + nb) % CBInc;
      }                  
      i += nb; r += r_; 				
    } while ((i < N)  && FAST_MOD(r_, M)==0);		
#else
  /*this function should not have been called & following assertion will fail*/;
#endif  
  } /* else (non-trivial stride in B required */
 
  COMMON_ASSERT_TR_X(
    (TraceStr, " recv: exp=%d, act=%d", M*N, r),
    r==M*N, COMMUN_ERROR_FLAG, "RecvBrd2BlkMat", 5, TRACE_COMMUN
  );
} /* RecvBrd2BlkMat */


