/* @(#) All Right Reserved, Copyright (c) 1997 Australian National University */
/* @(#)                     Copyright (c) 1997 Fujitsu Laboratories Ltd. */
/*  */
/*        */
/*                        Permission Notice */
/*  */
/*                Australian National University */
/*                Canberra ACT 0200 Australia */
/*                           and */
/*                Fujitsu Laboratories Ltd. */
/*                4-1-1 Kamikodanaka Nakahara */
/*                Kawasaki 211-88 Japan */
/*  */
/*   Permission to use, copy, modify and distribute this software (including  */
/*   its documentation) without fee is hereby granted under the following terms */
/*   and conditions: */
/*  */
/*   1. This permission notice and all of the copyright notices in this */
/*      software must appear in all copies and derived versions of all */
/*      or a part of this software. */
/*  */
/*   2. Sales of this software, whether incorporated in other products as a */
/*      part thereof or not, is prohibited. */
/*  */
/*   3. On publication or presentation in any form of the results obtained */
/*      by use of this software, an explicit reference to this software and */
/*      its copyright owner shall be made. */
/*  */
/*   4. THIS SOFTWARE IS PROVIDED 'AS IS' AND THE AUSTRALIAN NATIONAL UNIVERSITY */
/*      AND FUJITSU LABORATORIES LTD. MAKE NO WARRANTIES, EITHER EXPRESSED OR */
/*      IMPLIED, AS TO ANY MATTER WHATSOEVER REGARDING THIS SOFTWARE,  */
/*      INCLUDING, WITHOUT LIMITATION, WARRANTIES OF MERCHANTABILITY OR FITNESS */
/*      FOR ANY PARTICULAR PURPOSE. */
/*      IN NO EVENT SHALL THE AUSTRALIAN NATIONAL UNIVERSITY OR FUJITSU */
/*      LABORATORIES LTD. BE LIABLE FOR ANY CLAIM FROM A THIRD PARTY, OR */
/*      FOR SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES WHATSOEVER RESULTING */
/*      FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN ACTION OF CONTRACT, */
/*      NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION */
/*      WITH THE USE, COPYING OR MODIFICATION OF THIS SOFTWARE. */
/*  */

/************************************************************************/
/*       	FDATAMalloc: FDATA workspace meory allocation 		*/
/************************************************************************/
/* SCCS INFO @(#)FDATAMalloc.h 1.1         last modified P. Strazdins, 97/02/17
   first created: P. Strazdins 97/02
*/
/* essentially a front-end to the standard malloc library.
 * Due to serious call overheads observed in the implementation 
 * of this library, this module is required to improve 
 * performance for small workspaces.
 */

#ifndef __FDATAMalloc_INCL__
#define __FDATAMalloc_INCL__  

#include "FDATASetUp.h"

/********************* exported symbols  ***********************************/

#if FDATATYPE==SINGLE
#define MatAllocInit	DSMatAllocInit
#define MatFreeBuf	DSMatFreeBuf
#define MatMalloc_f	DSMatMalloc_f
#define MatFree		DSMatFree
#else
#define MatAllocInit	DDMatAllocInit
#define MatFreeBuf	DDMatFreeBuf
#define MatMalloc_f	DDMatMalloc_f
#define MatFree		DDMatFree
#endif

/****************************************************************/  

/* the following refer to a `pre-malloced' workspace W, from which 
 * `small' requests for memory will be serviced where posible.
 * `Small' means the size < FastWSSize.
 * Default values:  
 *	FastWSSize = FastWSSize_DEF
 *	|W| = 4*FastWSSize
 */

#define FastWSSize_DEF	512	

void MatMallocInit STDC_PARAMS( (int ParamType, int ParamVal) );
/* pre:  ParamType IN {0, 1}; ParamVal > 0
 * post: if ParamType=0, |W| = ParamVal
 *	 if ParamType=1, FastWSSize = ParamVal
 * notes: MallocInit(0, _) will only have effect if currently no unfreed
 *        allocations are from W
 */

void MatFreeBuf STDC_PARAMS( (void) );
/* post:  storage associated with W is freed.
 * notes: any unfreed blocks in W must not be used after this call 
 */


#define FD_ALLOC_FALSE	0
#define FD_ALLOC_TRUE	(~FD_ALLOC_FALSE)

/* MatMalloc() is recommended interface from outside the DBLAS;
 * of for inside the DBLAS wherever s/w overheads is not an issue,
 * ie. Size is always large 
 */
#define MatMalloc(Where, CrucialAlloc, A, Size)			\
	MatMalloc_f(Where, CrucialAlloc, A, Size, 		\
		    FD_ALLOC_FALSE, FD_ALLOC_FALSE)
		    
/* MatMalloc_F() is for fast allocation where Size may be < FastWSSize,
 * ie. a vector or small matrix 
 */	   	   
#define MatMalloc_F(Where, CrucialAlloc, A, Size)         	\
	MatMalloc_f(Where, CrucialAlloc, A, Size, 		\
		    FD_ALLOC_TRUE, FD_ALLOC_FALSE)

/* MatMalloc_FT() is for fast allocation where Size may be < |W|
 * and it can be guaranteed the coresponding MatFree() will be called
 * before any future malloc.
 */ 			   
#define MatMalloc_FT(Where, CrucialAlloc, A, Size)         	\
	MatMalloc_f(Where, CrucialAlloc, A, Size, 		\
		    FD_ALLOC_TRUE, FD_ALLOC_TRUE)
			   
int MatMalloc_f STDC_PARAMS( 
	(char *Where, int CrucialAlloc, FDATA **A, int Size,
	int UseFastMalloc, int ImmedFree) 
);
/* pre:  UseFastMalloc, UseFastMalloc IN {FD_ALLOC_FALSE, FD_ALLOC_TRUE}
 * post: returns whether (Size>0) FDATA elements allocated at *A;
 *	 if allocation fails, an error messgage is printed; 
 *	 if CrucialAlloc, DBLAS_EXIT() is called also.
 * notes:if UseFastMalloc==FD_ALLOC_TRUE, fast allocation methods
 *	 will be used if there is sufficient space insside the workspace.
 *	
 */ 

void MatFree STDC_PARAMS( (FDATA *A) );
/* pre:  a matrix (of base address) A was allocated from MatMalloc()
 * post: frees matrix which begins at A
 */ 

#endif /*__FDATAMalloc_INCL__*/
