/* @(#) All Right Reserved, Copyright (c) 1997 Australian National University */
/* @(#)                     Copyright (c) 1997 Fujitsu Laboratories Ltd. */
/*  */
/*        */
/*                        Permission Notice */
/*  */
/*                Australian National University */
/*                Canberra ACT 0200 Australia */
/*                           and */
/*                Fujitsu Laboratories Ltd. */
/*                4-1-1 Kamikodanaka Nakahara */
/*                Kawasaki 211-88 Japan */
/*  */
/*   Permission to use, copy, modify and distribute this software (including  */
/*   its documentation) without fee is hereby granted under the following terms */
/*   and conditions: */
/*  */
/*   1. This permission notice and all of the copyright notices in this */
/*      software must appear in all copies and derived versions of all */
/*      or a part of this software. */
/*  */
/*   2. Sales of this software, whether incorporated in other products as a */
/*      part thereof or not, is prohibited. */
/*  */
/*   3. On publication or presentation in any form of the results obtained */
/*      by use of this software, an explicit reference to this software and */
/*      its copyright owner shall be made. */
/*  */
/*   4. THIS SOFTWARE IS PROVIDED 'AS IS' AND THE AUSTRALIAN NATIONAL UNIVERSITY */
/*      AND FUJITSU LABORATORIES LTD. MAKE NO WARRANTIES, EITHER EXPRESSED OR */
/*      IMPLIED, AS TO ANY MATTER WHATSOEVER REGARDING THIS SOFTWARE,  */
/*      INCLUDING, WITHOUT LIMITATION, WARRANTIES OF MERCHANTABILITY OR FITNESS */
/*      FOR ANY PARTICULAR PURPOSE. */
/*      IN NO EVENT SHALL THE AUSTRALIAN NATIONAL UNIVERSITY OR FUJITSU */
/*      LABORATORIES LTD. BE LIABLE FOR ANY CLAIM FROM A THIRD PARTY, OR */
/*      FOR SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES WHATSOEVER RESULTING */
/*      FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN ACTION OF CONTRACT, */
/*      NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION */
/*      WITH THE USE, COPYING OR MODIFICATION OF THIS SOFTWARE. */
/*  */

/************************************************************************/
/*                     FDATAOps.h		                        */
/************************************************************************/
/* SCCS INFO @(#)FDATAOps.h 1.2         last modified P. Strazdins, 96/04/29
   first created: P. Strazdins 91/12
*/
/* header file for BLAS Floating Point Data Operations (macros)
	the FDATA data type should only be directly accessed through
	these macros; in this way, C code can be indepenent of FDATA  
 
  	assumes FDATATYPE, FDATA, FD_BASE are already defined 
*/
/************************************************************************/

#ifndef __FDATAOps_INCL__

#define __FDATAOps_INCL__

#include "FDATASetUp.h"

#if (FDATATYPE==SINGLE || FDATATYPE==DOUBLE)

/* the FD_ALPHA macros are designed to be more efficient than their
   (equivalent) FD_MULSET() & FD_SCALE() counterparts; 
   to use in a procedure, declare and initialize a register variable 
   FD_ALPHA using FD_ALPHA_REGS(); then  FD_ALPHA_SCALE() &  
   FD_ALPHA_MULSET() can be used within that procedure
*/

#define FD_ALPHA_REGS(a) 	register FD_BASE FD_ALPHA=a;
	/* FDATA a; */

#define FD_ALPHA_SCALE(c) 	{ (c) *=  FD_ALPHA; }
	/* FDATA c; */

#define FD_ALPHA_MULSET(c,op,b) { (c) op  FD_ALPHA * (b); }
	/* FDATA c, b; 
	   op is =, +=, *=. -= */
                
#define FD_ALPHA_SCALEADD(c, a, b) { (c) = (a) + FD_ALPHA * (b); }
	/* FDATA c, b; */
                
#define FD_MULSET(c,op,a,b) 	{ (c) op (a) * (b); }
	/* FDATA c, a, b; 
	   op is =, +=, *=. -= */

#define FD_SCALE(c,a) 		{ (c) *= (a); }
	/* FDATA c, a; */

#define FD_RECIP(c,a) 		{ (c) = (c) / (a); }
	/* FDATA c, a; */
                        
#define FD_NEG(c,a) 		{ (c) = -(a); }
	/* FDATA c, a; */

#define FD_SET(c,op,a) 		{ (c) op (a); }
	/* FDATA c, a; 
	   op is =, +=, *=. -= */

#define FD_SETOP(c, b, op, a) 	{ (c) = (b) op (a); }
	/* FDATA c, a, b; 
	   op is +, - */

#define FD_SET_REAL(c, s) 	{ (c) = s; }
	/* FDATA c; FD_BASE s; */

#define FD_SET_IMAG(c, s) 
	/* FDATA c; FD_BASE s; */

#define FD_EQ(a,b) 		((a) == (b))
	/* int FD_EQ(); 
	   FDATA a, b; */ 

#define FD_ABS(a) 		((a) < 0.0? -(a): (a))
	/* FD_BASE FD_ABS();
	   FDATA a; */

#define FD_REAL(a)              (a)
        /* FD_BASE FD_REAL();
           FDATA a; */

#define FD_IMAG(a)              (0.0)
        /* FD_BASE FD_IMAG();
           FDATA a; */


#else 	/* define equivalent complex operations */


#define FD_ALPHA_REGS(a) 	register FD_BASE FD_ALPHAr=(a).r, \
					 FD_ALPHAi=(a).i;
        /* FDATA a; */

#define FD_ALPHA_SCALE(c) 	{ register FD_BASE cr_=(c).r, ci_=(c).i; \
        			  (c).r = cr_*FD_ALPHAr - ci_*FD_ALPHAi; \
                		  (c).i = ci_*FD_ALPHAr + cr_*FD_ALPHAi; }
        /* FDATA c, b;
           op is =, +=, *=. -= */
                   
#define FD_ALPHA_MULSET(c,op,b) { register FD_BASE br_=(b).r, bi_=(b).i;  \
       		 		  (c).r op br_*FD_ALPHAr - bi_*FD_ALPHAi; \
                		  (c).i op bi_*FD_ALPHAr + br_*FD_ALPHAi; }
	/* FDATA c, a, b; 
	   op is =, +=, *=. -= */
                
                
#define FD_ALPHA_SCALEADD(c, a, b) { register FD_BASE br_=(b).r, bi_=(b).i; \
       		 		  (c).r = (a).r + br_*FD_ALPHAr - bi_*FD_ALPHAi; \
                		  (c).i = (a).i + bi_*FD_ALPHAr + br_*FD_ALPHAi; }
	/* FDATA c, b; */

#define FD_MULSET(c,op,a,b) 	{ register FD_BASE ar_=(a).r, br_=(b).r, \
						   ai_=(a).i, bi_=(b).i; \
				  (c).r op ar_*br_ - ai_*bi_; 		 \
                		  (c).i op ai_*br_ + ar_*bi_; 		 }
	/* FDATA c, a, b; 
	   op is =, +=, *=. -= */
                		  
#define FD_SCALE(c,a) 		{ register FD_BASE ar_=(a).r, cr_=(c).r, \
					 	   ai_=(a).i, ci_=(c).i; \
        			  (c).r = ar_*cr_ - ai_*ci_;		 \
                		  (c).i = ai_*cr_ + ar_*ci_;		 }
	/* FDATA c, a; */

#define FD_RECIP(c,a) 		{ register FD_BASE ar_=(a).r, ai_=(a).i, t_; \
        			  t_ = ar_*ar_ + ai_*ai_;		 \
                		  (c).r = ar_ / t_;			 \
                        	  (c).i = - ai_ / t_; 			 }
	/* FDATA c, a; */
                        
#define FD_NEG(c,a) 		{ (c).r = -(a).r; (c).i = -(a).i; }
	/* FDATA c, a; */

#define FD_SET(c,op,a) 		{ (c).r op (a).r; (c).i op (a).i; }
	/* FDATA c, a; */

#define FD_SETOP(c, b, op, a) 	{ (c).r = (b).r op (a).r; \
				  (c).i = (b).i op (a).i; }
	/* FDATA c, a, b; 
	   op is +, - */

#define FD_SET_REAL(c, s) 	{ (c).r = s; }
	/* FDATA c; FD_BASE s; */

#define FD_SET_IMAG(c, s) 	{ (c).i = s; }
	/* FDATA c; FD_BASE s; */

#define FD_EQ(a,b) 		((a).r == (b).r && (a).i == (b).i)
	/* int FD_EQ(); 
	   FDATA a, b; */ 

#define FD_ABS(a) 		(((a).r < 0.0? -(a).r: (a).r) + \
				 ((a).i < 0.0? -(a).i: (a).i)) 
	/* FD_BASE FD_ABS();
	   FDATA a; */

#define FD_REAL(a)              ((a).r)
        /* FD_BASE FD_REAL();
           FDATA a; */

#define FD_IMAG(a)              ((a).i)
        /* FD_BASE FD_IMAG();
           FDATA a; */
#endif /*(FDATAYPE...)*/

#endif /*__FDATAOps_INCL__*/
