/* @(#) All Right Reserved, Copyright (c) 1997 Australian National University */
/* @(#)                     Copyright (c) 1997 Fujitsu Laboratories Ltd. */
/*  */
/*        */
/*                        Permission Notice */
/*  */
/*                Australian National University */
/*                Canberra ACT 0200 Australia */
/*                           and */
/*                Fujitsu Laboratories Ltd. */
/*                4-1-1 Kamikodanaka Nakahara */
/*                Kawasaki 211-88 Japan */
/*  */
/*   Permission to use, copy, modify and distribute this software (including  */
/*   its documentation) without fee is hereby granted under the following terms */
/*   and conditions: */
/*  */
/*   1. This permission notice and all of the copyright notices in this */
/*      software must appear in all copies and derived versions of all */
/*      or a part of this software. */
/*  */
/*   2. Sales of this software, whether incorporated in other products as a */
/*      part thereof or not, is prohibited. */
/*  */
/*   3. On publication or presentation in any form of the results obtained */
/*      by use of this software, an explicit reference to this software and */
/*      its copyright owner shall be made. */
/*  */
/*   4. THIS SOFTWARE IS PROVIDED 'AS IS' AND THE AUSTRALIAN NATIONAL UNIVERSITY */
/*      AND FUJITSU LABORATORIES LTD. MAKE NO WARRANTIES, EITHER EXPRESSED OR */
/*      IMPLIED, AS TO ANY MATTER WHATSOEVER REGARDING THIS SOFTWARE,  */
/*      INCLUDING, WITHOUT LIMITATION, WARRANTIES OF MERCHANTABILITY OR FITNESS */
/*      FOR ANY PARTICULAR PURPOSE. */
/*      IN NO EVENT SHALL THE AUSTRALIAN NATIONAL UNIVERSITY OR FUJITSU */
/*      LABORATORIES LTD. BE LIABLE FOR ANY CLAIM FROM A THIRD PARTY, OR */
/*      FOR SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES WHATSOEVER RESULTING */
/*      FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN ACTION OF CONTRACT, */
/*      NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION */
/*      WITH THE USE, COPYING OR MODIFICATION OF THIS SOFTWARE. */
/*  */

/************************************************************************/
/*                     FDATASetUp.h		                        */
/************************************************************************/
/* SCCS INFO @(#)FDATASetUp.h 4.3         last modified P. Strazdins, 97/02/24
   first created: P. Strazdins 91/12
*/
/* header file for BLAS-3 Floating Point Data file 
   	assumes FDATATYPE is already defined 
  	written by P. Strazdins, from 11/12/91				*/
/************************************************************************/

#ifndef __FDATASetUp_INCL__ 

#define __FDATASetUp_INCL__

#include "GenMacros.h"
#include "IntOps.h"

#define SINGLE 		0	/* code for FDATA=float			*/
#define DOUBLE 		1	/* code for FDATA=double                */
#define COMPLEX 	2	/* code for FDATA=complex               */
#define DOUBLE_COMPLEX 	3	/* code for FDATA=double complex        */

/*  FD_BASE is base type of floating point data */
#if (FDATATYPE==SINGLE || FDATATYPE==COMPLEX)
#define FD_BASE float
#else
#define FD_BASE double
#endif

#ifndef FDATA			/* dont redefine FDATA & numeric constants */
/* FDATA is numeric type of floating point data */
# if (FDATATYPE==SINGLE || FDATATYPE==DOUBLE)
#  define FDATA FD_BASE
# else
typedef struct FD_complex {  
		FD_BASE r, 	/* real part				*/
			i; 	/*imaginary part			*/
		} FD_complex_struct;
#  define FDATA FD_complex_struct
# endif
#endif

typedef FDATA *FDATAPtr; 

#if (FDATATYPE==SINGLE)		/* needed as cannot use sizeof(FDATA) in #if */
# define FD_SIZE 4
# define LG_FD_SIZE 2
#elif (FDATATYPE==DOUBLE || FDATATYPE==COMPLEX)
# define FD_SIZE 8
# define LG_FD_SIZE 3
#else
# define FD_SIZE 16
# define LG_FD_SIZE 4
#endif /*FDATATYPE*/

#if (FDATATYPE==SINGLE)
# define FD_1	SFD_1
# define FD_M1	SFD_M1
# define FD_0	SFD_0
#elif (FDATATYPE==DOUBLE)
# define FD_1	DFD_1
# define FD_M1	DFD_M1
# define FD_0	DFD_0
#endif

extern FDATA FD_1 /*1.0*/, FD_M1 /*-1.0*/, FD_0 /*0.0 */;

#define FD_ADR(A) (((int) (A))/sizeof(FDATA))

/* convention that a pointer to a scaling factor represents unity if NULL */
#define pFD_UNITY(pa) 		((pa) == NULL || *(pa) == FD_1 )
	/* int pFD_UNITY(); 
	   FDATA *pa; */ 
#define pFD_VAL(pa)		((pa) == NULL? FD_1: *(pa))


#endif /*__FDATASetUp_INCL__*/
