/* @(#) All Right Reserved, Copyright (c) 1997 Australian National University */
/* @(#)                     Copyright (c) 1997 Fujitsu Laboratories Ltd. */
/*  */
/*        */
/*                        Permission Notice */
/*  */
/*                Australian National University */
/*                Canberra ACT 0200 Australia */
/*                           and */
/*                Fujitsu Laboratories Ltd. */
/*                4-1-1 Kamikodanaka Nakahara */
/*                Kawasaki 211-88 Japan */
/*  */
/*   Permission to use, copy, modify and distribute this software (including  */
/*   its documentation) without fee is hereby granted under the following terms */
/*   and conditions: */
/*  */
/*   1. This permission notice and all of the copyright notices in this */
/*      software must appear in all copies and derived versions of all */
/*      or a part of this software. */
/*  */
/*   2. Sales of this software, whether incorporated in other products as a */
/*      part thereof or not, is prohibited. */
/*  */
/*   3. On publication or presentation in any form of the results obtained */
/*      by use of this software, an explicit reference to this software and */
/*      its copyright owner shall be made. */
/*  */
/*   4. THIS SOFTWARE IS PROVIDED 'AS IS' AND THE AUSTRALIAN NATIONAL UNIVERSITY */
/*      AND FUJITSU LABORATORIES LTD. MAKE NO WARRANTIES, EITHER EXPRESSED OR */
/*      IMPLIED, AS TO ANY MATTER WHATSOEVER REGARDING THIS SOFTWARE,  */
/*      INCLUDING, WITHOUT LIMITATION, WARRANTIES OF MERCHANTABILITY OR FITNESS */
/*      FOR ANY PARTICULAR PURPOSE. */
/*      IN NO EVENT SHALL THE AUSTRALIAN NATIONAL UNIVERSITY OR FUJITSU */
/*      LABORATORIES LTD. BE LIABLE FOR ANY CLAIM FROM A THIRD PARTY, OR */
/*      FOR SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES WHATSOEVER RESULTING */
/*      FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN ACTION OF CONTRACT, */
/*      NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION */
/*      WITH THE USE, COPYING OR MODIFICATION OF THIS SOFTWARE. */
/*  */

/************************************************************************/
/*		FtnBLACS: C prototypes for limited Fortran cell BLACS 	*/ 
/************************************************************************/
/* SCCS INFO @(#)FtnBLACS.h 1.5         last modified P. Strazdins, 97/04/18
   first created: P. Strazdins 96/01/19
*/
/* compilation:	FDATATYPE must be defined
 *		if C_CALL_FTN_NO_CHANGE defined (eg for R8000's), strip `_'
 *		off FORTRAN procedure name		
 */

#ifndef __FtnBLACS_INCL__
#define __FtnBLACS_INCL__

#include "FDATASetUp.h"
   
#if FDATATYPE==SINGLE
# define FTN_GESD2D	sgesd2d_
# define FTN_GERV2D	sgerv2d_ 
# define FTN_GEBS2D	sgebs2d_
# define FTN_GEBR2D	sgebr2d_
# define FTN_GSUM2D	sgsum2d_
# define FTN_GAMX2D	sgamx2d_
#else
# define FTN_GESD2D	dgesd2d_
# define FTN_GERV2D	dgerv2d_ 
# define FTN_GEBS2D	dgebs2d_
# define FTN_GEBR2D	dgebr2d_
# define FTN_GSUM2D	dgsum2d_
# define FTN_GAMX2D	dgamx2d_
#endif /*FDATATYPE*/
#define  FTN_BLACS_GRIDINFO	blacs_gridinfo_
#define  FTN_BLACS_BARRIER	blacs_barrier_
#define  FTN_BLACS_ABORT	blacs_abort_
#define  FTN_BLACS_GET		blacs_get_
#define  FTN_BLACS_GRIDINIT	blacs_gridinit_
#define	 FTN_DWALLTIME00	dwalltime00_
#define  FTN_BLACS_PINFO	blacs_pinfo_
#define  FTN_IGEBS2D		igebs2d_
#define  FTN_IGEBR2D		igebr2d_

#ifdef C_CALL_FTN_NO_CHANGE
# if FDATATYPE==SINGLE
# define sgesd2d_	sgesd2d
# define sgerv2d_	sgerv2d
# define sgebs2d_	sgebs2d
# define sgebr2d_	sgebr2d
# define sgsum2d_	sgsum2d
# define sgamx2d_	sgamx2d
#else
# define dgesd2d_	dgesd2d
# define dgerv2d_	dgerv2d
# define dgebs2d_	dgebs2d
# define dgebr2d_	dgebr2d
# define dgsum2d_	dgsum2d
# define dgamx2d_	dgamx2d
# endif /*FDATATYPE*/
# define blacs_gridinfo_ blacs_gridinfo
# define blacs_barrier_ blacs_barrier
# define blacs_abort_	blacs_abort
# define blacs_get_	blacs_get 
# define blacs_gridinit_ blacs_gridinit 
# define blacs_pinfo_	blacs_pinfo
# define dwalltime00_	dwalltime00
# define igebs2d_	igebs2d
# define igebr2d_	igebr2d
#endif /*C_CALL_FTN_NO_CHANGE*/


void FTN_GESD2D STDC_PARAMS( 
	(int *ICONTXT, int *M, int *N, FDATA A[], int *LDA, 
	 int *RDEST, int *CDEST) 
);

void FTN_GERV2D STDC_PARAMS( 
	(int *ICONTXT, int *M, int *N, FDATA A[], int *LDA, 
	 int *RDEST, int *CDEST) 
);

void FTN_GEBS2D STDC_PARAMS( 
	(int *ICONTXT, char *SCOPE, char *TOP, int *M, int *N, 
	 FDATA A[], int *LDA)
);

void FTN_GEBR2D STDC_PARAMS( 
	(int *ICONTXT, char *SCOPE, char *TOP, int *M, int *N, 
	 FDATA A[], int *LDA, int *RSRC, int *CSRC)
);

void FTN_GSUM2D STDC_PARAMS(
	(int *ICONTXT, char *SCOPE, char *TOP, int *M, int *N, 
	 FDATA A[], int *LDA, int *RDEST, int *CDEST)
);

void FTN_GAMX2D STDC_PARAMS(
	(int *ICONTXT, char *SCOPE, char *TOP, int *M, int *N, FDATA A[], 
	 int *LDA, int *RA, int *CA, int *RCFLAG, int *RDEST, int *CDEST)
);

void FTN_BLACS_GRIDINFO STDC_PARAMS(
        (int *ICONTXT, int *NPROW, int *NPCOL, int *MYPROW, int *MYPCOL)
);

void FTN_BLACS_BARRIER STDC_PARAMS( (int *ICONTXT, char *SCOPE ) );

void FTN_BLACS_ABORT STDC_PARAMS( (int *ICONTXT, int *ERRORNUM ) );

void FTN_BLACS_GET STDC_PARAMS( (int *ICONTXT, int *WHAT, int *VAL ) );

void FTN_BLACS_GRIDINIT STDC_PARAMS( 
        (int *ICONTXT, char *ORDER, int *NPCOL, int *MYPROW)
);

double FTN_DWALLTIME00();

void FTN_IGEBS2D STDC_PARAMS( 
	(int *ICONTXT, char *SCOPE, char *TOP, int *M, int *N, 
	 int A[], int *LDA)
);

void FTN_IGEBR2D STDC_PARAMS( 
	(int *ICONTXT, char *SCOPE, char *TOP, int *M, int *N, 
	 int A[], int *LDA, int *RSRC, int *CSRC)
);


#endif /*__FtnBLACS_INCL__*/
