/* @(#) All Right Reserved, Copyright (c) 1997 Australian National University */
/* @(#)                     Copyright (c) 1997 Fujitsu Laboratories Ltd. */
/*  */
/*        */
/*                        Permission Notice */
/*  */
/*                Australian National University */
/*                Canberra ACT 0200 Australia */
/*                           and */
/*                Fujitsu Laboratories Ltd. */
/*                4-1-1 Kamikodanaka Nakahara */
/*                Kawasaki 211-88 Japan */
/*  */
/*   Permission to use, copy, modify and distribute this software (including  */
/*   its documentation) without fee is hereby granted under the following terms */
/*   and conditions: */
/*  */
/*   1. This permission notice and all of the copyright notices in this */
/*      software must appear in all copies and derived versions of all */
/*      or a part of this software. */
/*  */
/*   2. Sales of this software, whether incorporated in other products as a */
/*      part thereof or not, is prohibited. */
/*  */
/*   3. On publication or presentation in any form of the results obtained */
/*      by use of this software, an explicit reference to this software and */
/*      its copyright owner shall be made. */
/*  */
/*   4. THIS SOFTWARE IS PROVIDED 'AS IS' AND THE AUSTRALIAN NATIONAL UNIVERSITY */
/*      AND FUJITSU LABORATORIES LTD. MAKE NO WARRANTIES, EITHER EXPRESSED OR */
/*      IMPLIED, AS TO ANY MATTER WHATSOEVER REGARDING THIS SOFTWARE,  */
/*      INCLUDING, WITHOUT LIMITATION, WARRANTIES OF MERCHANTABILITY OR FITNESS */
/*      FOR ANY PARTICULAR PURPOSE. */
/*      IN NO EVENT SHALL THE AUSTRALIAN NATIONAL UNIVERSITY OR FUJITSU */
/*      LABORATORIES LTD. BE LIABLE FOR ANY CLAIM FROM A THIRD PARTY, OR */
/*      FOR SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES WHATSOEVER RESULTING */
/*      FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN ACTION OF CONTRACT, */
/*      NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION */
/*      WITH THE USE, COPYING OR MODIFICATION OF THIS SOFTWARE. */
/*  */

/************************************************************************/
/*		FtnBLAS: C prototypes for limited Fortran cell BLAS 	*/ 
/************************************************************************/
/* SCCS INFO @(#)FtnBLAS.h 1.3         last modified P. Strazdins, 97/05/07
   first created: P. Strazdins 96/01/19
*/
/* compilation:	FDATATYPE must be defined
 *		if C_CALL_FTN_NO_CHANGE defined (eg for R8000's), strip `_'
 *		off FORTRRAN procedure name		
 */
#ifndef __FtnBLAS_INCL__
#define __FtnBLAS_INCL__
   
 
#include "FDATASetUp.h"
   
#if FDATATYPE==SINGLE
# define FTN_ASUM        sasum_
# define FTN_AXPY        saxpy_
# define FTN_COPY        scopy_
# define FTN_IAMAX       isamax_
# define FTN_SCAL        sscal_
# define FTN_SWAP        sswap_
# define FTN_GEMV        sgemv_
# define FTN_GER         sger_
# define FTN_SYRK	 ssyrk_
# define FTN_GEMM        sgemm_
# define FTN_TRMM        strmm_
# define FTN_TRSM        strsm_
#else
# define FTN_ASUM        dasum_
# define FTN_AXPY        daxpy_   
# define FTN_COPY        dcopy_ 
# define FTN_IAMAX       idamax_
# define FTN_SCAL        dscal_   
# define FTN_SWAP        dswap_
# define FTN_GEMV        dgemv_
# define FTN_GER         dger_
# define FTN_SYRK	 dsyrk_
# define FTN_GEMM        dgemm_
# define FTN_TRMM        dtrmm_
# define FTN_TRSM        dtrsm_
#endif /*FDATATYPE*/

#ifdef C_CALL_FTN_NO_CHANGE
# if FDATATYPE==SINGLE
#   define sasum_	sasum
#   define saxpy_	saxpy
#   define scopy_	scopy
#   define isamax_	isamax
#   define sscal_	sscal
#   define sswap_	sswap
#   define sgemv_	sgemv
#   define sger_	sger
#   define ssyrk_	ssyrk
#   define sgemm_	sgemm
#   define strmm_	strmm
#   define strsm_	strsm
# else
#   define dasum_	dasum
#   define daxpy_	daxpy
#   define dcopy_	dcopy
#   define idamax_	dsamax
#   define dscal_	dscal
#   define dswap_	dswap
#   define dgemv_	dgemv
#   define dger_	dger
#   define dsyrk_	dsyrk
#   define dgemm_	dgemm
#   define dtrmm_	dtrmm
#   define dtrsm_	dtrsm
# endif /*FDATATYPE*/
#endif /*C_CALL_FTN_NO_CHANGE*/

FDATA FTN_ASUM STDC_PARAMS( (int *N, FDATA X[], int *INCX) );

int FTN_IAMAX STDC_PARAMS( (int *N, FDATA X[], int *INCX) );

void FTN_AXPY STDC_PARAMS(
	(int *N, FDATA *ALPHA, FDATA X[], int *INCX, FDATA Y[], int *INCY) 
);

void FTN_COPY STDC_PARAMS( 
	(int *N, FDATA X[], int *INCX, FDATA Y[], int *INCY) 
);

void FTN_SWAP STDC_PARAMS( 
	(int *N, FDATA X[], int *INCX, FDATA Y[], int *INCY) 
);

void FTN_SCAL STDC_PARAMS( 
	(int *N, FDATA *ALPHA, FDATA X[], int *INCX) 
);

void FTN_GER STDC_PARAMS( 
	(int *M, int *N, FDATA *ALPHA, FDATA X[], int *INCX, 
	 FDATA Y[], int *INCY, FDATA A[], int *LDA) 
);

void FTN_GEMV STDC_PARAMS( 
	(char *TRANS, int *M, int *N, FDATA *ALPHA, FDATA A[], int *LDA, 
	 FDATA X[], int *INCX, FDATA *BETA, FDATA Y[], int *INCY) 
);

void FTN_SYRK STDC_PARAMS( 
	(char *UPLO, char *TRANS, int *N, int *K, 
	 FDATA *ALPHA, FDATA A[], int *LDA, FDATA *BETA, FDATA C[], int *LDC)
);

void FTN_GEMM STDC_PARAMS( 
	(char *TRANSA, char *TRANSB, int *M, int *N, int *K, 
	 FDATA *ALPHA, FDATA A[], int *LDA, FDATA B[], int *LDB, 
	 FDATA *BETA,  FDATA C[], int *LDC)
);

void FTN_TRSM STDC_PARAMS(
        (char *SIDE, char *UPLO, char *TRANSA, char *DIAG, int *M, int *N, 
        FDATA *ALPHA, FDATA A[], int *LDA, FDATA B[], int *LDB)
);      
                            
void FTN_TRMM STDC_PARAMS(
        (char *SIDE, char *UPLO, char *TRANSA, char *DIAG, int *M, int *N, 
        FDATA *ALPHA, FDATA A[], int *LDA, FDATA B[], int *LDB)
);                                          
                           
#endif /*__FtnBLAS_INCL__*/

