/* @(#) All Right Reserved, Copyright (c) 1997 Australian National University */
/* @(#)                     Copyright (c) 1997 Fujitsu Laboratories Ltd. */
/*  */
/*        */
/*                        Permission Notice */
/*  */
/*                Australian National University */
/*                Canberra ACT 0200 Australia */
/*                           and */
/*                Fujitsu Laboratories Ltd. */
/*                4-1-1 Kamikodanaka Nakahara */
/*                Kawasaki 211-88 Japan */
/*  */
/*   Permission to use, copy, modify and distribute this software (including  */
/*   its documentation) without fee is hereby granted under the following terms */
/*   and conditions: */
/*  */
/*   1. This permission notice and all of the copyright notices in this */
/*      software must appear in all copies and derived versions of all */
/*      or a part of this software. */
/*  */
/*   2. Sales of this software, whether incorporated in other products as a */
/*      part thereof or not, is prohibited. */
/*  */
/*   3. On publication or presentation in any form of the results obtained */
/*      by use of this software, an explicit reference to this software and */
/*      its copyright owner shall be made. */
/*  */
/*   4. THIS SOFTWARE IS PROVIDED 'AS IS' AND THE AUSTRALIAN NATIONAL UNIVERSITY */
/*      AND FUJITSU LABORATORIES LTD. MAKE NO WARRANTIES, EITHER EXPRESSED OR */
/*      IMPLIED, AS TO ANY MATTER WHATSOEVER REGARDING THIS SOFTWARE,  */
/*      INCLUDING, WITHOUT LIMITATION, WARRANTIES OF MERCHANTABILITY OR FITNESS */
/*      FOR ANY PARTICULAR PURPOSE. */
/*      IN NO EVENT SHALL THE AUSTRALIAN NATIONAL UNIVERSITY OR FUJITSU */
/*      LABORATORIES LTD. BE LIABLE FOR ANY CLAIM FROM A THIRD PARTY, OR */
/*      FOR SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES WHATSOEVER RESULTING */
/*      FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN ACTION OF CONTRACT, */
/*      NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION */
/*      WITH THE USE, COPYING OR MODIFICATION OF THIS SOFTWARE. */
/*  */

/************************************************************************/
/*       GenAssert: General error checking / debugging module     	*/
/************************************************************************/
/* SCCS INFO @(#)GenAssert.h 1.4         last modified P. Strazdins, 97/04/25
   first created: P. Strazdins 96/01/23
*/
/* assertion checking routines which can also be used for call tracing. 
 * Here, every call to assertion checking routine CommonAssert() can also be
 * traced, according to its trace type and level.
 */
#ifndef __GenAssert_INCL__
#define __GenAssert_INCL__  

#include "GenMacros.h"
#include <stdio.h>

/********************* exported symbols  ***********************************/

#define Error_		GError_
#define CommonAssert_   GCommonAssert_   
#define TraceLevel	GTraceLevel  
#define TraceType	GTraceType
#define BaseTraceLevel	GBaseTraceLevel
#define CurTraceDepth	GCurTraceDepth
#define CallerStack	GDSCallerStack
#define TraceStr	GTraceStr
#define TmpTraceStrs	GTmpTraceStrs
#define AssertCId	GAssertCId
#define PrintRes	GPrintRes
#define ForceAlgVers	GForceAlgVers
#define NoOuterErrCheck GNoOuterErrCheck

/* values of TraceType; calls to CommonAssert() are tagged with one of these */
#define TRACE_ALL       0
#define TRACE_COMPUTE   1	/* calls from cell computation routines */
#define TRACE_COMMUN    2	/* calls from communication routines */
#define TRACE_MALLOC    3	/* calls from memory allocation code */
#define TRACE_OUTER	4	/* calls from outer level procedures */

extern int TraceType;		/* intially TRACE_ALL */
				/* print out traces only if TraceLevel:  */
extern int TraceLevel;    	/* 	>= caller's given trace level + */
extern int BaseTraceLevel;	/*	   BaseTraceLevel	&& 	*/
extern int CurTraceDepth;    	/*      >= <current procedure call depth>*/

extern char TraceStr[];		/* to form trace strings for CommonAssert[] */
extern char *TmpTraceStrs[];	/*3 x 64 char work strings, can be used to form TraceStr*/

#define MAX_CALL_DEPTH		16
extern char * CallerStack[];	

extern int AssertCId[];		/*if set to +ve values, error msgs / traces will
				  be prefixed with AssertCId[0]"."AssertCId[1]*/

#ifdef TRACE
# define ASSERT	
# define TRACE_VALS(VAL_SPRINTF_PARAMS) 	sprintf VAL_SPRINTF_PARAMS
	 /* VAL_SPRINTF_PARAMS of form: (TraceStr, "formatstr", v1, v2, ...) */
# else
# define TRACE_VALS(VAL_SPRINTF_PARAMS)
#endif
#ifndef NO_OUTER_TRACE
# define TRACE_VALS_(VAL_SPRINTF_PARAMS) 	sprintf VAL_SPRINTF_PARAMS
#else
# define TRACE_VALS_(VAL_SPRINTF_PARAMS)
#endif

#ifdef ASSERT
# define TRACE_ENTRY				CurTraceDepth++
# define TRACE_EXIT				CurTraceDepth--
# define COMMON_ASSERT(Cond, ErrFlag, Where, Level, Type)	\
		CommonAssert(Cond, ErrFlag, Where, Level, Type)
# define TRACE_RETURN	TRACE_EXIT; return	
# define TRACE_LEVEL_SHIFT(n)			BaseTraceLevel += (n)
# define ADD_TR_CALL_STK(Where)					\
	if (CurTraceDepth < MAX_CALL_DEPTH) 				\
	  CallerStack[CurTraceDepth] = Where;

#else
# define COMMON_ASSERT(Cond, ErrrFlag, Where, Level, Type)
# define TRACE_ENTRY				
# define TRACE_EXIT				
# define TRACE_RETURN	return	
# define TRACE_LEVEL_SHIFT(n)
# define ADD_TR_CALL_STK(Where)					
#endif

#ifdef TRACE
# define WILL_TRACE_PRINT(Level)					\
	(TraceLevel >= BaseTraceLevel+(Level) && TraceLevel >= CurTraceDepth)
# else
# define WILL_TRACE_PRINT(Level) 0	
#endif

#ifdef ASSERT
#define COMMON_ASSERT_TR(TRACE_PARAMS, Cond, EFlag, Where, Level, Type)	\
	{ int b = (Cond);						\
	  if (!b || (EFlag) || WILL_TRACE_PRINT(Level)) {		\
	    TRACE_VALS(TRACE_PARAMS);					\
	    COMMON_ASSERT(b, EFlag, Where, Level, Type);	    	\
	  }								\
	} 
#else
#define COMMON_ASSERT_TR(TRACE_PARAMS, Cond, EFlag, Where, Level, Type)
#endif

#define COMMON_ASSERT_TR_E(TRACE_PARAMS, Cond, EFlag, Where, Level, Type) \
	{ COMMON_ASSERT_TR(TRACE_PARAMS, Cond, EFlag, Where, Level, Type);\
	  ADD_TR_CALL_STK(Where);					\
	  TRACE_ENTRY;							\
	}
	
#define COMMON_ASSERT_TR_X(TRACE_PARAMS, Cond, EFlag, Where, Level, Type) \
	{ COMMON_ASSERT_TR(TRACE_PARAMS, Cond, EFlag, Where, Level, Type);\
	  TRACE_EXIT;						\
	}

#define ERROR_CHECK_TR(TRACE_PARAMS, ParamErr, Where, Level)		\
	if ((ParamErr)!=0 || WILL_TRACE_PRINT(Level)) {			\
	    TRACE_VALS_(TRACE_PARAMS);					\
	    if (ParamErr) {						\
	      Error(Where, ParamErr);					\
	    } else							\
	      COMMON_ASSERT(1, 0, Where, Level, TRACE_OUTER);  		\
	}								\
	ADD_TR_CALL_STK(Where);						\
	TRACE_ENTRY;						
	
/**********************************************************************/

#ifndef CALLER_PREFIX
# define CALLER_PREFIX "G"
#endif

#define CommonAssert(Cond, ErrFlag, Caller, CallTrLevel, CallTrType)	\
	CommonAssert_(Cond, ErrFlag, CALLER_PREFIX, Caller, CallTrLevel, CallTrType)
	
#define Error(Caller, ParamNum)				\
	Error_(CALLER_PREFIX, Caller, ParamNum)


void PrintLinePrefix STDC_PARAMS((void));
/* post: if AssertCId[0..1]>0, print out AssertCId[0]"."AssertCId[1]": "
 * note: this is useful if the stdout of all cells are interleaved;
 *	 by setting AssertCId[] to the current cell's (natural) 2-dim id,
 *	 the concatenated output can later be separated if all lines
 *	 of output have this cell-dependent prefix.
 *       Both Error() and CommonAssert() follow such a convention.
 *
 */
  	 
void Error_ STDC_PARAMS( 
	( char *CallerPrefix, char *Caller, int ParamNum ) 
);
/* pre:  TraceStr contains parameters of call to Caller in printed form
 *	 if ParamNum > 0, ParamNum = 0x pdn, else ParamNum = -p (and d=n=0)
 * post: prints out error message with name of procedure (CallerPrefix, Caller)
 *	 and the parameter number of the 1st parameter at error (p).
 *	 if d,n!= 0, the parameter was incorrect with 
 *	 relation to parameters numbers d and n, which are laso printed out.
 *	 The process is then aborted. 
 * note: this should only be called from outer-level or external procedures;
 *	 internal procuedures should use CommonAssert()
 */
 
void CommonAssert_ STDC_PARAMS( 
	(int Cond, int ErrFlag, 
	 char *CallerPrefix, char *Caller, int CallTrLevel, int CallTrType)
);
/* common assertion testing procedure; also can be used for tracing;
 * if (!Cond || ErrFlag), write error message and EXIT()
 * else if (TraceLevel >= CallTrLevel && TraceType in {TRACE_ALL, CallTrType}
 *	    && TraceDepth >= CurTraceDepth)
 *     write trace info (CallerPrefix, Caller, TraceStr) to stdout.
 * set TraceStr to ""
 * conventions: Cond is a general asert condition; case CallTrType of:
 *	TRACE_COMMUN: ErrFlag signifies a general communication error
 *	TRACE_COMPUTE: ErrFlag signifies matrix overlap error
 */

/*************** miscellanous variables used for testing / debugging ********/

/* these can be used by any module importing from this one. 
 * If used, the only required convention is that values of 0 will have no effect.
 * This is their default value
 */
   
extern int PrintRes;		/*max local size of matrices to print out for debug*/
extern int ForceAlgVers;	/*if 1..3, can be used to force the choice
				  of (high-level) DBLAS algorithm used.
				  eg. choose variant of alg to spread along
				  ith matrix dimension i=1 (M), 2 (N), 3 (K)
				 */ 
extern int NoOuterErrCheck;	 /* if != 0, suppress error checking in
				   outer-level procedures */

#endif /*__GenAssert_INCL__*/

