/* @(#) All Right Reserved, Copyright (c) 1997 Australian National University */
/* @(#)                     Copyright (c) 1997 Fujitsu Laboratories Ltd. */
/*  */
/*        */
/*                        Permission Notice */
/*  */
/*                Australian National University */
/*                Canberra ACT 0200 Australia */
/*                           and */
/*                Fujitsu Laboratories Ltd. */
/*                4-1-1 Kamikodanaka Nakahara */
/*                Kawasaki 211-88 Japan */
/*  */
/*   Permission to use, copy, modify and distribute this software (including  */
/*   its documentation) without fee is hereby granted under the following terms */
/*   and conditions: */
/*  */
/*   1. This permission notice and all of the copyright notices in this */
/*      software must appear in all copies and derived versions of all */
/*      or a part of this software. */
/*  */
/*   2. Sales of this software, whether incorporated in other products as a */
/*      part thereof or not, is prohibited. */
/*  */
/*   3. On publication or presentation in any form of the results obtained */
/*      by use of this software, an explicit reference to this software and */
/*      its copyright owner shall be made. */
/*  */
/*   4. THIS SOFTWARE IS PROVIDED 'AS IS' AND THE AUSTRALIAN NATIONAL UNIVERSITY */
/*      AND FUJITSU LABORATORIES LTD. MAKE NO WARRANTIES, EITHER EXPRESSED OR */
/*      IMPLIED, AS TO ANY MATTER WHATSOEVER REGARDING THIS SOFTWARE,  */
/*      INCLUDING, WITHOUT LIMITATION, WARRANTIES OF MERCHANTABILITY OR FITNESS */
/*      FOR ANY PARTICULAR PURPOSE. */
/*      IN NO EVENT SHALL THE AUSTRALIAN NATIONAL UNIVERSITY OR FUJITSU */
/*      LABORATORIES LTD. BE LIABLE FOR ANY CLAIM FROM A THIRD PARTY, OR */
/*      FOR SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES WHATSOEVER RESULTING */
/*      FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN ACTION OF CONTRACT, */
/*      NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION */
/*      WITH THE USE, COPYING OR MODIFICATION OF THIS SOFTWARE. */
/*  */

/************************************************************************/
/* GenMult: DBLAS General Matrix Multiplication procedures		*/
/************************************************************************/
/* SCCS INFO @(#)GenMult.h 4.1         last modified P. Strazdins, 97/02/19
   first created: P. Strazdins 93/15/01
*/
/* compilation: macro FDATATYPE must be defined;
 *              if macro ASSERT defined, assertion checking turned on
 *              if macro TRACE defined, parameter value tracing enabled
 *		if macro PRINTMAT define, matrices are printed for debugging
 */
#ifndef __GenMult_INCL__
#define __GenMult_INCL__

#include "FDATASetUp.h"
#include "DistMat.h"

#if FDATATYPE==SINGLE
#define DistGenMult	DSDistGenMult
#define DistRankKMult	DSDistRankKMult
#define DistVecMatMult	DSDistVecMatMult
#else
#define DistGenMult	DDDistGenMult
#define DistRankKMult	DDDistRankKMult
#define DistVecMatMult	DDDistVecMatMult
#endif

void DistGenMult STDC_PARAMS( 
	(int M, int N, int K, FDATA *alpha, DistMat A, DistMat B, 
	 FDATA *beta, DistMat C)
);
/* pre:  all cells are executing DistGenMult() with 
 *	 M, N, K, *alpha, *beta,  FormC parameters and (D[0..1], Trans) 
 * 	 fields for A, B & C consistent across all cells;
 *	 M, N, K >= 0; 
 *	 if FormC != Rectangular, C is a triangular matrix of form FormC;
 * post: performs global operation  C = (*beta)*C + (*alpha)*AB,
 * notes:chooses between DistRankKMult() & the 2 variants of DistVecMatMult();
 *	 this corresponds to spreading across the dimension K, M, or N 
 *	 on the basis of (in priority): 
 *		1) block incompatability: spread across the (at most 1)
 *		   dimension that is not block aligned; this includes a dim.
 *		   of A or B that has been (non-trivially) `pre-spread'.
 *		   This is because the spread will hide the misalignment.
 *		2) the greatest load imbalance, as spreading across this
 *		   dimension will spread the imbalance also.
 *		3) the smallest (parallel) communication volume.
 *		   If an explicit transpose is required with a spread operation,
 *		   this (rather crudely) is taken to effectively double the
 *		   volume of that spread. 
 *		   Currently, rotation requirements for spreads are ignored
 *		   (as generally smaller than the spread proper).
 *		   Also, the `weighting' of all spread communication, whether 
 *		   a broadcast or a summnation, is taken to be unity;
 *		   this is a simplification, as these can be different or
 *		   vary with matrix sizes.	
 */


void DistRankKMult STDC_PARAMS( 
	(int M, int N, int K, FDATA *alpha, DistMat A, DistMat B, 
	 FDATA *beta, DistMat C)
);
/* pre:  as for DistGenMult();
 * post: as for DistGenMult() 
 * notes:efficient for K << M, N.
 */

void DistVecMatMult STDC_PARAMS( 
	(int M, int N, int K, FDATA *alpha, DistMat A, DistMat B, 
	 FDATA *beta, DistMat C)
);
/* pre:  as for DistGenMult();
 * post: as for DistGenMult() 
 * notes:efficient for M << K, N; 
 * 	 !DM_Trans(B) is the `normal' variant for this compuation 
 *	 (vector-matrix multiply). To achieve a matrix-vector multiply, 
 *	 transpose all DistMat parameters, and interchange A,B and M,N.
 */

#endif /*__GenMult_INCL__*/

