/* @(#) All Right Reserved, Copyright (c) 1997 Australian National University */
/* @(#)                     Copyright (c) 1997 Fujitsu Laboratories Ltd. */
/*  */
/*        */
/*                        Permission Notice */
/*  */
/*                Australian National University */
/*                Canberra ACT 0200 Australia */
/*                           and */
/*                Fujitsu Laboratories Ltd. */
/*                4-1-1 Kamikodanaka Nakahara */
/*                Kawasaki 211-88 Japan */
/*  */
/*   Permission to use, copy, modify and distribute this software (including  */
/*   its documentation) without fee is hereby granted under the following terms */
/*   and conditions: */
/*  */
/*   1. This permission notice and all of the copyright notices in this */
/*      software must appear in all copies and derived versions of all */
/*      or a part of this software. */
/*  */
/*   2. Sales of this software, whether incorporated in other products as a */
/*      part thereof or not, is prohibited. */
/*  */
/*   3. On publication or presentation in any form of the results obtained */
/*      by use of this software, an explicit reference to this software and */
/*      its copyright owner shall be made. */
/*  */
/*   4. THIS SOFTWARE IS PROVIDED 'AS IS' AND THE AUSTRALIAN NATIONAL UNIVERSITY */
/*      AND FUJITSU LABORATORIES LTD. MAKE NO WARRANTIES, EITHER EXPRESSED OR */
/*      IMPLIED, AS TO ANY MATTER WHATSOEVER REGARDING THIS SOFTWARE,  */
/*      INCLUDING, WITHOUT LIMITATION, WARRANTIES OF MERCHANTABILITY OR FITNESS */
/*      FOR ANY PARTICULAR PURPOSE. */
/*      IN NO EVENT SHALL THE AUSTRALIAN NATIONAL UNIVERSITY OR FUJITSU */
/*      LABORATORIES LTD. BE LIABLE FOR ANY CLAIM FROM A THIRD PARTY, OR */
/*      FOR SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES WHATSOEVER RESULTING */
/*      FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN ACTION OF CONTRACT, */
/*      NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION */
/*      WITH THE USE, COPYING OR MODIFICATION OF THIS SOFTWARE. */
/*  */

/************************************************************************/
/* GetBlockFactor: calulate blocking factors for blocked algorithms   	*/
/************************************************************************/
/* SCCS INFO GetBlockFactor.c@(#)         	last modified P. Strazdins,  97/03/12
   first created: P. Strazdins 94/10/11					
   
*/

#include "GenMacros.h"
#include "GenAssert.h"
#include "GetBlockFactor.h"
#include "MatDim.h"
#include "IntOps.h"   

/*#define BLOCK_CELL_ADJUST_W*/

double sqrt STDC_PARAMS( (double) );

static inline int OptW(N, w_scale)
  int N; double w_scale; 
/* pre:  N is the current size of the  column of matrix to be zeroed
 * post: returns the optimal blocking factor (not considering CACHE)
 *  	 multiplied by w_scale (generally a small integer)
 */
{ int w = 2 * (int) (w_scale * sqrt((double) (N/2)));
  if (w==0) w=1;
  return (w);
} /* OptW */

int GenGetW(N, w_scale, w_min, w_max, WDim, j, jInc)
  int N, w_min, w_max, j, jInc;
  pMatDim WDim; double w_scale; 
{ int w;
  if (w_max <= w_min)
    w = w_max;
  else {
    int w_l, w_u;	     
    unsigned int B=d_B(WDim), P=d_P(WDim), lgB=d_lgB(WDim);
    int B0, k, kP, wB;		
  
    w = OptW(N, w_scale);
    if (w>=w_max) {
      w = w_u = w_max; 
    } else
      w_u = MAX(w_min, MIN((3*w)/2, w_max));
    
    if (w<=w_min)
      w = w_l = w_min;
    else
      w_l = MAX(w_min, (3*w)/4);

    COMMON_ASSERT( w_min<=w_l && w_l<=w && w<=w_u && w_u <= w_max,
    		   0, "GenGetW: internal", 4, TRACE_ALL);

    /* seek to adjust for better load bal.*/

    B0 = SubMatB0d(j, WDim);
    if ((jInc < 0) && (B0 > 0)) 
      B0 = B-B0-1;    
    if (VALID_LG(lgB)) {  
      k = (w + B0)>>lgB;		wB = (k<<lgB) - B0;
    } else { 
      k = (w + B0)/B;			wB = k*B - B0;
    }    					
    if (w_l<=wB)			/* adjust to nearest block boundary */
      w = wB;

#ifdef BLOCK_CELL_ADJUST_W
    kP = FAST_MODP2(k, P);
    if (P>1 && kP<=P/2) {     		/* possibly a few too many blocks */
      int wP = wB - (VALID_LG(lgB)? (kP<<lgB): kP*B);
      int wP1 = wP + (VALID_LG(lgB)? (P<<lgB): P*B);
      if (kP && wP1<=w_u)
        w = wP1;	     		/* increase to next biggest one */
      else if (w_l<=wP && k>=P) 
        w = wP;				/* decrease to get better load balance*/
    } else if (P>1) {			/* assert kP > 0 */
      /* try to increase # blocks to get better load balance */
      while (wB+B<=w_u && kP<P) {
        wB = wB+B;				
        kP++;
      }
      if (w_l<=wB)
        w = wB;
    }
#endif /*BLOCK_CELL_ADJUST_W*/
  } /* if */

  COMMON_ASSERT_TR(
    (TraceStr, "(%d,%1.1f,%d..%d,%d %d,%d %d:%d %d)=%d", N, w_scale, 
     w_min, w_max, j, jInc, d_B0(WDim), d_B(WDim), d_P0(WDim), d_P(WDim), w),
    N>=0 && w_max>0 && w_min>0 && w_scale>0.0 && ABS(jInc)==1, 
    0, "GenGetW", 2, TRACE_ALL);

  COMMON_ASSERT((w_max < w_min || w_min<=w) && w<=w_max, 0, "GenGetW: internal", 4, TRACE_ALL);
  return(w);
} /* GenGetW() */
