/* @(#) All Right Reserved, Copyright (c) 1997 Australian National University */
/* @(#)                     Copyright (c) 1997 Fujitsu Laboratories Ltd. */
/*  */
/*        */
/*                        Permission Notice */
/*  */
/*                Australian National University */
/*                Canberra ACT 0200 Australia */
/*                           and */
/*                Fujitsu Laboratories Ltd. */
/*                4-1-1 Kamikodanaka Nakahara */
/*                Kawasaki 211-88 Japan */
/*  */
/*   Permission to use, copy, modify and distribute this software (including  */
/*   its documentation) without fee is hereby granted under the following terms */
/*   and conditions: */
/*  */
/*   1. This permission notice and all of the copyright notices in this */
/*      software must appear in all copies and derived versions of all */
/*      or a part of this software. */
/*  */
/*   2. Sales of this software, whether incorporated in other products as a */
/*      part thereof or not, is prohibited. */
/*  */
/*   3. On publication or presentation in any form of the results obtained */
/*      by use of this software, an explicit reference to this software and */
/*      its copyright owner shall be made. */
/*  */
/*   4. THIS SOFTWARE IS PROVIDED 'AS IS' AND THE AUSTRALIAN NATIONAL UNIVERSITY */
/*      AND FUJITSU LABORATORIES LTD. MAKE NO WARRANTIES, EITHER EXPRESSED OR */
/*      IMPLIED, AS TO ANY MATTER WHATSOEVER REGARDING THIS SOFTWARE,  */
/*      INCLUDING, WITHOUT LIMITATION, WARRANTIES OF MERCHANTABILITY OR FITNESS */
/*      FOR ANY PARTICULAR PURPOSE. */
/*      IN NO EVENT SHALL THE AUSTRALIAN NATIONAL UNIVERSITY OR FUJITSU */
/*      LABORATORIES LTD. BE LIABLE FOR ANY CLAIM FROM A THIRD PARTY, OR */
/*      FOR SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES WHATSOEVER RESULTING */
/*      FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN ACTION OF CONTRACT, */
/*      NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION */
/*      WITH THE USE, COPYING OR MODIFICATION OF THIS SOFTWARE. */
/*  */

/************************************************************************/
/* GetBlockFactor: calulate blocking factors for blocked algorithms     */
/************************************************************************/
/* SCCS INFO GetBlockFactor.h@(#)         	last modified P. Strazdins,  97/02/24
   first created: P. Strazdins 94/10/11					   
*/

#ifndef __GetBlockFactor_INCL__
#define __GetBlockFactor_INCL__

#include "GenMacros.h"
#include "MatDim.h"

#define GenGetW	GGenGetW


int GenGetW STDC_PARAMS(
  (int N, double w_scale, int w_min, int w_max, pMatDim WDim, int j, int jInc)
);
/* pre:  N is the size of column j  of an algorithm where j changes with dj
 *	 WDim is dimension relative to row 0 of the columns.
 * post: returns largest blocking factor w such that:
 *	   MIN(w_max, w_min) <= w <= w_max
 *	 for w' = 2* (int) (w_scale * sqrt(N/2)), if w_min <= w' <= w_max,
 *	 w' will be used as an initial value for w, with the following 
 *	 strategies to be applied: 
 *	   1. w is to be chosen so that
 *		j+jInc*w
 *	      is on a block boundary (wrt WDim) (suppress by setting WDim.B=1)
 *	   2. w is to be chosen so that  j+jInc*w is on a cell-block boundary
 *	      (suppress by setting d_P(WDim)=1).
 *	 these are only applied if the resulting value of w satisifes:
 *	    MAX(w_min, 3w'/4) <= w <= MIN(1.5*w', w_max)
 */

#endif /*__GetBlockFactor_INCL__*/
