/* @(#) All Right Reserved, Copyright (c) 1997 Australian National University */
/* @(#)                     Copyright (c) 1997 Fujitsu Laboratories Ltd. */
/*  */
/*        */
/*                        Permission Notice */
/*  */
/*                Australian National University */
/*                Canberra ACT 0200 Australia */
/*                           and */
/*                Fujitsu Laboratories Ltd. */
/*                4-1-1 Kamikodanaka Nakahara */
/*                Kawasaki 211-88 Japan */
/*  */
/*   Permission to use, copy, modify and distribute this software (including  */
/*   its documentation) without fee is hereby granted under the following terms */
/*   and conditions: */
/*  */
/*   1. This permission notice and all of the copyright notices in this */
/*      software must appear in all copies and derived versions of all */
/*      or a part of this software. */
/*  */
/*   2. Sales of this software, whether incorporated in other products as a */
/*      part thereof or not, is prohibited. */
/*  */
/*   3. On publication or presentation in any form of the results obtained */
/*      by use of this software, an explicit reference to this software and */
/*      its copyright owner shall be made. */
/*  */
/*   4. THIS SOFTWARE IS PROVIDED 'AS IS' AND THE AUSTRALIAN NATIONAL UNIVERSITY */
/*      AND FUJITSU LABORATORIES LTD. MAKE NO WARRANTIES, EITHER EXPRESSED OR */
/*      IMPLIED, AS TO ANY MATTER WHATSOEVER REGARDING THIS SOFTWARE,  */
/*      INCLUDING, WITHOUT LIMITATION, WARRANTIES OF MERCHANTABILITY OR FITNESS */
/*      FOR ANY PARTICULAR PURPOSE. */
/*      IN NO EVENT SHALL THE AUSTRALIAN NATIONAL UNIVERSITY OR FUJITSU */
/*      LABORATORIES LTD. BE LIABLE FOR ANY CLAIM FROM A THIRD PARTY, OR */
/*      FOR SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES WHATSOEVER RESULTING */
/*      FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN ACTION OF CONTRACT, */
/*      NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION */
/*      WITH THE USE, COPYING OR MODIFICATION OF THIS SOFTWARE. */
/*  */

/************************************************************************/
/*                     MachineParams		                        */
/************************************************************************/
/* SCCS INFO @(#)MachineParams.h 1.6         last modified P. Strazdins, 97/02/24
   first created: P. Strazdins 92/01
*/
/* header file for cell architecture dependent constants,
 * currently defaults appropriate for  Fujitsu AP1000. 
 * 
 * compilation: 
 *	if TUNE is not defined, parameters are defined in this file as
 *	constants. This enables compiler to optimize thier use better.
 *	Otherwise, they are defined to be elements of an exported parameter 
 *	vector which is dynamically settable (initial values being the
 *	corresponding constants). This is useful for performance tuning, 
 *	and also testing/debugging.
 */	
/************************************************************************/

#ifndef __MachineParams_INCL__
#define __MachineParams_INCL__

#include "FDATASetUp.h"

#if FDATATYPE==SINGLE
#define MachParamVec	DSMachParamVec
#define MachParamVal	DSMachParamVal
#else
#define MachParamVal    DDMachParamVal
#define MachParamVec	DDMachParamVec
#endif

extern int MachParamVec[];
int MachParamVal STDC_PARAMS( ( char *ParamMnem) );
/* post: converts MachineParam paramemter menemonics to their corresp.
 *	 indices, ie.
 *	 if *ParamMnem = one of CACHESIZE_MNEM, ..., PAR_OUTER_VECSIZE_MNEM
 *	 returns the _NUM index corresp  to *ParamMnem; else returns -1
 */ 
 
					/* define indexes for each parameter */
#define CACHESIZE_NUM			0
#define CACHE_ASSOC_NUM			1
#define CACHE_LINE_SIZE_NUM		2
#define OPT_MSG_SIZE_NUM		3
#define MAX_DBLAS_WS_SIZE_NUM		4		
#define MIN_VECSIZE_NUM			5
#define INNER_VECSIZE_NUM		6
#define OUTER_VECSIZE_NUM		7
#define MINOR_OUTER_VECSIZE_NUM		8
#define K0_MINOR_OUTER_VECSIZE_NUM 	9
#define MINOR_OUTER_VECSIZE_K0_NUM	10
#define PAR_INNER_VECSIZE_NUM		11
#define PAR_INNER_VECSIZE_TRSM_NUM	12
#define PAR_OUTER_VECSIZE_NUM		13
#define PAR_INNER_VECSIZE_TRSM_MIN_NUM 	14
#define NUM_MACH_PARAMS			15	

#define CACHESIZE_MNEM			"cs"
#define CACHE_ASSOC_MNEM		"ca"
#define CACHE_LINE_SIZE_MNEM		"cl"
#define OPT_MSG_SIZE_MNEM		"ms"
#define MAX_DBLAS_WS_SIZE_MNEM		"ws"
#define MIN_VECSIZE_MNEM                "Mv"		
#define INNER_VECSIZE_MNEM		"Iv"
#define OUTER_VECSIZE_MNEM		"Ov"
#define MINOR_OUTER_VECSIZE_MNEM	"ov"
#define K0_MINOR_OUTER_VECSIZE_MNEM 	"ko"
#define MINOR_OUTER_VECSIZE_K0_MNEM 	"ok"
#define PAR_INNER_VECSIZE_MNEM		"PI"
#define PAR_INNER_VECSIZE_TRSM_MNEM	"PT"
#define PAR_OUTER_VECSIZE_MNEM		"PO"
#define PAR_INNER_VECSIZE_TRSM_MIN_MNEM "MT"

#ifdef TUNE		/* define parameters as elts of the parameter vector */
#define MachParamSet(ParamNum, ParamVal)				\
	{ if (0<=(ParamNum) && (ParamNum) < NUM_MACH_PARAMS)		\
	    MachParamVec[ParamNum] = (ParamVal); }		

#define CACHESIZE		MachParamVec[CACHESIZE_NUM]
#define CACHE_ASSOC		MachParamVec[CACHE_ASSOC_NUM]
#define CACHE_LINE_SIZE		MachParamVec[CACHE_LINE_SIZE_NUM]
#define OPT_MSG_SIZE		MachParamVec[OPT_MSG_SIZE_NUM]
#define MAX_DBLAS_WS_SIZE	MachParamVec[MAX_DBLAS_WS_SIZE_NUM]
#define MIN_VECSIZE		MachParamVec[MIN_VECSIZE_NUM]
#define INNER_VECSIZE		MachParamVec[INNER_VECSIZE_NUM]
#define OUTER_VECSIZE		MachParamVec[OUTER_VECSIZE_NUM]
#define MINOR_OUTER_VECSIZE	MachParamVec[MINOR_OUTER_VECSIZE_NUM]
#define K0_MINOR_OUTER_VECSIZE	MachParamVec[K0_MINOR_OUTER_VECSIZE_NUM]
#define MINOR_OUTER_VECSIZE_K0	MachParamVec[MINOR_OUTER_VECSIZE_K0_NUM]
#define PAR_INNER_VECSIZE	MachParamVec[PAR_INNER_VECSIZE_NUM]
#define PAR_INNER_VECSIZE_TRSM	MachParamVec[PAR_INNER_VECSIZE_TRSM_NUM]
#define PAR_INNER_VECSIZE_TRSM_MIN	\
				MachParamVec[PAR_INNER_VECSIZE_TRSM_MIN_NUM]
#define PAR_OUTER_VECSIZE	MachParamVec[PAR_OUTER_VECSIZE_NUM]

#else			/* define default values of parameters as constants */
#define MachParamSet(ParamNum, ParamVal) /* no effect */

/* cache parameters; if none exists, CACHESIZE should bet set to infinity */
#ifndef CACHE
# define CACHE			16	/* (primary) cache size in Kbytes */	
#endif 

#ifndef CACHESIZE
# define CACHESIZE 		(CACHE * 1024 / FD_SIZE)
#endif

#ifndef CACHE_ASSOC
# define CACHE_ASSOC		1	/* 1 for direct-mapped; n for n-way assoc */
#endif

#ifndef CACHE_LINE_SIZE
# define CACHE_LINE_SIZE 	16	/* cache line size in bytes */
#endif

/* there may be an optimum or maximum message size; if none exists, set to inf.*/
#ifndef OPT_MSG_SIZE
# define OPT_MSG_SIZE 		16384		
#endif			

/* can restrict size of workspace allocated within any DBLAS call.
 * This is only useful in an application already running out of heap memory.
 * Restriction  may result in loss of speed however.
 */
#ifndef MAX_DBLAS_WS_SIZE
# define MAX_DBLAS_WS_SIZE	116777216 /* 2^24, ie effectively infinity */
#endif

/* the following parameters refer to the (minimum) vector size for optimal
 * cell performance, based on a cell matrix multiply C = A*B, C is M x N. 
 * `OUTER' [`INNER'] refers to the outer [inner] dimension of a matrix multiply
 * (eg M or N) [(eg. K )]. 
 *
 * On a scalar cell, [INNER,OUTER]_VECSIZE would rever to the cache level 
 * of the memory hierarchy (with 2*INNER_VECSIZE*OUTER_VECSIZE ~ CACHESIZE
 * and OUTER_VECSIZE <= INNER_VECSIZE)
 * and MINOR_OUTER_VECSIZE referring to the register level.
 * On a vector cell, using an GAXPY-based multiply, OUTER_VECSIZE 
 * [MINOR_OUTER_VECSIZE] would refer to the length [number] of C vectors held in vector registers.
 * These parameters are used for rectangular/triangular cell MM partitioning. 
 *
 * MINOR_OUTER_VECSIZE may change for low K, ie K < K0_MINOR_OUTER_VECSIZE, 
 * (ie. for scalar cells, when the cell MM switches from inner-product
 * to rank-1 update based); in this case, MINOR_OUTER_VECSIZE_K0 should be set
 * to this different value.
 * 
 * All the above parameters are dependent on the implementation of cell BLAS
 * being used. If these are not precisely known, over-estimating will probably
 * yield better performance than underestimating. In any case,  experimentation
 * can be used to determine the best values.
 *
 * The prefix `PAR' refers to the partitioning size on the outer-level DBLAS
 * algorithms, which generally occurs over the smallest of the global 
 * dimensions. For most situations, they will be the same as the corresp.
 * macors without the prefix.
 */

#ifndef INNER_VECSIZE		/* ~ sqrt( CACHESIZE / 2) */
#if (CACHESIZE >= 32768)
# define INNER_VECSIZE 128
#elif (CACHESIZE >= 16384)
# define INNER_VECSIZE 88
#elif (CACHESIZE >= 8192)
# define INNER_VECSIZE 64
#elif (CACHESIZE >= 4096)
# define INNER_VECSIZE 44
#elif (CACHESIZE >= 2048)
# define INNER_VECSIZE 32
#elif (CACHESIZE >= 1024)
# define INNER_VECSIZE 24
#elif (CACHESIZE >= 512)
# define INNER_VECSIZE 16
#else
# define INNER_VECSIZE 16
#endif /*CACHESIZE*/
#endif /*INNER_VECSIZE*/

#ifndef OUTER_VECSIZE		
# define OUTER_VECSIZE 		(CACHESIZE / 2 / INNER_VECSIZE) 
					/* suitable for scalar cells only */
#endif 

#ifndef MINOR_OUTER_VECSIZE
#define MINOR_OUTER_VECSIZE	4	/* prob. valid for 16-32 reg scalar */
#endif

#ifndef K0_MINOR_OUTER_VECSIZE
#define K0_MINOR_OUTER_VECSIZE	4	/* prob. valid for 16-32 reg scalar */
#endif

#ifndef MINOR_OUTER_VECSIZE_K0
#define MINOR_OUTER_VECSIZE_K0	6	/* 6-8  prob. valid for 16-32 reg scalar */
#endif

#ifndef PAR_INNER_VECSIZE	/* partitioning for || Matrix Mult routines */
# define PAR_INNER_VECSIZE	INNER_VECSIZE
#endif

/* useful for TRSM proedures to have a seperately tunable parameter */
#ifndef PAR_INNER_VECSIZE_TRSM		
# define PAR_INNER_VECSIZE_TRSM	PAR_INNER_VECSIZE
#endif

#ifndef PAR_INNER_VECSIZE_TRSM_MIN /* minimum useful blocking factor for TRSM */
#define PAR_INNER_VECSIZE_TRSM_MIN	4
#endif

#ifndef PAR_OUTER_VECSIZE
# define PAR_OUTER_VECSIZE	OUTER_VECSIZE
#endif

#ifndef MIN_VECSIZE		/* min vector size for efficient iteration */
# define MIN_VECSIZE 		16
#endif

#endif /*TUNE*/

#endif /*__MachineParams_INCL__*/
