/* @(#) All Right Reserved, Copyright (c) 1997 Australian National University */
/* @(#)                     Copyright (c) 1997 Fujitsu Laboratories Ltd. */
/*  */
/*        */
/*                        Permission Notice */
/*  */
/*                Australian National University */
/*                Canberra ACT 0200 Australia */
/*                           and */
/*                Fujitsu Laboratories Ltd. */
/*                4-1-1 Kamikodanaka Nakahara */
/*                Kawasaki 211-88 Japan */
/*  */
/*   Permission to use, copy, modify and distribute this software (including  */
/*   its documentation) without fee is hereby granted under the following terms */
/*   and conditions: */
/*  */
/*   1. This permission notice and all of the copyright notices in this */
/*      software must appear in all copies and derived versions of all */
/*      or a part of this software. */
/*  */
/*   2. Sales of this software, whether incorporated in other products as a */
/*      part thereof or not, is prohibited. */
/*  */
/*   3. On publication or presentation in any form of the results obtained */
/*      by use of this software, an explicit reference to this software and */
/*      its copyright owner shall be made. */
/*  */
/*   4. THIS SOFTWARE IS PROVIDED 'AS IS' AND THE AUSTRALIAN NATIONAL UNIVERSITY */
/*      AND FUJITSU LABORATORIES LTD. MAKE NO WARRANTIES, EITHER EXPRESSED OR */
/*      IMPLIED, AS TO ANY MATTER WHATSOEVER REGARDING THIS SOFTWARE,  */
/*      INCLUDING, WITHOUT LIMITATION, WARRANTIES OF MERCHANTABILITY OR FITNESS */
/*      FOR ANY PARTICULAR PURPOSE. */
/*      IN NO EVENT SHALL THE AUSTRALIAN NATIONAL UNIVERSITY OR FUJITSU */
/*      LABORATORIES LTD. BE LIABLE FOR ANY CLAIM FROM A THIRD PARTY, OR */
/*      FOR SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES WHATSOEVER RESULTING */
/*      FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN ACTION OF CONTRACT, */
/*      NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION */
/*      WITH THE USE, COPYING OR MODIFICATION OF THIS SOFTWARE. */
/*  */

/************************************************************************/
/*	MatDim: distributed block-cyclic matrices definition		*/
/************************************************************************/
/* SCCS INFO @(#)MatDim.c 4.4         last modified P. Strazdins, 97/03/14
   first created: P. Strazdins 94/05/17
*/

#include "GenMacros.h"
#include "MatDim.h"
#include "IntOps.h"
#include "GenAssert.h"

/************* major indexing procedures *******************************/

#ifndef CELLBLAS
int G2Llg(N, B0, B, P0, P, lgB, lgP)
  unsigned int N; unsigned int B0, B, P0, P, lgB, lgP;
{ register int NBlks, NB, NBp; 
  if (B==1) 				/* important special case */
    if (N==1 || P==1)
      return( P0==0? N: 0 );		/* fast exit */
    else if (VALID_LG(lgP)) 
      return ((N - P0 + P-1) >> lgP);
    else
      return ((N - P0 + P-1) / P);
  if (Blk1(N, B0, B) || P==1)
    return( P0==0? N: 0 );		/* fast exit */
  N += B0;  
  if (VALID_LG(lgB))  
    NBlks = N >> lgB;
  else 
    NBlks = N / B;	
  if (VALID_LG(lgP)) {			/* #elts on cell,assuming full blocks*/
    NB = ((NBlks - P0 + P-1) >> lgP);
    NBp = MOD_P2(NBlks - P0 + P-1, P); 
  } else {
    DIV_MOD(NBlks - P0 + P-1, P, NB, NBp);
  }
  if (VALID_LG(lgB)) {
    NB = NB<<lgB;	
    if (NBp==P-1) 
      NB += MOD_P2(N, B);
  } else 
    if (NBp==P-1)	
      NB = N - (NBlks - NB)*B;
    else
      NB = NB*B;  
  if (P0==0)				/* adjust for partial block 1st cell */
    NB -= B0;
  return (NB);
} /* G2Llg() */


int L2Glg(i, B0, B, P0, P, lgB, lgP)
  unsigned int i; unsigned int B0, B, P0, P, lgB, lgP;
{ int ig, nB;
  if (B==1) 
    return ((VALID_LG(lgP)? (i<<lgP): (i*P)) + P0);    
  if (P0==0) i += B0;
  if (VALID_LG(lgB)) {
    nB = (i>>lgB);
    ig = (VALID_LG(lgP)? (nB<<lgP): (nB*P)) + P0;
    ig = (ig<<lgB) + MOD_P2(i, B);
  } else {
    nB = (i/B);
    ig = (VALID_LG(lgP)? (nB<<lgP): (nB*P)) + P0;
    ig = ((ig-nB)*B + i);
  }  
  /* note: ig*B + i%B = ig*B + i - nB*B = (ig-nB)*B + i */
  return( ig - B0 );
} /* L2Gd_() */
#endif /*CELLBLAS*/

/******************************* MatDim operations ************************/

int G2LSubDim(i, D)
  unsigned int i; pMatDim D;
{ unsigned int B = d_B(D),   		P0 = d_P0(D),
               P = d_P(D);
  int iP0;				/* new value of d_P0(D) */
  if (B==1) {				/* affords reduced arithmetic */
    if (i==1) {				/* level 2 optimization */
      if (P0!=0) i = 0;          	iP0 = CYCLIC_DEC(P0, P);
    } else {
      unsigned int lgP = d_lgP(D);
      if (VALID_LG(lgP)) {
       	iP0 = MOD_P2(i-P0 + P-1, P);	i = (i-P0 + P-1) >> lgP;
      } else 
       	DIV_MOD(i-P0 + P-1, P, i, iP0);
      iP0 = P-1 - iP0;
    }
  } else {
    unsigned int B0 = d_B0(D);
    int iB0;				/* new value d_B0(D) */
    if (i+B0 <= B) {			/* `within block' optimization */
      iB0 = i+B0;			iP0 = P0;
      if (P0!=0) i = 0;		
      if (iB0==B) {
        iB0 = 0;       	    		iP0 = CYCLIC_DEC(P0, P);	
      }
    } else {
      unsigned int lgB = d_lgB(D),	lgP = d_lgP(D);
      if (VALID_LG(lgB)) {  
        iB0 = MOD_P2(i+B0, B);		i = (i+B0) >> lgB;  
      } else { 
        DIV_MOD(i+B0, B, i, iB0);  	
      } 
      if (VALID_LG(lgP)) {
	iP0 = MOD_P2(i-P0 + P-1, P); 	i = (i-P0 + P-1) >> lgP;
      } else {
	DIV_MOD(i-P0 + P-1, P, i, iP0);
      }
      iP0 = P-1 - iP0;
      i = VALID_LG(lgB)? (i<<lgB): i*B;
      if (P0==0)
 	i -= B0;
      if (iP0==0)
	i += iB0;
    }
    d_B0(D) = iB0;
   } /* else (B>1) */

  d_P0(D) = iP0;
  return (i);
} /* G2LSubDim() */

int G2LSubDimM1(D)
 pMatDim D;
{ unsigned int B = d_B(D),   		P = d_P(D),	
    	       P0 = d_P0(D);
  int i=-1;
  if (B==1) {				/* affords reduced arithmetic */
    d_P0(D) = CYCLIC_INC(P0, P);
  } else {
    int B0 = i+d_B0(D);
    if (B0==-1) {
        B0 += B;       	    		d_P0(D) = CYCLIC_INC(P0, P);
    }
    d_B0(D) = B0;
  }  	
  if (d_P0(D)!=0) i = 0;  
  return (i);
} /* G2LSubDimM1() */


static unsigned char lgB_LU[] = { 
   D_INVAL_LG, 0, 1, 1, 2, 2, 2, 2,
	    3, 3, 3, 3, 3, 3, 3, 3};  

static inline unsigned int getLg(B)
{ 
  if (IS_P2(B)) {
    int lgB=0;	
    while (B>=16) {		/* optimized for small B, eg <= 256 */
      B /= 16;
      lgB += 4;
    }
    return(lgB + lgB_LU[B]); 
  } else {
    return(D_INVAL_LG);
  }  
} /* getLg() */


MatDim UnitDim = {0, 1, 0, 0, 0, 1};

pMatDim pInitDim(D, B0, B, P0, P)
  pMatDim D; int B0, B, P0, P; 
{ TRACE_VALS((TraceStr, "(_,%d,%d,%d,%d)", B0, B, P0, P));
  COMMON_ASSERT(0<=B0 && B0<B && 0<=P0 && P0<P, 0, "pInitDim", 6, TRACE_COMPUTE);
  d_B0(D) = B0;				d_B(D) = B;	
  d_P0(D) = P0;    			d_P(D) = P;
  d_lgB(D) = getLg(B);			d_lgP(D) = getLg(P);   
  return (D);
} /* pInitDim() */


pMatDim pSubDim_(D, D1, i)
  pMatDim D, D1; int i;
{ if (D1 != D)
    MatDimCp(D, D1);
  if (i != 0) {
    unsigned int B = d_B(D),
                 P0 = d_P0(D),     	P = d_P(D);
    if (B!=1) {
      int B0  = d_B0(D);    
      int lgB = d_lgB(D);
      i = i+B0;
      if (VALID_LG(lgB)) {
      	d_B0(D) = MOD_P2(i, B);
      	i = (i>>lgB);
      } else 
      	DIV_MOD(i, B, i, d_B0(D));
    }
    i = FAST_MODP2(i, P);
    d_P0(D) = CYCLIC_SUB(P0, i, P); 
  }
  return (D);
} /* pSubDim_() */

MatDim SubDim(DI, i)
  MatDim DI; int i;
{ pSubDim_(&DI, &DI, i);
  return (DI);
} /* SubDim() */


/*********************** major DBLAS checking procedure *****************/

static int HexSelect (pix1, pix2, ParamNums)
  int pix1, pix2, ParamNums;
{ int SelectVal = 0;			
  static int PShift[] = {8, 4, 0};
  if (ParamNums < (1<<8))
    ParamNums <<= 4;			/* can drop 3rd digit, esp if DNB=NULL*/
  if (pix1 > 0) 
    SelectVal += ((ParamNums >> PShift[pix1-1]) & 0xf) << 4;
  if (pix2 > 0) 
    SelectVal += ((ParamNums >> PShift[pix2-1]) & 0xf);
  return (SelectVal);  
} /* HexSelect() */

int CommonDimAlignCheck(N, MIncA, DNA, MIncB, DNB, ParamNums)
  int N, MIncA, MIncB, *ParamNums; 	pMatDim DNA, DNB;
{ int Aligned=1; 			int ParamNumsPre = *ParamNums;
  int n=-1;
  *ParamNums = 0;  
  if (N < 0)
    *ParamNums = HexSelect(0, 1, ParamNumsPre);
  else if (!Dim_OK(DNA) || MIncA<1)       
    *ParamNums = HexSelect(0, 2, ParamNumsPre);
  else if (N>1 && MIncA>1 && ((n = G2Ld(N, DNA)) > MIncA))
    *ParamNums = HexSelect(1, 2, ParamNumsPre); 
  else if (DNB == NULL)
    return (Aligned);  
  else if (!Dim_OK(DNB) || MIncB<1)       
    *ParamNums = HexSelect(0, 3, ParamNumsPre);
  else {
    Aligned = Dim_BCompat(DNA, DNB);    	
    if (N>1 && MIncB>1) {
      if (n==-1 || !Aligned || d_P0(DNA)!=d_P0(DNB) || d_P(DNA)!=d_P(DNB))
        n = G2Ld(N, DNB);		/* G2Ld(N, DNB) != G2Ld(N, DNA) */
      if (n > MIncB)
        *ParamNums = HexSelect(1, 3, ParamNumsPre);
    }       
  }  
  return (Aligned);	
}  /* CommonDimAlignCheck() */
