/* @(#) All Right Reserved, Copyright (c) 1997 Australian National University */
/* @(#)                     Copyright (c) 1997 Fujitsu Laboratories Ltd. */
/*  */
/*        */
/*                        Permission Notice */
/*  */
/*                Australian National University */
/*                Canberra ACT 0200 Australia */
/*                           and */
/*                Fujitsu Laboratories Ltd. */
/*                4-1-1 Kamikodanaka Nakahara */
/*                Kawasaki 211-88 Japan */
/*  */
/*   Permission to use, copy, modify and distribute this software (including  */
/*   its documentation) without fee is hereby granted under the following terms */
/*   and conditions: */
/*  */
/*   1. This permission notice and all of the copyright notices in this */
/*      software must appear in all copies and derived versions of all */
/*      or a part of this software. */
/*  */
/*   2. Sales of this software, whether incorporated in other products as a */
/*      part thereof or not, is prohibited. */
/*  */
/*   3. On publication or presentation in any form of the results obtained */
/*      by use of this software, an explicit reference to this software and */
/*      its copyright owner shall be made. */
/*  */
/*   4. THIS SOFTWARE IS PROVIDED 'AS IS' AND THE AUSTRALIAN NATIONAL UNIVERSITY */
/*      AND FUJITSU LABORATORIES LTD. MAKE NO WARRANTIES, EITHER EXPRESSED OR */
/*      IMPLIED, AS TO ANY MATTER WHATSOEVER REGARDING THIS SOFTWARE,  */
/*      INCLUDING, WITHOUT LIMITATION, WARRANTIES OF MERCHANTABILITY OR FITNESS */
/*      FOR ANY PARTICULAR PURPOSE. */
/*      IN NO EVENT SHALL THE AUSTRALIAN NATIONAL UNIVERSITY OR FUJITSU */
/*      LABORATORIES LTD. BE LIABLE FOR ANY CLAIM FROM A THIRD PARTY, OR */
/*      FOR SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES WHATSOEVER RESULTING */
/*      FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN ACTION OF CONTRACT, */
/*      NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION */
/*      WITH THE USE, COPYING OR MODIFICATION OF THIS SOFTWARE. */
/*  */

/************************************************************************/
/*         PBLAS interface to the DBLAS                                 */
/************************************************************************/
/* SCCS INFO @(#)PBLAS.h 1.3         last modified P. Strazdins, 97/04/22
   first created: P. Strazdins 96/10
 */
/* 
Header file for PBLAS interface to the DBLAS
for compilation, the macro FDATATYPE (having the values SINGLE or DOUBLE)
must be defined:
	if FDATATYPE=SINGLE, FDATA=float
	if FDATATYPE=DOUBLE, FDATA=double
Below, the underscores in the procedure name should be replaced by 'S',
if FDATATYPE=SINGLE (the PSBLAS procedures), and 'D', if FDATATYPE=DOUBLE
(the PDBLAS procedures). eg. for FDATATYPE=SINGLE, P_GEMM becomes psgemm_.

See ScaL:APACK documentation on the meaning of the procedures.
*/   

#ifndef __P_BLAS_INCL__
#define __P_BLAS_INCL__

#include "FDATASetUp.h"

#ifndef ALTNAMES			/* use standard PBLAS names */
#if (FDATATYPE==SINGLE)		/*define procedure identifiers and names*/
# define P_GEMM  	psgemm_
# define P_SYMM  	pssymm_
# define P_SYRK  	pssyrk_
# define P_SYR2K 	psyr2k_
# define P_TRMM  	pstrmm_
# define P_TRSM  	pstrsm_

# define P_GEMV	 	psgemv_	 
# define P_SYMV	 	pssymv_	 
# define P_TRMV	 	pstrmv_	 
# define P_TRSV	 	pstrsv_	 
# define P_GER	 	psger_	 
# define P_SYR	 	pssyr_	 
# define P_SYR2	 	pssyr2_	 

# define P_SWAP	 	psswap_	 
# define P_SCAL	 	psscal_	 
# define P_COPY	 	pscopy_	 
# define P_AXPY	 	psaxpy_	 
# define P_DOT	 	psdot_	 
# define P_NRM2	 	psnrm2_	 
# define P_ASUM	 	psasum_	 
# define P_AMAX		psamax_	 

# define P_TRAN		pstran_

#elif (FDATATYPE==DOUBLE)
# define P_GEMM  	pdgemm_
# define P_SYMM  	pdsymm_
# define P_SYRK  	pdsyrk_
# define P_SYR2K 	pdsyr2k_
# define P_TRMM  	pdtrmm_
# define P_TRSM  	pdtrsm_

# define P_GEMV	 	pdgemv_	 
# define P_SYMV	 	pdsymv_	 
# define P_TRMV	 	pdtrmv_	 
# define P_TRSV	 	pdtrsv_	 
# define P_GER	 	pdger_	 
# define P_SYR	 	pdsyr_	 
# define P_SYR2	 	pdsyr2_	 

# define P_SWAP	 	pdswap_	 
# define P_SCAL	 	pdscal_	 
# define P_COPY	 	pdcopy_	 
# define P_AXPY	 	pdaxpy_	 
# define P_DOT	 	pddot_	 
# define P_NRM2	 	pdnrm2_	 
# define P_ASUM	 	pdasum_	 
# define P_AMAX		pdamax_	 

# define P_TRAN		pdtran_
#endif /*FDATATYPE*/

#ifdef C_CALL_FTN_NO_CHANGE
# define psgemm_	psgemm
# define pssymm_	pssymm
# define pssyrk_	pssyrk
# define pssyr2k_	pssyr2k
# define pstrmm_	pstrmm
# define pstrsm_	pstrsm
# define psgemv_	psgemv
# define pssymv_	pssymv
# define pstrmv_	pstrmv
# define psger_		psger
# define pssyr_		pssyr
# define pssyr2_	pssyr2
# define psswap_	psswap
# define psscal_	psscal
# define pscopy_	pscopy
# define psaxpy_	psaxpy
# define psdot_		psdot
# define psnrm2_	psnrm2
# define psasum_	psasum
# define psamax_	psamax
# define pstran_	pstran

# define pdgemm_	pdgemm
# define pdsymm_	pdsymm
# define pdsyrk_	pdsyrk
# define pdsyr2k_	pdsyr2k
# define pdtrmm_	pdtrmm
# define pdtrsm_	pdtrsm
# define pdgemv_	pdgemv
# define pdsymv_	pdsymv
# define pdtrmv_	pdtrmv
# define pdger_		pdger
# define pdsyr_		pdsyr
# define pdsyr2_	pdsyr2
# define pdswap_	pdswap
# define pdscal_	pdscal
# define pdcopy_	pdcopy
# define pdaxpy_	pdaxpy
# define pddot_		pddot
# define pdnrm2_	pdnrm2
# define pdasum_	pdasum
# define pdamax_	pdamax
# define pdtran_	pdtran
#endif /* C_CALL_FTN_NO_CHANGE */

#else				/* use names prepended with a `d' to 
				   avoid clashes with the UTK PBLAS */
#if (FDATATYPE==SINGLE)		/* define procedure identifiers and names*/
# define P_GEMM  	dpsgemm_
# define P_SYMM  	dpssymm_
# define P_SYRK  	dpssyrk_
# define P_SYR2K 	dpsyr2k_
# define P_TRMM  	dpstrmm_
# define P_TRSM  	dpstrsm_

# define P_GEMV	 	dpsgemv_	 
# define P_SYMV	 	dpssymv_	 
# define P_TRMV	 	dpstrmv_	 
# define P_TRSV	 	dpstrsv_	 
# define P_GER	 	dpsger_	 
# define P_SYR	 	dpssyr_	 
# define P_SYR2	 	dpssyr2_	 

# define P_SWAP	 	dpsswap_	 
# define P_SCAL	 	dpsscal_	 
# define P_COPY	 	dpscopy_	 
# define P_AXPY	 	dpsaxpy_	 
# define P_DOT	 	dpsdot_	 
# define P_NRM2	 	dpsnrm2_	 
# define P_ASUM	 	dpsasum_	 
# define P_AMAX		dpsamax_	 

# define P_TRAN		dpstran_

#elif (FDATATYPE==DOUBLE)
# define P_GEMM  	dpdgemm_
# define P_SYMM  	dpdsymm_
# define P_SYRK  	dpdsyrk_
# define P_SYR2K 	dpdsyr2k_
# define P_TRMM  	dpdtrmm_
# define P_TRSM  	dpdtrsm_

# define P_GEMV	 	dpdgemv_	 
# define P_SYMV	 	dpdsymv_	 
# define P_TRMV	 	dpdtrmv_	 
# define P_TRSV	 	dpdtrsv_	 
# define P_GER	 	dpdger_	 
# define P_SYR	 	dpdsyr_	 
# define P_SYR2	 	dpdsyr2_	 

# define P_SWAP	 	dpdswap_	 
# define P_SCAL	 	dpdscal_	 
# define P_COPY	 	dpdcopy_	 
# define P_AXPY	 	dpdaxpy_	 
# define P_DOT	 	dpddot_	 
# define P_NRM2	 	dpdnrm2_	 
# define P_ASUM	 	dpdasum_	 
# define P_AMAX		dpdamax_	 

# define P_TRAN		dpdtran_
#endif /*FDATATYPE*/

#ifdef C_CALL_FTN_NO_CHANGE
# define psgemm_	dpsgemm
# define pssymm_	dpssymm
# define pssyrk_	dpssyrk
# define pssyr2k_	dpssyr2k
# define pstrmm_	dpstrmm
# define pstrsm_	dpstrsm
# define psgemv_	dpsgemv
# define pssymv_	dpssymv
# define pstrmv_	dpstrmv
# define psger_		dpsger
# define pssyr_		dpssyr
# define pssyr2_	dpssyr2
# define psswap_	dpsswap
# define psscal_	dpsscal
# define pscopy_	dpscopy
# define psaxpy_	dpsaxpy
# define psdot_		dpsdot
# define psnrm2_	dpsnrm2
# define psasum_	dpsasum
# define psamax_	dpsamax
# define pstran_	dpstran

# define pdgemm_	dpdgemm
# define pdsymm_	dpdsymm
# define pdsyrk_	dpdsyrk
# define pdsyr2k_	dpdsyr2k
# define pdtrmm_	dpdtrmm
# define pdtrsm_	dpdtrsm
# define pdgemv_	dpdgemv
# define pdsymv_	dpdsymv
# define pdtrmv_	dpdtrmv
# define pdger_		dpdger
# define pdsyr_		dpdsyr
# define pdsyr2_	dpdsyr2
# define pdswap_	dpdswap
# define pdscal_	dpdscal
# define pdcopy_	dpdcopy
# define pdaxpy_	dpdaxpy
# define pddot_		dpddot
# define pdnrm2_	dpdnrm2
# define pdasum_	dpdasum
# define pdamax_	dpdamax
# define pdtran_	dpdtran

#endif /* C_CALL_FTN_NO_CHANGE */

#endif /* ALTNAMES */				   			   

#ifdef C_CALL_FTN_NO_CHANGE

# define pberror_	pberror
# define ptopget_	ptopget_

#endif /* C_CALL_FTN_NO_CHANGE */


/*********************** PBLAS descriptor indices ************************/

#define    DLEN_        9                     /* Length of a descriptor */
#define    DT_          0                     /* Descriptor Type        */
#define    CTXT_        1                              /* BLACS context */
#define    M_           2                      /* Global Number of Rows */
#define    N_           3                   /* Global Number of Columns */
#define    MB_          4                          /* Row Blocking Size */
#define    NB_          5                       /* Column Blocking Size */
#define    RSRC_        6                     /* Starting Processor Row */
#define    CSRC_        7                  /* Starting Processor Column */
#define    LLD_         8                    /* Local Leading Dimension */

/*********************** level 3 procedures ******************************/

void P_GEMM STDC_PARAMS(
	(char *TransA, char *TransB, int *m, int *n, int *k, FDATA *alpha, 
	 FDATA A[], int *ia, int *ja, int desc_A[],
	 FDATA B[], int *ib, int *jb, int desc_B[], FDATA *beta, 
	 FDATA C[], int *ic, int *jc, int desc_C[])
);

void P_SYMM STDC_PARAMS(
	(char *Side, char *UpLo, int *m, int *n, FDATA *alpha, 
	 FDATA A[], int *ia, int *ja, int desc_A[], 
	 FDATA B[], int *ib, int *jb, int desc_B[], FDATA *beta, 
	 FDATA C[], int *ic, int *jc, int desc_C[])
);	

void P_SYRK STDC_PARAMS(
	(char *UpLo, char *Trans, int *n, int *k, FDATA *alpha, 
	 FDATA A[], int *ia, int *ja, int desc_A[], FDATA *beta, 
	 FDATA C[], int *ic, int *jc, int desc_C[])
);	

void P_SYR2K STDC_PARAMS(
	(char *UpLo, char *Trans, int *n, int *k, FDATA *alpha, 
	 FDATA A[], int *ia, int *ja, int desc_A[], 
	 FDATA B[], int *ib, int *jb, int desc_B[], FDATA *beta, 
	 FDATA C[], int *ic, int *jc, int desc_C[])
);	


void P_TRMM STDC_PARAMS(
	(char *Side, char *UpLo, char *TransA, char *Diag, int *m, int *n, 
	 FDATA *alpha, FDATA A[], int *ia, int *ja, int desc_A[], 
	 FDATA B[], int *ib, int *jb, int desc_B[])
); 


void P_TRSM STDC_PARAMS(
	(char *Side, char *Uplo, char *TransA, char *Diag, int *m, int *n, 
	 FDATA *alpha, FDATA A[], int *ia, int *ja, int desc_A[], 
	 FDATA B[], int *ib, int *jb, int desc_B[])
); 

/*********************** level 2 procedures ******************************/ 

void P_GEMV STDC_PARAMS(
	(char  *Trans, int *m, int *n, FDATA *alpha, 
	 FDATA A[], int *ia, int *ja, int desc_A[], 
	 FDATA X[], int *ix, int *jx, int desc_X[], int *incx, FDATA *beta, 
	 FDATA Y[], int *iy, int *jy, int desc_Y[], int *incy)
);

void P_SYMV STDC_PARAMS(
	(char  *UpLo, int *n, FDATA *alpha, 
	 FDATA A[], int *ia, int *ja, int desc_A[], 
	 FDATA X[], int *ix, int *jx, int desc_X[], int *incx, FDATA *beta, 
	 FDATA Y[], int *iy, int *jy, int desc_Y[], int *incy)
);

void P_TRMV STDC_PARAMS(
        (char *UpLo, char *Trans, char *Diag, int *n, 
         FDATA A[], int *ia, int *ja, int desc_A[], 
         FDATA X[], int *ix, int *jx, int desc_X[], int *incx)
);

void P_TRSV STDC_PARAMS(
        (char *UpLo, char *Trans, char *Diag, int *n, 
        FDATA A[], int *ia, int *ja, int desc_A[], 
        FDATA X[], int *ix, int *jx, int desc_X[], int *incx)
);

void P_GER STDC_PARAMS(
	(int *m, int *n, FDATA *alpha, 
	FDATA X[], int *ix, int *jx, int desc_X[], int *incx, 
	FDATA Y[], int *iy, int *jy, int desc_Y[], int *incy, 
	FDATA A[], int *ia, int *ja, int desc_A[])
);

void P_SYR STDC_PARAMS(
	(char *UpLo, int *n, FDATA *alpha, 
	 FDATA X[], int *ix, int *jx, int desc_X[], int *incx, 
	 FDATA A[], int *ia, int *ja, int desc_A[])
);	

void P_SYR2 STDC_PARAMS(
	(char *UpLo, int *n, FDATA *alpha, 
	 FDATA X[], int *ix, int *jx, int desc_X[], int *incx, 
	 FDATA Y[], int *iy, int *jy, int desc_Y[], int *incy, 
	 FDATA A[], int *ia, int *ja, int desc_A[])
);	

/*********************** level 1 procedures ******************************/ 

void P_SWAP STDC_PARAMS( 
	(int *n, FDATA X[], int *ix, int *jx, int desc_X[], int *incx, 
		 FDATA Y[], int *iy, int *jy, int desc_Y[], int *incy) 
);

void P_SCAL STDC_PARAMS( 
	(int *n, FDATA *alpha, 
	 FDATA X[], int *ix, int *jx, int desc_X[], int *incx) 
);

void P_COPY STDC_PARAMS(
	(int *n, FDATA X[], int *ix, int *jx, int desc_X[], 
	 int *incx, FDATA Y[], int *iy, int *jy, int desc_Y[], int *incy)
);

void P_AXPY STDC_PARAMS(  
	(int *n, FDATA *alpha, 
	 FDATA X[], int *ix, int *jx, int desc_X[], int *incx, 
	 FDATA Y[], int *iy, int *jy, int desc_Y[], int *incy)
);

void P_DOT STDC_PARAMS( 
	(int *n, FDATA *Dot, 
	 FDATA X[], int *ix, int *jx, int desc_X[], int *incx, 
	 FDATA Y[], int *iy, int *jy, int desc_Y[], int *incy) 
);

void P_NRM2 STDC_PARAMS( 
	(int *n, FDATA *Nrm2, 
	 FDATA X[], int *ix, int *jx, int desc_X[], int *incx) 
);	

void P_ASUM STDC_PARAMS( 
	(int *n, FDATA *ASum,
	 FDATA X[], int *ix, int *jx, int desc_X[], int *incx) 
);

void P_AMAX STDC_PARAMS( 
	(int *n, FDATA *Max, int *MaxIndx,
	FDATA X[], int *ix, int *jx, int desc_X[], int *incx) 
);

/********************** new parallel level 1 procedures  ********************/

void P_TRAN STDC_PARAMS(
	(int *m, int *n, FDATA *alpha, 
	 FDATA A[], int *ia, int *ja, int desc_A[], FDATA *beta, 
	 FDATA C[], int *ic, int *jc, int desc_C[])
);


/********************** PBLAS tools  procedures  ********************/

void pberror_ STDC_PARAMS( (int *cntxt, char *Msg, int *info ) );

void ptopget_ STDC_PARAMS( (int *cntxt, char *toptype, char *rowcol, char *top ) );

#endif /*__P_BLAS_INCL__*/
