/* @(#) All Right Reserved, Copyright (c) 1997 Australian National University */
/* @(#)                     Copyright (c) 1997 Fujitsu Laboratories Ltd. */
/*  */
/*        */
/*                        Permission Notice */
/*  */
/*                Australian National University */
/*                Canberra ACT 0200 Australia */
/*                           and */
/*                Fujitsu Laboratories Ltd. */
/*                4-1-1 Kamikodanaka Nakahara */
/*                Kawasaki 211-88 Japan */
/*  */
/*   Permission to use, copy, modify and distribute this software (including  */
/*   its documentation) without fee is hereby granted under the following terms */
/*   and conditions: */
/*  */
/*   1. This permission notice and all of the copyright notices in this */
/*      software must appear in all copies and derived versions of all */
/*      or a part of this software. */
/*  */
/*   2. Sales of this software, whether incorporated in other products as a */
/*      part thereof or not, is prohibited. */
/*  */
/*   3. On publication or presentation in any form of the results obtained */
/*      by use of this software, an explicit reference to this software and */
/*      its copyright owner shall be made. */
/*  */
/*   4. THIS SOFTWARE IS PROVIDED 'AS IS' AND THE AUSTRALIAN NATIONAL UNIVERSITY */
/*      AND FUJITSU LABORATORIES LTD. MAKE NO WARRANTIES, EITHER EXPRESSED OR */
/*      IMPLIED, AS TO ANY MATTER WHATSOEVER REGARDING THIS SOFTWARE,  */
/*      INCLUDING, WITHOUT LIMITATION, WARRANTIES OF MERCHANTABILITY OR FITNESS */
/*      FOR ANY PARTICULAR PURPOSE. */
/*      IN NO EVENT SHALL THE AUSTRALIAN NATIONAL UNIVERSITY OR FUJITSU */
/*      LABORATORIES LTD. BE LIABLE FOR ANY CLAIM FROM A THIRD PARTY, OR */
/*      FOR SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES WHATSOEVER RESULTING */
/*      FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN ACTION OF CONTRACT, */
/*      NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION */
/*      WITH THE USE, COPYING OR MODIFICATION OF THIS SOFTWARE. */
/*  */

/************************************************************************/
/*      common macro definitions and headers for files implementing     *
 *      the PBLAS interface to the DBLAS. Includes error checking prcoedures*/
/************************************************************************/
/* SCCS INFO @(#)PBLASCheck.c 2.2         last modified P. Strazdins, 97/04/30
   first created: P. Strazdins 96/10
 */
#include "PBLASCheck.h"

/******************** error checking area ******************************/

static int DecSelect(ParamIx, ParamNums)
/* pre:  0 < ParamIx-1 <= lg_16(ParamNums)
 *       ParamNums is a double hex-encoded digit list of positions,
 * post: returns the ParamIx'th element encoded in the sequence ParamNums
 */
{ int SelectVal; 
  int psize = 1,			p = ParamNums;
  
  while (p >= 100)
    p /= 100, psize *= 100;  
  while (ParamIx > 1 && psize >= 100) 
    psize /= 100, ParamIx--;    
  SelectVal = (ParamNums / psize) % 100;
  return (SelectVal);    	
} /* DecSelect() */


/* with the conventions assume that ParamNums exists in the scope of the call:
 *	ParamNums = PpQqRr (in decimal coding)
 * and the parameter positions:
 *	 DescA, DescB & DescC 	are Pp, Qq, Rr
 * respectively. The following macros can then be used to extract
 * the parameter posiiton of the desired parameters from its encoding in 
 * ParamNums. The motivation for this is to enable efficent checking
 * for parameters that are correct.
 */
#define A_pix 		1
#define B_pix 		2
#define C_pix		3
#define ParamNumI(k)                    (DecSelect(k, ParamInfo)-2)
#define ParamNumJ(k)                    (DecSelect(k, ParamInfo)-1)
#define ParamNumDesc(k,x) 		(DecSelect(k, ParamInfo)*100 + (x) + 1)

static int PBCkeckMat(Cntxt, P, desc, ParamInfo, pix)
  int Cntxt, desc[], P[], ParamInfo, pix;
{ 
/*printf("desc a%d=%d %d %d %d %d %d %d %d\n", (int) desc, desc[0], desc[1], desc[2], desc[3],
	 desc[4], desc[5], desc[6], desc[7], desc[8]);*/
  if (P[0]==-1)				/* invalid context */	 
    return(ParamNumDesc(pix, CTXT_));
  if (desc[DT_] != 1)
    return(ParamNumDesc(pix, DT_)); 
  if (desc[CTXT_] != Cntxt)  
    return(ParamNumDesc(pix, CTXT_));
  if (desc[M_] < 1)
    return(ParamNumDesc(pix, M_)); 
  if (desc[N_] < 1)  
    return(ParamNumDesc(pix, N_)); 
  if (desc[MB_] < 1)  
    return(ParamNumDesc(pix, MB_)); 
  if (desc[NB_] < 1)  
    return(ParamNumDesc(pix, NB_));
  if ((desc[RSRC_] < 0) | (desc[RSRC_] >= P[0]))
    return(ParamNumDesc(pix, RSRC_));
  if ((desc[CSRC_] < 0) | (desc[CSRC_] >= P[1]))
    return(ParamNumDesc(pix, CSRC_));
  if (desc[LLD_] < 1)
    return(ParamNumDesc(pix, LLD_));
  return(0);  
} /* PBCkeckMat() */
 

static inline int PBCheckIx0(i, j, ParamInfo, pix)
  int i, j, ParamInfo, pix;
{  
  if ((i < 0))
    return(ParamNumI(pix));  
  if ((j < 0))
    return(ParamNumJ(pix));
  return(0);      
} /*PBCheckIx0() */
  
static inline int PBCheckIx(M, N, Trans, i, j, desc, ParamInfo, pix)
  int M, N, Trans, i, j, desc[], ParamInfo, pix;
{ 
  if (Trans)
    SWAP(int, M, N);
  if ((i < 0) | (i+M > desc[M_]))
    return(ParamNumI(pix));  
  if ((j < 0) | (j+N > desc[N_]))
    return(ParamNumJ(pix));
  return(0);  
} /* PBCheckIx() */


#define PBLLDCheck(dC, desc, pix)	\
	if (G2Ld((desc)[M_], DM_D0(dC)) > (desc)[LLD_]) return (ParamNumDesc(pix, LLD_));

/********************** inter PBLAS call cache definition *******************/
/* cache recently used PBLAS descriptors & their DistMat equivalents.
 * This can be used to avoid relatively costly MatDim construction. 
 * PBLASDescCacheOn() can be called within a computation such as a matrix
 * factorization where all PBLAS operands share the same descriptor.
 * At the end of this computation, PBLASDescCacheOff() should be called.
 */
static DistMatStore DM_cache;

static struct {				
  int On; 				/* only cache if = 1 */
  int *Desc;				/* last PBLAS desciptor cached */
  DistMat DM;				/* coorresp DistMat reporesentation */
} MatDescCache = { 0, NULL, &DM_cache};    

void PDescCacheOn()
{ MatDescCache.On = 1;
}
  
void PDescCacheOff()
{ MatDescCache.On = 0; 			MatDescCache.Desc = NULL;
}

#define PDESC_EQ(DescA, DescB)	(DescA==DescB)

/********************** common checking / construction macros **************/

#define InfoCheckRet(v)   { int info_ = (v); if (info_ > 0) return (info_); }

#define ConvDesc2DistMat(DescA, I, P, dA)				\
  pInitDim(DM_D0(dA), 0, DescA[MB_], CYCLIC_SUB(I[0],DescA[RSRC_],P[0]), P[0]),\
  pInitDim(DM_D1(dA), 0, DescA[NB_], CYCLIC_SUB(I[1],DescA[CSRC_],P[1]), P[1])
    
#define CopyDim(dA, dB)							\
  MatDimCp(DM_D0(dB), DM_D0(dA));	 MatDimCp(DM_D1(dB), DM_D1(dA));

#define pColMajorLocSubMat(plB, i, j)	pLM_ADR(plB) += (i) + pLM_CInc(plB)*(j)


#define dA DMV[0]
#define dB DMV[1]
#define dC DMV[2]

static int PBCntxt = 0;			/* BLACS Context for the call; should
					   be set by P2DBLAS_COMMSETUP() */

#define P2DBLAS_COMMSETUP(desc_A, Cntxt, I, P)  			\
        { char RowTop, ColTop;          /* set up broadcast topologies */\
	  char BroadCastC = 'B', CombineC = 'C', RowC = 'R', ColC = 'C';\
          (Cntxt) = (desc_A)[CTXT_];                         		\
          FTN_BLACS_GRIDINFO(&(Cntxt), &(P)[0], &(P)[1], &(I)[0], &(I)[1]);\
/*printf("Cntxt=%d P=%d %d, I=%d %d\n", Cntxt, P[0], P[1], I[0], I[1]);*/\
          if ((Cntxt)!=-1 && (P)[0]!=-1) {				\
            BLACSCntxtSet(Cntxt);         CommSetUp((P)[0], (P)[1]);    \
            ptopget_(&(Cntxt), &BroadCastC, &RowC, &RowTop);            \
            BLACSTopSet(ROW, &BroadCastC, &RowTop);                     \
            ptopget_(&(Cntxt), &BroadCastC, &ColC, &ColTop);            \
            BLACSTopSet(COL, &BroadCastC, &ColTop);                     \
            ptopget_(&(Cntxt), &CombineC, &RowC, &RowTop);              \
            BLACSTopSet(ROW, &CombineC, &RowTop);                       \
            ptopget_(&(Cntxt), &CombineC, &ColC, &ColTop);             	\
            BLACSTopSet(COL, &CombineC, &ColTop); }          		\
          /*TraceLevel = -1;*/						\
        }

int PDesc3ChkConv( M, N, K, iA, jA, DescA, iB, jB, DescB, iC, jC, DescC,
		   ParamInfo, DMV, Align )
  int M, N, K; int iA, jA, DescA[];  int iB, jB, DescB[]; int iC, jC, DescC[]; 
  int ParamInfo; DistMat DMV[3]; int Align[3];
{
  int Cntxt, P[2], I[2];
  int iiA, jjA, iiB, jjB, iiC, jjC;
  int TrA, TrB, TrC;

  P2DBLAS_COMMSETUP(DescA, Cntxt, I, P);
    
  if ((1-MatDescCache.On) | (1-PDESC_EQ(DescA, MatDescCache.Desc))) 
    InfoCheckRet(PBCkeckMat(Cntxt, P, DescA, ParamInfo, A_pix))  
  iA = iA-1, jA = jA-1;	TrA = DM_Trans(dA); 
  InfoCheckRet(PBCheckIx(M, K, TrA, iA, jA, DescA, ParamInfo, A_pix));

  if (!PDESC_EQ(DescB, DescA)) 
    InfoCheckRet(PBCkeckMat(Cntxt, P, DescB, ParamInfo, B_pix))    
  iB = iB-1, jB = jB-1;	TrB = DM_Trans(dB);
  InfoCheckRet(PBCheckIx(K, N, TrB, iB, jB, DescB, ParamInfo, B_pix));

  if (!PDESC_EQ(DescC, DescA)) 
    InfoCheckRet(PBCkeckMat(Cntxt, P, DescC, ParamInfo, C_pix))
  iC = iC-1, jC = jC-1;	TrC = DM_Trans(dC);	 
  InfoCheckRet(PBCheckIx(M, N, TrC, iC, jC, DescC, ParamInfo, C_pix));

  if ((1-MatDescCache.On) | (1-PDESC_EQ(DescA, MatDescCache.Desc))) {
    ConvDesc2DistMat(DescA, I, P, dA);    
    if (MatDescCache.On) {
      MatDescCache.Desc = DescA;
      CopyDim(dA, MatDescCache.DM);
    }  
    PBLLDCheck(dA, DescA, A_pix);
  } else {
    CopyDim(MatDescCache.DM, dA);
  }
  
  if (PDESC_EQ(DescB, DescA)) {
    CopyDim(dA, dB);
  } else {
    ConvDesc2DistMat(DescB, I, P, dB); 
    PBLLDCheck(dB, DescB, B_pix);
  }        
  if (PDESC_EQ(DescC, DescA)) {
    CopyDim(dA, dC);
  } else if (PDESC_EQ(DescC, DescB)) {
    CopyDim(dB, dC);
  } else {
    ConvDesc2DistMat(DescC, I, P, dC);    
    PBLLDCheck(dC, DescC, C_pix);
  }
  
  if (TrA)
    SWAP(int, iA, jA);
  iiA = G2LSubDim(iA, DM_DM(dA));	
  jjA = G2LSubDim(jA, DM_DN(dA));

  if (TrB)   
    SWAP(int, iB, jB);
  if (PDESC_EQ(DescB, DescA) & (TrA^TrB) & (iB==jA)) {
    MatDimCp(DM_DM(dB), DM_DN(dA));
    iiB = jjA;				AlignK = 1;        
  } else {
    iiB = G2LSubDim(iB, DM_DM(dB));  
    AlignK = Dim_BCompat(DM_DN(dA), DM_DM(dB));
  } 
  jjB = G2LSubDim(jB, DM_DN(dB));  
    
  if (TrC)   
    SWAP(int, iC, jC);
  if (PDESC_EQ(DescC, DescA) & (TrA==TrC) & (iC==iA)) {
    MatDimCp(DM_DM(dC), DM_DM(dA));
    iiC = iiA;				AlignM = 1;        
  } else {
    iiC = G2LSubDim(iC, DM_DM(dC));  
    AlignM = Dim_BCompat(DM_DM(dA), DM_DM(dC));
  }
  if (PDESC_EQ(DescC, DescB) & (TrB==TrC) & (jC==jB)) {
    MatDimCp(DM_DN(dC), DM_DN(dB));
    jjC = jjB;				AlignN = 1; 
  } else {        
    jjC = G2LSubDim(jC, DM_DN(dC));	
    AlignN = Dim_BCompat(DM_DN(dB), DM_DN(dC));
  }  

  if (TrA)
    SWAP(int, iiA, jjA);
  pColMajorLocSubMat(DM_pLoc(dA), iiA, jjA);
  if (TrB)
    SWAP(int, iiB, jjB);
  pColMajorLocSubMat(DM_pLoc(dB), iiB, jjB);
  if (TrC)
    SWAP(int, iiC, jjC);
  pColMajorLocSubMat(DM_pLoc(dC), iiC, jjC);

  return(0);
} /* PDesc3ChkConv() */


int PDesc2ChkConv( M, N, iA, jA, DescA, iB, jB, DescB, ParamInfo, DMV, Align )
  int M, N; int iA, jA, DescA[];  int iB, jB, DescB[]; 
  int ParamInfo; DistMat DMV[2]; int Align[1];
{ int nA = M;
  int Cntxt, P[2], I[2];
  int iiA, jjA, iiB, jjB;
  int TrA, TrB;

  P2DBLAS_COMMSETUP(DescA, Cntxt, I, P);
  if (N==-1) 				/* assume A, B are 2 vectors */
    nA = N = 1;   

  if ((1-MatDescCache.On) | (1-PDESC_EQ(DescA, MatDescCache.Desc))) 
    InfoCheckRet(PBCkeckMat(Cntxt, P, DescA, ParamInfo, A_pix))  
  iA = iA-1, jA = jA-1;	TrA = DM_Trans(dA); 
  InfoCheckRet(PBCheckIx(M, nA, TrA, iA, jA, DescA, ParamInfo, A_pix));

  if (!PDESC_EQ(DescB, DescA)) 
    InfoCheckRet(PBCkeckMat(Cntxt, P, DescB, ParamInfo, B_pix))    
  iB = iB-1, jB = jB-1;	TrB = DM_Trans(dB);
  InfoCheckRet(PBCheckIx(M, N, TrB, iB, jB, DescB, ParamInfo, B_pix));

  if ((1-MatDescCache.On) | (1-PDESC_EQ(DescA, MatDescCache.Desc))) {
    ConvDesc2DistMat(DescA, I, P, dA);    
    if (MatDescCache.On) {
      MatDescCache.Desc = DescA;
      CopyDim(dA, MatDescCache.DM);
    }  
    PBLLDCheck(dA, DescA, A_pix);
  } else {
    CopyDim(MatDescCache.DM, dA);
  }
  
  if (PDESC_EQ(DescB, DescA)) {
    CopyDim(dA, dB);
  } else {
    ConvDesc2DistMat(DescB, I, P, dB); 
    PBLLDCheck(dB, DescB, B_pix);      
  }

  if (TrA)
    SWAP(int, iA, jA);
  iiA = G2LSubDim(iA, DM_DM(dA));	
  jjA = G2LSubDim(jA, DM_DN(dA));
  
  if (TrB)   
    SWAP(int, iB, jB);
  if (PDESC_EQ(DescB, DescA) & (TrA==TrB) & (iB==iA)) {
    MatDimCp(DM_DM(dB), DM_DM(dA));
    iiB = iiA;				AlignM = 1;        
  } else {
    iiB = G2LSubDim(iB, DM_DM(dB));  
    AlignM = Dim_BCompat(DM_DM(dA), DM_DM(dB));
  } 
  jjB = G2LSubDim(jB, DM_DN(dB));
    
  if (TrA)
    SWAP(int, iiA, jjA);
  pColMajorLocSubMat(DM_pLoc(dA), iiA, jjA);
  if (TrB)
    SWAP(int, iiB, jjB);
  pColMajorLocSubMat(DM_pLoc(dB), iiB, jjB);

/*printf("%d.%d: Align=%d A =%s B=%s\n", I[0], I[1], AlignM, DM_TRACE(0, dA), DM_TRACE(1, dB));*/
  return(0);
} /* PDesc2ChkConv() */


int PDesc1ChkConv( M, N, iA, jA, DescA, ParamInfo, DMV)
  int M, N; int iA, jA, DescA[];  
  int ParamInfo; DistMat DMV[2];
{
  int Cntxt, P[2], I[2];
  int iiA, jjA;
  int TrA;

  P2DBLAS_COMMSETUP(DescA, Cntxt, I, P);
    
  if ((1-MatDescCache.On) | (1-PDESC_EQ(DescA, MatDescCache.Desc))) 
    InfoCheckRet(PBCkeckMat(Cntxt, P, DescA, ParamInfo, A_pix))  
  iA = iA-1, jA = jA-1;	TrA = DM_Trans(dA); 
  InfoCheckRet(PBCheckIx(M, N, TrA, iA, jA, DescA, ParamInfo, A_pix));

  if ((1-MatDescCache.On) | (1-PDESC_EQ(DescA, MatDescCache.Desc))) {
    ConvDesc2DistMat(DescA, I, P, dA);    
    if (MatDescCache.On) {
      MatDescCache.Desc = DescA;
      CopyDim(dA, MatDescCache.DM);
    }  
  } else {
    CopyDim(MatDescCache.DM, dA);
  }
  
  iiA = G2LSubDim(iA, DM_D0(dA));	
  jjA = G2LSubDim(jA, DM_D1(dA));
  pColMajorLocSubMat(DM_pLoc(dA), iiA, jjA);

  return(0);
} /* PDesc1VChkConv() */


void PBError_(CallerPrefix, Caller, ParamNum )
  char *CallerPrefix, *Caller; int ParamNum;
{ char PBCaller[16];
  PrintLinePrefix();
  sprintf(PBCaller, "%s%s", CallerPrefix, Caller);
  printf("*** Error detected in call %s%s (cntxt=%d, err=%d)\n",  PBCaller, TraceStr, PBCntxt, ParamNum);
  pberror_(&PBCntxt, PBCaller, &ParamNum);
} /* PBError_() */


/************************************************************************/
/*      procedure for getting/setting DBLAS internal parameters from    */
/*      Fortran programs. These control tracing & tuning values         */
/************************************************************************/
 
#include "MachineParams.h"

void DBLASSetGet(GetSet, ParamCat, ParamNum, Val)
  int *GetSet, *ParamCat, *ParamNum, *Val;
{ int *v = NULL;
  if (*ParamCat==0) {
    if (0 <= *ParamNum  &&  *ParamNum < NUM_MACH_PARAMS)
      v = &MachParamVec[*ParamNum]; 		
  } else if (*ParamCat==1) {
    if (*ParamNum==0)
      v = &TraceLevel;
    else if (*ParamNum==1) 
      v = &TraceType;
  } else if (*ParamCat==2) {
    if (*ParamNum==0)
      v = &PrintRes;
  } else if (*ParamCat==3) {
    if (*ParamNum==0)
      v = &ForceAlgVers;
  }
  if (v == NULL) *Val = -999;
  else if (*GetSet) *Val = *v;
  else *v = *Val;
} /* dblasgetset */ 
