/* @(#) All Right Reserved, Copyright (c) 1997 Australian National University */
/* @(#)                     Copyright (c) 1997 Fujitsu Laboratories Ltd. */
/*  */
/*        */
/*                        Permission Notice */
/*  */
/*                Australian National University */
/*                Canberra ACT 0200 Australia */
/*                           and */
/*                Fujitsu Laboratories Ltd. */
/*                4-1-1 Kamikodanaka Nakahara */
/*                Kawasaki 211-88 Japan */
/*  */
/*   Permission to use, copy, modify and distribute this software (including  */
/*   its documentation) without fee is hereby granted under the following terms */
/*   and conditions: */
/*  */
/*   1. This permission notice and all of the copyright notices in this */
/*      software must appear in all copies and derived versions of all */
/*      or a part of this software. */
/*  */
/*   2. Sales of this software, whether incorporated in other products as a */
/*      part thereof or not, is prohibited. */
/*  */
/*   3. On publication or presentation in any form of the results obtained */
/*      by use of this software, an explicit reference to this software and */
/*      its copyright owner shall be made. */
/*  */
/*   4. THIS SOFTWARE IS PROVIDED 'AS IS' AND THE AUSTRALIAN NATIONAL UNIVERSITY */
/*      AND FUJITSU LABORATORIES LTD. MAKE NO WARRANTIES, EITHER EXPRESSED OR */
/*      IMPLIED, AS TO ANY MATTER WHATSOEVER REGARDING THIS SOFTWARE,  */
/*      INCLUDING, WITHOUT LIMITATION, WARRANTIES OF MERCHANTABILITY OR FITNESS */
/*      FOR ANY PARTICULAR PURPOSE. */
/*      IN NO EVENT SHALL THE AUSTRALIAN NATIONAL UNIVERSITY OR FUJITSU */
/*      LABORATORIES LTD. BE LIABLE FOR ANY CLAIM FROM A THIRD PARTY, OR */
/*      FOR SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES WHATSOEVER RESULTING */
/*      FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN ACTION OF CONTRACT, */
/*      NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION */
/*      WITH THE USE, COPYING OR MODIFICATION OF THIS SOFTWARE. */
/*  */
 
/************************************************************************/
/*	common macro definitions and headers for files implementing 	*
 *      the PBLAS interface to the DBLAS. Includes error checking macros etc*/ 
/************************************************************************/
/* SCCS INFO @(#)PBLASCheck.h 2.2         last modified P. Strazdins, 97/04/30
   first created: P. Strazdins 96/12
*/
/* compilation: if USE_PBERROR_ defined, use pberror_() as the PBLAS error
 *		routine (default: use GenAssert.Error()).
 */ 
#define USE_PBERROR_
    
#ifndef __PBLASCheck_INCL__
#define __PBLASCheck_INCL__

/******************************  common include files *********************/

#include "BLASOptions.h"
#include "FDATASetUp.h"
#include "IntOps.h"
#if (FDATATYPE==SINGLE)
#define  CALLER_PREFIX "PS"
#else
#define  CALLER_PREFIX "PD"
#endif
#include "GenAssert.h"
 
#include "FtnBLACS.h"
#include "LocMat.h"
#include "FDATACommun.h"
#include "MatDim.h"
#include "DistMat.h"

#include "PBLAS.h"
#include "D_BLAS.h"
#include "D_BLASCheck.h"

/******************************  common procedures *********************/


#if (FDATATYPE==SINGLE)         /*define procedure identifiers and names*/
# define PDesc3ChkConv	PSDesc3ChkConv
# define PDesc2ChkConv	PSDesc2ChkConv
# define PDesc1ChkConv	PSDesc1ChkConv
# define PBError_	PSBError_
# define PDescCacheOn   psdesccacheon_	
# define PDescCacheOff  psdesccacheoff_	
# define DBLASSetGet	dsblassetget_
#elif (FDATATYPE==DOUBLE)
# define PDesc3ChkConv	PDDesc3ChkConv
# define PDesc2ChkConv	PDDesc2ChkConv
# define PDesc1ChkConv	PDDesc1ChkConv
# define PBError_	PDBError_
# define PDescCacheOn   pddesccacheon_	
# define PDescCacheOff  pddesccacheoff_	
# define DBLASSetGet	ddblassetget_
#endif /*FDATATYPE*/


#ifdef C_CALL_FTN_NO_CHANGE
# define psdesccacheon_		psdesccacheon
# define psdesccacheoff_	psdesccacheoff
# define dsblassetget_		dsblassetget
# define pddesccacheon_		pddesccacheon
# define pddesccacheoff_	pddesccacheoff
# define ddblassetget_		ddblassetget
#endif 

#define DistMatAlloc(n)							\
	  DistMatStore DMS[n];                  DistMat DMV[n]  
#define AlignM  Align[0]
#define AlignN  Align[1]
#define AlignK  Align[2]

int PDesc3ChkConv STDC_PARAMS(
	( int M, int N, int K, int iA, int jA, int DescA[], int iB, int jB, 
	  int DescB[], int iC, int jC, int DescC[], int ParamInfo, 
	  DistMat DMV[3], int Align[3] )
);
/* pre:  A' = (_, iA, jA, DescA) is an M x K matrix, B' = (_, iB, jB, DescB)
 *	 is a K x N matrix, C' = (_, iC, jC, DescC) is an M x N matrix.
 *	 ParamNums is a 2-decimal-digit encoded list of parameter positions
 *	 of DescA, DescB & DescC (those of iA & jA are assumed to immediately 
  *	 precede that of DescA, etc).
 *	 DM_Loc({dA, dB, dC}) give the local matrix storage specifies of A', B' 
 *	 and C'; DM_Trans({dA, dB, dC}) indicates whether A', B' and C' are 
 *  	 transposed.
 * post: checks for validity of A', B', C'; if invalid, returns the 
 *	 parameter number of the first faulty parameter found; o.w. returns 0.
 *	 dA, dB & dC are the DistMat equivalents of A', B' and C';
 *	 AlignM, AlignN & AlignK indicate block-alignment in the dimensions
 *	 of  dA, dB & dC corresponding to M, N & K respectively.
 */

int PDesc2ChkConv STDC_PARAMS(
	( int M, int N, int iA, int jA, int DescA[], int iB, int jB, 
	  int DescB[], int ParamInfo, DistMat DMV[2], int Align[1] )
);
/* pre:  A' = (_, iA, jA, DescA) is an M x M' matrix, B' = (_, iB, jB, DescB)
 *	 is an M x N' matrix, where (M', N') = (N==-1? (1, 1): (M, N).
 *	 ParamNums is a 2-decimal-digit encoded list of parameter positions
 *	 of DescA & DescB (those of iA & jA are assumed to immediately 
 *	 precede that of DescA, etc).
 *	 DM_Loc({dA, dB}) give the local matrix storage specifies of A' 
 *	 and B'; DM_Trans({dA, dB}) indicates whether A', B' are 
 *  	 transposed.
 * post: checks for validity of A', B'; if invalid, returns the 
 *	 parameter number of the first faulty parameter found; o.w. returns 0.
 *	 dA & dB  are the DistMat equivalents of A' & B';
 *	 AlignM indicates block-alignment in the dimension
 *	 of dA & dB corresponding to M  
 */

int PDesc1ChkConv STDC_PARAMS(
	( int M, int N, int iA, int jA, int DescA[], 
	  int ParamInfo, DistMat DMV[1])
);
/* pre:  A' = (_, iA, jA, DescA) is an M x N matrix, 
 *	 ParamNums is the paramter position of DescA (those of iA & jA are 
 *	 assumed to immediately precede that of DescA, etc).
 *	 DM_Loc(dA) gives the local matrix storage specifies of A';
 *	 DM_Trans(dA) indicates whether A' is transposed.
 * post: checks for validity of A''; if invalid, returns the 
 *	 parameter number of the first faulty parameter found; o.w. returns 0.
 *	 dA is the DistMat equivalents of A';
 */
 
void PBError_ STDC_PARAMS(
        ( char *CallerPrefix, char *Caller, int ParamNum )
);
/* pre:	CallerPrefix##Caller is the procedure name
 *	GenAssert.TraceStr has the vaues of the procedure's paramters.
 * post:call the PBLAS (PTOOLS) error routine, with the current BLACS context, 
 *	procedure name, and ParamNum.
 *      A trace of the procedure call is printed before this.
 */

/****** Fortran callable procedures for controlling tracing/tuning *******/

void PDescCacheOn STDC_PARAMS( (void) );
/* post: turn on PBLAS matrix descriptor - DistMat cache between PBLAS calls
 */
         
void PDescCacheOff STDC_PARAMS( (void) );
/* post: turn off PBLAS matrix descriptor - DistMat cache between PBLAS calls
 */

void DBLASSetGet STDC_PARAMS( 
	(int *GetSet, int *ParamCat, int *ParamNum, int *Val)
);
/* post: case (*ParamCat, *ParamNum) of:
 *	 	(0, 0..MachineParams.NUM_MACH_PARAMS): 
 *			v = &MachineParams.MachParamVec[*ParamNum]
 *		(1, 0): v = &GenAssert.TraceLevel 		
 *		(1, 1): v = &GenAssert.TraceType		
 *		(2, 0): v = &GenAssert.PrintRes		
 *		(3, 0): v = &GenAssert.AlgVers
 *	 else v = NULL; 
 *	 if (v = NULL) *Val = -999;
 *	 else if (*GetSet==1) *Val = *v else *v = *Val
 * notes: 1. *ParamCat==0 only effective for DBLAS compiled with -DTUNE.
 *	     This can be used for tuning architecture-depedent parameters
 *	     used in DBLAS algorithms. See MachineParams.h.
 *        2. *ParamCat==1 only effective for DBLAS compiled with -DTRACE.
 *	     Setting TraceLevel = 0 will cause tracing of outer-level 
 *	     PBLAS calls; setting TraceLevel > 0 will also cause tracing
 *	     of internal DBLAS procedures up to depth TraceLevel.
 *	     TraceType can be used for restriciting which categories of 
 *	     routines are to be traced.	See GenAssert.h for details.
 *	  3. *ParamCat==2 can be used to cause printing of up to *Val x *Val
 *	     matrices in the DBLAS SpreadMatrix module. This can display
 *	     useful intermediate results in most computations.
 *	  4. ParamCat==3 can be used to force a particular algorithm variant
 *	     within the DBLAS to be executed.
 */  	     
      
/************** macros used in error checking ********************************/

#define DMLocSetUp(Trans, Ax, desca, Aval, A)				\
	(A) = &(Aval), LM_Init(DM_Loc(A), Ax,  1, (desca)[LLD_]), 	\
	DM_SetTrans(Trans, A)
#endif /* __PBLASCheck_INCL__ */

#define PBalphaval(alpha)	(alpha)
#define PBbetaval(beta)		(beta)

#define VIncOK(inc, desc)	(((inc)==1) | ((inc)==(desc)[M_]))

#ifdef USE_PBERROR_
# undef Error_
# define Error_(pref, call, v)	PBError_(pref, call, v); return
#endif

#define DBLAS1_CHECK_FLAG	0	
/* whether to do outer level checking in the DBLAS level 1 routines,
   called by the PBLAS level 1 implementation */ 

#define PBRowVec(inc, desc)	((inc)==(desc)[M_])
#define PBColVec(inc, desc)	((inc)!=(desc)[M_])
