/* @(#) All Right Reserved, Copyright (c) 1997 Australian National University */
/* @(#)                     Copyright (c) 1997 Fujitsu Laboratories Ltd. */
/*  */
/*        */
/*                        Permission Notice */
/*  */
/*                Australian National University */
/*                Canberra ACT 0200 Australia */
/*                           and */
/*                Fujitsu Laboratories Ltd. */
/*                4-1-1 Kamikodanaka Nakahara */
/*                Kawasaki 211-88 Japan */
/*  */
/*   Permission to use, copy, modify and distribute this software (including  */
/*   its documentation) without fee is hereby granted under the following terms */
/*   and conditions: */
/*  */
/*   1. This permission notice and all of the copyright notices in this */
/*      software must appear in all copies and derived versions of all */
/*      or a part of this software. */
/*  */
/*   2. Sales of this software, whether incorporated in other products as a */
/*      part thereof or not, is prohibited. */
/*  */
/*   3. On publication or presentation in any form of the results obtained */
/*      by use of this software, an explicit reference to this software and */
/*      its copyright owner shall be made. */
/*  */
/*   4. THIS SOFTWARE IS PROVIDED 'AS IS' AND THE AUSTRALIAN NATIONAL UNIVERSITY */
/*      AND FUJITSU LABORATORIES LTD. MAKE NO WARRANTIES, EITHER EXPRESSED OR */
/*      IMPLIED, AS TO ANY MATTER WHATSOEVER REGARDING THIS SOFTWARE,  */
/*      INCLUDING, WITHOUT LIMITATION, WARRANTIES OF MERCHANTABILITY OR FITNESS */
/*      FOR ANY PARTICULAR PURPOSE. */
/*      IN NO EVENT SHALL THE AUSTRALIAN NATIONAL UNIVERSITY OR FUJITSU */
/*      LABORATORIES LTD. BE LIABLE FOR ANY CLAIM FROM A THIRD PARTY, OR */
/*      FOR SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES WHATSOEVER RESULTING */
/*      FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN ACTION OF CONTRACT, */
/*      NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION */
/*      WITH THE USE, COPYING OR MODIFICATION OF THIS SOFTWARE. */
/*  */

/************************************************************************/
/*         DBLAS interface to P_ASUM() 					*/
/************************************************************************/
/* SCCS INFO @(#)P_ASUM.c 1.1         last modified P. Strazdins, 97/04/18
   first created: P. Strazdins 97/04
 */

#include "PBLASCheck.h"
#include "D_BLAS.h"

#ifndef ALTNAMES
static int UseDBLAS_P_ASUM=0;		/* to indicate if P_ASUM is linked in*/
#endif

#define dA DMV[0]
#define dA_ DMS[0]

void P_ASUM(n, Asum, A, ia, ja, desc_A, inca)
  int *n;
  FDATA *Asum, A[];  			int *ia, *ja;
  int desc_A[], *inca;
{ int ParamErr = 0;			int Trans = PBRowVec(*inca, desc_A);
  DistMatAlloc(1);
  int NoOuterErrCheck_ = NoOuterErrCheck;
  int P0; 

  if (!DimOK(*n))			ParamErr = 1; 
  else if (!VIncOK(*inca, desc_A))	ParamErr = 7;
  else {      
    DMLocSetUp(Trans, A, desc_A, dA_, dA);
    ParamErr = PDesc1ChkConv( *n, 1, *ia, *ja, desc_A, 6, DMV);
    P0 = d_P0(DM_DN(dA));
  }
    
  ERROR_CHECK_TR(
    (TraceStr, "(%d,_,a%d,%d,%d,_,%d)", 
      *n, FD_ADR(A), *ia, *ja, *inca), 
    -ParamErr, "ASUM", 0
  );

  if (P0==0) {				/* A & B over same grid row or col */
    DistMat_RVAlloc(1);			/* ColVec() ensures result of D_ASUM()
    					   is kept within this scope */
    NoOuterErrCheck = DBLAS1_CHECK_FLAG;/* suppress checking in D_ASUM() */
    *Asum = D_ASUM(*n, ColVec(dA));
    NoOuterErrCheck = NoOuterErrCheck_; 
  }
  TRACE_EXIT;
} /* P_ASUM */


