/* @(#) All Right Reserved, Copyright (c) 1997 Australian National University */
/* @(#)                     Copyright (c) 1997 Fujitsu Laboratories Ltd. */
/*  */
/*        */
/*                        Permission Notice */
/*  */
/*                Australian National University */
/*                Canberra ACT 0200 Australia */
/*                           and */
/*                Fujitsu Laboratories Ltd. */
/*                4-1-1 Kamikodanaka Nakahara */
/*                Kawasaki 211-88 Japan */
/*  */
/*   Permission to use, copy, modify and distribute this software (including  */
/*   its documentation) without fee is hereby granted under the following terms */
/*   and conditions: */
/*  */
/*   1. This permission notice and all of the copyright notices in this */
/*      software must appear in all copies and derived versions of all */
/*      or a part of this software. */
/*  */
/*   2. Sales of this software, whether incorporated in other products as a */
/*      part thereof or not, is prohibited. */
/*  */
/*   3. On publication or presentation in any form of the results obtained */
/*      by use of this software, an explicit reference to this software and */
/*      its copyright owner shall be made. */
/*  */
/*   4. THIS SOFTWARE IS PROVIDED 'AS IS' AND THE AUSTRALIAN NATIONAL UNIVERSITY */
/*      AND FUJITSU LABORATORIES LTD. MAKE NO WARRANTIES, EITHER EXPRESSED OR */
/*      IMPLIED, AS TO ANY MATTER WHATSOEVER REGARDING THIS SOFTWARE,  */
/*      INCLUDING, WITHOUT LIMITATION, WARRANTIES OF MERCHANTABILITY OR FITNESS */
/*      FOR ANY PARTICULAR PURPOSE. */
/*      IN NO EVENT SHALL THE AUSTRALIAN NATIONAL UNIVERSITY OR FUJITSU */
/*      LABORATORIES LTD. BE LIABLE FOR ANY CLAIM FROM A THIRD PARTY, OR */
/*      FOR SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES WHATSOEVER RESULTING */
/*      FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN ACTION OF CONTRACT, */
/*      NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION */
/*      WITH THE USE, COPYING OR MODIFICATION OF THIS SOFTWARE. */
/*  */

/************************************************************************/
/*         DBLAS interface to P_GEMV() 					*/
/************************************************************************/
/* SCCS INFO @(#)P_GEMV.c 1.2         last modified P. Strazdins, 97/04/25
   first created: P. Strazdins 97/04
 */

#include "PBLASCheck.h"
#include "GenMult.h"

#ifndef ALTNAMES
static int UseDBLAS_P_GEMV=0;		/* to indicate if P_GEMV is linked in*/
#endif

#define dA DMV[0]
#define dB DMV[1]
#define dC DMV[2]
#define dA_ DMS[0]
#define dB_ DMS[1]
#define dC_ DMS[2]

void P_GEMV( TransA, m, n, alpha, A, ia, ja, desc_A,
              B, ib, jb, desc_B, incb, beta, C, ic, jc, desc_C, incc )
  char *TransA;  			int *m, *n;

  FDATA *alpha, A[], B[], *beta, C[];  	int *ia, *ja, *ib, *jb, *ic, *jc;
  int desc_A[], desc_B[], *incb, desc_C[], *incc;
{ char Trans = UpperCase(*TransA);
  int ParamErr = 0;			
  DistMatAlloc(3); 			int Align[3];
  int TrA = Trans!=NoTranspose,		TrB = PBRowVec(*incb, desc_B),
      TrC = PBRowVec(*incc, desc_C);
  int M = TrA? *n: *m,                	N = TrA? *m: *n;
  
  if (!TransOK(Trans))	ParamErr = 1;
  else if (!DimOK(*m))	ParamErr = 2; 
  else if (!DimOK(*n))	ParamErr = 3; 
  else if (!VIncOK(*incb, desc_B))	ParamErr = 13; 
  else if (!VIncOK(*incc, desc_C))	ParamErr = 19; 
  else {
    DMLocSetUp(TrA, A, desc_A, dA_, dA);
    DMLocSetUp(TrB, B, desc_B, dB_, dB);
    DMLocSetUp(TrC, C, desc_C, dC_, dC);
    ParamErr = PDesc3ChkConv( M, 1, N, *ia, *ja, desc_A, *ib, *jb, desc_B,
                 	      *ic, *jc, desc_C, 81218, DMV, Align);
    if (ParamErr==0) {                 	      
      if (!AlignM)			ParamErr = 12;    
      else if (!AlignK)			ParamErr = 8;    
    }
  }

  ERROR_CHECK_TR(
    (TraceStr, "(%c,%d,%d,%1.1f,a%d,%d,%d,_,a%d,%d,%d,_,%d,%1.1f,a%d,%d,%d,_,%d)",
      *TransA, *m, *n, *alpha, FD_ADR(A), *ia, *ja,  
      FD_ADR(B),*ib, *jb, *incb, *beta, FD_ADR(C), *ic, *jc, *incc), 
    -ParamErr, "GEMV", 0
  );

  alpha = PBalphaval(alpha);		beta = PBbetaval(beta);
  DistVecMatMult(1, M, N, alpha, Trans(dB), Trans(dA), beta, Trans(dC));

  TRACE_EXIT;
} /* P_GEMV */

