/* @(#) All Right Reserved, Copyright (c) 1997 Australian National University */
/* @(#)                     Copyright (c) 1997 Fujitsu Laboratories Ltd. */
/*  */
/*        */
/*                        Permission Notice */
/*  */
/*                Australian National University */
/*                Canberra ACT 0200 Australia */
/*                           and */
/*                Fujitsu Laboratories Ltd. */
/*                4-1-1 Kamikodanaka Nakahara */
/*                Kawasaki 211-88 Japan */
/*  */
/*   Permission to use, copy, modify and distribute this software (including  */
/*   its documentation) without fee is hereby granted under the following terms */
/*   and conditions: */
/*  */
/*   1. This permission notice and all of the copyright notices in this */
/*      software must appear in all copies and derived versions of all */
/*      or a part of this software. */
/*  */
/*   2. Sales of this software, whether incorporated in other products as a */
/*      part thereof or not, is prohibited. */
/*  */
/*   3. On publication or presentation in any form of the results obtained */
/*      by use of this software, an explicit reference to this software and */
/*      its copyright owner shall be made. */
/*  */
/*   4. THIS SOFTWARE IS PROVIDED 'AS IS' AND THE AUSTRALIAN NATIONAL UNIVERSITY */
/*      AND FUJITSU LABORATORIES LTD. MAKE NO WARRANTIES, EITHER EXPRESSED OR */
/*      IMPLIED, AS TO ANY MATTER WHATSOEVER REGARDING THIS SOFTWARE,  */
/*      INCLUDING, WITHOUT LIMITATION, WARRANTIES OF MERCHANTABILITY OR FITNESS */
/*      FOR ANY PARTICULAR PURPOSE. */
/*      IN NO EVENT SHALL THE AUSTRALIAN NATIONAL UNIVERSITY OR FUJITSU */
/*      LABORATORIES LTD. BE LIABLE FOR ANY CLAIM FROM A THIRD PARTY, OR */
/*      FOR SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES WHATSOEVER RESULTING */
/*      FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN ACTION OF CONTRACT, */
/*      NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION */
/*      WITH THE USE, COPYING OR MODIFICATION OF THIS SOFTWARE. */
/*  */

/************************************************************************/
/*         DBLAS interface to P_SYMM()                                  */
/************************************************************************/
/* SCCS INFO @(#)P_SYMM.c 2.1         last modified P. Strazdins, 97/04/22
   first created: P. Strazdins 96/10
 */

#include "PBLASCheck.h"
#include "SYMMult.h"

#ifndef ALTNAMES
static int UseDBLAS_P_SYMM=0;		/* to indicate if P_SYMM is linked in*/
#endif

#define dA DMV[0]
#define dB DMV[1]
#define dC DMV[2]
#define dA_ DMS[0]
#define dB_ DMS[1]
#define dC_ DMS[2]

void P_SYMM(Side, UpLo, m, n, alpha, A, ia, ja, desc_A, B, ib, jb,
            desc_B, beta, C, ic, jc, desc_C)
  char *Side; char *UpLo; 		int *m, *n; 
  FDATA *alpha, A[], B[], *beta, C[]; 	int  *ia, *ja, *ib, *jb, *ic, *jc;	
  int desc_A[], desc_B[], desc_C[];
{ int ParamErr = 0;
  DistMatAlloc(3);                       int Align[3];
  char SideA = UpperCase(*Side),	UpLoA = UpperCase(*UpLo);
  int Trans = SideA!=Left;		
  int M = Trans? *n: *m,		N = Trans? *m: *n;

  if (!SideOK(SideA)) 			ParamErr = 1;
  else if (!UpLoOK(UpLoA)) 		ParamErr = 2;
  else if (!DimOK(*m)) 			ParamErr = 3;
  else if (!DimOK(*n)) 			ParamErr = 4;
  else {
    DMLocSetUp(Trans, A, desc_A, dA_, dA);    
    DMLocSetUp(Trans, B, desc_B, dB_, dB);    
    DMLocSetUp(Trans, C, desc_C, dC_, dC);    
    ParamErr = PDesc3ChkConv(M, N, M, *ia, *ja, desc_A, *ib, *jb, desc_B,
                 	    *ic, *jc, desc_C, 91318, DMV, Align);
    if (ParamErr==0) {
      if (!Dim_BCompat(DM_D0(dA), DM_D1(dA))) 	ParamErr = 9;
      else if (!AlignK) 		ParamErr = 13;
      else if (!AlignM) 		ParamErr = 18;
    }
  }
    
  ERROR_CHECK_TR(
    (TraceStr, "(%c,%c,%d,%d,%1.1f,a%d,%d,%d,_,a%d,%d,%d,_,%1.1f,a%d,%d,%d)",
      *Side, *UpLo, *m, *n,  *alpha, FD_ADR(A), *ia, *ja, 
      FD_ADR(B), *ib, *jb,  *beta, FD_ADR(C), *ic, *jc),
    -ParamErr, "SYMM", 0
  );

  if (Trans) 
    UpLoA = TransForm(UpLoA);
  alpha = PBalphaval(alpha);		beta = PBbetaval(beta);
  SYMMult(M, N, alpha, UpLoA, dA, dB, beta, dC);
  TRACE_EXIT;

} /* P_SYMM() */


