/* @(#) All Right Reserved, Copyright (c) 1997 Australian National University */
/* @(#)                     Copyright (c) 1997 Fujitsu Laboratories Ltd. */
/*  */
/*        */
/*                        Permission Notice */
/*  */
/*                Australian National University */
/*                Canberra ACT 0200 Australia */
/*                           and */
/*                Fujitsu Laboratories Ltd. */
/*                4-1-1 Kamikodanaka Nakahara */
/*                Kawasaki 211-88 Japan */
/*  */
/*   Permission to use, copy, modify and distribute this software (including  */
/*   its documentation) without fee is hereby granted under the following terms */
/*   and conditions: */
/*  */
/*   1. This permission notice and all of the copyright notices in this */
/*      software must appear in all copies and derived versions of all */
/*      or a part of this software. */
/*  */
/*   2. Sales of this software, whether incorporated in other products as a */
/*      part thereof or not, is prohibited. */
/*  */
/*   3. On publication or presentation in any form of the results obtained */
/*      by use of this software, an explicit reference to this software and */
/*      its copyright owner shall be made. */
/*  */
/*   4. THIS SOFTWARE IS PROVIDED 'AS IS' AND THE AUSTRALIAN NATIONAL UNIVERSITY */
/*      AND FUJITSU LABORATORIES LTD. MAKE NO WARRANTIES, EITHER EXPRESSED OR */
/*      IMPLIED, AS TO ANY MATTER WHATSOEVER REGARDING THIS SOFTWARE,  */
/*      INCLUDING, WITHOUT LIMITATION, WARRANTIES OF MERCHANTABILITY OR FITNESS */
/*      FOR ANY PARTICULAR PURPOSE. */
/*      IN NO EVENT SHALL THE AUSTRALIAN NATIONAL UNIVERSITY OR FUJITSU */
/*      LABORATORIES LTD. BE LIABLE FOR ANY CLAIM FROM A THIRD PARTY, OR */
/*      FOR SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES WHATSOEVER RESULTING */
/*      FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN ACTION OF CONTRACT, */
/*      NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION */
/*      WITH THE USE, COPYING OR MODIFICATION OF THIS SOFTWARE. */
/*  */

/************************************************************************/
/*         DBLAS interface to P_SYR2K()                                  */
/************************************************************************/
/* SCCS INFO @(#)P_SYR2K.c 2.1         last modified P. Strazdins, 97/04/22
   first created: P. Strazdins 96/10
 */

#include "PBLASCheck.h"
#include "SYR2KMult.h"

#ifndef ALTNAMES
static int UseDBLAS_P_SYR2K=0;		/* to indicate if P_SYR2K is linked in*/
#endif

#define dA DMV[0]
#define dB DMV[1]
#define dC DMV[2]
#define dA_ DMS[0]
#define dB_ DMS[1]
#define dC_ DMS[2]

void P_SYR2K (UpLo, Trans, n, k, alpha, A, ia, ja, desc_A, B, ib, jb, desc_B, 
	     beta,  C, ic, jc, desc_C )
  char *UpLo; char *Trans;              int *n, *k;
  FDATA *alpha, A[], B[], *beta, C[];   int *ia, *ja, *ib, *jb, *ic, *jc;
  int desc_A[], desc_B[], desc_C[];
{ char TransAB = UpperCase(*Trans), 	UpLoC = UpperCase(*UpLo); 
  int ParamErr = 0;		
  DistMatAlloc(3);                      int Align[3];
         
  if (!UpLoOK(UpLoC)) 			ParamErr = 1;
  else if (!TransOK(TransAB)) 		ParamErr = 2;
  else if (!DimOK(*n)) 			ParamErr = 3;
  else if (!DimOK(*k)) 			ParamErr = 4;
  else {
    DMLocSetUp(TransAB!=NoTranspose, A, desc_A, dA_, dA);
    DMLocSetUp(TransAB!=NoTranspose, B, desc_B, dB_, dB);
    DMLocSetUp(0, C, desc_C, dC_, dC);
    ParamErr = PDesc3ChkConv(*n, *n, *k, *ia, *ja, desc_A, *ib, *jb, desc_B,
                 	      *ic, *jc, desc_C, 91318, DMV, Align);
    if (ParamErr==0) {             	      
      if (!AlignM) 			ParamErr = 13;
      else if (!Dim_BCompat(DM_D0(dC), DM_D1(dC))) ParamErr = 18;
    }
  }
    
  ERROR_CHECK_TR(
      (TraceStr, "(%c,%c,%d,%d,%1.1f,a%d,%d,%d,_,a%d,%d,%d,_,a%d,%d,%d,_)",
       *UpLo, *Trans, *n, *k, *alpha, FD_ADR(A), *ia, *ja, 
       FD_ADR(B), *ib, *jb, *beta, FD_ADR(C), *ic, *jc),
       -ParamErr, "SYR2K", 0
  );

  alpha = PBalphaval(alpha);           beta = PBbetaval(beta);
  DistSYR2KMult(*n, *k, alpha, dA, dB, beta, UpLoC, dC);
  TRACE_EXIT;
} /* D_SYR2K()*/


