/* @(#) All Right Reserved, Copyright (c) 1997 Australian National University */
/* @(#)                     Copyright (c) 1997 Fujitsu Laboratories Ltd. */
/*  */
/*        */
/*                        Permission Notice */
/*  */
/*                Australian National University */
/*                Canberra ACT 0200 Australia */
/*                           and */
/*                Fujitsu Laboratories Ltd. */
/*                4-1-1 Kamikodanaka Nakahara */
/*                Kawasaki 211-88 Japan */
/*  */
/*   Permission to use, copy, modify and distribute this software (including  */
/*   its documentation) without fee is hereby granted under the following terms */
/*   and conditions: */
/*  */
/*   1. This permission notice and all of the copyright notices in this */
/*      software must appear in all copies and derived versions of all */
/*      or a part of this software. */
/*  */
/*   2. Sales of this software, whether incorporated in other products as a */
/*      part thereof or not, is prohibited. */
/*  */
/*   3. On publication or presentation in any form of the results obtained */
/*      by use of this software, an explicit reference to this software and */
/*      its copyright owner shall be made. */
/*  */
/*   4. THIS SOFTWARE IS PROVIDED 'AS IS' AND THE AUSTRALIAN NATIONAL UNIVERSITY */
/*      AND FUJITSU LABORATORIES LTD. MAKE NO WARRANTIES, EITHER EXPRESSED OR */
/*      IMPLIED, AS TO ANY MATTER WHATSOEVER REGARDING THIS SOFTWARE,  */
/*      INCLUDING, WITHOUT LIMITATION, WARRANTIES OF MERCHANTABILITY OR FITNESS */
/*      FOR ANY PARTICULAR PURPOSE. */
/*      IN NO EVENT SHALL THE AUSTRALIAN NATIONAL UNIVERSITY OR FUJITSU */
/*      LABORATORIES LTD. BE LIABLE FOR ANY CLAIM FROM A THIRD PARTY, OR */
/*      FOR SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES WHATSOEVER RESULTING */
/*      FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN ACTION OF CONTRACT, */
/*      NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION */
/*      WITH THE USE, COPYING OR MODIFICATION OF THIS SOFTWARE. */
/*  */

/************************************************************************/
/*         DBLAS interface to P_SYRK()                                  */
/************************************************************************/
/* SCCS INFO @(#)P_SYRK.c 2.2         last modified P. Strazdins, 97/04/24
   first created: P. Strazdins 96/10
 */

#include "PBLASCheck.h"
#include "SYRKMult.h"

#ifndef ALTNAMES
static int UseDBLAS_P_SYRK=0;		/* to indicate if P_SYRK is linked in*/
#endif

#define dA DMV[1]
#define dC DMV[0]
#define dA_ DMS[1]
#define dC_ DMS[0]

void P_SYRK (UpLo, Trans, n, k, alpha, A, ia, ja, desc_A, beta, 
	     C, ic, jc, desc_C )
  char *UpLo; char *Trans; 		int *n, *k;
  FDATA *alpha, A[], *beta, C[];  	int *ia, *ja, *ic, *jc;
  int desc_A[], desc_C[];
{ char TransA = UpperCase(*Trans), 	UpLoC = UpperCase(*UpLo); 
  int ParamErr = 0;
  DistMatAlloc(2);                      int Align[1];

  if (!UpLoOK(UpLoC)) 			ParamErr = 1;    
  else if (!TransOK(TransA)) 		ParamErr = 2;
  else if (!DimOK(*n)) 			ParamErr = 3;   
  else if (!DimOK(*k)) 			ParamErr = 4;
  else {
    DMLocSetUp(TransA!=NoTranspose, A, desc_A, dA_, dA);
    DMLocSetUp(0, C, desc_C, dC_, dC);
    ParamErr = PDesc2ChkConv(*n, *k, *ic, *jc, desc_C, *ia, *ja, desc_A, 
  	     	             1409, DMV, Align);
    if (ParamErr==0) {
      if (!AlignM) 			ParamErr = 14;
      else if (!Dim_BCompat(DM_D0(dC), DM_D1(dC)))	ParamErr = 9;
    }
  }
            
  ERROR_CHECK_TR(
    (TraceStr, "(%c,%c,%d,%d,%1.1f,a%d,%d,%d,_,%1.1f,a%d,%d,%d,_)", 
     *UpLo, *Trans, *n, *k, *alpha, FD_ADR(A), *ia, *ja, *beta, FD_ADR(C), *ic, *jc), 
    -ParamErr, "SYRK", 0
  );


  alpha = PBalphaval(alpha);            beta = PBbetaval(beta);
  DistSYRKMult(*n, *k, alpha, dA, beta, UpLoC, dC);
  TRACE_EXIT;
}  /* P_SYRK()*/

