/* @(#) All Right Reserved, Copyright (c) 1997 Australian National University */
/* @(#)                     Copyright (c) 1997 Fujitsu Laboratories Ltd. */
/*  */
/*        */
/*                        Permission Notice */
/*  */
/*                Australian National University */
/*                Canberra ACT 0200 Australia */
/*                           and */
/*                Fujitsu Laboratories Ltd. */
/*                4-1-1 Kamikodanaka Nakahara */
/*                Kawasaki 211-88 Japan */
/*  */
/*   Permission to use, copy, modify and distribute this software (including  */
/*   its documentation) without fee is hereby granted under the following terms */
/*   and conditions: */
/*  */
/*   1. This permission notice and all of the copyright notices in this */
/*      software must appear in all copies and derived versions of all */
/*      or a part of this software. */
/*  */
/*   2. Sales of this software, whether incorporated in other products as a */
/*      part thereof or not, is prohibited. */
/*  */
/*   3. On publication or presentation in any form of the results obtained */
/*      by use of this software, an explicit reference to this software and */
/*      its copyright owner shall be made. */
/*  */
/*   4. THIS SOFTWARE IS PROVIDED 'AS IS' AND THE AUSTRALIAN NATIONAL UNIVERSITY */
/*      AND FUJITSU LABORATORIES LTD. MAKE NO WARRANTIES, EITHER EXPRESSED OR */
/*      IMPLIED, AS TO ANY MATTER WHATSOEVER REGARDING THIS SOFTWARE,  */
/*      INCLUDING, WITHOUT LIMITATION, WARRANTIES OF MERCHANTABILITY OR FITNESS */
/*      FOR ANY PARTICULAR PURPOSE. */
/*      IN NO EVENT SHALL THE AUSTRALIAN NATIONAL UNIVERSITY OR FUJITSU */
/*      LABORATORIES LTD. BE LIABLE FOR ANY CLAIM FROM A THIRD PARTY, OR */
/*      FOR SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES WHATSOEVER RESULTING */
/*      FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN ACTION OF CONTRACT, */
/*      NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION */
/*      WITH THE USE, COPYING OR MODIFICATION OF THIS SOFTWARE. */
/*  */

/************************************************************************/
/*         DBLAS interface to P_TRSM()                                  */
/************************************************************************/
/* SCCS INFO @(#)P_TRSM.c 2.1         last modified P. Strazdins, 97/04/22
   first created: P. Strazdins 96/10
 */

#include "PBLASCheck.h"
#include "TRSM.h"
#include "TriMatSolve.h"

#ifndef ALTNAMES
static int UseDBLAS_P_TRSM=0;		/* to indicate if P_TRSM is linked in*/
#endif

#define dA DMV[0]
#define dB DMV[1]
#define dA_ DMS[0]
#define dB_ DMS[1]

void P_TRSM( Side, UpLo, TransA, Diag, m, n, alpha, A, ia, ja, desc_A,
              B, ib, jb, desc_B )              
  char *Side; char *UpLo; char *TransA; char *Diag; int *m; int *n;
  FDATA *alpha, A[], B[];		int *ia, *ja, *ib, *jb; 
  int desc_A[], desc_B[];
{ char SideA = UpperCase(*Side), 	UpLoA = UpperCase(*UpLo), 
       DiagA = UpperCase(*Diag), 	Trans = UpperCase(*TransA); 
  int ParamErr = 0;
  DistMatAlloc(2);                      int Align[1], Align_K;
  int TrB = (SideA!=Left); 
  int M = TrB? *n: *m,                  N = TrB? *m: *n;           

  if (!SideOK(SideA))                   ParamErr = 1;
  else if (!UpLoOK(UpLoA))              ParamErr = 2;
  else if (!TransOK(Trans))             ParamErr = 3;
  else if (!DiagOK(DiagA))              ParamErr = 4;
  else if (!DimOK(*m))                  ParamErr = 5;
  else if (!DimOK(*n))                  ParamErr = 6;
  else {
    DMLocSetUp(TrB^(Trans!=NoTranspose), A, desc_A, dA_, dA);
    DMLocSetUp(TrB, B, desc_B, dB_, dB);
    ParamErr = PDesc2ChkConv(M, N, *ia, *ja, desc_A, *ib, *jb, desc_B, 
                             1115, DMV, Align);
    Align_K = Dim_BCompat(DM_DM(dB), DM_DN(dA));
    if ((ParamErr==0) & ((1-AlignM)|(1-Align_K)))     ParamErr = 15;
  }

  ERROR_CHECK_TR(   
    (TraceStr, "(%c,%c,%c,%c,%d,%d,%1.1f,a%d,%d,%d,_,a%d,%d,%d,_)",
     *Side, *UpLo, *TransA, *Diag, *m, *n, *alpha, FD_ADR(A), *ia, *ja, 
     FD_ADR(B), *ib, *jb),
    -ParamErr, "TRSM", 0
  );
    
  alpha = PBalphaval(alpha);
  if (TrB)
    UpLoA = TransForm(UpLoA);
    
  if (!AlignM || (Align_K && 4*N<=M))
    if (Blk1d(N, DM_DN(dB)) || !AlignM)
      TriMatSolve1Blk(UpLoA, DiagA, M, N, alpha, dA, dB);
    else
      TriMatSolve(UpLoA, DiagA, M, N, alpha, dA, dB);
  else
    TR_M(1, UpLoA, DiagA, M, N, alpha, dA, dB);
  TRACE_EXIT;
} /* P_TRSM() */

