/* @(#) All Right Reserved, Copyright (c) 1997 Australian National University */
/* @(#)                     Copyright (c) 1997 Fujitsu Laboratories Ltd. */
/*  */
/*        */
/*                        Permission Notice */
/*  */
/*                Australian National University */
/*                Canberra ACT 0200 Australia */
/*                           and */
/*                Fujitsu Laboratories Ltd. */
/*                4-1-1 Kamikodanaka Nakahara */
/*                Kawasaki 211-88 Japan */
/*  */
/*   Permission to use, copy, modify and distribute this software (including  */
/*   its documentation) without fee is hereby granted under the following terms */
/*   and conditions: */
/*  */
/*   1. This permission notice and all of the copyright notices in this */
/*      software must appear in all copies and derived versions of all */
/*      or a part of this software. */
/*  */
/*   2. Sales of this software, whether incorporated in other products as a */
/*      part thereof or not, is prohibited. */
/*  */
/*   3. On publication or presentation in any form of the results obtained */
/*      by use of this software, an explicit reference to this software and */
/*      its copyright owner shall be made. */
/*  */
/*   4. THIS SOFTWARE IS PROVIDED 'AS IS' AND THE AUSTRALIAN NATIONAL UNIVERSITY */
/*      AND FUJITSU LABORATORIES LTD. MAKE NO WARRANTIES, EITHER EXPRESSED OR */
/*      IMPLIED, AS TO ANY MATTER WHATSOEVER REGARDING THIS SOFTWARE,  */
/*      INCLUDING, WITHOUT LIMITATION, WARRANTIES OF MERCHANTABILITY OR FITNESS */
/*      FOR ANY PARTICULAR PURPOSE. */
/*      IN NO EVENT SHALL THE AUSTRALIAN NATIONAL UNIVERSITY OR FUJITSU */
/*      LABORATORIES LTD. BE LIABLE FOR ANY CLAIM FROM A THIRD PARTY, OR */
/*      FOR SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES WHATSOEVER RESULTING */
/*      FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN ACTION OF CONTRACT, */
/*      NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION */
/*      WITH THE USE, COPYING OR MODIFICATION OF THIS SOFTWARE. */
/*  */

/************************************************************************/
/*         DBLAS interface to P_TRSV()                                  */
/*         PBLAS interface to the DBLAS					*/
/************************************************************************/
/* SCCS INFO @(#)P_TRSV.c 1.1         last modified P. Strazdins, 97/04/22
   first created: P. Strazdins 97/04
 */

#include "PBLASCheck.h"
#include "TriMatSolve.h"

#ifndef ALTNAMES
static int UseDBLAS_P_TRSV=0;		/* to indicate if P_TRSV is linked in*/
#endif


#define dA DMV[0]
#define dB DMV[1]
#define dA_ DMS[0]
#define dB_ DMS[1]

void P_TRSV( UpLo, TransA, Diag, n, A, ia, ja, desc_A,
             B, ib, jb, desc_B, incb )              
  char *UpLo; char *TransA; 		char *Diag; int *n;
  FDATA A[], B[];			int *ia, *ja, *ib, *jb; 
  int desc_A[], desc_B[], *incb;
{ char Trans = UpperCase(*TransA), 	UpLoA = UpperCase(*UpLo), 
       DiagA = UpperCase(*Diag);
  int ParamErr;
  DistMatAlloc(2);                      int Align[1];
  int TrB = PBRowVec(*incb, desc_B);

  if (!UpLoOK(UpLoA))			ParamErr = 1;
  else if (!TransOK(Trans))		ParamErr = 2;
  else if (!DiagOK(DiagA))		ParamErr = 3;
  else if (!DimOK(*n))			ParamErr = 4;
  else if (!VIncOK(*incb, desc_B))	ParamErr = 13;
  else {
    DMLocSetUp((Trans!=NoTranspose), A, desc_A, dA_, dA);
    DMLocSetUp(TrB, B, desc_B, dB_, dB);
    ParamErr = PDesc2ChkConv(*n, 1, *ia, *ja, desc_A, *ib, *jb, desc_B,
                             812, DMV, Align);                    
    if ((ParamErr==0) & (1-Dim_BCompat(DM_DM(dB), DM_DN(dA))))	
    					ParamErr = 12;
  }
  
  ERROR_CHECK_TR(   
    (TraceStr, "(%c,%c,%c,%d,a%d,%d,%d,_,a%d,%d,%d,_,%d)",
     *UpLo, *TransA, *Diag, *n, FD_ADR(A), *ia, *ja, FD_ADR(B), *ib, *jb, *incb),
    -ParamErr, "TRSV", 0
  );
    
  if (TrB)
    UpLoA = TransForm(UpLoA);
  TriMatSolve1Blk(UpLoA, DiagA, *n, 1, &FD_1, dA, dB);
      
  TRACE_EXIT;
} /* P_TRSV() */

