/* @(#) All Right Reserved, Copyright (c) 1997 Australian National University */
/* @(#)                     Copyright (c) 1997 Fujitsu Laboratories Ltd. */
/*  */
/*        */
/*                        Permission Notice */
/*  */
/*                Australian National University */
/*                Canberra ACT 0200 Australia */
/*                           and */
/*                Fujitsu Laboratories Ltd. */
/*                4-1-1 Kamikodanaka Nakahara */
/*                Kawasaki 211-88 Japan */
/*  */
/*   Permission to use, copy, modify and distribute this software (including  */
/*   its documentation) without fee is hereby granted under the following terms */
/*   and conditions: */
/*  */
/*   1. This permission notice and all of the copyright notices in this */
/*      software must appear in all copies and derived versions of all */
/*      or a part of this software. */
/*  */
/*   2. Sales of this software, whether incorporated in other products as a */
/*      part thereof or not, is prohibited. */
/*  */
/*   3. On publication or presentation in any form of the results obtained */
/*      by use of this software, an explicit reference to this software and */
/*      its copyright owner shall be made. */
/*  */
/*   4. THIS SOFTWARE IS PROVIDED 'AS IS' AND THE AUSTRALIAN NATIONAL UNIVERSITY */
/*      AND FUJITSU LABORATORIES LTD. MAKE NO WARRANTIES, EITHER EXPRESSED OR */
/*      IMPLIED, AS TO ANY MATTER WHATSOEVER REGARDING THIS SOFTWARE,  */
/*      INCLUDING, WITHOUT LIMITATION, WARRANTIES OF MERCHANTABILITY OR FITNESS */
/*      FOR ANY PARTICULAR PURPOSE. */
/*      IN NO EVENT SHALL THE AUSTRALIAN NATIONAL UNIVERSITY OR FUJITSU */
/*      LABORATORIES LTD. BE LIABLE FOR ANY CLAIM FROM A THIRD PARTY, OR */
/*      FOR SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES WHATSOEVER RESULTING */
/*      FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN ACTION OF CONTRACT, */
/*      NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION */
/*      WITH THE USE, COPYING OR MODIFICATION OF THIS SOFTWARE. */
/*  */

/************************************************************************/
/* SYR2KMult: DBLAS Symmetric Rank-2K Multiplication procedures		*/
/************************************************************************/
/* SCCS INFO @(#)SYR2KMult.h 4.1         last modified P. Strazdins, 97/02/19
   first created: P. Strazdins 95/04/02
*/
/* compilation: macro FDATATYPE must be defined;
 *              if macro ASSERT defined, assertion checking turned on
 *              if macro TRACE defined, parameter value tracing enabled
 *		if macro PRINTMAT define, matrices are printed for debugging
 */
#ifndef __SYR2KMult_INCL__
#define __SYR2KMult_INCL__

#include "FDATASetUp.h"
#include "DistMat.h"

#if FDATATYPE==SINGLE
# define DistSYR2KMult	DSDistSYR2KMult
#else
# define DistSYR2KMult	DDDistSYR2KMult
#endif

void DistSYR2KMult STDC_PARAMS( 
	(int N, int K, FDATA *alpha, DistMat A, DistMat B, 
	 FDATA *beta, char FormC, DistMat C)
);
/* pre:  all cells are executing DistRankKMult() with 
 *	 M, N, K, *alpha, *beta,  FormC parameters and (D[0..1], Trans) 
 * 	 fields for A, B & C consistent across all cells;
 *	 M, N, K >= 0; DM_Trans(A)==DM_Trans(B) && UpLoOK(FormC)
 * post: performs global operation  C = (*beta)*C + (*alpha)*(AB^T + BA^T),
 *	 where C is a triangular matrix of form FormC,
 * notes: efficient for K << N
 */

#endif /*__SYR2KMult_INCL__*/

