/* @(#) All Right Reserved, Copyright (c) 1997 Australian National University */
/* @(#)                     Copyright (c) 1997 Fujitsu Laboratories Ltd. */
/*  */
/*        */
/*                        Permission Notice */
/*  */
/*                Australian National University */
/*                Canberra ACT 0200 Australia */
/*                           and */
/*                Fujitsu Laboratories Ltd. */
/*                4-1-1 Kamikodanaka Nakahara */
/*                Kawasaki 211-88 Japan */
/*  */
/*   Permission to use, copy, modify and distribute this software (including  */
/*   its documentation) without fee is hereby granted under the following terms */
/*   and conditions: */
/*  */
/*   1. This permission notice and all of the copyright notices in this */
/*      software must appear in all copies and derived versions of all */
/*      or a part of this software. */
/*  */
/*   2. Sales of this software, whether incorporated in other products as a */
/*      part thereof or not, is prohibited. */
/*  */
/*   3. On publication or presentation in any form of the results obtained */
/*      by use of this software, an explicit reference to this software and */
/*      its copyright owner shall be made. */
/*  */
/*   4. THIS SOFTWARE IS PROVIDED 'AS IS' AND THE AUSTRALIAN NATIONAL UNIVERSITY */
/*      AND FUJITSU LABORATORIES LTD. MAKE NO WARRANTIES, EITHER EXPRESSED OR */
/*      IMPLIED, AS TO ANY MATTER WHATSOEVER REGARDING THIS SOFTWARE,  */
/*      INCLUDING, WITHOUT LIMITATION, WARRANTIES OF MERCHANTABILITY OR FITNESS */
/*      FOR ANY PARTICULAR PURPOSE. */
/*      IN NO EVENT SHALL THE AUSTRALIAN NATIONAL UNIVERSITY OR FUJITSU */
/*      LABORATORIES LTD. BE LIABLE FOR ANY CLAIM FROM A THIRD PARTY, OR */
/*      FOR SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES WHATSOEVER RESULTING */
/*      FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN ACTION OF CONTRACT, */
/*      NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION */
/*      WITH THE USE, COPYING OR MODIFICATION OF THIS SOFTWARE. */
/*  */
  
/************************************************************************/
/* TRSM: DBLAS update B with (inverse of) triangular matrix A		*/
/************************************************************************/
/* SCCS INFO @(#)TRSM.c 4.4         last modified P. Strazdins, 97/04/09
   first created: P. Strazdins 93/15/01
*/
/* compilation: macro FDATATYPE must be defined;
 *              if macro ASSERT defined, assertion checking turned on
 *              if macro TRACE defined, parameter value tracing enabled
 */

#include "TRSM.h"
#include "FDATAOps.h"
#include "FDATAAssert.h"
#include "BLASOptions.h"
 
#include "LocMat.h"
#include "MatDim.h"
#include "MachineParams.h"
#include "FDATACellMatrix.h"
#include "FDATACommun.h"
#include "SpreadMatrix.h"
#include "GetBlockFactor.h"


void SpreadL2TR_M(
	DoTRSM, UpLo, Diag, RowCol, w, w_loc, n, BDM, AsMScale, AsM, BsM, 
	B, brdB
)
  int DoTRSM, UpLo; char Diag; int RowCol, w, w_loc, n; 
  pMatDim BDM; FDATA *AsMScale; pLocMat AsM, BsM, B;
  int brdB;
{ int iw;
  int i_loc; 				MatDim BDM_; 
  COMMON_ASSERT_TR_E(
    (TraceStr, "(%d,%d%c,%d,%dx%d,%d %d:%d %d,%1.1f,a%d %d %d,a%d %d %d,a%d %d %d,%d)",
    DoTRSM, UpLo, Diag, RowCol, w, n, d_B0(BDM), d_B(BDM), d_P0(BDM), d_P(BDM), 
    pFD_VAL(AsMScale), FD_ADR(pLM_ADR(AsM)), pLM_RInc(AsM), pLM_CInc(AsM), 
    FD_ADR(pLM_ADR(BsM)), pLM_RInc(BsM), pLM_CInc(BsM), 
    FD_ADR(pLM_ADR(B)),   pLM_RInc(B),   pLM_CInc(B), brdB),
    w>=1 && n>=0 && (brdB || Blk1d(w, BDM)), 0, "SpreadL2TR_M", 2, TRACE_ALL
  );
  if (n==0) {
    TRACE_RETURN;
  }  
  TRACE_LEVEL_SHIFT(2);
  
  BDM = pSubDim_(&BDM_, BDM, (DoTRSM^UpLo)? 0: w-1);
  i_loc = (DoTRSM^UpLo)? 0: (w_loc - (d_P0(BDM)==0));  
  
  for (iw=0; iw < w; iw++) {
    /* invariant: i_loc = G2Ld(i, BDM_pre) */	
    int i = (DoTRSM^UpLo)? iw: w-iw-1;	/* global relative index */
    int BDMiP0 = d_P0(BDM);   		/* cell-offset of diag cell */
    LocMat AsMi, BsMi, Bi, Bid;
    int i_locd, m;
    pLocSubMat_(&Bi, B, i_loc, 0);	pLocSubMat_(&BsMi, BsM, i, 0);
    if (DoTRSM && BDMiP0==0 && Diag!=Unit) {
      FDATA Aii_inv = - pFD_VAL(AsMScale) / *pLM_Index(AsM, i_loc, i);
      MatScale(1, n, &Aii_inv, &Bi);
    }       
    if (DoTRSM && brdB) {
      FDATA *Bx = LM_ADR(BsMi);
      if (BDMiP0==0) {
        if (LM_Inc(1, Bi)==1)		/* optimization: send directly from B */
	  Bx = LM_ADR(Bi);
	else if (LM_ADR(Bi)!=LM_ADR(BsMi)) 
	  VecCopy(n, LM_ADR(Bi), LM_CInc(Bi), LM_ADR(BsMi), LM_CInc(BsMi));
      } else {
        COMMON_ASSERT(LM_ADR(Bi)!=Bx, 0, "SpreadL2TR_M:BsM=B", 6, TRACE_COMMUN);
      }
      BrdMat(RowCol, BDMiP0, 1, n, Bx, n);
      if (BDMiP0==0 && LM_ADR(Bi)!=LM_ADR(BsMi) && LM_Inc(1, Bi)==1)
	VecCopy(n, LM_ADR(Bi), LM_CInc(Bi), LM_ADR(BsMi), LM_CInc(BsMi));
      if (PrintRes)
        MatPrint("Bi", 1, MIN(PrintRes, n), BsMi);      	
    } 

    /* B[i1..i2,:] -= A[i1..12,i]*B[i,:], i1..i2 = UpLo? 0..i-1: i+1..w-1 */
    i_locd = i_loc;
    if (!UpLo && BDMiP0==0) {		/* advance past diagonal element */
      i_locd++;		
      pLocSubMat(&Bi, 1, 0);
    }  		
    pLocSubMat_(&AsMi, AsM, UpLo? 0: i_locd, i);
    m = UpLo? i_locd: w_loc - i_locd;
    if (m > 0)
      MatMult(m, n, 1, AsMScale, &AsMi, &BsMi, NULL, UpLo? B: &Bi);

    if (!DoTRSM && BDMiP0==0 && Diag!=Unit) 
      MatScale(1, n, pLM_Index(AsM, i_loc, i), pLocSubMat_(&Bi, B, i_loc, 0));
          	      
    i_loc += (DoTRSM^UpLo)? G2LSubDim(1, BDM): G2LSubDimM1(BDM);
  } /* for (iw...) */
  TRACE_LEVEL_SHIFT(-2);
  TRACE_EXIT;
} /* SpreadL2TR_M() */ 


#define lA 	DM_Loc(A)
#define lB 	DM_Loc(B)
#define BDM 	DM(B)
#define DM(B) 	DM_D(  Side, B)
#define BDN 	DM_D(1-Side, B)

void TR_M(DoTRSM, UpLo, Diag, M, N, alpha, A, B)
  int DoTRSM; char UpLo, Diag; int M, N;
  FDATA *alpha; DistMat A, B;
{ int j, jF, dj;			/* iterate over j: (;j!=jF;j+=dj) */
  int Side = DM_Trans(B);		/* 0 = `Left'; 1 = `Right' */
  int UpLoTr = ((UpLo==Upper)^DM_Trans(A))^Side;
  LocMat lBt;

  COMMON_ASSERT_TR_E(
    (TraceStr, "(%d,%c%c,%dx%d,%1.1f,%d a%d %d %d,%d a%d %d %d)",
     DoTRSM, UpLo, Diag, M, N, *alpha, DM_Trans(A), FD_ADR(LM_ADR(lA)), LM_RInc(lA), 
     LM_CInc(lA), DM_Trans(B), FD_ADR(LM_ADR(lB)), LM_RInc(lB), LM_CInc(lB)),
    M>=0 && N>=0 && UpLoOK(UpLo) && DiagOK(Diag), 0, "TR_M", 1, TRACE_ALL
  );

  if (M==0 || N==0) {
    TRACE_RETURN;
  } 
  if (M==1 && Diag==Unit /* A is identity matrix */ || FD_EQ(*alpha, FD_0)) { 
    LocMat lBt;				LM_Copy(Side, lBt, lB);
    MatScale(G2Ld(M, BDM), G2Ld(N, BDN), alpha, &lBt);
    TRACE_RETURN;
  }

  if (UpLoTr==DoTRSM) { 		/* backward column/row iteration */  
    j = M-1; jF = -1; dj = -1; 
  } else {          			/* forward column/row iteration */
    j = 0;   jF = M;  dj =  1; 
  }

  while (j != jF) { 
    /* update B[j0..j1,:] by A[j0..j1,jL..jR];  [jL..jR] = spread of jL..jR */
    DistMat AjL, BjL;			/* SubMat(A, j0, jL), SubMat(B, jL, 0)*/
    DistMat_RVAlloc(2);
    int w, jL, jR;			/* w is blocking factor; jL+w-1 = jR  */
    int m; 				/* local length in AsM for level 3 mm */
    int nB;				/* local length for B */
    int w_loc;				/* local length for w */
    LocMat AsM, 			/* init. A[j0..j1, [jL..jR]] 	      */
	   BsM;                    	/* B[[jL..jR], :]		      */
    int AsMn, BsMn;    			/* = workspace lengths for AsM,BsM*/
    FDATA AsMScale = DoTRSM? FD_M1: FD_1;
    pLocMat AsMr1, 			/* A[jL..jR, [jL..jR]]		      */
   	    AsMmm; 			/* A[jL..jR, [jL..jR]]		      */
    LocMat AsM_,
	   lBr1, 			/* B[jL..jR, :]  		*/
	   lBmm;			/* B[j0mm..j1mm, :] 		*/
    int j0, j1/*, j0mm , j1mm*/;	/* see defn below; m ~ j1mm-j0mm+1 */
    MatDim BDMj0;			/* SubDim(BDM, j0) */

    w = GenGetW(UpLoTr? j+1: M-j, 1.0, PAR_INNER_VECSIZE_TRSM_MIN,  
    		MIN(ABS(jF-j), PAR_INNER_VECSIZE_TRSM), BDM, j, dj); 
    jL = dj>0? j: j-w+1;		jR = dj>0? j+w-1: j;
    if (UpLoTr) {
      j0 = 0; 		j1 = jR;	/* j0mm = 0;		j1mm = jL-1; */
    } else {
      j0 = jL; 		j1 = M-1; 	/* j0mm = jR+1;		j1mm = M-1; */
    }      

    BjL = SubMat(B, jL, 0);		
    if (Side)
      DM_Transpose(BjL); 		/* no transpose in SpreadMat() */
    w_loc = G2Ld(w, DM(BjL));
    LM_Copy(Side, lBr1, DM_Loc(BjL));	/* lBr1 is w_loc x n */
    if (UpLoTr) {
      LM_Copy(Side, lBmm, lB);		/* lBmm is m x nB, above lBr1 */
    } else {
      LM_Copy(Side, lBmm, DM_Loc(BjL));	
      pLocSubMat(&lBmm, w_loc, 0); 	/* lBmm is m x nB, below lBr1 */
    }  
    pSubDim_(&BDMj0, BDM, j0); 
    
    AjL = SubMat(A, j0, jL); 
    if (Side)                      	/* SpreadMat() requires transpose ...*/
      DM_Transpose(AjL);   		/* ... iff DM_Trans(A)==Side	     */
    LM_ADR(AsM) = NULL;
    AsMn = SpreadMat(!Side, w, j1-j0+1, AjL, &AsMScale, &BDMj0, &AsM, 0, N, BDN);
    m = ABS(AsMn) - w_loc;

    LM_Trans(Side, AsM);		/* AsM is (m+w_loc) x w */
    if (UpLoTr) {
      AsMr1 = pLocSubMat_(&AsM_, &AsM, m, 0); 
      AsMmm = &AsM;
    } else {
      AsMr1 = &AsM;
      AsMmm = pLocSubMat_(&AsM_, &AsM, w_loc, 0);
    }  	  				/* AsM m x w, A[j0mm..j1mm, [jL..jR]]*/

    LM_ADR(BsM) = NULL;
    if (!DoTRSM) {
      if (!pFD_UNITY(alpha))	
        MatScale(w_loc, G2Ld(N, BDN), alpha, &lBr1);
      BsMn = SpreadMat(Side, w, N, BjL, &FD_1, BDN, &BsM, 0, j1-j0+1, &BDMj0);
      nB = ABS(BsMn);
      LM_Trans(Side, BsM);		/* BsM is w x nB */	
      if (m > 0)
        MatPartMult(m, nB, w, NULL, AsMmm, &BsM, NULL, &lBmm);
    } else {
      BsMn = InitSpreadWS(Side, w, N, BjL, BDN, &BsM, 0, j1-j0+1, &BDMj0);
      nB = ABS(BsMn);
      LM_Trans(Side, BsM);		/* BsM is w x nB */	
    }

    /* level 3 MM B[j0mm..j1mm,:] -= A[j0mm..j1,[jL..jR]] * B[[jL..jR],:]; 
       tri. update B[jL..jR,:] by A[jL..jR, [jL..jR]], B[[jL..jR],:]	*/ 

    SpreadL2TR_M(DoTRSM, UpLoTr, Diag, Side, w, w_loc, nB, DM(BjL), 
    		 &AsMScale, AsMr1, &BsM, &lBr1, 1-Blk1d(j1-j0+1, &BDMj0));

    if (DoTRSM) {
      if (m > 0)
        MatPartMult(m, nB, w, &AsMScale, AsMmm, &BsM, NULL, &lBmm);
      if (!pFD_UNITY(alpha))	
        MatScale(w_loc, nB, alpha, &lBr1);
    }
    if (BsMn>0)
      MatFree(LM_ADR(BsM));
    if (AsMn>0)
      MatFree(LM_ADR(AsM));
    j += w * dj;  
  } /* while (j...) */
  TRACE_EXIT;
} /* TR_M() */  

#undef lA 
#undef lB 
#undef BDM 
#undef BDN 
#undef DM
