/* @(#) All Right Reserved, Copyright (c) 1997 Australian National University */
/* @(#)                     Copyright (c) 1997 Fujitsu Laboratories Ltd. */
/*  */
/*        */
/*                        Permission Notice */
/*  */
/*                Australian National University */
/*                Canberra ACT 0200 Australia */
/*                           and */
/*                Fujitsu Laboratories Ltd. */
/*                4-1-1 Kamikodanaka Nakahara */
/*                Kawasaki 211-88 Japan */
/*  */
/*   Permission to use, copy, modify and distribute this software (including  */
/*   its documentation) without fee is hereby granted under the following terms */
/*   and conditions: */
/*  */
/*   1. This permission notice and all of the copyright notices in this */
/*      software must appear in all copies and derived versions of all */
/*      or a part of this software. */
/*  */
/*   2. Sales of this software, whether incorporated in other products as a */
/*      part thereof or not, is prohibited. */
/*  */
/*   3. On publication or presentation in any form of the results obtained */
/*      by use of this software, an explicit reference to this software and */
/*      its copyright owner shall be made. */
/*  */
/*   4. THIS SOFTWARE IS PROVIDED 'AS IS' AND THE AUSTRALIAN NATIONAL UNIVERSITY */
/*      AND FUJITSU LABORATORIES LTD. MAKE NO WARRANTIES, EITHER EXPRESSED OR */
/*      IMPLIED, AS TO ANY MATTER WHATSOEVER REGARDING THIS SOFTWARE,  */
/*      INCLUDING, WITHOUT LIMITATION, WARRANTIES OF MERCHANTABILITY OR FITNESS */
/*      FOR ANY PARTICULAR PURPOSE. */
/*      IN NO EVENT SHALL THE AUSTRALIAN NATIONAL UNIVERSITY OR FUJITSU */
/*      LABORATORIES LTD. BE LIABLE FOR ANY CLAIM FROM A THIRD PARTY, OR */
/*      FOR SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES WHATSOEVER RESULTING */
/*      FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN ACTION OF CONTRACT, */
/*      NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION */
/*      WITH THE USE, COPYING OR MODIFICATION OF THIS SOFTWARE. */
/*  */
  
/************************************************************************/
/* TRSM: DBLAS update B with (inverse of) triangular matrix A		*/
/************************************************************************/
/* SCCS INFO @(#)TRSM.h 1.2         last modified P. Strazdins, 97/04/09
   first created: P. Strazdins 93/15/01
*/
/* compilation: macro FDATATYPE must be defined;
 *              if macro ASSERT defined, assertion checking turned on
 *              if macro TRACE defined, parameter value tracing enabled
 *              if macro PRINTMAT defined, matrices are printed for debugging
 */
#ifndef __TRSM_INCL__
#define __TRSM_INCL__ 

#include "FDATASetUp.h" 
#include "DistMat.h"

#if (FDATATYPE==SINGLE)
#define SpreadL2TR_M	DSSpreadL2TR_M
#define TR_M		DSTR_M
#else
#define SpreadL2TR_M	DSSpreadL2TR_M
#define TR_M		DDTR_M
#endif 

void SpreadL2TR_M STDC_PARAMS(
        (int DoTRSM, int UpLo, char Diag, int RowCol, int w, int w_loc, int n, 
	 pMatDim BDM, FDATA * AsMScale, pLocMat AsM, pLocMat BsM, pLocMat B, 
	 int brdB)
);
/* pre:  AsM is a spread matrix of a global w x w matrix, of form UpLo
 *       and diagonal Diag. BsM is the spread matrix of B in dimension w;
 *       BDM specifies the dimension of B in the spread dimension.
 *       n is the local width of B and BsM; note w is a global size.
 *       w_loc = G2Ld(w, BDM)
 *       pFD_VAL(AsMScale)=+/-1; (-1 only if DoTRSM)
 * post: performs distributed update of w x n B by AsM and BsM
 *       if (DoTRSM), b/c of BsM rows only if brdB
 */


void TR_M STDC_PARAMS( 
	(int DoTRSM, char UpLo, char Diag, int M, int N, 
	 FDATA *alpha, DistMat A, DistMat B)
);
/* pre:  all cells are executing TR_M() with consistent 
 *	 DoTRSM, UpLo, Diag M, N, alpha, A, & B;
 *	 UpLoOK(UpLo); DiagOK(Diag); 0<= M, N
 * post: performs B := *alpha f(A)B 
 *       where f(A) = DoTRSM? Inv(A): A, and A is of form UpLo.
 *       with MxN B, 
 *       if Diag=Unit, diagonal elements of A assumed unity and not accessed
 * 	 The `left' variant of parallel TRSM/TRMM is the normal one here;
 *	 to achieve the `right' variant, interchange M & N and transpose A & B
 */

#endif /*__TRSM_INCL__*/
