/* @(#) All Right Reserved, Copyright (c) 1997 Australian National University */
/* @(#)                     Copyright (c) 1997 Fujitsu Laboratories Ltd. */
/*  */
/*        */
/*                        Permission Notice */
/*  */
/*                Australian National University */
/*                Canberra ACT 0200 Australia */
/*                           and */
/*                Fujitsu Laboratories Ltd. */
/*                4-1-1 Kamikodanaka Nakahara */
/*                Kawasaki 211-88 Japan */
/*  */
/*   Permission to use, copy, modify and distribute this software (including  */
/*   its documentation) without fee is hereby granted under the following terms */
/*   and conditions: */
/*  */
/*   1. This permission notice and all of the copyright notices in this */
/*      software must appear in all copies and derived versions of all */
/*      or a part of this software. */
/*  */
/*   2. Sales of this software, whether incorporated in other products as a */
/*      part thereof or not, is prohibited. */
/*  */
/*   3. On publication or presentation in any form of the results obtained */
/*      by use of this software, an explicit reference to this software and */
/*      its copyright owner shall be made. */
/*  */
/*   4. THIS SOFTWARE IS PROVIDED 'AS IS' AND THE AUSTRALIAN NATIONAL UNIVERSITY */
/*      AND FUJITSU LABORATORIES LTD. MAKE NO WARRANTIES, EITHER EXPRESSED OR */
/*      IMPLIED, AS TO ANY MATTER WHATSOEVER REGARDING THIS SOFTWARE,  */
/*      INCLUDING, WITHOUT LIMITATION, WARRANTIES OF MERCHANTABILITY OR FITNESS */
/*      FOR ANY PARTICULAR PURPOSE. */
/*      IN NO EVENT SHALL THE AUSTRALIAN NATIONAL UNIVERSITY OR FUJITSU */
/*      LABORATORIES LTD. BE LIABLE FOR ANY CLAIM FROM A THIRD PARTY, OR */
/*      FOR SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES WHATSOEVER RESULTING */
/*      FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN ACTION OF CONTRACT, */
/*      NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION */
/*      WITH THE USE, COPYING OR MODIFICATION OF THIS SOFTWARE. */
/*  */
 
/************************************************************************/
/* TriMatSolve: DBLAS Tri Mat Solve alg					*/
/************************************************************************/
/* SCCS INFO @(#)TriMatSolve.h 1.3         last modified P. Strazdins, 97/04/25
   first created: P. Strazdins 97/03
*/

#ifndef __TriMatSolve_INCL__
#define __TriMatSolve_INCL__
   
#include "LocMat.h"
#include "DistMat.h"

#if FDATATYPE==SINGLE
# define TriMatSolve		DSTriMatSolve
# define TriMatSolve1Blk	DSTriMatSolve1Blk
#else
# define TriMatSolve		DDTriMatSolve
# define TriMatSolve1Blk	DDTriMatSolve1Blk
#endif

void TriMatSolve STDC_PARAMS( 
  (char UpLo, char Diag, int M, int N, FDATA *alpha, DistMat A, DistMat B)
);
/* pre:  all cells are executing TriMatSolve with consistent
 *       UpLo, Diag M, N, alpha, A, & B;
 *       UpLoOK(UpLo); DiagOK(Diag); 0<= M, N
 *	 Dim_BCompat(DM_DM(B), DM_DM(A)), Dim_BCompat(DM_DM(B), DM_DN(A))
 * post: performs B := *alpha A^-1 B
 *       where A is of form UpLo with MxN B,
 *       if Diag=Unit, diagonal elements of A assumed unity and not accessed
 * notes: efficient for N<<M
 */


void TriMatSolve1Blk STDC_PARAMS( 
  (char UpLo, char Diag, int M, int N, FDATA *alpha, DistMat A, DistMat B)
);
/* pre:  as for TriMatSolve(), except Dim_BCompat(DM_DM(B), DM_DM(A))
 * 	 not required
 * post: as for TriMatSolve()
 * notes: more efficient than TriMatSolve() for Blk1d_(N, DM_DN(B)), esp N=1
 */

#endif /*__TriMatSolve_INCL__*/

