/* @(#) All Right Reserved, Copyright (c) 1997 Australian National University */
/* @(#)                     Copyright (c) 1997 Fujitsu Laboratories Ltd. */
/*  */
/*        */
/*                        Permission Notice */
/*  */
/*                Australian National University */
/*                Canberra ACT 0200 Australia */
/*                           and */
/*                Fujitsu Laboratories Ltd. */
/*                4-1-1 Kamikodanaka Nakahara */
/*                Kawasaki 211-88 Japan */
/*  */
/*   Permission to use, copy, modify and distribute this software (including  */
/*   its documentation) without fee is hereby granted under the following terms */
/*   and conditions: */
/*  */
/*   1. This permission notice and all of the copyright notices in this */
/*      software must appear in all copies and derived versions of all */
/*      or a part of this software. */
/*  */
/*   2. Sales of this software, whether incorporated in other products as a */
/*      part thereof or not, is prohibited. */
/*  */
/*   3. On publication or presentation in any form of the results obtained */
/*      by use of this software, an explicit reference to this software and */
/*      its copyright owner shall be made. */
/*  */
/*   4. THIS SOFTWARE IS PROVIDED 'AS IS' AND THE AUSTRALIAN NATIONAL UNIVERSITY */
/*      AND FUJITSU LABORATORIES LTD. MAKE NO WARRANTIES, EITHER EXPRESSED OR */
/*      IMPLIED, AS TO ANY MATTER WHATSOEVER REGARDING THIS SOFTWARE,  */
/*      INCLUDING, WITHOUT LIMITATION, WARRANTIES OF MERCHANTABILITY OR FITNESS */
/*      FOR ANY PARTICULAR PURPOSE. */
/*      IN NO EVENT SHALL THE AUSTRALIAN NATIONAL UNIVERSITY OR FUJITSU */
/*      LABORATORIES LTD. BE LIABLE FOR ANY CLAIM FROM A THIRD PARTY, OR */
/*      FOR SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES WHATSOEVER RESULTING */
/*      FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN ACTION OF CONTRACT, */
/*      NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION */
/*      WITH THE USE, COPYING OR MODIFICATION OF THIS SOFTWARE. */
/*  */
 
/************************************************************************/
/* TriMatUpdate: DBLAS Tri Mat Update (Multiply) alg				*/
/************************************************************************/
/* SCCS INFO @(#)TriMatUpdate.c 1.1         last modified P. Strazdins, 97/04/08
   first created: P. Strazdins 97/03
*/
/* compilation: macro FDATATYPE must be defined;
 *              if macro ASSERT defined, assertion checking turned on
 *              if macro TRACE defined, parameter value tracing enabled
 */

#include "TriMatUpdate.h"

#include "BLASOptions.h"
#include "FDATAOps.h"
#include "FDATAAssert.h"

#include "LocMat.h"
#include "MatDim.h"
#include "DistMat.h"
#include "MachineParams.h"
#include "FDATACellMatrix.h"
#include "FDATACellTri1Matrix.h"
#include "SpreadMatrix.h"

#define ADM DM_DM(A)
#define ADN DM_DN(A)
#define BDM DM_DM(B)
#define BDN DM_DN(B)  
#define lA DM_Loc(A)
#define lB DM_Loc(B)

void TriMatUpdate(UpLo, Diag, M, N, alpha, A, B)
  char UpLo, Diag; int M, N;
  FDATA *alpha; DistMat A, B; 
{ int n;                         	
  int RowCol = DM_Trans(A);
  DistMatStore BN;

  COMMON_ASSERT_TR_E(
    (TraceStr, "(%c%c,%dx%d,%1.1f,%d a%d %d %d,%d a%d %d %d)",
     UpLo, Diag, M, N, *alpha, DM_Trans(A), FD_ADR(LM_ADR(lA)), LM_RInc(lA), 
     LM_CInc(lA), DM_Trans(B), FD_ADR(LM_ADR(lB)), LM_RInc(lB), LM_CInc(lB)),
     CompatBlk(BDM, ADN) && M>=0 && N>=0 , 0, "TriMatUpdate", 1, TRACE_ALL
  ); 
 
  if (M==0 || N==0) {
    TRACE_RETURN;
  }

  if (DM_Trans(B))
    UpLo = TransForm(UpLo);
  n=0;
  while (n < N) {
    int mA /*= G2Ld(M, ADM)*/, 		kA /*= G2Ld(M, ADN)*/;
    FDATA ScAlpha = *alpha;
    int dN = MIN(PAR_OUTER_VECSIZE, N-n);
    DistMat Bt = RowCol==0? Trans(B): B, Ct = RowCol==0? B: Trans(B);	
    LocMat BsN, CsN;                	/* spread workspaces of B */
    int BsNn, CsNn;

    LM_ADR(CsN) = NULL;
    CsNn = InitReduceWS(1-RowCol, dN, M, Ct, &FD_0, ADM, &CsN, 1, M, ADN);    
    mA = ABS(CsNn);

    LM_ADR(BsN) = NULL;
    BsNn = SpreadMat(RowCol, dN, M, Bt, &ScAlpha, ADN, &BsN, 0, M, ADM);
    kA = ABS(BsNn);			

    LM_Trans(1-RowCol, BsN);		/* ensure BsN is kA x dN */
    LM_Trans(RowCol, CsN);		/* ensure CsN is mA x dN */
    Tri1MatPartMult(mA, dN, kA, &ScAlpha, UpLo, Diag, A, &BsN, NULL, &CsN); 
    if (BsNn>0)
      MatFree(LM_ADR(BsN));
    LM_Trans(RowCol, CsN);		/* restore CsN */

    ReduceMat(1-RowCol, dN, M, Ct, &FD_0, ADM, &CsN, M, ADN);
    if (CsNn>0)
      MatFree(LM_ADR(CsN));
      
    n += dN;
    if (n < N)
      B = SubMat_(&BN, B, 0, dN);
  } /* while (n...) */
  TRACE_EXIT;
} /*TriMatUpdate()*/
